.\"
.\" gated.man,v 1.9 1992/12/04 23:40:13 jch Exp
.\"
.\" Gated Release 3.0
.\" Copyright (c) 1990,1991,1992,1993 by Cornell University. All rights reserved.
.\" Refer to Particulars and other Copyright notices at the end of this file.
.\"
.TH GATED 8 "7 October 1992" "Gated Release 3.0" "Cornell GateDaemon Project"
.SH NAME
gated \- gateway routing daemon
.SH SYNOPSIS
.I gated
[
.B \-c
] [
.B \-n
] [
.B \-N
] [
.BR \-t\fR<trace_options>
] [
.BR \-f <config_file>
] [trace_file]
.SH DESCRIPTION
.I Gated
is a routing daemon that handles multiple routing protocols and replaces
routed(8), egpup(8), and any routing daemon that
speaks the HELLO routing protocol.
.I Gated
currently handles the RIP, BGP, EGP,
HELLO, and OSPF routing protocols.  The
.I gated
process can be configured to perform all
routing protocols or any 
combination of the five.
.PP
The command-line options are:
.IP -c
Specifies that the configuration file will be parsed for syntax errors
and then gated will exit, if there were no errors, gated will leave a
dump file in /usr/tmp/gated_dump.  Gated does not need to be run as
the super user to use the -c option but it may not be possible to read
the kernel's routing table if not run as super user.  The -c option
implies -tierk.
.IP -n
Specifies that gated will not modify the kernel's
routing table.  This is used for testing gated
configurations with actual routing data.
.IP -n
Specifies that gated will not daemonize.
Normally, if tracing to the console is
not specified gated will daemonize if
the parent process ID is not 1.  This
allows the use of an /etc/inittab-like
method of invoking gated that does not
have a PID of 1.
.IP -t
Specifies a comma separated list of
trace options to be enabled on startup.
If no flags are specified, "general" is
assumed.  No space is allowed
between this option and it's arguments.
.sp
This option must be used to trace events
that take place before the config file
is parsed, such as determining the
interface configuration and reading
routes from the kernel.
.PP
The trace options are explained in greater detail in
.BR gated-config (5).
.RE
.IP -f
Use an alternate config file.
By default,
.I gated
uses
.IR /etc/gated.conf .
.PP
If on the command line a trace file is specified, or no trace
flags are specified, gated detaches from the terminal
and runs in the background.  If trace flags are
specified without specifying a trace file, gated
assumes that tracing is desired to the console and
remains in the foreground.
.SH "SIGNAL PROCESSING"
.PP
.I Gated
catches a the following signals and does special processing.
.PD 1.5v
.IP SIGHUP: 16
Re-read configuration.
.sp
A SIGHUP causes gated to reread the configuration
file.  Gated first performs a clean-up of all
allocated policy structures.  All BGP and EGP peers
are flagged for deletion and the configuration file is
re-parsed. 
.sp
If the re-parse is successful, any BGP and EGP peers
that are no longer in the configuration are shut down,
and new peers are started.  Gated attempts to
determine if changes to existing peers require a
shutdown and restart.
.sp
It should also be possible to enable/disable any
protocol without restarting gated.
.sp
Reconfiguration is currently disabled when OSPF
is enabled, this will hopefully be fixed in a
future release.
.IP SIGINT:
Snap-shot of current state.
.sp
The current state of all gated tasks, timers,
protocols and tables are written to
/usr/tmp/gated_dump.
.sp
On systems supporting fork(), this is done by
forking a subprocess to dump the table information so
as not to impact gated's routing functions.  On
systems where memory management does not support
copy-on-write, this will cause the gated address space
to be duplicated; this may be cause a noticeable impact
on the system.  On system not supporting fork(), the
main process immediately processes the dump, which may
impact gated's routing functions.
.IP SIGTERM:
Graceful shutdown.
.sp
On receipt of a SIGTERM, gated attempts a graceful
shutdown.  All tasks and protocols are asked to
shutdown.  Most will terminate immediately, the
exception being EGP peers which wait for confirmation.
It may be necessary to repeat the SIGTERM once or
twice if it this process takes too long.
.sp
All exterior routes (BGP and EGP) are removed from the
kernel's routing table on receipt of a SIGTERM.
Interior routes (all others) remain.  To terminate
gated with the exterior routes intact, use SIGKILL or
SIGQUIT (which causes a core dump).
.IP SIGUSR1:
Toggle tracing.
.sp
On receipt of a SIGUSR1, gated will close the trace
file.  A subsequent SIGUSR1 will cause it to be
reopened.  This will allow the file to be moved
regularly.
.sp
It is not possible to use SIGUSR1 if a trace file has
not been specified, or tracing is being performed to
the console.
.IP SIGUSR2:
Check for interface changes.
.sp
On receipt of a SIGUSR2, gated will rescan the kernel
interface list looking for changes.
.PD
.SH AUTHORS
Mark Fedor <fedor@psi.com>
.br
Jeffrey C Honig <jch@gated.cornell.edu>
.br
Rob Coltun <rcoltun@ni.umd.edu>
.SH SEE ALSO
gated-config(5),
gated-copyright(8),
arp(8),
ifconfig(8),
netstat(8),
ripquery(8)
routed(8)
.PP
.PD 0
.IP RFC\ 891 16
DCN Local-Network Protocols (HELLO)
.IP RFC\ 904
Exterior Gateway Protocol Formal Specification
.IP RFC\ 1058
Routing Information Protocol
.IP RFC\ 1163
A Border Gateway Protocol (BGP)
.IP RFC\ 1164
Application of the Border Gateway Protocol in the Internet
.IP RFC\ XXXX
OSPF Specification, Version 2
.PD
.SH COPYRIGHT INFORMATION
This software and associated documentation is Copyright \*(CO 1990,
1991, 1992 Cornell University., all rights reserved.
.PP
This daemon contains code that is Copyright \*(CO 1988 Regents of the
University of California., all rights reserved.; it also contains code
that is Copyright \*(CO 1989, 1990, 1991 The University of Maryland,
College Park, Maryland., all rights reserved; and also contains code that 
is Copyright 1991 D.L.S. Associates., all rights reserved.
.\"
.\" ------------------------------------------------------------------------
.\" 
.\" 	GateD, Release 3
.\" 
.\" 	Copyright (c) 1990,1991,1992,1993 by Cornell University
.\" 	    All rights reserved.
.\" 
.\" 	THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT ANY
.\" 	EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT
.\" 	LIMITATION, THE IMPLIED WARRANTIES OF MERCHANTABILITY
.\" 	AND FITNESS FOR A PARTICULAR PURPOSE.
.\" 
.\" 	Royalty-free licenses to redistribute GateD Release
.\" 	3 in whole or in part may be obtained by writing to:
.\" 
.\" 	    GateDaemon Project
.\" 	    Information Technologies/Network Resources
.\" 	    200 CCC
.\" 	    Cornell University
.\" 	    Ithaca, NY  14853-2601  USA
.\" 
.\" 	GateD is based on Kirton's EGP, UC Berkeley's routing
.\" 	daemon	 (routed), and DCN's HELLO routing Protocol.
.\" 	Development of GateD has been supported in part by the
.\" 	National Science Foundation.
.\" 
.\" 	Please forward bug fixes, enhancements and questions to the
.\" 	gated mailing list: gated-people@gated.cornell.edu.
.\" 
.\" 	Authors:
.\" 
.\" 		Jeffrey C Honig <jch@gated.cornell.edu>
.\" 		Scott W Brim <swb@gated.cornell.edu>
.\" 
.\" ------------------------------------------------------------------------
.\" 
.\"       Portions of this software may fall under the following
.\"       copyrights:
.\" 
.\" 	Copyright (c) 1988 Regents of the University of California.
.\" 	All rights reserved.
.\" 
.\" 	Redistribution and use in source and binary forms are
.\" 	permitted provided that the above copyright notice and
.\" 	this paragraph are duplicated in all such forms and that
.\" 	any documentation, advertising materials, and other
.\" 	materials related to such distribution and use
.\" 	acknowledge that the software was developed by the
.\" 	University of California, Berkeley.  The name of the
.\" 	University may not be used to endorse or promote
.\" 	products derived from this software without specific
.\" 	prior written permission.  THIS SOFTWARE IS PROVIDED
.\" 	``AS IS'' AND WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES,
.\" 	INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
.\" 	MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
.\"
