#ifndef _idrp_globals_h
#define _idrp_globals_h

/* Options */

/* flow control */

extern int acked_something;		/* True if last received packet acknowledged something */

/* task interface */

extern int doing_idrp;		/* Are we running IDRP protocols? */
extern int idrp_reparsing;	/* idrp is reparsing */

extern task *idrp_master_task;		/* IDRP master task block */
extern flag_t idrp_trace_flags;	/* Trace flags from parser */

/* idrp variables for unix sockets */

extern u_short idrp_port;	/* IDRP well known port */

/* global addresses for IP and OSI */
/* addresses for this box */
/* ip addresses - rt_ip_rdi - must have match to AS */
/* look at this again prior to moving to IP */

extern struct iso_net_addr rt_rdi;	/* Our routing domain identifier */
extern struct iso_net_addr rt_net;	/* Our network entity title %% */
extern struct iso_net_addr rt_as_rdi;	/* our Mapping from AS to RDI */ 
extern snpa_list	rt_snpa;	/* snpa for this node */

/* IP information */

extern struct iso_net_addr rt_ip_rdi;	/* Our Routing Domain identifier for IP */
					/* only if we are running SIN idrp */
					/* for IP and OSI */ 
extern struct sockaddr_un rt_ip_net;	/* our IP address for NET */


/* task block pointers for allocation of space in task block initialization */

/* local route information */

extern int	idrp_local_route_cnt;		 	/* count of local route blocks */
extern idrp_attribute_record *local_route_attr_rec;	/* local routes attribute record */
	
extern int local_route_id;	/* our local route id - negative */
extern int local_hop_count;	/* our local hop count - zero */


/* link list of idrp neighbors - IP and OSI depends only on the address inside */

extern int	idrp_neighbors;		/* Count of neighbors */
extern int	idrp_ext_neighbors;		/* Count of external neighbors */
extern int	idrp_n_disabled;		/* configured but disabled neighbors */
extern int	idrp_n_established;		/* established IDRP neighbors */ 
extern int 	idrp_n_ext_established;	/* Unused? */

extern idrpPeer	*idrp_peers;		/* Linked list of IDRP peers */
extern idrpPeer	*idrp_ext_peers;	/* pointer to first external peer */
extern idrpPeer	idrp_this_node;		/* peer structure for this node */
					/* used in local routes */

extern RD_path		local_rd_path;		/* fake RD pathway structure */ 
						/* for local routes */

extern int	consistency_errors;		/* consistency errors */

/* attribute lists */

extern idrp_attribute_record *idrp_attr_list;	/* linked list of attributes */
extern idrp_canon_rdpath *canon_rd_list;	/* linked list of rd paths */

/* stuff for best external routes pseudo-RIB */

extern idrp_best_ext *ber_rib_head;	/* points to beginning of best external */

/* add structure for adding to table */
/* idrp pdu buffer */

extern idrpPdu	*idrp_send_buffer;		/* buffer for sending idrp pdu */

/* Policy definitions for idrp */
/* below comment carried over from idrp_init.c */
/* why are policy definitions not here??? */
/* why no definition of idrp_import_list, idrp_export_list */
/* idrp_import_paths, idrp_export_paths */

extern int notifyLengths[];		/* Lengths of notifyData */
extern metric_t idrp_default_metric;	/* Default IDRP metric to use */

extern adv_entry *idrp_import_list;	/* List of IDRP advise entries */
extern adv_entry *idrp_export_list;	/* List of IDRP export entries */
extern adv_entry *idrp_import_paths;	/* List of IDRP import paths */
extern adv_entry *idrp_export_paths;	/* List of IDRP export paths */

/* idrp protocol states */
 
extern bits idrpStates[];		/* Values of the states */

/* should this change to bits??? */

extern char *idrpManagementEvents[];	/* Names of IDRP management events */
extern char *idrpTypes[]; 		/* Names of IDRP PDU types */

/* Maybe this should be replaced by a shift of these bits
 * Must match the sequence of ATTRIBUTE types
 */

extern char idrpAttrType[];		/* Type code for IDRP Attributes */ 

/* attribute flags array */

extern int idrpAttrFlags[]; 		/* attribute flag values */
extern int idrpAttrBits[];		/* attribute bit mask */
extern bits idrp_flag_bits[];		/* pretty print bits */ 
extern char *idrpChangeTypes[];		/* Names of IDRP change field values */
extern nlri_proto_info idrp_nlri_proto_id[];	/* protocol id info array for nlri in update */ 
extern nlri_fam nlri_family[];		/* nlri family */
extern u_int	proto_header_len[];	/* header length for each protocol  for update pdu */


/* these arrays are not set-up */
 
extern bits idrpEvents[];			/* Event values */

/* jgs -- the below are never referenced or declared */
extern bits idrpOptions[];
extern bits idrpAsDirs[];               /* Values of the directions */
extern bits idrpOpenType[];             /* Values of the open types */
extern bits idrpPduType[];              /* Values of the packet types */
extern bits idrpErrors[];               /* Values of error codes */
extern bits idrpUpdateErrors[];         /* Values of update error subcodes */
extern bits idrpFlags[];                 /* Values of the flags */
/* jgs -- the above are never referenced or declared */

#endif /* not defined _idrp_globals_h */
