/* idrp_init.c

 *
 *  Initial Coding by Dave Katz, Merit/NSFNET
 *  Re-write by John Scudder, Susan Hares Merit/NSFNET. 
 *
 */

#include "include.h"
#include "idrp.h"

/* temporary inclusion for sending to TCP port */
#include "inet.h"

/* timer */
static timer *retransmission_timer;

flag_t	 idrp_default_trace_flags;	/* DEFAULT  trace flags */
int	 idrp_default_preference;	/* DEFAULT  trace flags */


/* The Idrp_Init Code saga begins here */

/* gated requires that the following sequence of initialization take place 
 * (If you want to play the idrp-gated blues, you  must play this intros 
 * This is the minimal overture to the task movement  
 *    * Initialization call sequence:
         * idrp_cleanup (only on reconfig)
         * idrp_var_init
	 * parse
	 * idrp_init
	 * idrp_re-init
	 * idrp_flash
	 * idrp_new_policy

* bgp's intro does a symphony intro - it plays all the main themes of: 
* (gotta keep the beat with Mr Jones Q. BGP  or loose internet status)
	 *
         * bgp_cleanup(only on reconfig)
	 * bgp_var_init
         * parse
	 * bgp_init
	 * bgp_reinit - called after config file re-read and bgp_init done
	 * bgp_flash
	 * bgp_newpolicy
*/

/*
 *    idrp_cleanup before re-parse
 *    when called: before configuration file is to be re-read
 *    what done: loop through all import/export marking groups 
 *               to be deleted.  Since currently we only
 *               have global lists the pointers are set to zero.
 */

static void
idrp_cleanup(tp)
task *tp;
{
idrpPeer 	*peer;


	/* flag each peer as delete so initialization will clean up 
	 * each idrp_peer task will clean up structure
	 * idrp_cleanup - must loop through idrp peers and reset the globals
	 */


	idrp_reparsing = TRUE;

	/* clear the counters of established peers */

	idrp_n_disabled = 0;
	idrp_n_established = 0;


    /* set peers to delete */
 	 
    IDRP_LIST(peer)
	{
	BIT_SET(peer->idrp_flags,IDRPF_DELETE);
	}	

    /* free local_route list flag on local chain
     * let new parsing fill in loca routes once again.
     */ 

	idrp_local_route_reset();
	 

    /* BGP cleans up the import lists by groups */
    /* once we get to policy we will need to do */
    /* this too.  Right now the import and export */
    /* list is a simple global policy */ 
 
    adv_free_list(idrp_import_list);
    idrp_import_list = (adv_entry *) 0;
    adv_free_list(idrp_export_list);
    idrp_export_list = (adv_entry *) 0;
 
    adv_free_list(idrp_import_paths);
    idrp_import_paths = (adv_entry *) 0;
    adv_free_list(idrp_export_paths);
    idrp_export_paths = (adv_entry *) 0;
    idrp_var_init();
}


/* 
 * idrp_var_init - initialize variables prior to gated
 *                 parsing the configuration file.
 * set up global idrp processing (doing_idrp or better
 * known as "doing the idrp blues"), trace flags
 * default metric, and default preference within gated 
*/

void idrp_var_init()
{
	doing_idrp = FALSE;
	idrp_default_trace_flags = trace_flags;
	idrp_default_metric = IDRP_METRIC_NONE;
	idrp_default_preference = RTPREF_IDRP_EXT;
	
	/* reset the local IDRP structures  prior to config run
	 * - reset count of local routes
	 * - pointer to local attribute record 
	 * - reset the local rd path length - since no one is using it 
 	 */

	if (!idrp_reparsing)
		{ 
		idrp_last_local_att == (idrp_attribute_record *) 0;
		}
	idrp_local_route_cnt = 0;
	bzero(&local_rd_path,sizeof(RD_path)); 
	
	/* set-up the Peer structure for this node */

	idrp_this_node.next = (idrpPeer *) 0;

}



sockaddr_un *idrp_find_interface(type)
int	type;
{
sockaddr_un	*localaddr;
idrpPeer	*peer;
if_addr		*ifap;

	/* walk down the list fo IDRP peers and try to find a configured interface */

	IDRP_LIST(peer) {

		/* IP interface search  */ 
		if (type == AF_INET)
			{
		    	if (peer->ip_intf_addr) 
				{
				localaddr = peer->ip_intf_addr;
				break;
		    		}
			}

		/* ISO interface search */

		if (type == AF_ISO)
			{
			if (peer->iso_intf_addr.isoa_len)
				{
				localaddr = (sockaddr_un *) &peer->iso_intf_addr;
				break;
				}
			}
	} IDRP_LIST_END
	
	/* find an active interface address to send this pdu to the neighbor */
	
	if (!localaddr) {
	
	    IF_ADDR(ifap) { 
	        if (!BIT_TEST(ifap->ifa_state, IFS_LOOPBACK) &&
			(socktype(ifap->ifa_addr) == type))
			{
			localaddr = ifap->ifa_addr;
			break;
			}
    	      } IF_ADDR_END(ifap); 
	}

return (localaddr);
}



/* terminate the main reception task 
 *
 */


void
idrp_terminate(tp)
task *tp;
{
idrpPeer	*peer;

/* terminate the master task */
/* since we can't receive -terminate all the other tasks */ 

	IDRP_LIST(peer) {
		if (peer->idrp_flags == IDRPF_CONNECT)
			{
			idrp_peer_terminate(peer->task);
			}
		/* delete the idrp master task *
		 * skh -- need to remove the locally configured routes
		 *   from the route table 
		 */

		idrp_local_route_clean();
		task_delete(idrp_this_node.task);	
		idrp_this_node.task = (task *) 0;
 
	} IDRP_LIST_END;
	

}
 
 
/*
 *      shut down a peer 
 * 	delete the task, and delete
 *	the idrp structure
 */
void
idrp_peer_terminate(tp)
task *tp;
{
	idrpPeer *peer = (idrpPeer *) tp->task_data;
	if (peer->idrp_flags == IDRPF_CONNECT)
		{
		idrp_stop_event(peer);
		timer_delete(tp->task_timer[IDRPTIMER_CLOSEWAIT]);
		timer_delete(tp->task_timer[IDRPTIMER_HOLDTIME]);
		}
	task_delete(tp);
    	idrp_delete(peer);
}


 
/*
 *      Cleanup for a peer
 */
void
idrp_peer_cleanup(tp)
task *tp;
{
	/* make sure delete flags are set-up for peer structure */
 
	idrpPeer *peer = (idrpPeer *) tp->task_data;
	peer->idrp_flags |= IDRPF_DELETE;
 
}
 
 
/*
 *      Reinit a peer
 */
static void
idrp_peer_reinit(tp)
task *tp;
{
    idrpPeer *peer = (idrpPeer *) tp->task_data;
 
    if (!doing_idrp || (peer->idrp_flags & IDRPF_DELETE))
	{
	/* idrp turned off - so delete any peer reaching
	 * this re-init routine 
	 */

	peer->idrp_flags |= IDRPF_DELETE;
       	idrp_peer_terminate(tp);
	}
    else
	{
	/* idrp functioning 
	 * handle changes in peers due to re-config
	 * peer delete - means was in config/ now not in config
	 * peer UNCONFIGURED - means error in parsing routines
	 * peer TRY_CONNECT - normal state for peer
	 *		  should be new peer/peer with no connection 
	 *
	 * peer CONNECT - 
	 * 	problem with parsing - no configuration
	 *	delete peer
	 * 
	 * peer CONNECT & TRY_CONNECT  - if we are reparsing idrp
	 *		  then leave this peer alone.
	 *		  (skh- hopefully this is correct gated behavior)
	 * 
	 * peer CONNECT & WRITEFAILED 
	 *		- path_down on peer 
	 *		- reinit the peer
	 *
	 * peer CONNECT & IDLED 
	 *		- path down on peer
	 *		- leave the task and timers on the peer
	 *		skh--- if space becomes a problem (3/7/93)
	 *		skh--- we can delete this structure 
 	 * 
	 * peer WRITE_FAILED - socket problems 
	 *			try after gated re-init
	 * 
	 * peer IDLED - no connect configured peer off.
	 * 		leave timers intact and task intact
	 *		skh--- if space becomes a problem (3/7/93)
	 *		skh--- we can delete this structure 
	 */



	if (peer->idrp_flags & IDRPF_DELETE)
		{
		trace (TR_INT,LOG_ERR, "peer %s deleted from configuration file status = %x",peer->idrp_flags);
		peer->idrp_flags = IDRPF_DELETE; 	
		}

	switch (peer->idrp_flags) {
		case IDRPF_UNCONFIGURED:
			/* send error message */
			trace(TR_INT, LOG_ERR, "idrp_peer_reinit: re-init unconfigured neighbor %s", peer->name);
			break;

		case IDRPF_DELETE:
			/* send error message */
			trace(TR_INT, LOG_ERR, "idrp_peer_reinit: deleted neighbor %s", peer->name);
			idrp_peer_terminate(tp);
			break;

		case IDRPF_CONNECT:
			/* send error messages */

			trace(TR_INT, LOG_ERR, "idrp_peer_reinit:connected neighbor %s unconfigured so deleted",peer->name); 
			idrp_peer_terminate(tp);
			break;

		case IDRPF_TRY_CONNECT:
		case (IDRPF_CONNECT | IDRPF_WRITEFAILED):

			/* initialize unconnected */

			trace(TR_INT,LOG_ERR,"idrp_peer_reinit: peer %s reinit last flags %x",peer->name,peer->idrp_flags); 
			peer->idrp_flags = IDRPF_TRY_CONNECT;
			if (!peer->start_timer_running)
				{
				idrp_release_hash_tbl(peer);
				idrp_peer_ete_reinit(peer);
				idrp_sm(peer,IDRP_START_EVENT);
				}
			break;
 

		case (IDRPF_CONNECT | IDRPF_TRY_CONNECT):

			/* clear TRY and let connection run */

			peer->idrp_flags = IDRPF_CONNECT;
			idrp_reinit_ete_timers(peer);	
			trace(TR_INT,LOG_ERR,"idrp_peer_reinit: connected peer %s left connected",peer->name);
			break;

		case (IDRPF_CONNECT | IDRPF_IDLED):

			/* clear down this peer
			 * - leave task structure intact
			 */
			begin_close(peer);
			idrp_sm(peer,IDRP_STOP_EVENT);
			peer->idrp_flags = IDRPF_IDLED;	
			idrp_n_disabled++;
			trace (TR_INT,LOG_ERR,"idrp re-configuration idles peer %s",peer->name);
			break;
 
	
		case IDRPF_WRITEFAILED:
			/* not real sure about this - write failed --skh 3/7/93
			 */
  
			trace (TR_INT,LOG_ERR,"idrp reinit on write failed port peer %s",peer->name);
			idrp_release_hash_tbl(peer);
			idrp_peer_ete_reinit(peer);
			idrp_sm(peer,IDRP_START_EVENT);
			break;

		case IDRPF_IDLED:	
			/* send message */
	   	 	trace(TR_INT, LOG_ERR, "idrp_peer_reinit: config permanently idled neighbor %s", peer->name);
			idrp_n_disabled++;
			break;
		} /* end of switch */	
        } /* end of else */
}

/* idrp initialization task */ 
 
void
idrp_init()
{
idrpPeer	*peer;

    if (!doing_idrp)
	{ 
        /* IDRP is not running, but it may have been 
	 *  Delete all peers and associated tasks
 	 */ 

        IDRP_LIST(peer) {
		if (!peer->task)
			{
			task_delete(peer->task);
			peer->task=NULL;
			}
			/* here free up the adv_entry for the peer 
			 *  when we go to peer by peer policy 
			 */

        	idrp_delete(peer);
        } IDRP_LIST_END
 
      idrp_cleanup((task *) 0);
      return;
      } 

      	/* gated is "doing_idrp"  so start the task blues  
	 * here's the bgp receipe for initialization task 
	 *      - 1.) get rid of groups 
	 *        2.) free any policy  
	 *        3.) init the mib      
         *        4.) start up ability to receive on sockets 
	 * We are leaving groups, some policy, and the
	 * mib until phase2.  
	 *    4.) start up receive on sockets - 
	 * 	  This is done by opening up UDP/CLNP/IP/IDRP
	 *	  socket.  Or any combination of these. 
 	 * 
	 * skh -- 3/7/93 - still need to hack in CLNP/IP/IDRP sockets
	 */
 

	/* initialize things that could have changed on the
	 * local peer during a policy re-configuration
	 * - this may change during a re-init as well
	 */

	idrp_local_peer_init();


	/* initialize the local node peer  and the master task
	 * - the master task structure is the focus for all
	 *   received BISPDUs from any socket
	 * - the local peer structure holds the pointer to
	 *   the master task structures,
	 *   the minimum route advertisement for this RD timer
 	 * - is the peer structure for all local idrp routes
	 *
	 * - this initialization is done each time
	 *   the peer is activated.  Where it the above
	 *   re-configuration of local pathways may
	 *   only be done after a re-configuration bid.
	 *   I still need to find out how you disable
	 *   a peer without -usr2 to gated 
	 */

	idrp_init_this_node();


	/* now add in the routes we read in the parsing structure. 
	 * - call idrp_init_local_routes
	 * - this can be called during the re-initialization
	 *   after re-configuration or new policy 
	 */

	idrp_init_local_rt();

	/* Now create the tasks for each peer.
	 * - task have timers for peer connections
	 *   include mininum route Advertisement timer
	 *   for that peer
	 * - RD minimum route advertisement timer is
	 *   on an RD basis.
	 */  

        IDRP_LIST(peer) {

	    /* if task does not already exists */

            if (!peer->task) {
		task *tsk;

		/* initialize the configuration portion of
		 * the peer's structure
		 * - the can be configured after re-parse
		 */ 

		idrp_peer_config_init(peer);	


		/* initialize the ETE structure
		 * - this structure can be re-init
		 *   when you enable peer
		 */

		idrp_peer_ete_init(peer);	

		/* skh --- 3/7/93 do idrp task initialization here 
		 * the task is open for ip_intf
		 * if the proto_sock = clnp, this will have to change 
		 */ 

		tsk = peer->task = task_alloc("IDRP",TASKPRI_EXTPROTO);
		peer->iso_gw.gw_task = tsk;
		peer->gw.gw_task = tsk;
       		tsk->task_trace_flags = peer->trace_flags;
		tsk->task_addr = sockdup(peer->ip_intf_addr);
		sock2port(tsk->task_addr) = idrp_port;

		/* peer task routines */ 
		/* note - peer tasks take care of timers */
		/* all data reception is on master task */
 
                tsk->task_rtproto = RTPROTO_IDRP;
		tsk->task_dump = idrp_peer_dump;
                tsk->task_terminate = idrp_peer_terminate;
                tsk->task_cleanup = idrp_peer_cleanup;
                tsk->task_reinit = idrp_peer_reinit;
                tsk->task_data = (caddr_t) peer;

		/* timers */
                tsk->task_n_timers = IDRPTIMER_MAX;
                if (!task_create(tsk)) {
                    task_quit(EINVAL);
                }
		(void) timer_create(tsk,
				    IDRPTIMER_KEEPALIVE,
				    "Keepalive",
				    0,
				    (time_t) 0,
				    (time_t) 0,
				    idrp_event_keepalivetimer,
				    (caddr_t) 0);
		(void) timer_create(tsk,
				    IDRPTIMER_CLOSEWAIT,
				    "CloseWait",
				    0,
				    (time_t) 0,
				    (time_t) 0,
				    idrp_event_closetimer,
				    (caddr_t) 0);
		(void) timer_create(tsk,
				    IDRPTIMER_HOLDTIME,
				    "Holdtime",
				    0,
				    (time_t) 0,
				    (time_t) 0,
				    idrp_event_holdtimetimer,
				    (caddr_t) 0);
		(void) timer_create(tsk,
				    IDRPTIMER_REXMIT,
				    "Retransmit",
				    0,
				    (time_t) 0,
				    (time_t) 0,
				    idrp_event_rexmittimer,
				    (caddr_t) 0);

		if (peer->idrp_flags == IDRPF_IDLED)
			{
			/* do not start start timer  */
			peer->start_timer_running = FALSE;
			(void) timer_create(tsk,
				    IDRPTIMER_START,
				    "Start",
				    0,
				    (time_t) 0 ,
				    (time_t) 0 ,
				    idrp_event_starttimer,
				    (caddr_t) 0);
			}
		else
			{
			/* start the start connection timer */

			peer->start_timer_running = TRUE;

			(void) timer_create(tsk,
				    IDRPTIMER_START,
				    "Start",
				    0,
				    (time_t) IDRP_START_TIME,
				    (time_t) IDRP_START_TIME,
				    idrp_event_starttimer,
				    (caddr_t) 0);

			}
		(void) timer_create(tsk,
				    IDRPTIMER_OPENSENT,
				    "Open Sent",
				    0,
				    (time_t) 0,
				    (time_t) 0,
				    idrp_event_opensenttimer,
				    (caddr_t) 0);
		(void) timer_create(tsk,
				    IDRPTIMER_MIN_ADV,
				    "Mininum Route Advertisement ",
				    0,
				    (time_t) 0,
				    (time_t) 0,
				    idrp_process_minadv,
				    (caddr_t) 0);
#ifdef ECHOKLUDGE
 		/* %% */
		(void) timer_create(tsk,
				    IDRPTIMER_ECHO,
				    "Echo",
				    0,
				    (time_t) 0,
				    (time_t) 0,
				    idrp_event_echotimer,
				    (caddr_t) 0);
#endif
             	} 
	else
		{
		/* task already exists for this peer 
		 * - we are in the midst of reparsing
		 * here's where the proto socket has
	 	 * to change if peer->old_proto_sock
		 * is different than peer->proto_sock 
		 * initialized one should be zero or IP 
		 */

                peer->task->task_trace_flags = peer->trace_flags;
		}
        } IDRP_LIST_END;
}

void
idrp_init_this_node()
{
task		*idrp_master_task;
idrpPeer	*peer = &idrp_this_node;
sockaddr_un	*localaddr = NULL;


	/* initialize this node's peer structure and
	 * - check to make sure master task
	 *   has not been configured before
	 * 
	 * the master task associated with it
 	 * - all receive packets come to this
	 *   set of structures
	 * - all local routes associated with
	 *   idrp configuration or EXT_INFO
	 *   routes are associated with this 
	 *   set of structures
 	 * - minimum advertisement of routes
	 *   within RD timer associated with
	 *   this structure 
	 */


 
	/* Allocate the send/receive task. */

	if (idrp_reparsing)
		{
		/* check to see if the proto_sock has changed
		 * - re-init the task socket and
		 * close old socket
		 * - empty until I attack ISO
		 */
		return;
		}
 
        idrp_master_task = task_alloc("IDRP",TASKPRI_EXTPROTO);
	localaddr = idrp_find_interface(AF_INET);

	idrp_this_node.ip_neighbor = localaddr;
	idrp_this_node.ip_intf_addr = localaddr;

	/* set-up the master task with routines for: 
	 * idrp_recv - receive data from socket 
	 * idrp_terminate - terminating all idrp peer sessions 
         *   and clean-out tasks 
	 * idrp_cleanup - cleanup all peer sessions deleted, reset globals
 	 */

        idrp_master_task->task_trace_flags = idrp_trace_flags;
        idrp_master_task->task_rtproto = RTPROTO_IDRP;
	idrp_master_task->task_recv = idrp_recv;
        idrp_master_task->task_terminate = idrp_terminate;
        idrp_master_task->task_cleanup = idrp_cleanup;
        idrp_master_task->task_reinit = idrp_reinit;
	idrp_master_task->task_dump = idrp_master_dump;
	idrp_master_task->task_flash = idrp_flash; 
	idrp_master_task->task_newpolicy = idrp_newpolicy;
        idrp_master_task->task_data = (caddr_t) 0;

        idrp_master_task->task_n_timers =  IDRP_MASTER_TIMERS_MAX;	/* note peer structure identical 

/* parameters the following tasks queued with 
 *  idrp_recev - pdu off idrp IP port(now) or PID later or CLNP 
 *  idrp_terminate - set clear all idrp peers
 */ 

/* following things are not in master task vocabulary 
 *  not applicable because  IP and CLNP protocol
 * task_accept - process accepts 
 * task_connect - process connect complete
 * task_except - process exceptional 
 */

/* following are not included yet but applicable 
 *    * task_ifchange -
      * task_newpolicy
*/ 
	/* Find an interface to use. */


	/* parse doesn't have the interface - so add it in now */
	switch(idrp_this_node.proto_sock)
		{
		case IDRP_PDU_PROTO_IP_RAW:
			localaddr = idrp_find_interface(AF_INET);
			if (!localaddr) {
				trace(TR_INT, LOG_ERR, "idrp_init: Can't determine interface for neighbor %s", peer->name);
				task_quit(EDESTADDRREQ);
				}

			idrp_ip_pid_sock_init(idrp_master_task,localaddr);
			break;

		case IDRP_PDU_PROTO_UDP:
			localaddr = idrp_find_interface(AF_INET);
			if (!localaddr) {
				trace(TR_INT, LOG_ERR, "idrp_init: Can't determine interface for neighbor %s", peer->name);
				task_quit(EDESTADDRREQ);
				}
			idrp_udp_sock_init(idrp_master_task,localaddr);
			break;

		case IDRP_PDU_PROTO_CLNP:
			localaddr = idrp_find_interface(AF_ISO);
			if (!localaddr) {
				trace(TR_INT, LOG_ERR, "idrp_init: Can't determine interface for neighbor %s", peer->name);
				task_quit(EDESTADDRREQ);
				}
			idrp_clnp_raw_sock_init(idrp_master_task,localaddr);
			break;

		case IDRP_PDU_PROTO_IDRP:
			localaddr = idrp_find_interface(AF_ISO);
			if (!localaddr) {
				trace(TR_INT, LOG_ERR, "idrp_init: Can't determine interface for neighbor %s", peer->name);
				task_quit(EDESTADDRREQ);
				}
			idrp_idrp_sock_init(idrp_master_task,localaddr);
			break;
		}		


	/* ---skh should we add in a TASKOPTION_SENDBUF call here 
	 * --- egp has this call in it init sequence
	 * --- should we have the TASKOPTION non-blocking???
	 * --- again EGP has this
	 */
 
	task_alloc_recv(idrp_master_task, IDRP_PKTSIZE);
	task_alloc_send(idrp_master_task, IDRP_PKTSIZE);
	idrp_master_task->task_rtbit = rtbit_alloc(idrp_master_task,IDRP_TSI_SIZE,(caddr_t) &idrp_this_node,idrp_peer_rtbit_dump);

	/* put the local routes as part of the master task 
	 * it's since there are no timers to release these routes
  	 */  

	idrp_this_node.task = idrp_master_task;
	idrp_this_node.gw.gw_task = idrp_master_task;
	idrp_this_node.iso_gw.gw_task = idrp_master_task;

	/* set-up the timer for the minimum Route Advertisement timer
	 * for this routing domain 
	 * - Even though this could be global in the current concept
	 *   of routers, it is likely under the QOS or  Internet  Route
	 *   Server model this timer will need to be set as the
	 *   master task per group of idrp speakers . 
	 *    
	 */  

	(void) timer_create(idrp_master_task,
			    IDRP_MASTER_TIMER_MIN_ADVRD,
			    "Minimum Route Adver this RD",
			    0,
			    (time_t) 0,
			    (time_t) 0,
			    idrp_process_minadvRD, /* jgs */
			    (caddr_t) 0);
}

void
idrp_init_local_rt()
{
int i;
idrp_attribute_record *p_att;
idrpRoute	*p_idrp_rt;
idrpRoute	*p_last_irt;
idrpRoute	*p_irt;
idrpRoute_entry	*p_ire;
int		changes = 0 ;


	/* first clean up any local routes left over
	 * from the parsing routines -
	 */

	rt_open(idrp_this_node.task);
	changes = idrp_local_route_clean();  
	 
 
	/* By now, the gated routing structure should be able
	 * handle route additions
	 * walk the attribute list looking for new local routes until you get
	 * to the end of the local attributes 
 	 */


	p_att = idrp_attr_list;
	while (p_att)
		{
		for (p_ire = p_att->route_id_list; p_ire; p_ire = p_ire->p_next)
		    {
		     NLRI_FAMILY_LOOP(i) 
			{
			p_idrp_rt = p_ire->p_route[i];
			p_last_irt = (idrpRoute *) 0;
			for (p_irt = p_idrp_rt; p_irt != (idrpRoute *) 0; p_irt = p_irt->p_next_nlri)
				{
				if (p_irt->status & IDRP_STATUS_LOCAL_NEW_ROUTE)
					{
					if (!idrp_add_local_rt_gated(p_irt))
						{
					 	/* route could not be added to gated 
					 	 * flag it and  pull it out of the list 
						 */
			
						trace (TR_IDRP,0,"local route %s could not be added to gated table, deleted ",
						iso_ptoa(&p_idrp_rt->nlri));
				

						/* relink the attribute record to get around this route */
	 
						if (p_last_irt)
							{
							/* a previous route exists */
							p_last_irt->p_next_nlri = p_idrp_rt->p_next_nlri;
							}
						else
							{
							/* first one on list */	
							p_att->route_id_list->p_route[i] = p_idrp_rt->p_next_nlri;
							}
						}
					else
						{
						/* good route added to gated table */
	
						changes++; 
						p_last_irt = p_idrp_rt;
						}	
					p_irt->status = IDRP_STATUS_LOCAL_ROUTE;
					} /* end if new local route */ 
				} /* end of for loop */
			    } NLRI_FAMILY_LOOP_END;
			}  /* end of loop for route_id */

		if (p_att == idrp_last_local_att)
			{
			p_att = (idrp_attribute_record *) 0;
			}
		else
			{
			p_att = p_att->next;
			}
		}

	/* close the task and gateway for IP, but ISO will have
	 * happened as well.  Since one open aught to be OK,
	 * we will let this kludge continue
	 * skh--- /3/17/93  
	 */
	rt_close(idrp_this_node.task,&idrp_this_node.gw,changes,NULL);
}


void
idrp_peer_ete_init(peer)
idrpPeer *peer;
{
	/* set-up all the ete initialization for a peer
	 * - set close state
	 * - reset all timers to global values
	 * - clear bad AdjRib flags
	 * - reset xmit and rexmit queues
	 * - init the hold table
	 * - init the route_out table
	 */
 
	peer->send_next = 1;
	peer->send_unack = 0;
	idrp_peer_ete_reinit(peer);

}

void
idrp_peer_ete_reinit(peer)
idrpPeer	*peer;
{
        peer->state = IDRP_CLOSED;
	peer->hold_time = IDRP_HOLD_TIME;
	peer->rexmit_time = IDRP_REXMIT_TIME;
	peer->rcv_credit = IDRP_MAX_CREDIT;
	peer->pdu_maxsendsize = IDRP_PKTSIZE;
	peer->checksum = 0;
	peer->AdjRibBad = 0;
	peer->xmit_queue_head = NULL;
	peer->xmit_queue_tail = &peer->xmit_queue_head;
	peer->rexmit_queue_head = NULL;
	peer->rexmit_queue_tail = &peer->rexmit_queue_head;

	/* initialize the hash table again */
	idrp_init_hash_tbl(peer);
}
 
/*
 * Pretty-print an ISO prefix.
 * %% Needs to be able to handle non-semi-octet-aligned prefixes!
 */

#define ISO_PTOA_LENGTH 61
#define hexchar(x) ((x)>9?(x)-10+'A':(x)+'0')

char *iso_ptoa(prefix)
struct iso_prefix *prefix;
{
static char result[ISO_PTOA_LENGTH];
int resultix = 0;
int addrix;

	for (addrix = 0; addrix < PFX_BYTE_LENGTH(*prefix); addrix++) {
		if (addrix & 1) {
			if (resultix > ISO_PTOA_LENGTH - 4)
				break;
			result[resultix++] = '.';
		}
		if (resultix > ISO_PTOA_LENGTH - 3)
			break;
		result[resultix++] = hexchar((prefix->pfx[addrix] & 0xf0) >> 4);
		if (addrix == prefix->pfx_len/8 && ((prefix->pfx_len & 7) <= 4))
			break;
		result[resultix++] = hexchar(prefix->pfx[addrix] & 0x0f);
	}
	result[resultix++] = '\0';
	return(result);
}
 
#define hexchar(x) ((x)>9?(x)-10+'A':(x)+'0')

int isoaddrcmp(addr1, addr2)
struct iso_net_addr *addr1, *addr2;
{
	int i;
	if (addr1->isoa_len != addr2->isoa_len)
		return(1);

	return(bcmp((caddr_t) addr1->isoa_genaddr, (caddr_t) addr2->isoa_genaddr, addr1->isoa_len));
}

int
isoaddr_netiso_cmp(addr1, addr2)
struct iso_addr *addr1;
struct iso_net_addr *addr2;
{
	int i;
	if (addr1->isoa_len != addr2->isoa_len)
		return(1);

	return(bcmp((caddr_t) addr1->isoa_genaddr, (caddr_t) addr2->isoa_genaddr, addr1->isoa_len));
}

int
compare_iso_addr(addr1,addr2)
struct iso_net_addr *addr1, *addr2;
{
int len,i;

	/* min(addr1.isoa_len,addr2.isoa_len) */

	len = (addr1->isoa_len > addr2->isoa_len) ? addr2->isoa_len : addr1->isoa_len;

	for (i = 0; i < len; i++)
		{
		if (addr1->isoa_genaddr[i] != addr2->isoa_genaddr[i])
			{
			if (addr1->isoa_genaddr[i] < addr2->isoa_genaddr[i])
				{	
				return (ISO_ADDR_LESS_THAN);
				}
			else
				{
				return (ISO_ADDR_GREATER_THAN);
				}
			}
		}

	if (addr1->isoa_len == addr2->isoa_len)
		{
		return (ISO_ADDR_EQUAL);
		}
	if (addr1->isoa_len < addr2->isoa_len)
		{
		return (ISO_ADDR_LESS_THAN);
		}

	/* addr1 longer than address 2  - so greater than */

	return (ISO_ADDR_GREATER_THAN);
}


isopfxcompare(pfx1, pfx2)
struct iso_prefix *pfx1, *pfx2;
{
	int result;
	int i;
	for (i = 0; i < MIN(PFX_BYTE_LENGTH(*pfx1), PFX_BYTE_LENGTH(*pfx2)); i++) {
		if (pfx1->pfx[i] < pfx2->pfx[i])
			return(-1);
		if (pfx1->pfx[i] > pfx2->pfx[i])
			return(1);
	}
	if (pfx1->pfx_len == pfx2->pfx_len)
		return(0);

	if (pfx1->pfx_len < pfx2->pfx_len)
		return(-1);
	else
		return(1);
}


/*
 * idrp_reinit - this is called after the configuration file has been
 * reread, and after idrp_init() has been called to start the master
 * task has been set-up to deal with the reception of  BISPDU packets
 *  from gated.
 * 
 * skh---- I cannot find anything specific to do for the master
 * skh--- task in task reinit. I'm sure I'll find something later.
 * skh 3/7/93 - once receive tasks for multiple sockets (ip /clnp) 
 * 
 * Peer changes are handled by the idrp_peer_reinit routine.
 * 
*/

/* note that for now - we have no concept of the grouping of 
 * idrp peers.  We will need to add this to keep up with BGP Q JONES 
 * the pride of the internet family 
 */
 
static void
idrp_reinit(tp)
task *tp;
{
	trace (TR_IDRP,0,"idrp_reinit called, but null processing "); 
	return;
} 

 
void
idrp_newpolicy(tp,rtl)
task	*tp;
rt_list	*rtl;
{
idrp_ann_list	ann_list;
idrp_ann_list	*p_ann_list = &ann_list;
	/* well I'm not sure why the new policy get all
	 * the routes, but here it is after the initialization
	 * try calling phase3 processing to see what goes
	 */

	bzero(p_ann_list,sizeof(idrp_ann_list));
	p_ann_list = phase3_newpolicy(rtl,p_ann_list);

	if (idrp_n_established)
		{
		idrp_refresh_all_peers(p_ann_list);
		}
	free_ann_list(p_ann_list);
	return;
} 

void
idrp_reinit_ete_timers(peer)
idrpPeer	*peer;
{

	/* reset the tranport layer timers
	 * hopefully, this will kill off my cease
	 */ 
 
	timer_reset(TASK_TIMER(peer->task,IDRPTIMER_KEEPALIVE));
	timer_reset(TASK_TIMER(peer->task,IDRPTIMER_HOLDTIME));
	timer_set(TASK_TIMER(peer->task, IDRPTIMER_HOLDTIME), peer->hold_time, (time_t) 0);
	kill_rexmit_timer(peer);
	

}
