/* idrp_init_sock.c
 *
 *  routines to initialize the
 *  4 types of socket used by IDRP
 * 
 * - hopefully most of this code will
 *   go to the gated internals
 * idrp_ip_pid_sock_init
 * idrp_upd_sock_init
 * idrp_clnp_raw_sock_init
 * idrp_idrp_sock_init
 * idrp_clnp_packet_send
 * idrp_idrp_packet_send
 *
 */

#include "include.h"
#include "idrp.h"
#include "inet.h"
#ifdef ADN_PATCH
#	define IPPROTO_IDRP	128
#else
#include <netiso/tp_user.h>
#endif /* ADN PATCH */



/* 
 * Socket initialization routine for IP PIDs 
 */

void
idrp_ip_pid_sock_init(p_task,p_addr)
task		*p_task;
sockaddr_un	*p_addr;
{
	p_task->task_proto = IPPROTO_IDRP;
	p_addr->in.sin_port = 0;
	p_task->task_addr = sockdup(p_addr);
     	if ((p_task->task_socket = task_get_socket(AF_INET, SOCK_RAW, IPPROTO_IDRP)) < 0)
		{
		task_quit(errno);
		}
	task_set_socket(p_task,p_task->task_socket);
	if (!task_create(p_task))
                {
                task_quit(EINVAL);
		}

	/* set receive buffer size */

       if (task_set_option(p_task, TASKOPTION_RECVBUF, (caddr_t) (10*1500)) < 0) 
		{
		task_quit(errno);
		}
	/* set transmit buffer size for kernel *?

       if (task_set_option(p_task, TASKOPTION_SENDBUF, (caddr_t) (10*1500)) < 0) 
		{
		task_quit(errno);
		}

	/* set non-block IO */

	if (task_set_option(p_task,TASKOPTION_NONBLOCKING,(caddr_t) TRUE) < 0)
		{
		task_quit (errno);
		}


	/* bind the task to the PID socket */
 
	/* 	if (task_addr_local(p_task,p_addr))
	 *	{
	 *	task_quit(errno);
 	 *	}
	 */

	trace (TR_IDRP,0,"idrp master task now supports idrp pid on the ip raw socket"); 	

}

void idrp_udp_sock_init(p_task,p_addr)
task	*p_task;
sockaddr_un	*p_addr;
{ 
	/* here's the IDRP set-up for a single TCP port 
         * this must change to a protocol id
	 */

        p_task->task_addr = sockdup(inet_addr_any);
	sock2port(p_task->task_addr) = idrp_port;
	sock2port(p_addr) = idrp_port;

	if ((p_task->task_socket = task_get_socket(AF_INET, SOCK_DGRAM, 0)) < 0)
		{
		task_quit(errno);
		}

	if (!task_create(p_task))
                {
                task_quit(EINVAL);
		}

       if (task_set_option(p_task, TASKOPTION_RECVBUF, (caddr_t) (10*1500)) < 0) 
		{
		task_quit(errno);
		}

	/* bind the task to the  UDP socket */
 
	if (task_addr_local(p_task,p_addr))
		{
		task_quit(errno);
		}

	trace (TR_IDRP,0,"idrp master task now supports the udp socket"); 	
}

void idrp_clnp_raw_sock_init(p_task,p_addr)
task		*p_task;
sockaddr_un	*p_addr;
{
int	hold;

	/* need to complete this socket */

	/* ISOPROTO_CLNP - set in documents
	 * Marc says only 0 works for raw
	 * - use the following sequence for the
	 *   raw clnp packets to wind their way
	 *   through the pathway 
	 * task_get_sock -> which does a socket call
	 * direct call to setsockopts to replicate mark's call for CLNPPING
	 * 
	 *  
	 */
	
	/* need allocate with the iso_gw_addr for task 
	 *  - try for now.  If problems.  We must use
	 * older version of  sockaddr_iso
	 */

	p_task->task_addr = sockdup((sockaddr_un *) idrp_this_node.iso_gw.gw_addr);
#ifdef ADN_PATCH  
   	if ((p_task->task_socket = task_get_socket(AF_ISO, SOCK_DGRAM,0)) < 0)
#else
       if ((p_task->task_socket = task_get_socket(AF_ISO, SOCK_DGRAM,ISOPROTO_IDRP)) < 0)
#endif /* ADN_PATCH */
		{
		task_quit(errno);
		}

	task_set_socket(p_task,p_task->task_socket);

	/* Mark calls setsockopt to set the receive buffer to some
	 * reasonable number
	 * -0h.. dear I guess I'll try to get a buffer of
	 *  
	 * gated call I think would work
	 */

	if (task_set_option(p_task, TASKOPTION_RECVBUF, (caddr_t) (10*1500)) < 0) 
		{
	   		task_quit(errno);
	 	}
	

	/*  hold = 10*1500;	
	 * setsockopt(p_task->task_socket,SOL_SOCKET,SO_RCVBUF,(char *) &hold,sizeof(hold));
	 */
	 trace (TR_IDRP,0,"idrp master task now supports the raw clnp iso socket"); 	

	if (task_set_option(p_task,TASKOPTION_SENDBUF,(caddr_t) IDRP_PKTSIZE) < 0)
		{
		task_quit(errno);
		}

	/* I think we want to try non-blocking on the socket
	 * - so let's try it here.
	 * - the egp protocol tries it for a pid
	 * - I will try that later
	 */

	if (task_set_option(p_task,TASKOPTION_NONBLOCKING,(caddr_t) TRUE) < 0)
		{
		task_quit(errno);
		}

}

	

void idrp_idrp_sock_init(p_task,p_addr)
task		*p_task;
sockaddr_un	*p_addr;
{
	trace (TR_IDRP,0,"idrp master task now supports the idrp socket"); 	
}
