#include "include.h"
#include "idrp.h"
#include <stdio.h>


/* idrp_rib_refresh.c
 *
 * rib refresh routines:
 *	rib_refresh_pdu_release(peer) - release all the Rib Refresh updates
 *	refresh_pdu_link(peer,results,pdu)
 *	rib_refresh_update_process(peer) - parse all the updates by walking 
 *		the Rib Refresh chain
 */ 
 

void rib_refresh_pdu_release(peer) 
idrpPeer *peer;
{
/* release all the Rib Refresh updates */

idrpRefresh	*p_ref;
idrpRefresh	*p_ref2;
	/* walk the chain and release the refresh structures */

	p_ref = peer->refresh.pdu_list_start;
	while (p_ref) 
		{ 
		task_mem_free(peer->task,p_ref->results->withdraw);
		task_mem_free(peer->task,p_ref->results);
		p_ref2 = p_ref;
		p_ref= p_ref->p_next;
		task_mem_free(peer->task,p_ref2);
		}
}

void refresh_pdu_link(peer,results,pdu)
idrpPeer *peer;
parse_results *results;
idrpPdu *pdu;
{
idrpRefresh	*p_refresh;	

	p_refresh = (idrpRefresh *) task_mem_malloc(peer->task,sizeof(idrpRefresh));
	bzero(p_refresh,sizeof(idrpRefresh));
	p_refresh->results = results;

	/* if this is not the first pdu on the list */

	if (peer->refresh.pdu_list_start)
		{
		peer->refresh.pdu_list_end->p_next = p_refresh;
		peer->refresh.pdu_list_end = p_refresh;
		peer->refresh.cnt++;
		} 
	else
		{
		peer->refresh.pdu_list_end = peer->refresh.pdu_list_start = p_refresh;
		peer->refresh.cnt = 1;
		}
}


	
void rib_refresh_update_process(peer)
idrpPeer *peer;
{
idrpRefresh *p_ref,*p_ref2; 
idrpRoute   *p_rt;
rt_list		*rtl;
idrp_ann_list	*p_w_list,*p_ann_list;
idrp_ann_list	*p_rib_ann_list;
int		i;
int		changes = 0;	

/* parse all the updates by walking the Rib Refresh */
/* chain */

	if (peer->refresh.pdu_list_start == (idrpRefresh *) 0) 
		{
		/* nothing in the link list - just return */
		trace(TR_IDRP,0,"Rib Refresh stop with no update pdus for peer %s  ",
			peer->name);
		return;
		}

	/* go through full list and update the gated routing table 
	 *  - first release all routes associated with this
	 *  - by pulling a list of routes that need to be
	 *    deleted
	 */



	p_w_list = (idrp_ann_list *)  task_mem_malloc(peer->task,sizeof(idrp_ann_list));
	p_ann_list = (idrp_ann_list *)  task_mem_malloc(peer->task,sizeof(idrp_ann_list));

	bzero(p_w_list,sizeof(idrp_ann_list));	
	bzero(p_ann_list,sizeof(idrp_ann_list));

	rt_open(peer->task);

	/* one list per gateway - which is per protocol */

	rtl = rtlist_gw(&peer->gw);
	changes = idrp_peer_route_pull(rtl,p_w_list);
	rtl = rtlist_gw(&peer->iso_gw);
	changes += idrp_peer_route_pull(rtl,p_w_list);
	trace(TR_IDRP,0,"%d delete of routes for peer %s rib refresh delete of routes",changes,peer->name);


	for (p_ref = peer->refresh.pdu_list_start; peer->refresh.cnt; peer->refresh.cnt-- )
		{
		p_rib_ann_list = idrp_process_pdu_routes(peer,p_ref->results,IDRP_RIB_REFRESH_TYPE);
	
		/* add current announce list to the old announce and withdrawl list
		 *
		 */

		link_ann_list_ann_list(p_ann_list,peer,p_rib_ann_list);
		task_mem_free(peer->task,p_rib_ann_list);

		/* clean up the results array */

		task_mem_free(peer->task,p_ref->results->withdraw);
		task_mem_free(peer->task,p_ref->results);
		p_ref2 = p_ref;
		p_ref = p_ref->p_next;
		task_mem_free(peer->task,p_ref2);

		} /* loop for the rib refreshes */

	rib_refresh_remove_dup(peer,p_w_list,p_ann_list);	
	free_ann_list(p_w_list);

	if (peer->type == IDRP_PEER_EXTERNAL)
		{
		changes = phase1_external(p_ann_list);
		}
	else
		{
		changes = phase1_internal(p_ann_list);	
		}

	trace (TR_IDRP,0,"rib refresh for peer %s had %d changes to route table",peer->name,changes);

	free_ann_list(p_ann_list);
	rt_close(peer->task,&peer->gw,changes,NULL);
	return;
}

/* walk through the remove routes and the new routes,
 * comparing the nlris and weeding out 
 * all withdraws that are not permanent for 
 * final announce list to hand to phase1 processing
 */

void 
rib_refresh_remove_dup(peer,p_w_list,p_ann_list)
idrpPeer	*peer;
idrp_ann_list	*p_w_list;	/* withdrawls list */
idrp_ann_list	*p_ann_list;	/* rib refresh list */
{
idrp_ann_list	*p_atl;
idrpRoute	*p_irt;
int		i;


	/* walk the withdrawl list and try to weed out the
	 * announces
	 */

	ANN_LIST(p_atl,p_w_list)
		{
		NLRI_FAMILY_LOOP(i)
			{
			ANN_WITH_LIST_WALK(p_irt,i,p_atl)
				{
				if (find_nlri_in_ann_nlri(p_irt,p_ann_list))
				  {
				    /* set phase1 withdraw flag 
				     */
				    p_irt->status |= IDRP_STATUS_REPLACE;
				  }
				else 
				  { 
				    /* if not found in announce nlri, linked to withdraw list
				     */
				    
				    link_ann_list(p_irt,p_ann_list,NLRI_LINKED_P_WITH);
				  }
				} ANN_WITH_LIST_WALK_END;
			} NLRI_FAMILY_LOOP_END;
		} ANN_LIST_END;
}	
