#include "include.h"
#include "iso.h"
#include "inet.h"
#include "idrp.h"

/*  idrp_rt_gated.c
 *    
 * %% gated interaction routines
 * 
 * - mediate_pref_match - modify the gated preference match as phase2 desires
 * - idrp_add_rt_to_gated - add a route to gated tables
 * - idrp_repl_rt_gated - replace an idrpRoute structure in gated table
 *                      - call the rt_change routine to handle changes if needed
 * - idrp_mod_rt_gated - modify gated routing table
 * - find_idrp_gated_rt - find from an external peer the idrp
 * 			  route 
*/ 

/* %% gated related routines  */


/* mediate_pref_match - modify gated preference so that  
 * 		    gated can pick a route within idrp routes
 *		    and external route.  idrp routes 
 *		    have constant + idrp_pref *2.  add or
 *		    zero route to select a single idrp route 
 */ 


int
mediate_pref_match(p_pref,p_irt)
idrp_rt_chain_walk	*p_pref;
idrpRoute		*p_irt;
{
idrp_rt_chain_walk	*p_ch;
idrpRoute		*p_best_rt;
idrpRoute		*p_irt_new;
int			idrp_best_rt = 0;
int			multi_idrp_best_rt = 0;	
u_int			proto;
u_int			tie_result;		

	/* mediate gated preference matches
	 * the idrp route past matches the list of gated routes
	 * - Dispatch on protocol 
	 *   if IDRP vs IDRP route -
	 *		use method from standard
	 *		of comparing routes
 	 *   if IDRP vs another protocol
 	 *	call idrp_proto_pref routine
	 *	to determine if this protocols
	 *	should be used over idrp
	 *	if true, return FALSE so idrp route's preference
	 *	 is decremented
	 *	if false, continue to look for other idrp routes
	 *	to get best idrp route 
	 * 
	 * 
	 *   prefer an interior protocol over BGP or IDRP
	 *
	 */

	p_best_rt = p_irt; 
		
	IDRP_WALK_RT_CHAIN(p_pref,p_ch)
		{ 
		if (p_ch->p_rt->rt_gwp->gw_proto == RTPROTO_IDRP)
			{
			/* count how many idrp routes preference match */
			
			multi_idrp_best_rt++;

			/* idrp tie breaking - 
			 *  use idrp tie breaking methods specified in
			 *  ISO or idrp for IP specification 
			 */
		
			trace (TR_IDRP,0,"IDRP tie breaking multiple preference matches route from peer %s route %s",
				p_irt->peer,iso_ptoa(&p_irt->nlri));
 
			/*  The second argument is of type rt_data* as linked by gated tables.
			 *  However, gated expects you to re-cast this thing to your structure. 
			 *  This is recast above all comparisons so it will not need to
			 *  be recast in the code multiple times. 
			 *  (Thanks John for the suggestion!!)
			 */

			p_irt_new = (idrpRoute *) p_ch->p_rt->rt_data; 
			switch	(tie_break(p_best_rt,p_irt_new)) 
				{
				case PFX_ADDR_EQUAL:
					/* equal routes warn and use 2nd route idrp route */
					p_best_rt = p_irt_new;
					break;

				case PFX_ADDR_LESS_THAN:
					/* 1st route better than 2nd */
					break;
			
				case PFX_ADDR_GREATER_THAN:
					/* 2nd route better than 1st */
					p_best_rt = p_irt_new; 
					break; 
					}
			}
		proto = (u_int) p_ch->p_rt->rt_gwp->gw_proto;
		if (!idrp_proto_pref(proto,p_irt->peer))
			{
			/* other protocol prefered to idrp - return false */

			return (FALSE);
			}
	
		} IDRP_WALK_RT_CHAIN_END

	trace (TR_IDRP,0," peer %s has route for %s with %d idrp preference matches best pref from peer = %s", 
			p_irt->peer->name,iso_ptoa(&p_irt->nlri),multi_idrp_best_rt,p_best_rt->peer->name);

	if (p_best_rt == p_irt)
		{
		return (TRUE);
		}	
	return (FALSE);
}


int
idrp_add_rt_to_gated(p_idrp_rt,p_dest)
idrpRoute *p_idrp_rt;
sockaddr_un	*p_dest;				/* destination address for gated Route search */
{
#ifdef ADN_PATCH
rt_parms	rt_parms_adn;			/* rt parms for gated add */
#else
rt_parms	rt_parms;			/* rt parms for gated add */
#endif
sockaddr_un	*p_gw;				/* pointer to modified gateway for route */
						/* takes off NSEL off router address */

	/* brand new  route 
	 * add example for gated
	 */ 

#ifdef ADN_PATCH
	bzero ((caddr_t) &rt_parms_adn,sizeof(rt_parms));  /* zero params */
#else
	bzero ((caddr_t) &rt_parms,sizeof(rt_parms));  /* zero params */
#endif /* ADN_PATCH */
	p_gw = (sockaddr_un *) task_mem_malloc(p_idrp_rt->peer->task,ISO_SOCK_SIZE);

	/* set preference in new route to IDRP generic preference
	 * + 2 times local preference + no tie break (0)  
	 * - tie break allows idrp to mediate ties 
	 *   per idrp specification 
	 */ 

	

/* here we need to fix things to include - rt_data as the beginning of idrpRoute 
 * we need to add in a dump routine and a free routine if possible
 */ 
 
#ifdef ADN_PATCH 
 	rt_parms_adn.rtp_rtd = (rt_data *) p_idrp_rt; 
	rt_parms_adn.rtp_rtd->rtd_length = sizeof(idrpRoute);
	rt_parms_adn.rtp_rtd->rtd_data = (void_t) &p_idrp_rt->p_rt;
	rt_parms_adn.rtp_preference = idrp_to_gated_pref(p_idrp_rt);
	rt_parms_adn.rtp_dest = p_dest;
	
	rt_parms_adn.rtp_n_gw = 1;
#else
        rt_parms.rtp_rtd = (rt_data *) p_idrp_rt;
        rt_parms.rtp_rtd->rtd_length = sizeof(idrpRoute);
        rt_parms.rtp_rtd->rtd_data = (void_t) &p_idrp_rt->p_rt;
        rt_parms.rtp_preference = idrp_to_gated_pref(p_idrp_rt);
        rt_parms.rtp_dest = p_dest;
 
        rt_parms.rtp_n_gw = 1;
#endif /* ADN_PATCH */

/* set the bit for the idrp route added to gated */

	switch (p_idrp_rt->family)
		{
#ifndef ADN_PATCH
		case AF_ISO:
			rt_parms.rtp_dest_mask = iso_mask_prefix(p_idrp_rt->nlri.pfx_len);
			if (p_idrp_rt->p_loc_rt_peer)
				{
				rt_parms.rtp_routers[0] = (sockaddr_un *) p_idrp_rt->p_loc_rt_peer;
				}	
			else
				{
				/* sockaddr_un for remote peer must
				 * terminate the NSEL - a debate on what an
				 * NET really is.  
				 * Re-examine after EMAIL to cisco, IBM,
				 * proteon and 3COM 
				 * copy in the full NET for now
				 */
		 
				/* take off NSEL position */
				p_gw->iso.siso_len =  p_idrp_rt->peer->neighbor.isoa_len - 1;

				bcopy(&p_idrp_rt->peer->neighbor.isoa_genaddr,p_gw->iso.siso_addr,p_gw->iso.siso_len);
				p_gw->iso.siso_family = AF_ISO; 
				/* add two for the iso structure for sockaddr_un */ 
				p_gw->iso.siso_len +=2;
				rt_parms.rtp_routers[0] = (sockaddr_un *) p_gw;
				}
 
			rt_parms.rtp_gwp = &p_idrp_rt->peer->iso_gw;
			break;	
#endif /* ADN_PATCH */
		case AF_INET:
#ifdef ADN_PATCH
       rt_parms_adn.rtp_dest_mask = inet_mask_prefix(p_idrp_rt->nlri.pfx_len);
                        if (p_idrp_rt->p_loc_rt_peer)
                                {
                                /* local peer - so use address within system */
                                rt_parms_adn.rtp_routers[0] = (sockaddr_un *) p_idrp_rt->p_loc_rt_peer;
                                }
                        else
                                {
                                /* if not local address - try the IP neighbor address */
                                rt_parms_adn.rtp_routers[0] = p_idrp_rt->peer->ip_neighbor;
                                }
                        rt_parms_adn.rtp_gwp = &p_idrp_rt->peer->gw;
                        break;
                }
 
        /* set protocol metric to calculated metric for idrp
         * local routes
         */
 
        rt_parms_adn.rtp_metric = p_idrp_rt->pref_cal;
 
        /* set the no-age flag so that the local
         * routes do not time out.
         */
 
        rt_parms_adn.rtp_state = RTS_EXTERIOR | RTS_NOAGE;
        if (p_idrp_rt->p_rt_ext_info)
                {
                /* this is a idrp route created on the
                 * basis of another gated route
                 * Do NOT LET this become the
                 * kernel route   - set no install   
		 * I think we want advised - but we need
                 * to watch this
                 */
 
                rt_parms_adn.rtp_state |= RTS_NOTINSTALL;
                }
	p_idrp_rt->p_rt = rt_add(&rt_parms_adn);
#else
			rt_parms.rtp_dest_mask = inet_mask_prefix(p_idrp_rt->nlri.pfx_len);
			if (p_idrp_rt->p_loc_rt_peer)
				{
				/* local peer - so use address within system */
				rt_parms.rtp_routers[0] = (sockaddr_un *) p_idrp_rt->p_loc_rt_peer;
				}
			else
				{	
				/* if not local address - try the IP neighbor address */
				rt_parms.rtp_routers[0] = p_idrp_rt->peer->ip_neighbor;
				}
			rt_parms.rtp_gwp = &p_idrp_rt->peer->gw;
			break;
		}

	/* set protocol metric to calculated metric for idrp
	 * local routes
	 */

	rt_parms.rtp_metric = p_idrp_rt->pref_cal;


	/* set the no-age flag so that the local
	 * routes do not time out.
	 */

	rt_parms.rtp_state = RTS_EXTERIOR | RTS_NOAGE;  
	if (p_idrp_rt->p_rt_ext_info)
		{
		/* this is a idrp route created on the
	 	 * basis of another gated route
		 * Do NOT LET this become the
		 * kernel route   - set no install
		 * I think we want advised - but we need
		 * to watch this
		 */

		rt_parms.rtp_state |= RTS_NOTINSTALL;
		}	
		
	p_idrp_rt->p_rt = rt_add(&rt_parms);
#endif /* ADN_PATCH */


	if (p_idrp_rt->p_rt)
		{
		/* set the bit to indicate we are announcing this 
		 * - we will set this bit and see what
	 	 *   happens - at least we won't get the
		 *   null delete any more.
		 * - but then again, we may break the gated
		 *   state machine.   Sigh... life is
		 *   so much fun when the beat of gated
		 *   is a fleeting whisper in the wind, wafting
		 *   past our ears. 
		 *  
		 */

		return(TRUE);
		}
	else
		{
		return(FALSE);
		}
}

/*
 *  idrp_repl_rt_gated:
 *  
 * Replace the idrp route in the gated structure
 * destination has been replaced. 
 * 
 *
 */
idrpRoute *
idrp_repl_rt_gated(p_idrp_rt,p_rt)
idrpRoute	*p_idrp_rt;
rt_entry	*p_rt;
{
idrpRoute	*p_idrp_rt_old;

	/* The idrp route structure is pointed to by rt_data in the
	 * gated rt_entry
	 * We need to pull out the rt_data from the old idrp route
	 * and put in the new idrp route.  I think it is
	 * as simple as that.
	 *
	 * The steps in the process are as follows:
	 * 1.) grab a copy of the pointer to idrpRoute structure from the route
	 * 2.) copy over the header from the old idrp route
	 * 3.) set the new idrpRoute structure at rt_entry->rt_data 
 	 * 4.) return the copy of the old idrp structure  
	 */

	/* grab pointer */

	p_idrp_rt_old = (idrpRoute *) p_rt->rt_data; 
	p_idrp_rt->rt_header.rtd_refcount = p_idrp_rt_old->rt_header.rtd_refcount;
	p_idrp_rt->rt_header.rtd_length = p_idrp_rt_old->rt_header.rtd_length;
	
	/* copy the old rt_entry pointer */
 
	p_idrp_rt->p_rt = p_rt;

	/* set this data pointer to point to idrp route */

	p_rt->rt_data = (rt_data *) p_idrp_rt; 
	return (p_idrp_rt_old);
}


int 
idrp_mod_rt_gated(p_idrp_rt,pref)
idrpRoute *p_idrp_rt;
int		pref;
{
metric_t	metric;
metric_t	tag = 0;
sockaddr_un	*gw;

	metric = (metric_t ) p_idrp_rt->pref_cal;

	/* tag set to zero until we figure out from dennis what it is */
	
	if (p_idrp_rt->p_attr->next_hop)
		{
		gw = p_idrp_rt->p_attr->next_hop;
		}
	else
		{
		switch(p_idrp_rt->family)
			{
			case AF_ISO:
				gw = p_idrp_rt->peer->iso_gw.gw_addr;
				break;

			case AF_INET:
				gw = p_idrp_rt->peer->gw.gw_addr;
				break;
			}
		}

	if (!rt_change(p_idrp_rt->p_rt,metric,tag,pref,1,&gw))
		{
		trace (TR_IDRP | TR_UPDATE, 0, " peer %s route %s tried to change gated table but could  not", 
			p_idrp_rt->peer->name,iso_ptoa(&p_idrp_rt->nlri)); 
		return (FALSE);
		}
	return (TRUE);
}

/*
 * idrp_del_rt_gated(idrpRoute *p_idrp_rt)
 *
 * Delete the gated route (rt_entry) associated with the idrpRoute pointed to by
 * p_idrp_rt.
 */
idrp_del_rt_gated(p_idrp_rt)
idrpRoute *p_idrp_rt;
{
rt_entry	*p_rt;

	p_rt = p_idrp_rt->p_rt;

	/* once we get here - we best not be announcing the route
	 * so clear the gated bit
	 */ 

	if (!p_rt)
		{
		trace(TR_IDRP | TR_UPDATE, LOG_WARNING," Withdraw of idrp route without gated rt peer %s route %s", 
			p_idrp_rt->peer->name,iso_ptoa(&p_idrp_rt->nlri));
		return;
		} 	
		 
	rtbit_reset(p_rt,idrp_this_node.task->task_rtbit);
	if (p_rt->rt_state & RTS_DELETE) 
			{
			/* put warning flags in all over */
			trace (TR_IDRP | TR_UPDATE, LOG_WARNING, "Delete route from peer %s has route id %d delete flag set for nlri %s", 
			p_idrp_rt->peer->name,p_idrp_rt->route_id_in,iso_ptoa(&p_idrp_rt->nlri));
			if (!(p_rt->rt_state & RTS_EXTERIOR))
				{
				trace (TR_IDRP | TR_UPDATE, LOG_WARNING, "Plus IDRP route has no exterior flag"); 
				}
			return;
			}		
	
	rt_delete(p_rt);
}

/* look for a idrp route created from 
 * this gated route 
 *  (Can need EXT_INFO or IGP info)
 */

idrpRoute *
idrp_find_ext_info_gated_rt(rth,p_rt)
rt_head 	*rth;
rt_entry	*p_rt;
{
rt_entry	*rt;
idrpRoute	*p_idrp_rt;

	RT_ALLRT(rt,rth)
		{
		/* check for idrp route */
 
		if (rt->rt_gwp->gw_proto == RTPROTO_IDRP)
			{	 
			/* idrp route - check to see if
			 * this route contains pointer
			 * to original ext info route
			 */

			p_idrp_rt = (idrpRoute *) rt->rt_data;
			if (p_idrp_rt->p_rt_ext_info == p_rt)
				{
				return (p_idrp_rt);
				}
			}
		
		} RT_ALLRT_END(rt,rth);	

	/* no route found return a zero */

	p_idrp_rt = (idrpRoute *) 0;
	return (p_idrp_rt);
}
