#include "include.h"
#include "idrp.h"

/* 
 *  local routes configured by IDRP
 *
 * %% local routes set-up routines
 *
 * idrp_add_local_rt_gated(p_idrp_rt) - add local route to gated
 *
 * idrp_local_rt(nlri,family) -
 *      add new local route to peer and gated
 *              1.) att_record created if
 *               with peer structure this_node
 *              2.) create idrpRoute
 *              3.) fill idrpRoute structure
 *              4.) link to att_record route list
 *
 *
 * find_local_route_dest -
 *	1.) find local route with same destination and options
 *
 * create_local_attr_record-
 *              create attribute record for local routes
 *              set-up first route_id list
 *
 * create_local_ext_info_att_rec()
 *              create attribute record for local ext_info routes
 *
 * idrp_options_to_mask(p_atts,p_opts) - translate options to mask
 *
 * link_local_attr(p_atts) - link to attribute list
 *                                and update last local attribute pointer
 *
 * idrp_fill_local_man_attrib(p_att) -
 * idrp_fill_local_opt_attrib(p_atts,p_opts)
 * fill_next_hop(p_atts,p_opts)
 * fill_DIST_LIST(p_atts,p_rd,type)
 * link_local_attr_list (p_atts)
 * create_local_rd_path
 * idrp_create_local_gw(p_gw_str,gw_byte_cnt,family)
 * idrp_local_route_reset - reset all local routes prior to re-parsing
 * idrp_flag_local_att - flag routes associated for one attribute  
 * idrp_local_route_clean - cleanup routes after reparsing to 
 *			    get of any local routes not reconfigured 
 * idrp_clean_local_att(p_att) - clean up one attribute after reparsing
 * 
 * idrp_clean_local_route_chain(p_rte) - clean up a local route chain after
 *					 reconfiguration 
 * idrp_free_local_attrib - free local attribute record
 * idrp_free_local_idrpRoute_ent(p_rte)
 * idrp_free_local_rt(p_idrp_rt)
 * idrp_free_local_route_chain(p_idrp_rtl) -
 *      free the local route chain associated with
 *      this route id;

 * idrp_free_local_att - free attribute created out local
 * 
 * idrp_local_mem_fit - allocate memory out of the righ
 *			local memory size(small, medium, large
 *			just like the 3 bears)
 * 
 * rdi_canon_length - length of turning rdi chain to DIST_LIST_INCL
 */

/* %% local RD routines */

/* idrp_add_local_rt_gated(p_idrp_rt
 * - add local route to gated table during initialization
 *   or re-init
 */

extern int 
idrp_add_local_rt_gated(p_idrp_rt)
idrpRoute	*p_idrp_rt;
{
struct iso_net_addr	dest;                   /* destination address for gated Route search */
struct iso_net_addr	mask;                   /* mask for gated table Route search */
idrp_rt_chain_walk      gateway;                /* chain for route locate search for gateway */
idrp_rt_chain_walk      preference;             /* chain for route locate serach for perference */
idrpRoute		*p_best_ext = 0;
sockaddr_un		*p_dest;
sockaddr_un		*p_mask;


        /* add the route - warn if another route from gateway is there
         * calculates idrp preference in the routine
         */
	p_dest = (sockaddr_un *) &dest;

        idrp_add_route_locate(p_dest,p_idrp_rt,&gateway,&preference,&p_best_ext);

        if (gateway.p_rt)
                {
                /* problem - gateway has same value */
                return (FALSE);
                }
        if (preference.p_rt)
               {
                /* preference matches */
                /* add one to this node */
                p_idrp_rt->pref_cal++;
                }
	return(idrp_add_rt_to_gated(p_idrp_rt,p_dest));
}



/* 
 * idrp_local_rt - set up local routes into
 *  		   appropriate routing structures
 */

int idrp_local_rt(p_nlri_str,byte_cnt,bit_cnt,family,p_gw_str,gw_byte_cnt,p_opts)
u_char	*p_nlri_str;
int	byte_cnt;
int	bit_cnt;
int family;
u_char	*p_gw_str;
int	gw_byte_cnt;
idrpRoute_options	*p_opts;
{
rt_entry	*p_rt;				/* gated route entry */
rt_entry	*p_rt2;				/* gated route entry */
idrpRoute	*p_idrp_rt_old;			/* pointer to old route */
int		pref;				/* preference for gated */
idrpRoute	*p_idrp_rt,*p_irt;
idrpRoute_entry	*p_rtl = 0;
sockaddr_un	*p_next_hop;
idrp_attribute_record	*p_atts,*p_atts_old;
int			nlri_id;
idrpRoute_options	opts;

	if (!p_opts)
		{
		bzero(&opts,sizeof(idrpRoute_options));
		p_opts = &opts;
		}

	/* see if this location was configured once before */


	p_idrp_rt = idrp_find_local_dest(p_nlri_str,byte_cnt,bit_cnt,family,p_gw_str,gw_byte_cnt,p_opts);	
	if (p_idrp_rt)
		{
		IDRP_STATUS_BIT_CLEAR(p_idrp_rt,IDRP_STATUS_RECONFIGURE);
		return;
		}	
	
	/* No, so go off to try to create the locate attribute record or
	 * gateway 
	 */ 
		
	p_next_hop = idrp_create_local_gw(p_gw_str,gw_byte_cnt,family);
	p_atts = create_local_attr_record(p_next_hop,p_opts);	
	p_atts_old = find_attr_rec(&idrp_this_node,p_atts,TRUE);
	if (p_atts_old)
		{
		idrp_free_local_att(p_atts);
		p_atts = p_atts_old;
		}
	else
		{
		link_local_attr_list(p_atts);
		}

		
	/* set-up idrpRoute structure */ 

	p_idrp_rt = (idrpRoute *) idrp_med_blk_alloc(); 
	idrp_local_route_cnt++;
	bzero(p_idrp_rt,sizeof(idrpRoute));

	/* set the local variables 
	 * type of memory this is allocated from 
	 * nlri, route_id, links to next route, peer structure 
  	 * calculate the idrp preference
	 */ 

	p_idrp_rt->mem_type = IDRP_LOC_MEM; 
	bcopy(p_nlri_str,&p_idrp_rt->nlri.pfx,byte_cnt);
	p_idrp_rt->nlri.pfx_len = bit_cnt;	
	p_idrp_rt->p_attr = p_atts;
	p_idrp_rt->route_id_in = IDRP_LOCAL_ROUTE_ID;
	p_idrp_rt->p_next_nlri = (idrpRoute *) 0;
	p_idrp_rt->peer = &idrp_this_node;
	p_idrp_rt->pref_rcvd = 0;
	p_idrp_rt->family = family;
	p_idrp_rt->status =  IDRP_STATUS_LOCAL_NEW_ROUTE;
	p_idrp_rt->p_loc_rt_peer = (sockaddr_un *) idrp_small_blk_alloc();
	bcopy(p_gw_str,&p_idrp_rt->p_loc_rt_peer->iso.siso_addr,gw_byte_cnt);	 
	p_idrp_rt->p_loc_rt_peer->iso.siso_len = gw_byte_cnt + 2 ;	
	p_idrp_rt->p_loc_rt_peer->iso.siso_family = AF_ISO;
	
	/* set-up link between attribute and route structure */  
				
	nlri_id = family_to_nlri_id(family);
	p_atts->ref_cnt++;

	/* link into nlri chain off this route id for
	 * this attribute record
	 */

	if (p_atts->route_id_list->p_route[nlri_id])
		{
		/* already entries on the list */

		p_irt = p_atts->route_id_list->p_route[nlri_id];
		while (p_irt->p_next_nlri)
			{
			p_irt= p_irt->p_next_nlri;
			}
		p_irt->p_next_nlri = p_idrp_rt;
		}
	else
		{
		p_atts->route_id_list->p_route[nlri_id] = p_idrp_rt;
		}	

return(TRUE);
}

idrpRoute *
idrp_find_local_dest(p_nlri_str,byte_cnt,bit_cnt,family,p_gw_str,gw_byte_cnt,p_opts)	
u_char	*p_nlri_str;
int	byte_cnt;
int	bit_cnt;
int family;
u_char	*p_gw_str;
int	gw_byte_cnt;
idrpRoute_options	*p_opts;
{
 /* walk the idrp_attr_list until the end of the local attribute
  * to find the idrp_route structure
  */
int			i;
idrp_attribute_record	*p_att;
struct iso_prefix	dest;
idrpRoute		*p_irt;
idrpRoute_entry		*p_rte;

	bcopy(p_nlri_str,&dest.pfx,byte_cnt);
	dest.pfx_len = bit_cnt;

	ATTR_LIST(p_att)
		{
		for (p_rte = p_att->route_id_list;p_rte; p_rte = p_rte->p_next)
			{
			NLRI_FAMILY_LOOP(i)
				{
				for (p_irt = p_rte->p_route[i]; p_irt; p_irt = p_irt->p_next_nlri)
					{
					if (!isopfxcompare(&dest,&p_irt->nlri))
						{
						/* destination matches - do the options */
						if (local_gw_match(p_irt,p_gw_str,gw_byte_cnt) &&
							local_options_match(p_irt,p_opts))
							{
							return(p_irt);
							}
						}
					}
				}
			} NLRI_FAMILY_LOOP_END;
		} ATTR_LIST_END;

	return(0);
					
}


int
local_gw_match(p_irt,p_gw_str,byte_cnt)
idrpRoute	*p_irt;
u_char		*p_gw_str;
int		byte_cnt;
{

	/* compare against the sockaddr_un format which has two more bytes */
 
	if (p_irt->p_loc_rt_peer->iso.siso_len != (byte_cnt+2) )
		{
		return(FALSE);
		}
	if (bcmp(p_irt->p_loc_rt_peer->iso.siso_addr,p_gw_str,byte_cnt))
		{
		return(FALSE);
		}
	return(TRUE);
}


int 
local_options_match(p_irt,p_opts)
idrpRoute	*p_irt;
idrpRoute_options	*p_opts;
{
	/* for right now null out the options match */

	return(TRUE);


}



idrp_attribute_record *
create_local_attr_record(p_next_hop,p_opts)
sockaddr_un		*p_next_hop;
idrpRoute_options	*p_opts;
{
idrp_attribute_record	*att;
idrp_attribute_record	*p_atts,*p_last;

	/* create attribute record, idrpRoute route_id list,
	 * and rd canon path structure  from task_alloc memory
	 * - already assigned from memory is p_atts->next_hop
	 *
	 */ 


	p_atts = (idrp_attribute_record *) idrp_med_blk_alloc();
	bzero(p_atts,sizeof(idrp_attribute_record));
	p_atts->peer_alloc= &idrp_this_node;
	p_atts->mem_type = IDRP_LOC_MEM;
	p_atts->next_hop = p_next_hop;

	/* allocate the route_id_list */

	p_atts->route_id_list = (idrpRoute_entry *) idrp_small_blk_alloc();
	p_atts->route_id_list->mem_type = IDRP_LOC_MEM;
	bzero(p_atts->route_id_list,sizeof(idrpRoute_entry));

	/* 
	 * Allocate the canonical RD path  
	 */

	p_atts->rd_path = (idrp_canon_rdpath *) idrp_small_blk_alloc();
	p_atts->rd_path->p_next = 0;
	  
	/* set the local mask to reset RD path upon
	 * re-parsing of node
	 */

	p_atts->local_mask = IDRP_LOCAL_RDPATH_RESET_LENGTH | IDRP_LOCAL_NO_PDU; 

	/* fill local attributes and link to attribute list */

	idrp_fill_local_man_attrib(p_atts);
	idrp_fill_local_opt_attrib(p_atts,p_opts);
	return(p_atts);
}





idrp_attribute_record *
find_local_att_rec(p_next_hop,p_opts)
sockaddr_un	*p_next_hop;
idrpRoute_options	*p_opts;
{
idrpRoute_options	opts;
idrp_attribute_record	*p_att;

	
	ATTR_LIST(p_att)
		{
		if ((p_att->idrp_mask == IDRP_ATTBT_MANDATORY) ||
			(p_att->idrp_mask == (IDRP_ATTBT_MANDATORY+IDRP_ATTBT_NEXT_HOP)))
			{
			if (p_att->local_mask & IDRP_LOCAL_NO_PDU)
				{
				if (sockaddrcmp(p_next_hop,p_att->next_hop))
					{
					return(p_att);
					}
				}
			}
		}

	/* zero out the options in case non sent */
	bzero(&opts,sizeof(idrpRoute_options));

	/* create an idrp attribute record */ 
	p_att = create_task_att_rec(&idrp_this_node);
	p_att->next_hop = p_next_hop;

	/* fill mandatory attributes */

	p_att->local_mask = IDRP_LOCAL_NO_PDU; 
	idrp_fill_local_man_attrib(p_att);

	if (p_opts)	
		{
		idrp_fill_local_opt_attrib(p_att,p_opts);
		}
	else
		{
		idrp_fill_local_opt_attrib(p_att,&opts);
		}

	/* set up the RD path length */

	p_att->attrib[IDRP_ATTR_RD_PATH].length = rt_rdi.isoa_len + sizeof(idrp_path_segment_header) + RDI_LENGTH_BYTE_LEN; 	

	/* link on the general attribute list */
	
	link_att_list(p_att);
	return (p_att);
}	
			 



/* find the EXT_INFO attribute record with the assigned 
 * next hop - translate IP into ISO if idrp_this_node.proto_sock
 * iso 
 */

idrp_attribute_record *
find_local_ext_attr(p_next_gw,p_opts)
sockaddr_un		*p_next_gw;
idrpRoute_options	*p_opts;
{
idrp_attribute_record	*p_att;
sockaddr_un		*p_next_hop = 0;
idrpRoute_options	opts;

	/* Find the local external attribute
	 * with the right next hop
	 */

	ATTR_LIST(p_att)
		{
		/* walk attribute chain looking for the 
		 * attribute with EXT_INFO and local attribute
		 * and right next hops
		 */  

		if ((p_att->idrp_mask & IDRP_ATTBT_EXT_INFO) && (p_att->local_mask & IDRP_LOCAL_EXT_INFO))
			{
			/* found one - is it the same as the next hop */

			if (sockaddrcmp(p_next_gw,p_att->next_hop))
				{	
				return(p_att);
				} 
			} /* test for external infor record */
		} ATTR_LIST_END;	

	if (p_next_gw)
		{
		p_next_hop = idrp_sock_dup(&idrp_this_node,p_next_gw);
		}
	if (p_opts)
		{
		return(create_local_ext_info_att_rec(p_next_hop,p_opts));
		}
	else
		{
		bzero(&opts,sizeof(idrpRoute_options));
		return(create_local_ext_info_att_rec(p_next_hop,&opts));
		}
}

/* create an attribute record out of task memory */

idrp_attribute_record *
create_task_att_rec(peer)
idrpPeer	*peer;
{
idrp_attribute_record	*p_atts;

	p_atts = (idrp_attribute_record *) task_mem_malloc(idrp_this_node.task, sizeof(idrp_attribute_record));
	bzero(p_atts,sizeof(idrp_attribute_record));
	p_atts->peer_alloc = &idrp_this_node;
	p_atts->mem_type = IDRP_TASK_MEM;

	/* allocate route id list of of task memory */

	p_atts->route_id_list = (idrpRoute_entry *) task_mem_malloc(idrp_this_node.task,sizeof(idrpRoute_entry)); 

	bzero(p_atts->route_id_list,sizeof(idrpRoute_entry));
	p_atts->route_id_list->mem_type = IDRP_TASK_MEM;
	p_atts->rd_path = (idrp_canon_rdpath *) task_mem_malloc(idrp_this_node.task,sizeof(idrp_canon_rdpath));
	bzero(p_atts->rd_path,sizeof(idrp_canon_rdpath));
	return (p_atts);
}


idrp_attribute_record *
create_local_ext_info_att_rec(p_next_hop,p_opts)
sockaddr_un		*p_next_hop;
idrpRoute_options	*p_opts;
{
idrp_attribute_record	*att;
idrp_attribute_record	*p_atts,*p_last;

	/* create attribute record, idrpRoute record 
	 * and rt_entry record for route change list
	 * - This must be called after master task has been
	 *   initialized. 
	 * - allocated attribute record, route_list entry
	 *  and  cannonical RD path structure  
	 */

	p_atts = create_task_att_rec(&idrp_this_node);
	p_atts->next_hop = p_next_hop;

	/* set local mask to indicate
	 *   This is a local record without a PDU attached
	 *   Set in ext_info attribute to be true   
	 */
	
	idrp_fill_local_man_attrib(p_atts);

	/* fill in next_hop right for now 
	 * - later allow other options
	 * - by protocols - such as bgp
	 */

	idrp_fill_local_opt_attrib(p_atts,p_opts);

	p_atts->attr_len += IDRP_EXT_INFO_LENGTH + sizeof(idrp_path_attr_hdr); 

	/* set up the RD path length */

	p_atts->attrib[IDRP_ATTR_RD_PATH].length = rt_rdi.isoa_len + sizeof(idrp_path_segment_header) + RDI_LENGTH_BYTE_LEN; 	

	/* set up the  EXT_INFO specific entries:
	 * - route_id
	 * - mask bits with EXT_INFO
	 * - EXT_INFO attribute
	 */
 
	p_atts->route_id_list->route_id = IDRP_EXT_INFO_ROUTE_ID; 
	p_atts->local_mask = IDRP_LOCAL_NO_PDU; 
	p_atts->idrp_mask = IDRP_ATTBT_MANDATORY | IDRP_ATTBT_EXT_INFO;
	p_atts->attrib[IDRP_ATTR_EXT_INFO].present = TRUE;
	p_atts->attrib[IDRP_ATTR_EXT_INFO].flags = IDRP_ATTRFLG_EXT_INFO;
	p_atts->attrib[IDRP_ATTR_EXT_INFO].data = (u_char *) NULL; 
	p_atts->attrib[IDRP_ATTR_EXT_INFO].length = 0;
	link_att_list(p_atts);
	return (p_atts);
}


/* fill in the local_mandatory attributes */


void
idrp_fill_local_man_attrib(p_atts)
idrp_attribute_record *p_atts;
{
idrp_attribute_record	*att;
idrp_attribute_record	*p_last;

	/* set ups the route_id list to local route_id */
 
	p_atts->route_id_list->route_id = IDRP_LOCAL_ROUTE_ID;
	p_atts->route_id_list->peer = &idrp_this_node;
	p_atts->route_id_list->status = IDRP_ATTR_ROUTE_ID_NORM;

	/* set local route_id and pref to zero - should
	 * never go out
	 */

	bzero(&idrp_local_route_id.id,sizeof(route_id));
	idrp_local_route_id.loc_pref = 0;
	
	/* set up rib id */

	 p_atts->rib_id = 0;
	/* set-up local hop count */

	local_hop_count = 0;
	p_atts->hopcount = 0; 


	/* set up attribute mask 
	 * hopcount, route_id, rd path
	 */	

	p_atts->idrp_mask = IDRP_ATTBT_MANDATORY;
	p_atts->attrib[IDRP_ATTR_ROUTE_SEPARATOR].present = TRUE;
	p_atts->attrib[IDRP_ATTR_ROUTE_SEPARATOR].flags = IDRP_ATTRFLG_ROUTE_SEPARATOR;
	p_atts->attrib[IDRP_ATTR_ROUTE_SEPARATOR].data = (u_char *) &idrp_local_route_id; 
	p_atts->attrib[IDRP_ATTR_ROUTE_SEPARATOR].length = IDRP_ROUTE_SEPARATOR_LENGTH;


	p_atts->attrib[IDRP_ATTR_RD_PATH].present = TRUE;
	p_atts->attrib[IDRP_ATTR_RD_PATH].flags = IDRP_ATTRFLG_RD_PATH;
	p_atts->attrib[IDRP_ATTR_RD_PATH].data = (u_char *) &local_rd_path;

	/* note this must be filled in by the init code after the 
	 * the rdi for this box is read in
 	 */
 
	p_atts->attrib[IDRP_ATTR_RD_PATH].length = 0; 
	p_atts->rd_path->p_next = (idrp_canon_rdpath *) 0;
	bcopy(&rt_rdi,&p_atts->rd_path->rdi,sizeof(rt_rdi));

	/* set the hop count up */

	p_atts->attrib[IDRP_ATTR_HOP_COUNT].present = TRUE;
	p_atts->attrib[IDRP_ATTR_HOP_COUNT].flags = IDRP_ATTRFLG_HOP_COUNT;
	p_atts->attrib[IDRP_ATTR_HOP_COUNT].data = (u_char *) &local_hop_count;  
	p_atts->attrib[IDRP_ATTR_HOP_COUNT].length =IDRP_RD_HOP_COUNT_LENGTH; 

	/* set length in the attribute mandatory length
	 * withtout the rdi length
	 * - that will have to be adjusted 
	 *   after all configurations are done
	 */

	p_atts->attr_len = IDRP_LOCAL_MAN_ATT_LEN; 
	p_atts->p_attr_pdu = (u_char *) 0; 

}



void
idrp_fill_local_opt_attrib(p_atts,p_opts)
idrp_attribute_record	*p_atts;
idrpRoute_options	*p_opts;
{
int	len = 0;
int	len2 = 0;

	/* fill in next hop */
	
	if (p_atts->next_hop)
		{
		len2 = fill_next_hop(p_atts,p_opts);
		if (p_atts->local_mask & IDRP_ROUTE_SERVER_FLAG)
			{
			len += len2;
			}
		}
	

	/* fill in the DIST_LIST_INCL and DIST_LIST_EXCL parameters */

	if (p_opts->p_DIST_LIST_INCL)
		{
		len += fill_DIST_LIST(p_atts,p_opts->p_DIST_LIST_INCL,IDRP_ATTR_DIST_LIST_INCL);
		}	
		
 	if (p_opts->p_DIST_LIST_EXCL)
		{
		len += fill_DIST_LIST(p_atts,p_opts->p_DIST_LIST_EXCL,IDRP_ATTR_DIST_LIST_EXCL);
		}
	if (p_opts->multi_exit)
		{
		len += fill_MULTI_EXIT(p_atts,p_opts);
		}

	p_atts->attr_len += len;
}	

int
fill_next_hop(p_atts,p_opts)
idrp_attribute_record	*p_atts;
idrpRoute_options	*p_opts;
{
int	mem_need = 0;
u_short	hold;
u_char	*cp;
int	nlri_id;
int	i;
int	size;
snpa_entry	*p_snpa;
	

	/* fill in the idrp next_hop feature if there is a gateway
	 */  

	
	if (!p_atts->next_hop)
		{
		return(0);
		}

	/* set the flags and that NEXT_HOP is prsent */

	p_atts->idrp_mask |= IDRP_ATTBT_NEXT_HOP;
	p_atts->attrib[IDRP_ATTR_NEXT_HOP].present = TRUE;
	p_atts->attrib[IDRP_ATTR_NEXT_HOP].flags = IDRP_ATTRFLG_NEXT_HOP;

	/* default is not to send the next hop */ 
	p_atts->local_mask |= IDRP_DONT_SEND_NEXT_HOP;

	if (p_opts)
		{
		mem_need = p_opts->next_hop_snpas.pdu_len;
		}


	nlri_id = family_to_nlri_id(p_atts->next_hop->a.sa_family);
	hold = idrp_nlri_proto_id[nlri_id].proto_len; 

	/* length need - NET length, plus header bytes
	 * SNPA lengths plus count bytes
	 * + one for Route server byte
	 */

	mem_need = sizeof(struct iso_net_addr) + sizeof(nlri_proto_info) + (hold); 
	if (p_atts->mem_type == IDRP_TASK_MEM)
		{
		cp =(u_char *)  task_mem_malloc(p_atts->peer_alloc->task,mem_need);
		p_atts->attrib[IDRP_ATTR_NEXT_HOP].data = cp;
		}
	else
		{
		/* if local memory -find the best fit */

		cp = idrp_local_mem_fit(mem_need);
		p_atts->attrib[IDRP_ATTR_NEXT_HOP].data = cp; 
		}


	/* copy in the Route Server flag into the record
	 * - default to IDRP_SERVER_ALLOWED for now
	 */

	if (p_opts)
		{
		if (p_opts->route_server)
			{				
			p_atts->local_mask &= IDRP_DONT_SEND_NEXT_HOP_OFF;  
			p_atts->local_mask |= IDRP_ROUTE_SERVER_FLAG;
			}	
		*cp = p_opts->route_server;	 
		}
	else
		{
		*cp = IDRP_SERVER_NOT_ALLOWED;
		}
	cp++; 
	/* generate the protocol type for this NET */

	bcopy (&idrp_nlri_proto_id[nlri_id].proto_type,cp,1);
	cp++;
	bcopy (&idrp_nlri_proto_id[nlri_id].proto_len,cp,1);
	cp++;
	bcopy(&idrp_nlri_proto_id[nlri_id].proto_val,cp,idrp_nlri_proto_id[nlri_id].proto_len);
	hold = (u_int)idrp_nlri_proto_id[nlri_id].proto_len;
	cp += hold;

	/* copy in the NET now */

	switch	(nlri_id)	
		{
		case IDRP_NLRI_ISO:
			size = p_atts->next_hop->a.sa_len - 2;
			*cp = size;
			cp++;
			bcopy (&p_atts->next_hop->iso.siso_addr,cp,size);
			cp += size;
			break;
	
		case IDRP_NLRI_IP: 
			*cp = sizeof(struct in_addr);
			cp++;
			bcopy (&p_atts->next_hop->in.sin_addr,cp,sizeof(struct in_addr));
			cp += sizeof(struct in_addr);
			break;
		}

	/* add in the snpas */ 		
	/* there may be snpa configurations */

	if (p_opts)
		{
		/* if there is a options record get length */
 
		hold = p_opts->next_hop_snpas.pdu_len;
		}
	else
		{
		/* if not it must be zero */

		hold = 0;
		}
	*cp = hold;
	cp++;

	/* if there are any snpas */

	if (hold)
		{
		p_snpa = (snpa_entry *) &p_opts->next_hop_snpas.snpa1;
		for (i = 0; i < hold; i++)
			{
			/* copy in all the snpa octets */
			*cp = p_snpa->len;
			cp++;	
			bcopy(p_snpa->snpa,cp,p_snpa->len);
			cp += p_snpa->len;
			}
		}
	p_atts->attrib[IDRP_ATTR_NEXT_HOP].length = (cp - p_atts->attrib[IDRP_ATTR_NEXT_HOP].data);
	
	return (mem_need);
}


int
fill_DIST_LIST(p_atts,p_rd,type)
idrp_attribute_record	*p_atts;
idrp_canon_rdpath	*p_rd;
int			type;
{
idrp_canon_rdpath	*p_rd2;
idrp_rdi_mem		req;	
int			len = 0;
int			mem_need = 0; 
idrp_dist_list		*p_dist_list;
u_char			*cp;

	/* calculate statistics on rd path 
	 * - the amount of rdis in sequence
	 * and the memory need for DIST_LIST attribute
	 */ 
	

	/* add one to the memory for the count of rdis */

	mem_need = rdi_canon_length(p_rd,&req) + 1;
	if (p_atts->mem_type == IDRP_TASK_MEM)
		{
		p_dist_list = (idrp_dist_list *) task_mem_malloc(p_atts->peer_alloc->task,mem_need);
		}
	else
		{
		p_dist_list = (idrp_dist_list *) idrp_local_mem_fit(mem_need);
		}

	/* copy in the count of rdis in the
	 * pdus	 
	 */

	cp =  (u_char *) p_dist_list;
	cp = (u_char *) &req.cnt;
	cp++;

	p_rd2 = p_rd;
	while (p_rd2)
		{
		register int rdi_len;
		rdi_len = p_rd->rdi.isoa_len;	
		*cp = rdi_len;
		cp++;
		bcopy(&p_rd->rdi.isoa_len,cp,rdi_len);
		cp += rdi_len; 
		p_rd2 = p_rd2->p_next;
		}

	if (type == IDRP_ATTR_DIST_LIST_INCL)
		{
		/* distribution list include */

		p_atts->idrp_mask |= IDRP_ATTBT_DIST_LIST_INCL;
		p_atts->attrib[IDRP_ATTR_DIST_LIST_INCL].present = TRUE;
		p_atts->attrib[IDRP_ATTR_DIST_LIST_INCL].flags = IDRP_ATTRFLG_DIST_LIST_INCL;
		p_atts->attrib[IDRP_ATTR_DIST_LIST_INCL].length = mem_need;
		p_atts->attrib[IDRP_ATTR_DIST_LIST_INCL].data = (u_char *) p_dist_list;
		}
	if (type == IDRP_ATTR_DIST_LIST_EXCL)
		{
		/* distribution list exclude */

		p_atts->idrp_mask |= IDRP_ATTBT_DIST_LIST_EXCL;
		p_atts->attrib[IDRP_ATTR_DIST_LIST_EXCL].present = TRUE;
		p_atts->attrib[IDRP_ATTR_DIST_LIST_EXCL].flags = IDRP_ATTRFLG_DIST_LIST_EXCL;
		p_atts->attrib[IDRP_ATTR_DIST_LIST_EXCL].length = mem_need;
		p_atts->attrib[IDRP_ATTR_DIST_LIST_EXCL].data = (u_char *) p_dist_list;
		}

	return (mem_need);

}

int fill_MULTI_EXIT(p_atts,p_opts)
idrp_attribute_record 	*p_atts;
idrpRoute_options	*p_opts;
{

	p_atts->attrib[IDRP_ATTR_MULTI_EXIT_DISC].present = TRUE;
	p_atts->attrib[IDRP_ATTR_MULTI_EXIT_DISC].flags = IDRP_ATTRFLG_MULTI_EXIT_DISC;
	p_atts->attrib[IDRP_ATTR_MULTI_EXIT_DISC].length = IDRP_MULTI_EXIT_DISC_LENGTH;
	p_atts->attrib[IDRP_ATTR_MULTI_EXIT_DISC].data = &p_atts->multi_exit;
}
		
int		
link_local_attr_list(p_atts)
idrp_attribute_record	*p_atts;
{
idrp_attribute_record *p_atts_task;
	
	/* setup the following attribute list points
	 * idrp_attr_list - contains all attributes
	 * within that list:
	 * 1.) local attributes are linked first
	 *     because they are created first
	 */

	/* link into attribute list as first one */

	if (idrp_last_local_att)
		{
		if (idrp_last_local_att->next)
			{
			/* are there routes beyond the last local route?
		 	 * then  we are reparsing 
			 * link in the attribute as last local
			 * in doubly linked list
			 */
			p_atts_task = idrp_last_local_att->next;
			p_atts_task->prev = p_atts;
			idrp_last_local_att->next = p_atts;
			p_atts->prev = idrp_last_local_att;
			p_atts->next = p_atts_task;
			idrp_last_local_att = p_atts;
			return;
			}
		}

	/* still linking only local attributes - use normal methods */ 
					 
	link_att_list(p_atts);
	idrp_last_local_att = p_atts;	
}	

idrp_canon_rdpath *
create_local_rd_path(rdi_len,p_rdi_str)
int	rdi_len;
u_char	*p_rdi_str;
{
idrp_canon_rdpath	*p_rd;

	p_rd = (idrp_canon_rdpath *) idrp_small_blk_alloc();
	p_rd->rdi.isoa_len = rdi_len;
	bcopy(p_rdi_str,&p_rd->rdi.isoa_genaddr,rdi_len);
	p_rd->p_next = (idrp_canon_rdpath *) 0;
	return (p_rd);
}

sockaddr_un *
idrp_create_local_gw(p_gw_str,gw_byte_cnt,family)
u_char	*p_gw_str;
int	gw_byte_cnt;
int	family;
{
sockaddr_un	*p_gw;
	
	p_gw = (sockaddr_un *) idrp_small_blk_alloc();	
	p_gw->a.sa_family = family;
	p_gw->a.sa_len = gw_byte_cnt + 2;
	switch (family)
		{
		case AF_ISO:
			bcopy(p_gw_str,&p_gw->iso.siso_addr,gw_byte_cnt);
			break;
		case AF_INET:
			bcopy(p_gw_str,&p_gw->in.sin_addr,gw_byte_cnt);
			break;
		}
	return (p_gw);
}



void
idrp_local_route_reset()
{
idrp_attribute_record	*p_atts = idrp_attr_list;
idrp_attribute_record	*p_att2;

	/* loop flaging all the local attribute records
	 * and routes that reconfiguration has occurred 
	 */

	
	while (p_atts != idrp_last_local_att)
		{
		p_att2 = p_atts;
		p_atts = p_atts->next;
		idrp_flag_local_att(p_att2);
		}
	idrp_flag_local_att(p_atts);
}

	

int 
idrp_local_route_clean()
{
idrp_attribute_record	*p_atts = idrp_attr_list;
idrp_attribute_record	*p_att2;
int			changes = 0;


	/* if there is no local route set-up yet,
	 * then ignore this
	 */

	if (!idrp_last_local_att)	
		{
		return;
		}

	while (p_atts != idrp_last_local_att)
		{
		p_att2 = p_atts;
		p_atts = p_atts->next;
		changes += idrp_clean_local_att(p_att2);
		}
	changes += idrp_clean_local_att(p_atts);

}

/* free the attribute  space created for local attribute records */

void 
idrp_free_local_attrib_task(p_att)
idrp_attribute_record *p_att;
{
int i;
	/* do the remainder of the attribute space
	 * - IDRP_ATTR_ROUTE SEPARTOR has local data - do not free
	 * - IDRP_ATTR_RD_PATH - has local data - do not free
	 * - IDRP_MULTI_EXIT_DISC - has local data - do not free
	 * - IDRP_ATTR_HOP_COUNT - has local data - do not free
	 * That leaves:
	 * NEXT_HOP, IDRP_DIST_LIST_INCL, IDRP_DIST_LIST_EXCL
	 * in this version of idrp 
	 * And in the future-  TRANSIT_DELAY, EXPENSE, LOCALLY_DEF_QOS   
 	 * SECURITY, CAPACITY and PRIORITY
	 */
	
	for (i = IDRP_ATTR_NEXT_HOP; i < IDRP_ATTR_MULTI_EXIT_DISC; i++)
		{ 
		if (p_att->attrib[i].data)
			{
			task_mem_free(p_att->peer_alloc->task,p_att->attrib[i].data);	 
			}
		}

}
void
idrp_free_local_att(p_att)
idrp_attribute_record	*p_att;
{
idrpRoute_entry	*p_rte,*p_rte2;
	
        /* free the the attribute records pieces
         * 1) free all route_id_list entries
         * 2) free the next_hop entry
         * 3) free the snpa list 
         * 4) free any memory in attribute list
         * 5) free any rd_canon path
         */

 
	p_rte = p_att->route_id_list;

        /* loop through the local_route_attr_record
         * free the local route chains
         */

	while (p_rte)
		{
		p_rte2 = p_rte;
		p_rte = p_rte->p_next;  
		idrp_free_local_route_chain(p_rte2);
		}

        /* free any associated storage
         * - here we need to check on 
         * 1) next_hop gw
         * 2) snpa_list
         * 3) any associated lengths in the attribute records
         * 4) any rd_canon path
         */

	free_local_attr_record(p_att);
}


void 
idrp_flag_local_att(p_att)
idrp_attribute_record *p_att;
{
idrpRoute_entry	*p_rte;
idrpRoute	*p_irt;
int		i;

	/* walk all route_id_list and
	 * flag all routes as needing
	 * reconfiguration 
	 */
	p_rte = p_att->route_id_list;

	/* loop through the local_route_attr_record
	 * free the local route chains
	 */

	for (p_rte = p_att->route_id_list; p_rte; p_rte = p_rte->p_next)	
		{
		p_rte->status |= IDRP_LOCAL_ROUTE_RECONFIG; 
		NLRI_FAMILY_LOOP(i)
			{
			for (p_irt = p_rte->p_route[i]; p_irt; p_irt = p_irt->p_next_nlri)
				{
				p_irt->status |= IDRP_STATUS_RECONFIGURE;
				}
			} 
		}	
}


int
idrp_clean_local_att(p_att)
idrp_attribute_record *p_att;
{
idrpRoute_entry	*p_rte,*p_rte_last;
int		changes = 0;


	/* loop through the local_route_attr_record
	 * cleaning out the local route chains
	 * if no nlri exists on local route chain - free it
	 */

	p_rte_last = 0;
	p_rte = p_att->route_id_list;
	while (p_rte)
		{
		/* here we need to flag if we
		 * need to keep the route chain
		 */
		p_rte_last = p_rte;
		p_rte = p_rte->p_next;
		changes = clean_local_route_chain(p_rte_last);
		}
	return(changes);
}


void
free_local_attr_record(p_att)
idrp_attribute_record *p_att;
{
idrp_canon_rdpath       *p_rd,*p_rd2;
int             i;

	
	/* IDRP Attribute needs to be freed
	 *  free any associated storage
 	 * - here we need to check on 
	 * 1) next_hop gw
	 * 2) snpa_list
	 * 3) any associated lengths in the attribute records
	 * 4) any rd_canon path
	 */

	if (p_att->next_hop)
		{
		idrp_small_blk_free((u_char *) p_att->next_hop);
		}

	if (p_att->rd_path)
		{
		/* walk the idrp_canon_rdpath structure and free up blocks */
		p_rd = p_att->rd_path;
		while (p_rd)
			{
			p_rd2 = p_rd;
			p_rd = p_rd->p_next;
			idrp_small_blk_free((u_char *)p_rd2);
			}
		}

	/* do the remainder of the attribute space
	 * - IDRP_ATTR_ROUTE SEPARTOR has local data - do not free
	 * - IDRP_ATTR_RD_PATH - has local data - do not free
	 * - IDRP_MULTI_EXIT_DISC - has local data - do not free
	 * - IDRP_ATTR_HOP_COUNT - has local data - do not free
	 * That leaves:
	 * NEXT_HOP, IDRP_DIST_LIST_INCL, IDRP_DIST_LIST_EXCL
	 * in this version of idrp 
	 * And in the future-  TRANSIT_DELAY, EXPENSE, LOCALLY_DEF_QOS   
 	 * SECURITY, CAPACITY and PRIORITY
	 */
	
	for (i = IDRP_ATTR_NEXT_HOP; i < IDRP_ATTR_MULTI_EXIT_DISC; i++)
		{ 
		if (p_att->attrib[i].data)
			{
			idrp_mem_fit_free(&p_att->attrib[i]);	 
			}
		}

	 
	/* free the local route attribute record */
	idrp_med_blk_free((u_char *) p_att);

}

int
clean_local_route_chain(p_rte)
idrpRoute_entry	*p_rte;
{
idrpRoute	*p_irt,*p_last_rt;
int		i;
int		changes	= 0;	

	NLRI_FAMILY_LOOP(i)
		{
		p_irt = p_rte->p_route[i];
		while (p_irt)
			{
			p_last_rt = p_irt;
			p_irt = p_irt->p_next_nlri;
			if (p_last_rt->status & IDRP_STATUS_RECONFIGURE)
				{
				/* delete the gated route
				 * delete the route out of route lists and attribute record 
				 * delete idrpRoute
				 */

				idrp_del_rt_gated(p_last_rt);
				idrp_free_nlri_att_rec(p_last_rt,NLRI_LINKED_P_NEXT);	
				idrp_free_local_rt(p_last_rt);
				changes++;
				}
			}
		}

	return(changes);	
}
	
void
idrp_free_local_idrpRoute_ent(p_rte)
idrpRoute_entry	*p_rte;
{
	idrp_small_blk_free((u_char *) p_rte);
}


void
idrp_free_local_rt(p_idrp_rt)
idrpRoute	*p_idrp_rt;
{
	idrp_med_blk_free((u_char *) p_idrp_rt);
}

void
idrp_free_local_route_chain(p_idrp_rtl)
idrpRoute_entry	*p_idrp_rtl;
{
idrpRoute	*p_rt;
idrpRoute	*p_rt2;
int		i;

	NLRI_FAMILY_LOOP(i)
	{
		p_rt = p_idrp_rtl->p_route[i];

		/* use task block free when delete local routes */
		/* since local routes created before idrp_init */
 
		while (p_rt)
			{
			p_rt2 = p_rt;
			p_rt = p_rt->p_next_nlri; 

			/* don't worry about gated route -
			 * already freed 
			 */

			idrp_small_blk_free((u_char *) p_rt2);
			}
		p_idrp_rtl->p_route[i] = (idrpRoute *) 0;	
	} NLRI_FAMILY_LOOP_END;

	idrp_small_blk_free((u_char *) p_idrp_rtl);

	return;
}


u_char *
idrp_local_mem_fit(len)
int len;
{
u_char	*p;
	if (len < IDRP_SMALL_BLOCK_SIZE)
		{
		return(idrp_small_blk_alloc());
		}
	if (len < IDRP_MEDIUM_BLOCK_SIZE)
		{
		return(idrp_med_blk_alloc());
		}
	return (idrp_peer_alloc((idrpPeer*) 0));
}

void
idrp_mem_fit_free(p_attrib)
idrp_attribute_entry	*p_attrib;
{

	if (p_attrib->length < IDRP_SMALL_BLOCK_SIZE)
		{
		idrp_small_blk_free((u_char *) p_attrib->data);
		return;
		}
	if (p_attrib->length < IDRP_MEDIUM_BLOCK_SIZE)
		{
		idrp_med_blk_alloc((u_char *) p_attrib->data);
		return;
		}
	idrp_big_blk_free((idrpPeer *) p_attrib->data);
}

int rdi_canon_length(p_rdi,p_req)
idrp_canon_rdpath	*p_rdi;
idrp_rdi_mem		*p_req;
{
idrp_canon_rdpath	*p_rd;
int			len;
int			cnt;

	for (p_rd = p_rdi,cnt = 0, len = 0 ; p_rd; p_rd->p_next,cnt++)
		{
		/* add in the length of the rdi */
		len += p_rd->rdi.isoa_len + 1;
		}
	p_req->cnt = cnt;
	p_req->mem_need = len;

return (len);
}	
