#include "include.h"
#include "idrp.h"

#ifdef ADN_PATCH
/*	for timezone & timeval */
#	include <sys/time.h>
#endif /* ADN_PATCH */

/* idrp_rt_minadv.c
 * - routine to handle Minimum route advertisement for
 *   for external and local routes

 * %% idrp_add_min_Route
 *
 * idrp_minadv_annlist	- add all of announce list
 *			  nlri's to this announce list
 *
 * idrp_add_min_route_adv - add this route to minimum route
 *                          advertisement timer list
 * advmin_interval_calc   - for this run of routes, calculate
 *			    the interval after the next
 *			    min_route advertisement timer
 *			    goes off until the time for
 *			    this group of routes should go off.
 *			    for example if timer is 30 second
 *			    and 10 have gone by - 20 seconds
 *			    is the internval
 *			    Two types of timers handled - 
 *			    minimum route advertisement timer
 *			    (outside the domain)
 *			    and minimum route advertisment timer
 *			    (from within the domain)
 *                          to time after timer expires
 *                          or if no timer give full routines
 *
 * idrp_set_min_route_timer - set the minimum route timer to expire
 * idrp_process_minadv-   process list of routes after min route advertisement
 *                              timer expires
 * idrp_min_adv_rt              - process one route from without a domain
 * 
 * idrp_add_rt_min_advRD  - add this route to minimum adv
 *			    timer for this RD
 * idrp_set_min_advRD_time -  set the minimum advRD timer to expire 
 * idrp_process_minAdvRD - process routes damped by the min_advRd timer
 * idrp_min_advRd_rt		- process one route from within this RD 
 *     
 *   
 */

void 
idrp_minadv_annlist(p_ann)
idrp_ann_list	*p_ann;
{

/* walk through an nlri portion of the announce list
 *	1) setting the minimum advertisement timer
 *	   for routes from other domains
 *	2) setting minadvRD for routes
 *		from this domain 
 * 		
 * 
 * flag on each route announced,
 * the appropriate flags in each route
 */

idrp_ann_list 	*p_atl;
idrpRoute	*p_irt;
int		i;

	ANN_LIST(p_atl,p_ann)
		{		
		/* walk the announce list for each NLRI Family */

		NLRI_FAMILY_LOOP(i)
			{
			ANN_ANN_LIST_WALK(p_irt,i, p_atl)
				{	
				/* walk the announce nlri portion
				 * set the add min route timer
				 * for routes from External peers
				 */
	
				switch (p_irt->peer->type)
					{
					case IDRP_PEER_EXTERNAL:
						idrp_add_min_route_adv(p_irt);
						break;
	
					case IDRP_PEER_INTERNAL:
						trace (TR_IDRP,LOG_ERR,"Internal peer's route Phase3 announce list %s", 
							iso_ptoa(&p_irt->nlri));
						break;
	
					case IDRP_PEER_TEST:
						trace (TR_IDRP,LOG_ERR,"Test peer %s sent routes - Problem!! ",p_irt->peer->name);
						break;

					case IDRP_PEER_LOCAL:
						idrp_add_rt_min_advRD(p_irt);
						break;
					}

				}ANN_ANN_LIST_WALK_END;
			} NLRI_FAMILY_LOOP_END;

		}ANN_LIST_END;
}

/*  %% Minimum route advertisement timer related routines */

void
idrp_add_min_route_adv(p_idrp_rt)
idrpRoute	*p_idrp_rt;
{
idrpAdvRt	*p_advrt;	
idrpPeer	*peer;
struct	timezone tzp;
struct	timeval	tp;	
int	id;

	/* set flag to indicate that minimum  advertisement timer is running */


	IDRP_STATUS_BIT_SET(p_idrp_rt,IDRP_STATUS_MIN_ADV_RUN);

	/* look to see if the minimum route advertisement set-up for this
	 * phase processing is there yet 
  	 */

	peer = p_idrp_rt->peer;
	id = family_to_nlri_id(p_idrp_rt->family);
	if (!peer->start_min_adv)
		{
		/* No flag to start min route Advertisement timer
		 * - set the flag and calculate interval
		 * for the routes in this flash of routes
		 */
 
		peer->start_min_adv = TRUE;
		p_advrt = (idrpAdvRt *) task_mem_malloc(peer->task,sizeof(idrpAdvRt));	
		bzero(p_advrt,sizeof(idrpAdvRt));
		if (gettimeofday(&tp,&tzp))
			{
			trace (TR_IDRP, LOG_ERR,"idrp used gettime day  and failed, dropped min_adv_rt request ");
			return;
			}

		/* translate time of day here to the seconds to calculate next interval */

		p_advrt->interval = advmin_interval_calc(&tp,peer,IDRP_MIN_ADV_TIMER_TYPE);
		/* set up this adv route structure with head and tail of route */

		p_advrt->routes[id].head = p_idrp_rt;
		p_advrt->routes[id].tail = p_idrp_rt;

		/* set up this time's advertisment structure 
		 * - if peer has advertisement timer
	 	 *   set-up for this situation
		 */

		if (peer->p_minadv_head)
			{
			peer->p_minadv_tail->p_next = p_advrt;
			peer->p_minadv_tail = p_advrt;
			}
		else
			{
			peer->p_minadv_head = p_advrt;
			peer->p_minadv_tail = p_advrt;
			}
		}
	else
		{
		/* already set-up for routes on this interval
	 	 * - so just add routes to the current min adv timer structure 
		 */ 
	
		p_advrt = peer->p_minadv_tail;
		if (p_advrt->routes[id].tail)
			{
			p_advrt->routes[id].tail->p_min_adv = p_idrp_rt;
			p_advrt->routes[id].tail = p_idrp_rt;
			}
		else
			{
			p_advrt->routes[id].head = p_idrp_rt;
			p_advrt->routes[id].tail = p_idrp_rt;
			}
		}
	p_idrp_rt->p_min_adv = (idrpRoute *) 0; 

}


idrpAdvRt	*p_advrt;	
int
advmin_interval_calc(p_tp,peer,type)
struct timeval	*p_tp;
idrpPeer	*peer;
int		type;
{

	/* this code returns a delta if a timer is running 
	 * or the full value if the timer is not
	 */

	if (type == IDRP_MIN_ADV_TIMER_TYPE)
		{		 
		if (peer->min_adv_started)
			{
			return(p_tp->tv_sec - peer->min_adv_started);
			}
		/* if the timer is not already running - return min adv time */

		return(IDRP_MIN_ADV_TIME);

		}
	if (type == IDRP_MIN_ADVRD_TIMER_TYPE)
		{
		if (peer->min_advRD_started)
			{
			return(p_tp->tv_sec - peer->min_advRD_started);
			}

		/* if the timer is not already running - return min adv time */

		return(IDRP_MIN_ADVRD);
		}
	return (0);
}

/* set the timers by peer task for the routes associated with that peer */

int
idrp_set_min_route_timer()
{
idrpPeer	*peer;
struct	timeval	tp;
struct timezone	tzp;
		
	if (gettimeofday(&tp,&tzp))
		{
		trace (TR_IDRP,LOG_ERR,"idrp_set_min_route gettimeofday failed");
		return (FALSE);
		}

		
	IDRP_LIST(peer)
		{
		/* check to see if the min advertisement timer should be running */
		if (peer->start_min_adv)
			{
			if (!peer->min_adv_started)
				{
				/* timer is not running start it for this peer 
				 * zero out interval that starts this one 
				 */
 
				bcopy(&tp.tv_sec,&peer->min_adv_started,sizeof(time_t));
				timer_set(TASK_TIMER(peer->task,IDRPTIMER_MIN_ADV),peer->p_minadv_head->interval,(time_t) 0);
				peer->p_minadv_head->interval = 0;	
				}
			}
		} IDRP_LIST_END
	return (TRUE);
}
		

void
idrp_process_minadv(tip,interval)
timer	*tip;
time_t	interval;
{
idrpPeer	*peer;
idrpRoute	*p_irt;
idrpAdvRt	*p_adv;
idrpAdvRt	*p_adv_last = 0;
idrp_ann_list	ann_list;
int		i;
int		changes = 0;
int		cnt = 0;

	peer = (idrpPeer *) tip->timer_task->task_data;
	bzero(&ann_list,sizeof(ann_list));
	rt_open(peer->task);
	p_adv = peer->p_minadvRD_head;
	while (p_adv)
		{
		if (p_adv->interval)
			{
			break;
			}
		NLRI_FAMILY_LOOP(i)
			{
			for (p_irt = p_adv->routes[i].head; p_irt; p_irt->p_min_adv)
				{
				changes += idrp_min_adv_rt(p_irt,&ann_list);
				cnt++;
				}
			}
		p_adv_last = p_adv;
		p_adv = p_adv->p_next;
		task_mem_free(peer->task,p_adv_last);
		}	
	 
	rt_close(peer->task,&peer->gw,changes,NULL);
	
	/* must send the routes to the peers */

	idrp_send_phase3_routes(&ann_list);
		
	if (p_advrt)
		{
		peer->p_minadv_head = p_advrt;
		timer_set(TASK_TIMER(peer->task,IDRPTIMER_MIN_ADV),p_advrt->interval,(time_t) 0);
		p_advrt->interval = 0;
		}
	else
		{
		peer->p_minadv_head = peer->p_minadv_tail = (idrpAdvRt *) 0;
		peer->start_min_adv = FALSE;
		}
}	

int
idrp_min_adv_rt(p_idrp_rt,p_ann_list)
idrpRoute	*p_idrp_rt;
idrp_ann_list	*p_ann_list;
{
int		changes = 0;

	/* min route advertisement
	 * - status:
	 *
	 * 1) clear the IDRP_STATUS_MIN_ADV_RUN flag
	 * 2) if route delete - IDRP_STATUS_DELETE
	 *	pull the route and delete 
	 *
	 * 3) if route changed - announce the last change
	 * 	link to the 
 	 * 
	 *	
	 */
	 
	IDRP_STATUS_BIT_CLEAR(p_idrp_rt,IDRP_STATUS_MIN_RDADV_RUN);

	if (p_idrp_rt->status & IDRP_STATUS_EXT_INFO != 0  || p_idrp_rt->route_id_in == IDRP_EXT_INFO_ROUTE_ID)
		{
		/* is this ext_info  route to be deleted ? */
 
		if (p_idrp_rt->status & IDRP_STATUS_DELETE)
			{
			/* delete external info for this route 
			 * - I'm not sure about the only two status for this
			 * - being IDRP_STATUS_WITH or IDRP_STATUS_DELETE_SEND
			 * - must go through this as we delete IS-IS routes
			 * - or IP routes to review the logic 
			 */
			trace (TR_IDRP,0,"peer %s deleted route %s min_adv process", 
				p_idrp_rt->peer->name,iso_ptoa(&p_idrp_rt->nlri)); 
			if (p_idrp_rt->p_rt_ext_info)
				{
				rtbit_reset(p_idrp_rt->p_rt_ext_info,idrp_this_node.task->task_rtbit);	
				}
			idrp_del_rt_gated(p_idrp_rt);	
			idrp_free_nlri_att_rec(p_idrp_rt,NLRI_LINKED_P_ANN);
			idrp_free_idrpRoute(p_idrp_rt);
			changes++;
			}

		if (p_idrp_rt->status & IDRP_STATUS_MIN_ADV_CHG)
			{	
			/* changed so re-advertise this route
			 *
			 */
			link_ann_list(p_idrp_rt,p_ann_list,NLRI_LINKED_P_ANN);
			trace (TR_IDRP,0,"peer %s deleted route %s min_adv process", 
				p_idrp_rt->peer->name,iso_ptoa(&p_idrp_rt->nlri)); 
			}
		}
	return(changes);

}	

/* idrp_add_rt_min_advRD -
 * -  Add route to minimum advertisement timer
 */
 
void
idrp_add_rt_min_advRD(p_idrp_rt) 
idrpRoute	*p_idrp_rt;
{
idrpAdvRt	*p_advrt;	
idrpPeer	*peer;
struct	timezone tzp;
struct	timeval	tp;	
int	id;

	/* set flag to indicate that minimum  advertisement timer is running */

	IDRP_STATUS_BIT_SET(p_idrp_rt,IDRP_STATUS_MIN_RDADV_RUN);

	/* look to see if the minimum route advertisement set-up for this
	 * phase processing is there yet 
  	 */

	peer = p_idrp_rt->peer;
	id = family_to_nlri_id(p_idrp_rt->family);
	if (!peer->start_min_advRD)
		{
		/* No flag to start min route Advertisement timer
		 *  set up this flash's list 
		 * set-up the flags
		 * - allocate the timer structure
		 * - set the interval 
		 */
 
		peer->start_min_advRD = TRUE;
		p_advrt = (idrpAdvRt *) task_mem_malloc(peer->task,sizeof(idrpAdvRt));	
		/* zero out the structure so we don't have to
		 * worry what things get set - always zero
		 */
 
		bzero(p_advrt,sizeof(idrpAdvRt));

		if (gettimeofday(&tp,&tzp))
			{
			trace (TR_IDRP, LOG_ERR,"idrp used gettime day  and failed, dropped min_adv_rt request ");
			return;
			}

		/* translate time of day here to the seconds to calculate next interval */

		p_advrt->interval = advmin_interval_calc(&tp,peer->min_adv_started,IDRP_MIN_ADVRD_TIMER_TYPE);
		p_advrt->routes[id].head = p_advrt->routes[id].tail = p_idrp_rt;

		/* set -up the peer's min advterisment for this RD timer list */
		if (peer->p_minadvRD_head)
			{
			/* mininum advRD list already set-up just add this
			 * node to the end
			 */
		
			peer->p_minadvRD_tail->p_next = p_advrt;
			peer->p_minadvRD_tail = p_advrt;
			}
		else
			{
			/* minimum advRD list not set-up - create the list
			 */
			peer->p_minadvRD_head = p_advrt;
			peer->p_minadvRD_tail = p_advrt;
			}
			
		}
	else
		{	
		/* this is the normal call for
		 * the min_adv timer - just link the route at
		 * the end of the timer
		 */

		p_advrt = peer->p_minadvRD_tail;
		if (p_advrt->routes[id].tail)
			{
			p_advrt->routes[id].tail->p_min_advRD = p_idrp_rt;
			}
		else
			{
			p_advrt->routes[id].head = p_advrt->routes[id].tail = p_idrp_rt;
			}
		}

	p_idrp_rt->p_min_advRD = (idrpRoute *) 0;

}



/* set the minimum advertisement for this RD 
 * timer.  The only idrp peer structure affected
 * is the local node
 */ 

int
idrp_set_min_advRD_timer()
{
/* set the minimum advertisement for this RD 
 * for now we set this timer only on the local
 * node's peer structure 
 */

struct	timeval	tp;
struct timezone	tzp;
idrpPeer	*peer = &idrp_this_node; 	/* for now this nodes */
		
	if (gettimeofday(&tp,&tzp))
		{
		trace (TR_IDRP,LOG_ERR,"idrp_set_min_route gettimeofday failed");
		return (FALSE);
		}

	if (peer->start_min_advRD)
		{
		if (peer->min_advRD_started)
			{
			/* timer is not running start it for this peer 
			 * zero out interval that starts this one 
			 */
 
			bcopy(&tp.tv_sec,&peer->min_advRD_started,sizeof(time_t));
			timer_set(TASK_TIMER(peer->task,IDRP_MASTER_TIMER_MIN_ADVRD),idrp_this_node.min_advRD_time,(time_t) 0);
			peer->p_minadvRD_head->interval = 0;	
			}
		}
	return (TRUE);
}
		

void
idrp_process_minadvRD(tip,interval)
timer	*tip;
time_t	interval;
{
idrpPeer	*peer;
idrpAdvRt	*p_adv;
idrpAdvRt	*p_adv_last;
idrp_ann_list	ann_list;
int		changes = 0;
int		cnt = 0;
int		i;
idrpRoute	*p_irt;	

	peer = (idrpPeer *) tip->timer_task->task_data;

	/* for now the only peer processing this RD is the
	 * local node peer structure
	 * - the future depends on the route server function 
	 */


	/* process each of the routes to see if the route awaits
	 * - change of status 
 	 *	to delete
	 *	to new status change
 	 */

	trace (TR_IDRP,0," Minimum Route Advertisement per this RD is being processed ");     

	rt_open(idrp_this_node.task);
	p_adv = peer->p_minadvRD_head;
	while (p_adv)
		{
		if (p_adv->interval)
			{
			break;
			}
		NLRI_FAMILY_LOOP(i)
			{
			for (p_irt = p_adv->routes[i].head; p_irt; p_irt->p_min_advRD)
				{
				changes += idrp_min_advRD_rt(p_irt,&ann_list);
				cnt++;
				}
			}
		p_adv_last = p_adv;
		p_adv = p_adv->p_next;
		task_mem_free(peer->task,p_adv_last);
		}	
	rt_close(idrp_this_node.task,&idrp_this_node.gw,changes,NULL);
	
	trace (TR_IDRP,0,"peer %s Minimum Route Advertisement for this RD process %d routes ", 
			peer->name,cnt);
	trace (TR_IDRP,0,"peer %s MinAdvRD had %s routing table changes ",peer->name, changes);

	/* idrp send the phase tree routes sent out */

	idrp_send_phase3_routes(&ann_list);

	/* check to see  if we need to restart the timers
  	 */
	
	if (p_adv)
		{
		peer->p_minadv_head = peer->p_minadv_tail = (idrpAdvRt *) 0;
		trace (TR_IDRP,0,"peer %s has no more on minadvRD timer list",peer->name);
		return;
		}
	/* A new structure - does it have an interval timer? 
	 */ 
	if (p_adv->interval)
		{
		peer->p_minadv_head = p_adv;
		timer_set(TASK_TIMER(peer->task,IDRP_MASTER_TIMER_MIN_ADVRD),p_adv->interval,(time_t) 0);
		p_adv->interval = 0;
		return;
		}
	
	trace(TR_IDRP,0,"peer %s AdvMinRD timer is zero interval w/out catching earlier - task quit it",
		peer->name);
	task_quit(0);
	
}

int
idrp_min_advRD_rt(p_idrp_rt,p_ann_list)
idrpRoute	*p_idrp_rt;
idrp_ann_list	*p_ann_list;
{
int	changes;
	/* minimum route advertisement per RD
	 * - look for change flag with with adv_chng flag set
	 * - only one of the minimum route advertisement
	 * - flags should be set at any one time
	 * 
	 * 1) clear the IDRP_STATUS_MIN_RDADV_RUN flag
	 * 1) if route delete - pull the EXT_INFO or local route
	 *	test for flags EXT_INFO &  
	 *	delete flags: *	IDRP_STATUS_DELETE_SEND or
	 *	      - IDRP_STATUS_DELETE
	 *
	 * 2) if route changed - announce the last change
	 *	EXT_INFO or local route and  
	 *	IDRP_STATUS_CHG_MIN_ADV -
	 * 	- announce the current route to world
	 *	  using the idrp send routines
 	 * 
	 *	
	 */

	IDRP_STATUS_BIT_CLEAR(p_idrp_rt,IDRP_STATUS_MIN_RDADV_RUN);

	if (p_idrp_rt->status & IDRP_STATUS_EXT_INFO != 0  || p_idrp_rt->route_id_in == IDRP_EXT_INFO_ROUTE_ID)
		{
		/* is this ext_info  route to be deleted ? */
 
		if (p_idrp_rt->status & IDRP_STATUS_DELETE)
			{
			/* delete external info for this route 
			 * - I'm not sure about the only two status for this
			 * - being IDRP_STATUS_WITH or IDRP_STATUS_DELETE_SEND
			 * - must go through this as we delete IS-IS routes
			 * - or IP routes to review the logic 
			 */
	
			if (p_idrp_rt->p_rt_ext_info)
				{
				rtbit_reset(p_idrp_rt->p_rt_ext_info,idrp_this_node.task->task_rtbit);	
				}
			idrp_del_rt_gated(p_idrp_rt);	
			idrp_free_nlri_att_rec(p_idrp_rt,NLRI_LINKED_P_MIN_ADV_RD);
			idrp_free_idrpRoute(p_idrp_rt);
			trace (TR_IDRP,0,"peer %s delete idrp_minadvRD EXT_INFO local route %s",iso_ptoa(&p_idrp_rt->nlri));
			changes++;
			}

		if (p_idrp_rt->status & IDRP_STATUS_MIN_ADV_CHG)
			{	
			/* changed so re-advertise this route
			 *
			 */
			link_ann_list(p_idrp_rt,p_ann_list,NLRI_LINKED_P_ANN);
			trace (TR_IDRP,0,"peer %s announce idrp_minadvRD EXT_INFO local route %s",iso_ptoa(&p_idrp_rt->nlri));
			}
		return(changes);
		}

	if (p_idrp_rt->route_id_in == IDRP_LOCAL_ROUTE_ID)
		{
		if (p_idrp_rt->status & (IDRP_STATUS_DELETE | IDRP_STATUS_DEL_SEND))
			{
			/* delete local route */

			idrp_del_rt_gated(p_idrp_rt);	
			idrp_free_nlri_att_rec(p_idrp_rt,NLRI_LINKED_P_MIN_ADV_RD);
			idrp_free_local_rt(p_idrp_rt);
			changes++;	
			trace (TR_IDRP,0,"peer %s delete idrp_minadvRD local Route %s",iso_ptoa(&p_idrp_rt->nlri));
			}
		if (p_idrp_rt->status & IDRP_STATUS_MIN_ADV_CHG)
			{
			link_ann_list(p_idrp_rt,p_ann_list,NLRI_LINKED_P_ANN);
			trace (TR_IDRP,0,"peer %s announce idrp_minadvRD local Route %s",iso_ptoa(&p_idrp_rt->nlri));
			}
		}
	return(changes);
	
}
