#include "include.h"
#include "idrp.h"

/* idrp_rt_policy.c - 
 * 
 *  routines that interact with policy to 
 *  do
 *  1) preference calculation -
 *	gated import function
 *	PREF - from IDRP	
 *
 *  PREF(p_idrp_rt) - do import policy on the idrp - not in service now
 * 
 *  2) Distribution policy for routes:
 * 	DIST(p_idrp_rt,peer)
 *
 *  3) DIST_LIST_INCL and DIST_LIST_EXCL
 *	for attribute records:
 * 
 *	DIST_ATTR(p_att,peer) 
 *
 *  4) export policy for gated
 *
 * %% distribution policy routines  	 
 * 
 *  DIST(idrpRoute,peer) - run policy on a NLRI and peer 
 *  DIST_ATTR(p_att,peer) - run policy on Attribute list
 * 
 * 5) Any outbound aggregretion policy
 *    hopefull none of this will be idrp specific
 *
 *  6)  preference match mediation -
 *	idrp_vs_proto_pref(proto,peer)
 * 	given protocols with route with same preference
 *	as idrp route, should protocols route take
 *	preference for peer
 *	- use preference table in policy structures
 *	eventually, now hard coded routine for all peers	 
 */
 
/* routine is hardwired for now until
 * idrp_rt_policy structures   
 * - TRUE - prefer IDRP
 * - FALSE - prefer other protocol 
 */


int idrp_proto_pref(proto_rcv,peer)
u_int		proto_rcv;
idrpPeer	*peer;
{

	switch (proto_rcv)
		{
		case RTPROTO_IDRP:
		   trace (TR_IDRP,0,"idrp_vs_proto_pref called with IDRP as protocol - peer %s", peer->name);	
		   return(FALSE);	

		case RTPROTO_BGP:
			return (TRUE);
			break;

		case RTPROTO_ISIS:
		default:
			return (FALSE);
			break;
		}
}
 
/* preference on a route */

int PREF(p_rt,p_idrp_rt)
rt_entry 	*p_rt;
idrpRoute 	*p_idrp_rt;
{
	/* here's where the gated policy stuff will go
	 * along with the idrp policy structures for import
	 * policy will take the gateway received from
	 * plus the idrp policy 
	 */

	if (p_idrp_rt)
		{
		trace (TR_IDRP,0,"PREF called and no policy basic there for idrp peer %s route %s", 
			p_idrp_rt->peer->name,iso_ptoa(&p_idrp_rt->nlri));
		}
	
	return (0);

}	


int
DIST(p_idrp_rt,peer,type)
idrpRoute	*p_idrp_rt;
idrpPeer	*peer;
int		type;
{

	switch(type)
		{
		case IDRP_DIST_PEER_UP_DUMP:
		case IDRP_DIST_REFRESH:
			if (p_idrp_rt->peer == peer)
				{
				return(FALSE);
				}

		case IDRP_DIST_FLASH:
			/* here are simple distribution stuff
			 *
			 */

			if (peer->type == IDRP_PEER_INTERNAL)
				{
				/* don't send any internal routes
				 * to another peer
				 * - cut out redistribution here
				 */

				if (p_idrp_rt->peer->type == IDRP_PEER_INTERNAL)
					{
					return(FALSE);
					}
				}
			if (peer == p_idrp_rt->peer)
				{
				/* Do not send routes back to peer we learned
				 * them from.  A bad idea, although BGP does it
				 */ 
				return (FALSE);
				}	
			break;
		}
	return(TRUE);
}

idrp_attribute_record *
DIST_ATTR(p_att,peer)
idrp_attribute_record	*p_att;
idrpPeer		*peer;
{
	/* attribute list will be modified for this peer
	 * for DIST_LIST_INCL or DIST_LIST_EXCL 
	 * for idrp
	 */

	trace (TR_IDRP,0,"DIST_ATTR called - no code ");
	return (p_att);
}

void
AGGR(peer,p_ann_list)
idrpPeer	*peer;
idrp_ann_list	*p_ann_list;
{
	trace (TR_IDRP,0,"AGGR called to aggregated for peer %s ",peer->name);
}	
