#include "include.h"
#include "idrp.h"
#include <stdio.h>
#include "md4global.h"	/* RSAREF globals and constants */
#include "md4.h"	/* md4-specific includes */


/* idrp_validate.c
 *  
 * pdu validation routines
 *	checksum_ok(peer,pdu,length) - validate checksum for recvd pdu
 *      idrp_rib_id(att_mask) - calculate internal rib id from
 * 	              		idrp attribute record mask bits 
 * 	valid_header(peer,pdu,length) - validate header 
 */ 
  
 

/*
 * checksum_ok(peer,pdu,length)
 *    -  validate the received checksum
 * return - TRUE/FALSE
 */
int checksum_ok(peer,pdu,length)
idrpPeer *peer;
idrpPdu *pdu;
int length;
{
/* check which type of checksum we are working on */
u_char checksum_pdu[16];
u_char digest[16];
u_int *p_checksum;
u_int return_code = FALSE;

	/* md4 zeros out the digest.    */

	switch (peer->AuthType) 
		{
		case IDRP_AUTH_CODE_TEST:
			return_code = TRUE;
			break;
	
		case IDRP_AUTH_CODE_CLEAR:
			bcopy(&pdu->header.validation,&checksum_pdu,sizeof(pdu->header.validation));
			bzero(&pdu->header.validation,sizeof(pdu->header.validation));
			md4((u_char *)pdu,length,digest);
			if (!bcmp(&checksum_pdu, &digest, 16))
				{
				return_code = TRUE;
				}	 
			break;

		case IDRP_AUTH_CODE_CRYPT:
			trace(TR_IDRP,0,"Peer %s trying to do encrypt - no support",peer->name);
			break;

		case IDRP_AUTH_CODE_SIMPLE:
			trace(TR_IDRP,0,"Peer %s trying to do simple validation - no support",peer->name);
		 	break;	
		}

return(return_code);   
}


int idrp_rib_id(att_mask)
int att_mask;
{
	/* for now we only support default */
	/* so warn if any distinquishing attributes */
	/* are there */

	if (att_mask & IDRP_ATTBT_DISTINGUISH) 
		{
		/* we have non-zero attribute - return -1 */
		/* for no support  */

		return (IDRP_ATTBT_UNSUPPORTED);
		}
	return (0);

}


int valid_header(peer,pdu,length)
idrpPeer *peer;
idrpPdu *pdu;
int	length;
{

/* Below are checks from section 7.21.1 from the IDRP document.
 * These checks will result in the pdu being discarded
 * and the HEADER error event being generated 
 */
 
	switch(pdu->header.pdu_type)
		{
		case IDRP_OPEN_TYPE:
			if (length < IDRP_MIN_OPEN_PDU_SIZE || 
				length > IDRP_MAX_OPEN_PDU_SIZE)
				{
				trace(TR_IDRP,0,"failed open type header check peer %s",peer->name);
				return(FALSE);
				}
			else
				{
				return (TRUE);
				}	
			break;
			
		case IDRP_UPDATE_TYPE:
			if (length < IDRP_MIN_UPDATE_SIZE)
				{
				return (FALSE);
				trace(TR_IDRP,0,"failed update type header check peer %s",peer->name);
				}
			break;

		case IDRP_KEEPALIVE_TYPE:
			if (length != IDRP_KEEPALIVE_SIZE)
				{
				trace(TR_IDRP,0,"failed keepalive type header check peer %s",peer->name);
				return(FALSE);
				}
			break;

		case IDRP_ERROR_TYPE:
		case IDRP_CEASE_TYPE:
		case IDRP_RIB_REFRESH_TYPE:
#ifdef	ECHOKLUDGE
		case IDRP_ECHO_TYPE:
		case IDRP_ECHO_REPLY_TYPE:
#endif
			return(TRUE);
			break;

		default:
			/* the unknown type */
			trace(TR_IDRP,0,"failed unknown type header check peer %s",peer->name);
			return (FALSE);
			break;
		}

	/* 7.21.1 a) check that the PDU length does not exceed maximum size 
	 * that peer is suppose to send
 	 */

	if (length > peer->pdu_maxsendsize)
		{
		trace(TR_IDRP,0,"peer %s failed pdu over max pdu size ",peer->name);
		return (FALSE);
		}
	return(TRUE);

}
