#ifndef _idrp_h
#define _idrp_h

#include <netiso/clnp.h>
#include "osi.h"
#include "idrp_proto.h"

/* other includes at bottom */

#define	IDRP_ASPATH		TRUE	/* DEFINE to be AS path functional for now */			
					/* clever configuration needs to make this optional */
#define	IDRP_AS_SIZE		2	/* two octets for AS number */ 

#define IDRP_PORT 10747			/* Bogus UDP port number */
#define IDRP_PKTSIZE 200		/* Bogus packet size for now */
#define IDRP_HOLD_TIME 30		/* Croak if the this timer expires */
#define IDRP_MAX_CREDIT 5		/* Maximum credit offered */
#define IDRP_REXMIT_TIME 2		/* Rexmit timeout value */
#define IDRP_CLOSE_WAIT_TIME 30		/* Time to sit in close_wait state */
#define IDRP_START_TIME 10		/* Time to wait before initial start event */
#define	IDRP_MIN_ADV_TIME 	30	/* 30 seconds for mininum advertisement */	


#define	IDRP_TSI_SIZE	1		/* TSI size */

#ifdef ECHOKLUDGE
#define IDRP_ECHO_TIME 5		/* Time to wait before calling echo function %% */
#define ECHO_MAX 10			/* Max number of echos to send */
#endif

#define	IDRP_MAX_LOCAL_ROUTES	30	/* Maximum local routes */	
#define IDRP_MAX_PEERS 10 		/* Max number of peers */
#define IDRP_HASH_TABLE_SIZE 50	/* hash table size */
#define	IDRP_SMALL_BLOCK_SIZE	128	/* small block size */

/* here I need to add some cute assign stuff to make sure idrp_small_block = 
 * the greater of idrp_canon_rd_path, and idrpRoute_entry 
 * - John can help I think  
 */ 

#define	IDRP_MEDIUM_BLOCK_SIZE sizeof(idrp_attribute_record)+50 /* medium block size */

/* medium size block must be bigger than attribute record 
 * and the idrpRoute block - watch this
 * again some clever assign stuff needs to insure that 
 */
 
#define	IDRP_BIG_BLOCK_SIZE   sizeof(idrpPeer)   /* big block size */

/* policy definitions for now */

#define IDRP_METRIC_NONE 255		/* No metric in IDRP */

/* prefix Address comparison  */

#define	PFX_ADDR_EQUAL		0		/* addr1 = addr 2 */
#define PFX_ADDR_LESS_THAN	1		/* addr1 < addr 2 */
#define PFX_ADDR_GREATER_THAN	2		/* addr1 > addr 2 */ 	

#define	ISO_ADDR_EQUAL		0		/* addr1 = addr 2 */
#define ISO_ADDR_LESS_THAN	1		/* addr1 < addr 2 */
#define ISO_ADDR_GREATER_THAN	2		/* addr1 > addr 2 */ 	


/*
 *  Buffer structure
 */

typedef struct _idrpBuffer {
    struct _idrpBuffer *next;		/* Forward link */
    u_int seq;				/* Sequence number */
    idrp_proto_header	*pdu_proto;	/* beginning of PDU with protocol header */
    u_int proto_len;			/* length with protocol header */
    idrpPdu *pdu;			/* PDU */
    u_int length;			/* Length of buffer alone */
    u_int sequenced;			/* TRUE if PDU is sequenced (reliable) */
    u_int ref;				/* reference count */
} idrpBuffer;


/*
 * Route redistribution map.  A bit for each peer.
 */
typedef struct {
    u_int mapword[(IDRP_MAX_PEERS-1)/32+1];	/* Bitmap words */
} redist_map;

#define IDRP_MAPWORD(x) ((x)/32)
#define IDRP_MAPMASK(x) (1 << ((x)%32))

/*
 * Attributes on a route - entry per type of
 * idrp attribute 
 */

typedef struct {
	u_char present;			/* TRUE if attribute is present */
	u_char flags;			/* Attribute flags */
	u_short length;			/* Attribute length */
	u_char *data;			/* Pointer to attribute */
} idrp_attribute_entry;

/* here we are currently getting the idrp entry to have the */
/* attribute structure per record.   */

/* canonical IDRP paths */

typedef struct _idrp_canon_rdpath {
	struct	_idrp_canon_rdpath	*p_next;	/* next RD struct */
	struct	iso_net_addr		rdi;		/* RD prefix */
} idrp_canon_rdpath;
	 
/* next hop attribute information saved at top */

#define NO_NEXT_HOP		0x0	/* no next hop specified in pdu */	
#define	NEXT_HOP_AF_INET	0x1	/* next hop - has IP NET */ 
#define NEXT_HOP_AF_ISO		0x2	/* next hop - has AFI.ISO NET */

/* List of idrpRoute structure with IDs */
 
#define	IDRP_NLRI_FAMILIES_SUPPORTED	2 	/* NLRI supported */
#define	IDRP_NLRI_ISO			0	/* NLRI AF_ISO */
#define	IDRP_NLRI_IP			1	/* NLRI AF_INET */	
#define	IDRP_NLRI_SIP			3	/* NLRI AF_INET */	
#define	IDRP_NLRI_PIP			4	/* NLRI AF_INET */	

typedef	struct	_nlri_fam {
	byte	family;
	int	nlri_id;
} nlri_fam;
		


typedef struct _idrpRoute_entry {
	struct _idrpRoute_entry	*p_next;	/* link list of route_id structures */
	int	 		route_id;	/* route-ID */
	int			mem_type;	/* memory type */

#define	IDRP_TASK_MEM		0x0	 /* task memory */
#define	IDRP_LOC_MEM		0x1	 /* memory assigned before task */
 				
	int			loc_pref;	/* preference on this group of route-ids */
	int			status;		/* route-id status */

#define IDRP_ATTR_ROUTE_ID_WITH 0x1		/* deleting this route id */
#define IDRP_ATTR_ROUTE_ID_NORM 0x0		/* normal setting */
#define	IDRP_LOCAL_ROUTE_RECONFIG 0x2		/* reconfiguring this local route_id */

	struct _idrpPeer	*peer;		/* peer for this route_id */			
	struct _idrpRoute	*p_route[IDRP_NLRI_FAMILIES_SUPPORTED];	/* link list of route-id structures */
} idrpRoute_entry; 

#define IDRP_LOCAL_ROUTE_ID	-1		/* local route id is illegal */ 
#define	IDRP_EXT_INFO_ROUTE_ID	-2		/* IDRP ext_info route id */
#define IDRP_INIT_ROUTE_ID	-3		/* idrp init_route_id */

typedef struct _idrpRoute_list {
	idrpRoute_entry		*head;		/* route list head */
	idrpRoute_entry		*tail;		/* route list tail */
} idrpRoute_list;

typedef	struct	 _idrpRoute_link	{
	struct	_idrpRoute	*head;		/* head of idrpRoute list */
	struct	_idrpRoute	*tail;		/* tail of idrpRoute list */
} idrpRoute_link;

typedef	struct	_idrp_rt_chain_walk	{
	struct	_idrp_rt_chain_walk	*p_next;	/* next on list */
	rt_entry			*p_rt;		/* entry for chain */
} idrp_rt_chain_walk;

/* list types */

#define	IDRP_ROUTE_LIST_LINKED		0	/* linked route list */
#define IDRP_ROUTE_LIST_CIRCULAR	1	/* circular linked list */


typedef	struct	_idrpAdvRt {
	struct	_idrpAdvRt	*p_next;	/* next idrpAdvRt struct on list */
	int	interval;			/* second to set this route list */
	idrpRoute_link	routes[IDRP_NLRI_FAMILIES_SUPPORTED];
						/* routes */
} idrpAdvRt;

 
/* idrp for IP special - */
/* contained in idrp route is an AS_PATHS structure */

typedef struct _idrp_attribute_record {
	struct	_idrp_attribute_record 	*next;		/* linked list */
	struct	_idrp_attribute_record 	*prev;		/* linked list */
	int	ref_cnt;				/* reference count */	
	int	mem_type;				/* memory type - TASK or LOCAL */ 
	struct	_idrpPeer		*peer_alloc;	/* peer allocated struct*/	
	idrpRoute_entry			*route_id_list;	/* route id's attach to this attribute record */	
	u_int	rib_id;					/* internal ID of AdjRib */
	u_int	idrp_mask;				/* mask of attribute */
	u_int	local_mask;				/* mask of local route bits */ 

#define	IDRP_LOCAL_RDPATH_RESET_LENGTH	0x1		/* reset length */
#define	IDRP_LOCAL_NO_PDU		0x2		/* no pdu for local node route */
#define IDRP_LOCAL_EXT_INFO		0x4		/* EXT_INFO is local RD */
#define	IDRP_DONT_SEND_NEXT_HOP		0x8		/* don't sent the Next hop */
#define	IDRP_ROUTE_SERVER_FLAG		0x10		/* set route server flag on NEXT HOP */
#define IDRP_DONT_SEND_NEXT_HOP_OFF	0xF7		/*  send NEXT hop */ 

	byte			multi_exit;		/* multi exit discriminator value */
	proto_t			idrp_local_proto;	/* mask of local protocols  */
	sockaddr_un			*next_hop;	/* next hop for route */
	u_int				hopcount;	/* path length */
	idrp_canon_rdpath 		*rd_path;	/* cannonical rep of rd path */
#ifdef	IDRP_ASPATH
	struct _as_path			*as_path;	/* AS path pointer for path */
#endif /* IDRP_ASPATH */
	idrp_attribute_entry attrib[IDRP_ATTR_NO];	 /* Attributes */
	idrp_attribute_entry usr_attrib[IDRP_ATTR_USR_NO]; /* Attributes */
	u_char				*p_attr_pdu;	/* copy of attributes from pdu */
	int				attr_len;	/* length of attributes from pdu */	
	int                             ref_no;         /* a tmp # to reference in dump routines */
} idrp_attribute_record;


	

#define	IDRP_ASPATH_EXT_INFO		0		/* AS PATH for EXT_INFO from route */
#define	IDRP_ASPATH_BGP_XFR		1		/* AS PATH for BGP-4 transfer */


/* Withdrawl ID storage structure  */

#define	IDRP_WITHDRAW_MAX_IDs	(IDRP_PKTSIZE - MinPDULength)/4  

typedef struct {
        u_int   	count;		/* count of route_id's to withdrawl */
        int		withdraw[IDRP_WITHDRAW_MAX_IDs];	/* withdrawl array */
} Withdrawl;


/* Parse results from update pdu array */

typedef struct {
	int		route_id_in;	/* route_id */
	int		pref_rcvd;	/* preference received */
	idrp_attribute_record *p_att;	/* attribute record */
	struct _idrpRoute	*rt_chg[IDRP_NLRI_FAMILIES_SUPPORTED];
					/* list of nlris by family */ 
	Withdrawl	*withdraw; 	/* withdrawl structure */
} parse_results;  
 

/* idrp route id hash table entry */

/* use IDRP_INIT_ROUTE_ID - for initialized hash table */

typedef struct  _idrp_rid_hash {
        int   route_id;               /* route id */       
        struct  _idrp_rid_hash *next;	/* hash chain */ 
        struct _idrpRoute *p_idrp_rt[IDRP_NLRI_FAMILIES_SUPPORTED];
				        /* Route lists per nlri */
} idrp_rid_hash; 


/* idrp last error structure  */

typedef struct {
	u_char	error_code; 	/* error pdu error code */
	u_char	error_subcode;	/* error pdu error subcode */
	u_char	*data;		/* pointer to data */
     	u_int	data_len;	/* length of data */
	u_char	*data2;		/* pointer to secondary data */
	u_int	data2_len;	/* length of secondary data */
} error_info;

#define	SNPA_SIZE	8		/* 8 octet SNPAs */

typedef	struct	snpa_entry {
	struct	_snpa_entry	*p_next;	/* list of SNPAS for this next hop */	
	int	len;				/* length of snpa */	
	u_char	snpa[SNPA_SIZE];		/* array of SNPAs */
} snpa_entry;
	
typedef struct	_snpa_list {
	u_int   pdu_len;	/* length of list in the pdus */	
	u_int	cnt;		/* count of snpa's in list */
	snpa_entry	snpa1;	/* list of snap entryies */
} snpa_list;

 
typedef	struct	_idrpRoute_out {
	struct	_idrpRoute	*p_next;	/* next idrpRoute structure */
	int	route_id;			/* route id */
} idrpRoute_out;
	
/* idrp_best_ext structure
 * We use this track best external routes:  for efficiency, routes to the same dest
 * (nlri) are insertion-sorted in pref_cal order as we receive them, and an idrp_best_ext
 * structure is used to track the route with the best pref_cal for each nlri.
 * We also have invented the concept of a best external routes RIB (BER-Rib).  This 
 * consists of all of the idrp_best_ext structs strung together in unsorted order.  We
 * can use it to dump the entire set of best ext routes when we bring up an IIDRP peer.
 */
typedef struct _idrp_best_ext {
	struct	_idrpRoute	*p_best_ext;	/* link to best extern route for this destination */
	struct	_idrp_best_ext	*p_prev;	/* link to prev best ext struct in BER-Rib */
	struct	_idrp_best_ext	*p_next;	/* link to next best ext struct in BER-Rib */
	int	refcount;
} idrp_best_ext;


/* idrp Route structure  */
/* saved as rt_data (protocol specific data in rt_entry */
/* of gated */

typedef struct _idrpRoute {
	rt_data		rt_header;	/* header for rt_data stuff */
	rt_entry	*p_rt;		/* link to gated routed table */ 
	rt_entry	*p_rt_ext_info;	/* link to gated routed table */
					/* for ext info original entry */
	struct _idrpPeer *peer;		/* Peer from which we received */
					/* this route */
	sockaddr_un	*p_loc_rt_peer;	/* pointer to peer description */
					/* for local route */ 
	idrp_attribute_record *p_attr;	/* Attribute record */
					/* pointers to required attributes */
	int	mem_type;		/* type of memory structure allocated from - local or task */
	int	route_id_in;		/* Route ID received */
	int	pref_cal;		/* calculated idrp preference */
	int	pref_rcvd;		/* preference received from internal peer */

	struct	_idrpRoute	*p_next_nlri;	/* > 1 nlri/route id - link list */
	struct	_idrpRoute	*p_with;	/* next on withdrawl list processing */
						/* phase1 or phase 3 uses */
	struct	_idrpRoute	*p_ann_nlri;	/* announcing nlri */
    	struct	_idrpRoute	*p_min_adv;	/* link for min route advertisement timer */
    	struct	_idrpRoute	*p_min_advRD;	/* link for min route within RD timer */

	struct	_idrpRoute	*p_next_best;	/* link to next-best external route (if this route */
						/* is external to begin with) */
	struct	_idrpRoute	*p_better;	/* link to next-better external route (if this route */
						/* is external to begin with) */
	struct	_idrp_best_ext	*p_p_best_ext;	/* pointer to idrp_best_ext structure.  Reference */
						/* off of idrp_best_ext to find the best ext route */

#define NLRI_LINKED_P_NEXT	0		/* NLRI linked on P_NEXT */
#define	NLRI_LINKED_P_WITH	1		/* NLRI linked on WITH */
#define	NLRI_LINKED_P_ANN	2		/* NLRI LINKED on ANN */
#define	NLRI_LINKED_P_MINADV	3		/* NLRI LINKED on min_adv */
#define	NLRI_LINKED_P_MIN_ADV_RD 4		/* NLRI LINKED on min_adv_rd */
	
	idrpRoute_out	route_out[IDRP_MAX_PEERS];
					/* Array of outbound route ID */
					/* by peer for this route */
					/* also contains list of other */
					/* nlri announced in route-id */
					/* nlri linked - circular linked list */
	struct iso_prefix nlri;		/* NLRI stored in prefix format */ 
	int	family;			/* family route belongs to */
	int	nlri_id;		/* nlri id for to */
	int	status;			/* status */

#define	IDRP_STATUS_ADJ_RIB		0x1	/* idrp Route AdjRib */
#define	IDRP_STATUS_BEST_EXT		0x2	/* best exterior route */
#define	IDRP_STATUS_LOC_RIB		0x4	/* idrp loc_RIB */
#define	IDRP_STATUS_MIN_ADV_RUN 	0x8	/* min advertisement timer running */
#define	IDRP_STATUS_EXT_INFO		0x10	/* route entry handles ext_info */
					/* annoumeent for idrp */
#define IDRP_STATUS_WITH		0x20	/* withdraw set on this route */
#define	IDRP_STATUS_DEL_SEND 		0x40	/* delete after send route to peer */
#define	IDRP_STATUS_MIN_ADV_CHG		0x80	/* change during min advertisement*/
#define	IDRP_STATUS_REPLACE 		0x100	/* withdrawl/replace */
						/* pdu withdraw id/announce */
						/* timer duration for this route */
#define	IDRP_STATUS_MIN_RDADV_RUN	0x200	/* minimum advertisement of nlri from this RD */
#define IDRP_STATUS_DELETE		0x400	/* delete this route */
#define IDRP_STATUS_RECONFIGURE		0x800	/* reconfigure this local route */
#define	IDRP_STATUS_LOCAL_NEW_ROUTE	0x1000	/* local new route */
#define	IDRP_STATUS_WITH_EARLY_PROC	0x2000	/* early processing of withdraw nlri with others */
						/* of the same route_id outbound */
#define	IDRP_STATUS_LOCAL_ROUTE		0x0	/* local route should have no status bits set */

#define IDRP_STATUS_INT_MASK	(IDRP_STATUS_ADJ_RIB | IDRP_STATUS_BEST_EXT | IDRP_STATUS_LOC_RIB) 
} idrpRoute;

typedef struct _RD_path {
	idrp_path_segment_header	rd_header;	/* this nodes path */	
	struct iso_addr			rdi;		/* rdi value */ 
} RD_path;

#define RDI_LENGTH_BYTE_LEN		1

typedef struct _idrp_send_list {
	struct	_idrp_send_list	*p_next;	/* next structure on send list */
	idrp_attribute_record 	*p_attr;	/* attr record assoc */
	int	rib_id;				/* rib id of this list */
	u_int	pdu_bytes;			/* pdu bytes used nrli and withdraw id */
	u_int	pdu_bytes_attr;			/* pdu bytes used for attribute */
	byte	loc_pref;			/* local preference */
	idrpRoute_link	ann_nlri[IDRP_NLRI_FAMILIES_SUPPORTED];  
	Withdrawl	*withdraw;		/* withdraw structure */
} idrp_send_list;
		

typedef struct _idrp_ann_list {
	struct	_idrp_ann_list	*p_next;	/* next structure on announce list */
	int	rib_id;				/* rib id of this list */
	idrp_attribute_record 	*p_attr;	/* attr record assoc */
	struct	_idrpPeer	*peer;		/* peer received from */ 
	idrpRoute_link	ann_nlri[IDRP_NLRI_FAMILIES_SUPPORTED];  
	idrpRoute_link	with_nlri[IDRP_NLRI_FAMILIES_SUPPORTED];
} idrp_ann_list;
	
#define	RIB_ID_DEFAULT	0		/* default RIB */


typedef struct _idrpRefresh {
	struct _idrpRefresh	*p_next;	/* linked list */
	idrpPdu		*p_pdu;		/* pdu */ 
	parse_results	*results;	/* results array */
					/* for update parsing */
} idrpRefresh;  

/* idrp RIB REFRESH structure for peer */

typedef	struct {
	u_int	seq_start;	/* sequence number of RIB REFRESH Start */
	idrpRefresh	*pdu_list_start;	/* pdu list for UPDATEs */
	idrpRefresh	*pdu_list_end;		/* pdu list for Updates */	
	u_int	seq_stop;	/* sequence number of RIB REFRESH Stop */
	u_int	cnt;		/* count of pdus to process */  
	u_int	rib_no;		/* internal rib id */
} refresh_info;	 
		
 

/*
 *  IDRP peer structure
*/

typedef struct _idrpPeer {
    struct _idrpPeer *next;        	/* Pointer to next idrpPeer table */
    u_int type;				/* internal or external neighbor */
#define IDRP_PEER_EXTERNAL	0	/* external peer */
#define IDRP_PEER_INTERNAL	1	/* internal peer */
#define IDRP_PEER_TEST		2	/* debugging peer */
#define IDRP_PEER_LOCAL		3	/* local node's peer structure */

    u_int	proto_sock;		/* protocol IDRP runs over */
    u_int	last_proto_sock;	/* last protocol socket peer structure re-inited */

/* definitions for socket */
#define	IDRP_PDU_PROTO_IP_RAW	0	/* IP PID used for PDU transmission */
#define	IDRP_PDU_PROTO_UDP	1	/* IP UDP used for PDU transmission */
#define	IDRP_PDU_PROTO_CLNP	2	/* CLNP used for PDU transmission */
#define	IDRP_PDU_PROTO_IDRP	3	/* CLNP IDRP socket used for PDU transmission */

    struct iso_net_addr	neighbor;	/* Neighbor's address ISO */
    struct iso_net_addr iso_gw_addr;	/* ---skh to be change iso gw until we meld neighbor */	
    struct iso_net_addr iso_intf_addr;	/* Interface address for iso */ 
					/* IP address of the neighbor is kept in the gw_entry struct */
#define ip_neighbor	gw.gw_addr
    sockaddr_un		*ip_intf_addr;

    struct iso_net_addr	rdi;		/* Neighbor's RDI is always an ISO Address */
					/* be it IP or ISO */ 
#define rdi_addr	rdi.isoa_genaddr
#define rdi_length	rdi.isoa_len


    u_int	version;		/* version number */
    snpa_list	snpa;			/* SNPAs for this peer reception */ 
					/* RDC goes here */
    u_int	AuthType;		/* Authentication type */ 
    u_char	AuthCode[IDRP_AUTH_CODE_SIZE];  	/* Authentication code for this peer */
    u_int       Capacity;               /* capacity */
    u_int       rdLRE;                  /* configured error rate */
    u_int	Expense;                /* expense */
    u_int	Priority;		/* min priority to forward PDU */    
    u_short	pdu_maxsendsize;	/* Max PDU size to send */
    u_short	pdu_maxrcvsize;		/* Max PDU size to receive */
    u_int	listen_for_open;	/* listen for open in close state */
    u_int	max_open_sent;		/* max number open to send in seq */
					/* before going into close wait */ 
    u_int	multi_exit;		/* multi_exit set on neighbor */
    u_int	route_server;		/* router server allowed function */
 	
    u_int	nlri_supp[IDRP_NLRI_FAMILIES_SUPPORTED];	/* nlri supported */
								/* filled with AF_xxx */
							        /* of protocol supported */
/* policy information */
/* needs to be added */ 
	

/* stateInfo */

    u_int state;                   	/* Protocol State */
    u_int send_next;			/* Seq of next PDU to send */
    u_int upper_window_edge;		/* Seq+1 of last PDU we can send */
    u_int send_unack;			/* Seq of oldest unacked PDU */

    u_int rcv_seq_expected;		/* Expected seq of next PDU received */
    u_char rcv_credit;			/* Max outstanding PDUs to receive */
    u_char rcved_credit_avail;		/* Last received credit_available */
    u_int cred_avail_zero_seq;		/* Sequence # of last seqed PDU with cred_avail = 0 */
    u_int cred_avail_zero_seq_last;	/* Previous value of above */
    refresh_info	refresh;	/* RIB REFRESH information */

/* Counters */

    u_int	open_sent;		/* open bispdus sent since start event */
    u_int	updatesIn;		/* Update pdu received from peer */
    u_int	updatesOut;		/* Update pdu send to peer */
    u_int	totalBISPDUIn;		/* total BISPDUs received from peer */
    u_int	totalBISPDUOut;		/* total BISPDUs sent to peer */
    u_int	keepAlives_Since_last_Update;
					/* keepalives from peer */
 
/* information on AdjRibIn */ 
 
    u_int checksum;			/* checksum for AdjRibIn for peer */ 
    u_int AdjRibBad;			/* indication that AdjRibIn is bad */

/* general  helps for gated */
    flag_t idrp_flags;                 	/* Protocol Flags */
    u_int options;                 	/* User specified options */
    task *task;                    	/* Pointer to task */
    flag_t trace_flags;            	/* Trace flags from parser */

#define	IDRP_TRACE_PDUS	0x01		/* trace pdus send/xmt */

    char name[61];              	/* Name of this peer */
    gw_entry iso_gw;			/* iso gateway for this peer */ 
    gw_entry gw;                   	/* GW block for this peer */
 					/* note IP address kept in gw block */
					/* policy linked in gated structures */ 

/* idrp_addr = gw.gw_addr  - changed here to make sure I catch all the address assignments */
#define idrp_import     gw.gw_import
#define idrp_export     gw.gw_export
#define idrp_time       gw.gw_time


/* route ID out information */ 

    u_int route_id_out;			/* current route_id_out */
    u_int  id;				/* Peer id in idrpRouteout */

/* route ID inbound hash table */
    struct	_idrp_rid_hash	hash_table[IDRP_HASH_TABLE_SIZE];
					/* hash table for route_id in */ 
/* debugging for idrp */

    error_info	last_error_pdu;		/* error codes last ERROR pdu sent */ 
					/* used in parsing of pdus */

/* timers */
/* switches for  timers running */

    u_char rexmit_timer_running;	/* NZ if rexmit timer is running */
    u_char start_timer_running;		/* NZ if start timer is running */

/* time values for the timers */

    time_t max_rib_integ_check;		/* rib checksum time */	
    time_t ribcheck_timer;		/* rib checksum timer */	
    time_t keepalive_time;		/* Send keepalives more often than this */
    time_t hold_time;			/* Kill connection if this time goes by */
    time_t close_wait;			/* close wait time */
    time_t rexmit_time;			/* Retransmit outstanding packets after this time */
    time_t opensent_time;		/* time between open pdus sent to peer */ 
					/* if zero, 1 pdu and holdtime to close */
    time_t min_adv_time;		/* time for min_advertisement */
    time_t min_advRD_time;		/* time for min_advertisement within this RD. Only used */
    					/* local node. */
    time_t cpu_overload_time;		/* time between checking for CPU overload */
    	
    idrpBuffer *xmit_queue_head;	/* Transmit queue; can be kinda long */
    idrpBuffer **xmit_queue_tail;	/* Transmit queue tail */
    idrpBuffer *rexmit_queue_head;	/* Retransmit queue */
    idrpBuffer **rexmit_queue_tail;	/* Retransmit queue tail */

/* minimum route advertisement timer */
#define IDRP_MIN_ADV_TIMER_TYPE		0	/* minimum route advertisement timer  */
#define IDRP_MIN_ADVRD_TIMER_TYPE 	1	/* minimum route Adv timer with RD */

    time_t	min_adv_started;	/* time min adv started */ 
    int		start_min_adv;		/* flag for need to start min_adv with current time slot */	
    idrpAdvRt	*p_minadv_head;		/* minimum route advertisement list */
    idrpAdvRt	*p_minadv_tail;		/* current adding min adver list */

    time_t	min_advRD_started;	/* time min_advRD started */
    int		start_min_advRD;	/* flag for need to start min_adv with current time slot */
    idrpAdvRt	*p_minadvRD_head;	/* minimum rt AdvRd list head  */
    idrpAdvRt	*p_minadvRD_tail;	/* minimum rt advRD list tail  */	
} idrpPeer;
 
/* idrp status flags */
#define	IDRPF_UNCONFIGURED	0x00	/* incomplete configuration if zero */
#define	IDRPF_DELETE		0x01	/* delete this peer - it has been deleted from config */
#define	IDRPF_TRY_CONNECT	0x02	/* idrp trying to connect - but connection down */
#define	IDRPF_CONNECT		0x04	/* idrp peer connected */
#define	IDRPF_WRITEFAILED	0x10	/* idrp write to task socket failed */
#define IDRPF_IDLED             0x40    /* This peer is permanently idled - by off on config line */

/* Timers */
#define IDRPTIMER_KEEPALIVE     0       /* Fire when we should send a keepalive */
#define IDRPTIMER_HOLDTIME      1       /* Fire when hold time expires */
#define IDRPTIMER_REXMIT	2	/* Fire when we should retransmit a packet */
#define IDRPTIMER_CLOSEWAIT	3	/* Fire when close waiting is done */
#define IDRPTIMER_START		4	/* Fire when starting up the first time */
#define	IDRPTIMER_OPENSENT	5	/* Fire when open sent x times before restart */ 
#define IDRPTIMER_MIN_ADV	6	/* Fire when minimum route advertisement */
#ifdef ECHOKLUDGE
#define IDRPTIMER_ECHO		7	/* Fire when time to send echo %% */
#define IDRPTIMER_MAX           8       /* Number of timer slots to alloc */
#else
#define IDRPTIMER_MAX		7	
#endif
#define IDRP_MASTER_TIMER_MIN_ADVRD	0	/* Fire when minimum route advertisement */
 
#define IDRP_MASTER_TIMERS_MAX	1	/* timers for master task */

/* Protocol States--MUST BE THE SAME AS FSM_ERROR CODES!! */
#define IDRP_CLOSED		1	/* Connection closed */
#define IDRP_OPEN_RCVD		2	/* Open received */
#define IDRP_OPEN_SENT		3	/* Open sent */
#define IDRP_CLOSE_WAIT		4	/* Closing connection */
#define IDRP_ESTABLISHED	5	/* Connection established */
 
/* Events */
#define IDRP_START_EVENT	0	/* Start a connection */
#define IDRP_STOP_EVENT		1	/* Stop a connection */
#define IDRP_CLOSE_TIMEOUT	2	/* Close_wait timeout */
#define IDRP_HOLDTIME_TIMEOUT	3	/* Hold time timeout */
#define IDRP_HEADER_ERROR	4	/* BISPDU header error */
#define IDRP_KEEPALIVE		5	/* KEEPALIVE with no errors */
#define IDRP_CEASE		6	/* IDRP CEASE with no errors */
#define IDRP_OPEN_ERROR		7	/* IDRP OPEN with errors */
#define IDRP_OPEN		8	/* IDRP OPEN with no errors */
#define	IDRP_UPDATE_ERROR	9	/* IDRP UPDATE with errors */
#define	IDRP_UPDATE		10	/* IDRP UPDATE with no errors */
#define	IDRP_ERROR_ERROR	11	/* IDRP ERROR with errors */
#define IDRP_ERROR		12	/* IDRP ERROR with no errors */
#define IDRP_REFRESH_ERROR	13	/* IDRP RIB REFRESH with errors */
#define IDRP_REFRESH		14	/* IDRP RIB REFRESH with no errors */
#define IDRP_REXMIT_TIMEOUT	15	/* Retransmission timeout */
#define IDRP_KEEPALIVE_TIMEOUT	16	/* Keepalive timeout */
#define IDRP_SEND_PDU		17	/* Send a PDU */
#ifdef ECHOKLUDGE
#define IDRP_ECHO_TIMEOUT	19	/* Echo timeout %% */
#endif


/* Management event codes */

#define IDRP_BAD_LENGTH 	0	/* PDU length bad */
#define IDRP_BAD_PDU_TYPE	1	/* PDU type bad */
#define IDRP_PDU_HEADER_ERROR	2	/* PDU header bad */
#define IDRP_ERROR_PDU_ERROR	3	/* ERROR PDU has error */

/* Management event types */

#define IDRP_PDU_HEADER_ERR_BAD_TYPE	1	/* bad type on HEADER */

/* EXT_INFO information */

#define	IDRP_EXT_INFO_EXT_INFO	10	/* external information from other protocols */
#define	IDRP_EXT_INFO_IGP	5	/* external to idrp but internal to is-is (IGP) */  

/* idrp osi route option structure
 * - where we will read in the snpas,
 * - multi_EXIT value
 * - and temporarily the dist_lists
 *   until we go through gated export and import
 *   (bletch - but a workable set-up for a default
 *    DIST_LIST) 
 *
 */
 
typedef struct _idrpRoute_options {
	int		rib_id;			/* rib id of routes */
	snpa_list 	next_hop_snpas;		/* snpa list of routes */	
	int		multi_exit;		/* multi_exit value on this route */
	byte		route_server;		/* route server value for route */
	idrp_canon_rdpath 	*p_DIST_LIST_INCL;	/* list RDs to pass route to */
	idrp_canon_rdpath	*p_DIST_LIST_EXCL;	/* distribute list exclude */
} idrpRoute_options;


typedef struct _idrp_rdi_mem {
	int	mem_need;	/* DIST_LIST_INCL/EXCL memory needed for this path */
	int	cnt;		/* count of rdis in list */
} idrp_rdi_mem;


#include "idrp_globals.h"
#include "idrp_prototypes.h"
#include "idrp_macros.h"
#include "idrp_policy.h"

#endif /* not defined _idrp_h */



