#include "include.h"
#include "idrp.h"

/* Options */

/* flow control */

int acked_something;		/* True if last received packet acknowledged something */

/* task interface */

int doing_idrp = FALSE;		/* Are we running IDRP protocols? */
int idrp_reparsing = FALSE;	/* idrp is reparsing */

task *idrp_master_task;		/* IDRP master task block */
flag_t idrp_trace_flags;	/* Trace flags from parser */

/* idrp variables for unix sockets */

u_short idrp_port = IDRP_PORT;	/* IDRP well known port */

/* global addresses for IP and OSI */
/* addresses for this box */
/* ip addresses - rt_ip_rdi - must have match to AS */
/* look at this again prior to moving to IP */

struct iso_net_addr rt_rdi;	/* Our routing domain identifier */
struct iso_net_addr rt_net;	/* Our network entity title %% */
struct iso_net_addr rt_as_rdi;	/* our Mapping from AS to RDI */ 
snpa_list	rt_snpa;	/* snpa for this node */

/* IP information */

struct iso_net_addr rt_ip_rdi;	/* Our Routing Domain identifier for IP */
				/* only if we are running SIN idrp */
				/* for IP and OSI */ 
sockaddr_un rt_ip_net;		/* our IP address for NET */


/* task block pointers for allocation of space in task block initialization */

block_t		idrp_big_block_index;
block_t		idrp_small_blk_index;
block_t		idrp_med_blk_index;


/* local route information */

int	idrp_local_route_cnt;		 	/* count of local route blocks */


/* the local attribute records are to support idrp configured routes
 * in the parsing routing.
 */

idrp_attribute_record *idrp_last_local_att;     /* local routes attribute record */
                                                /* first is idrp_attr_list */


/* this attribute record chain supports the EXT_INFO
 * for the local routes
 */

idrp_attribute_record *idrp_ext_info_attr_head; /* idrp local EXT_INFO */
idrp_attribute_record *idrp_ext_info_attr_tail; /* idrp local EXT_INFO */


	
route_separator idrp_local_route_id;	/* our local route id - negative */
int local_hop_count;	/* our local hop count - zero */


/* link list of idrp neighbors - IP and OSI depends only on the address inside */

int	idrp_neighbors = 0;		/* Count of neighbors */
int	idrp_ext_neighbors = 0;		/* Count of external neighbors */
int	idrp_n_disabled = 0;		/* configured but disabled neighbors */
int	idrp_n_established = 0;		/* established IDRP neighbors */ 
int 	idrp_n_ext_established = 0;	/* Unused? */

idrpPeer	*idrp_peers;		/* Linked list of IDRP peers */
idrpPeer	*idrp_ext_peers;	/* pointer to first external peer */
idrpPeer	idrp_this_node;		/* peer structure for this node */
					/* used in local routes */

RD_path		local_rd_path;		/* fake RD pathway structure */ 
					/* for local routes */

int	consistency_errors;		/* consistency errors */

/* attribute lists */

idrp_attribute_record *idrp_attr_list;	/* linked list of attributes */
idrp_canon_rdpath *canon_rd_list;	/* linked list of rd paths */

/* stuff for best external routes pseudo-RIB */

idrp_best_ext *ber_rib_head = NULL;	/* points to beginning of best external */

/* add structure for adding to table */
/* idrp pdu buffer */

idrpPdu	*idrp_send_buffer;		/* buffer for sending idrp pdu */

/* Policy definitions for idrp */
/* below comment carried over from idrp_init.c */
/* why are policy definitions not here??? */
/* why no definition of idrp_import_list, idrp_export_list */
/* idrp_import_paths, idrp_export_paths */

int notifyLengths[];		/* Lengths of notifyData */
metric_t idrp_default_metric;	/* Default IDRP metric to use */

adv_entry *idrp_import_list;	/* List of IDRP advise entries */
adv_entry *idrp_export_list;	/* List of IDRP export entries */
adv_entry *idrp_import_paths;	/* List of IDRP import paths */
adv_entry *idrp_export_paths;	/* List of IDRP export paths */

/* idrp protocol states */
 
bits idrpStates[] =		/* Values of the states */
{
	{0, ""},
	{IDRP_CLOSED, "Closed"},
	{IDRP_OPEN_RCVD, "Open_received"},
	{IDRP_OPEN_SENT, "Open_sent"},
	{IDRP_CLOSE_WAIT, "Close_wait"},
	{IDRP_ESTABLISHED, "Established"},
	{-1}
};

/* should this change to bits??? */

char *idrpManagementEvents[] =	/* Names of IDRP management events */
{
	"bad length",
	"bad PDU type",
	"bad PDU header error",
	"bad ERROR pdu",

};

char *idrpTypes[] =		/* Names of IDRP PDU types */
{
	"Bogus",
	"Open",
	"Update",
	"Error",
	"Keepalive",
	"Cease",
	"RIB Refresh"
#ifdef ECHOKLUDGE
	,"Echo",
	"Echo reply"
#endif
};

/* Maybe this should be replaced by a shift of these bits
 * Must match the sequence of ATTRIBUTE types
 */

char idrpAttrType[] =		/* Type code for IDRP Attributes */ 
{
	IDRP_ATTR_LOCAL_RD,
	IDRP_ATTR_ROUTE_SEPARATOR,
	IDRP_ATTR_EXT_INFO,
	IDRP_ATTR_RD_PATH,
	IDRP_ATTR_NEXT_HOP,
	IDRP_ATTR_DIST_LIST_INCL,
	IDRP_ATTR_DIST_LIST_EXCL,
	IDRP_ATTR_MULTI_EXIT_DISC,
	IDRP_ATTR_TRANSIT_DELAY,
	IDRP_ATTR_RESIDUAL_ERROR,
	IDRP_ATTR_EXPENSE,
	IDRP_ATTR_LOCALLY_DEF_QOS,
	IDRP_ATTR_HIERARCH_REC,
	IDRP_ATTR_HOP_COUNT,
	IDRP_ATTR_SECURITY,
	IDRP_ATTR_CAPACITY,
	IDRP_ATTR_PRIORITY
};

char *idrpAttrTypeStrings[] =		/* Type code for IDRP Attributes */ 
{
	"IDRP_ATTR_LOCAL_RD",
	"IDRP_ATTR_ROUTE_SEPARATOR",
	"IDRP_ATTR_EXT_INFO",
	"IDRP_ATTR_RD_PATH",
	"IDRP_ATTR_NEXT_HOP",
	"IDRP_ATTR_DIST_LIST_INCL",
	"IDRP_ATTR_DIST_LIST_EXCL",
	"IDRP_ATTR_MULTI_EXIT_DISC",
	"IDRP_ATTR_TRANSIT_DELAY",
	"IDRP_ATTR_RESIDUAL_ERROR",
	"IDRP_ATTR_EXPENSE",
	"IDRP_ATTR_LOCALLY_DEF_QOS",
	"IDRP_ATTR_HIERARCH_REC",
	"IDRP_ATTR_HOP_COUNT",
	"IDRP_ATTR_SECURITY",
	"IDRP_ATTR_CAPACITY",
	"IDRP_ATTR_PRIORITY"
};

/* attribute flags array */

int idrpAttrFlags[] = 		/* attribute flag values */
{
	IDRP_ATTRFLG_LOCAL_RD,          /* local RD path - not in spec */
	IDRP_ATTRFLG_ROUTE_SEPARATOR,
	IDRP_ATTRFLG_EXT_INFO,
	IDRP_ATTRFLG_RD_PATH,
	IDRP_ATTRFLG_NEXT_HOP,
	IDRP_ATTRFLG_DIST_LIST_INCL,
	IDRP_ATTRFLG_DIST_LIST_EXCL,
	IDRP_ATTRFLG_MULTI_EXIT_DISC,
	IDRP_ATTRFLG_TRANSIT_DELAY,
	IDRP_ATTRFLG_RESIDUAL_ERROR,
	IDRP_ATTRFLG_EXPENSE,
	IDRP_ATTRFLG_LOCALLY_DEF_QOS,
	IDRP_ATTRFLG_HIERARCH_REC,
	IDRP_ATTRFLG_HOP_COUNT,
	IDRP_ATTRFLG_SECURITY,
	IDRP_ATTRFLG_CAPACITY,
	IDRP_ATTRFLG_PRIORITY
};

int idrpAttrBits[] =		/* attribute bit mask */
{
	IDRP_ATTBT_LOCAL_RD,
	IDRP_ATTBT_ROUTE_SEPARATOR,
	IDRP_ATTBT_EXT_INFO,
	IDRP_ATTBT_RD_PATH,
	IDRP_ATTBT_NEXT_HOP,
	IDRP_ATTBT_DIST_LIST_INCL,
	IDRP_ATTBT_DIST_LIST_EXCL,
	IDRP_ATTBT_MULTI_EXIT_DISC,
	IDRP_ATTBT_TRANSIT_DELAY,
	IDRP_ATTBT_RESIDUAL_ERROR,
	IDRP_ATTBT_EXPENSE,
	IDRP_ATTBT_LOCALLY_DEF_QOS,
	IDRP_ATTBT_HIERARCH_REC,
	IDRP_ATTBT_HOP_COUNT,
	IDRP_ATTBT_SECURITY,
	IDRP_ATTBT_CAPACITY,
	IDRP_ATTBT_PRIORITY
};

bits idrp_flag_bits[] =		/* pretty print bits */ 
{
	{IDRPF_UNCONFIGURED, "Unconfigured"},
	{IDRPF_DELETE, "Delete"},
	{IDRPF_TRY_CONNECT, "TryConnect"},
	{IDRPF_CONNECT, "ConnectTimer"},
	{IDRPF_WRITEFAILED, "WriteFailed"},
	{IDRPF_IDLED,"Idled"},
	{0}
};

char *idrpChangeTypes[] =	/* Names of IDRP change field values */
{
	"None",
	"Modified",
	"Deleted"
};

nlri_proto_info idrp_nlri_proto_id[] =	/* protocol id info array for nlri in update */ 
{
	{IDRP_PROTO_TYPE_TR_9577,IDRP_CLNP_PROTO_LEN,CLNP_NLPID},
	{IDRP_PROTO_TYPE_TR_9577,IDRP_IP_PROTO_LEN,IP_NLPID}
};

nlri_fam nlri_family[] =		/* nlri family */
{
	{AF_ISO,IDRP_NLRI_ISO},
	{AF_INET,IDRP_NLRI_IP}
};

u_int	proto_header_len[] =		/* header length for each protocol  for update pdu */
{
	IDRP_IP_HEADER_LEN,
	IDRP_UDP_HEADER_LEN,
	IDRP_CLNP_HEADER_LEN,
	IDRP_IDRP_HEADER_LEN
};


/* these arrays are not set-up */
 
bits idrpEvents[] =			/* Event values */
{
	{IDRP_START_EVENT,"idrp start event"},
	{IDRP_STOP_EVENT,"idrp stop event"},
	{IDRP_CLOSE_TIMEOUT,"idrp close timeout"},
	{IDRP_HOLDTIME_TIMEOUT,"idrp hold timeout"},
	{IDRP_HEADER_ERROR,"idrp bispdu header error"},
	{IDRP_KEEPALIVE,"idrp keepalive received"},
	{IDRP_CEASE,"idrp cease received with no errors"},
	{IDRP_OPEN_ERROR,"idrp open pdu with errors received"},
	{IDRP_OPEN,"idrp open with no errors received"},
	{IDRP_UPDATE_ERROR,"idrp update with errors received"},
	{IDRP_UPDATE,"idrp update w/no errors recieved"},
	{IDRP_ERROR_ERROR,"idrp error with errors recieved"},
	{IDRP_ERROR,"idrp error w/no errors recieved"},
	{IDRP_REFRESH_ERROR,"idrp refresh with errors recieved"},
	{IDRP_REFRESH,"idrp refresh w/no errors recieved"},
	{IDRP_REXMIT_TIMEOUT,"idrp re-transmit timeout occurred"},
	{IDRP_KEEPALIVE_TIMEOUT,"idrp keepalive timeout occurred"},
	{IDRP_SEND_PDU,"idrp_pdu sent"},
#ifdef ECHOKLUDGE
	{IDRP_ECHO_TIMEOUT,"idrp_echo_timeout"},
#endif
	{-1," "}
};
