#include "include.h"
#include "idrp.h"
#include <stdio.h>

/* idrp_timers.c
 *
 *  Timer routines: 
 *   start and stop timers
 *	start_rexmit_timer 
 *	kill_rexmit_timer
 * 	start_keepalive_timer 
 * 	start_opensent_timer
 *      start_echo_timer
 */ 
  
 
/*
 * Start the retransmission timer if it isn't already running.
 */
void start_rexmit_timer(peer)
idrpPeer *peer;
{
	if (!peer->rexmit_timer_running) {
		timer_set(TASK_TIMER(peer->task, IDRPTIMER_REXMIT), peer->rexmit_time, (time_t) 0);
		peer->rexmit_timer_running = TRUE;
	}
}

/*
 * Kill the retransmission timer if it isn't needed.
 */
void kill_rexmit_timer(peer)
idrpPeer *peer;
{
	if (peer->rexmit_timer_running) {
		if (peer->rexmit_queue_head == NULL && peer->cred_avail_zero_seq_last == 0) {
			timer_reset(TASK_TIMER(peer->task, IDRPTIMER_REXMIT));
			peer->rexmit_timer_running = FALSE;
		}
	}
}

/*
 * Start the keepalive timer. 
 */

void start_keepalive_timer(peer)
idrpPeer *peer;
{
	timer_set(TASK_TIMER(peer->task, IDRPTIMER_KEEPALIVE), peer->keepalive_time, (time_t) 0);
}

/*
 * Start the opensent timer
 */
static void start_opensent_timer(peer)
idrpPeer *peer;
{
	timer_set(TASK_TIMER(peer->task,IDRPTIMER_KEEPALIVE),peer->opensent_time,(time_t) 0);
}


#ifdef ECHOKLUDGE
/*
 * start echo timer (only if ECHOKLUDGE is def'd)
 */
static void start_echo_timer(peer)
idrpPeer *peer;
{
	timer_set(TASK_TIMER(peer->task, IDRPTIMER_ECHO), IDRP_ECHO_TIME, (time_t) 0);
}
#endif /* ECHOKLUDGE */
