%{
/*
 *  parser.y,v 1.60 1993/04/10 12:25:26 jch Exp
 */

/* Gated Release 3.0 */
/* Copyright (c) 1990,1991,1992,1993 by Cornell University. All rights reserved. */
/* Refer to Particulars and other Copyright notices at the end of this file. */


#include "include.h"
#ifdef	PROTO_INET
#include "inet.h"
#endif	/* PROTO_INET */
#ifdef	PROTO_ISO
#include "iso.h"
#endif	/* PROTO_ISO */
#include "parse.h"
#include "krt.h"
#ifdef	PROTO_RIP
#include "rip.h"
#endif	/* PROTO_RIP */
#ifdef	PROTO_HELLO
#include "hello.h"
#endif	/* PROTO_HELLO */
#ifdef	PROTO_EGP
#include "egp.h"
#endif	/* PROTO_EGP */
#ifdef	PROTO_BGP
#include "bgp.h"
#endif	/* PROTO_BGP */
#ifdef	PROTO_DVMRP
#include "dvmrp.h"
#endif	/* PROTO_DVMRP */
#ifdef	PROTO_ISODE_SNMP
#include "snmp_isode.h"
#endif	/* PROTO_ISODE_SNMP */
#ifdef	 PROTO_ISIS
#include "isis.h"
#include "isis_test.h"
#include "isis_globals.h"
#endif	/* PROTO_ISIS */
#ifdef	PROTO_IDRP
#include "idrp.h"
#endif /* PROTO_IDRP */
  
char parse_error[LINE_MAX] = { 0 };
char *parse_filename = 0;

static	proto_t	parse_proto;			/* For correctly tagging ADV lists */
static	proto_t	parse_export_proto;		/* For checking metrics */
static	gw_entry	**parse_gwlist;
#if	defined(PROTO_SLSP) || defined(PROTO_EGP) || defined(PROTO_BGP)
static	gw_entry	*parse_gwp;		/* To fake a list for EGP and BGP */
#endif
static  char *parse_serv_proto;
static	int	parse_contig_masks = TRUE;	/* Masks must be contiguous */
static	bgpPeer *bnp;
static	bgpPeerGroup *bgp;
static 	struct bgp_conf *bcp;
static	idrpPeer *idgp, idrp_group, *gr_idrp;
struct	iso_prefix *nlri; 
static	idrpRoute *p_idrp;
static	int idrp_parse_group_index;
static	idrpRoute_options 	idrp_rt_opts;	
static	idrp_canon_rdpath	*p_rdi_head,*p_rdi_tail;
static	idrp_canon_rdpath	*p_idrp_excl_rdi_tail, *p_idrp_excl_rdi_head;
static	idrp_canon_rdpath	*p_idrp_incl_rdi_tail, *p_idrp_incl_rdi_head;
static if_info parse_ifi;	/* Used for configuring interfaces that do not exist */

u_int parse_state = PS_INITIAL;
proto_t protos_seen = 0;

PROTOTYPE(yyerror,
	  static void,
	  (const char *));

#define	free_string(s)	task_mem_free((task *) 0, (caddr_t) s); s = 0
#define	free_charstr(s)	task_mem_free((task *) 0, (caddr_t) (s)->ptr); (s)->ptr = (char *) 0; (s)->len = (s)->type = 0
#define	free_bytestr(s)	task_mem_free((task *) 0, (caddr_t) (s)->ptr); (s)->ptr = (byte *) 0; (s)->len = (s)->type = 0

#define	PARSE_ERROR	yyerror(parse_error); yynerrs++; YYERROR;

#define	PROTO_SEEN(proto) \
	if (BIT_TEST(protos_seen, RTPROTO_BIT(proto))) { \
	    sprintf(parse_error, "parse_proto_seen: duplicate %s clause", gd_lower(trace_state(rt_proto_bits, proto))); \
	    PARSE_ERROR; \
	} else { \
	    BIT_SET(protos_seen, RTPROTO_BIT(proto)); \
	    parse_proto = proto; \
	} \

#ifdef	PROTO_ASPATHS
static unsigned	aspath_state;
static asmatch_t *aspath_current;

#define	ASPATH_NEXT_STATE	if (aspath_state++ == ASP_BITS) { \
	sprintf(parse_error, "too many AS path regexp states"); \
	PARSE_ERROR; \
    }
#endif	/* PROTO_ASPATHS */

%}

/* Global */
%union {
    long	num;
    u_long	u_num;
    bytestr	hex;
    charstr	str;
    flag_t	flag;
    time_t	time;
    as_t	as;
    proto_t	proto;
    pmet_t	metric;
    pref_t	pref;
    if_addr	*ifap;
    if_addr_entry	*ifae;
    adv_entry	*adv;
    gw_entry	*gwp;
    sockaddr_un	*sockaddr;
#ifdef	PROTO_INET
    struct in_addr in_addr;
#endif	/* PROTO_INET */
    dest_mask	dm;
    struct {
	pmet_t metric;
	adv_entry *adv;
    } metric_adv ;
    struct {
	proto_t proto;
	adv_entry **import_list;
	adv_entry **export_list;
	gw_entry **gw_list;
    } proto_lists;
    config_entry *config;
    config_list *conflist;
#ifndef	GATED_LEXER
#ifdef	PROTO_OSPF
    ospf_config_router *ospf_router;
#endif	/* PROTO_OSPF */
#if	defined(PROTO_ASPATHS)
    asmatch_t	*asp_match;
    struct {
	u_short	begin;
	u_short end;
    } asp_range;
    struct {
	flag_t	start[1 + ASP_ACC/ASP_WORD];
	flag_t	state[1 + ASP_ACC/ASP_WORD];
    } asp_stack;
#endif	/* PROTO_ASPATHS */
#endif	/* GATED_LEXER */
}

/* Lexer tokens */
%token			L_EOS L_ERROR
%token	<num>		L_NUMBER
%token	<str>		L_STRING L_HNAME L_KEYWORD
%token	<hex>		L_HEXSTRING L_OCTETSTRING L_HEXSTRING_ODD
%token	<hex>		L_ISOADDR

/* Keyword tokens */    
%token	<num>		T_DIRECT T_INTERFACE T_INTERFACES T_LCLADDR
%token	<num>		T_PROTO T_METRIC T_METRICIN T_METRICOUT T_INFINITY T_DOWN
%token	<num>		T_PARSE T_ADV T_DEFAULT T_DEFAULTS
%token	<num>		T_YYDEBUG T_YYSTATE T_YYQUIT T_DEBUG
%token	<num>		T_SYSLOG T_UPTO T_EMERG T_ALERT T_CRIT T_ERR T_WARNING T_NOTICE T_INFO
%token	<num>		T_ON T_OFF T_GATEWAY T_PREFERENCE T_DEFAULTMETRIC T_NONBROADCAST T_BROADCAST T_TYPE
%token	<num>		T_POINTOPOINT T_MULTICAST T_NETMASK T_DEFINE
%token	<num>		T_NOP T_NOP1	/* Filler for conditional clauses */

%type	<num>		syslog_level syslog_mask

/* BGP and EGP */
%token	<num>		T_PEERAS T_LOCALAS T_NEIGHBOR T_GROUP
%token	<num>		T_GENDEFAULT T_NOGENDEFAULT T_VERSION T_DEFAULTIN T_DEFAULTOUT

/* IDRP and BGP 4 */

%token	<num>		T_HOLDTIME

%token	<num>		T_BGP T_ALLOW T_TEST T_KEEPALL T_SENDBUF T_RECVBUF T_SPOOLBUF
%type	<num>		bgp_linktype bgp_kbufsize
%type	<flag>		bgp_trace_options
%type	<proto>		bgp_prop_init    
%type	<dm>		bgp_allow
%type	<adv>		bgp_allow_list    

/* Needed for both EGP and AS paths */    
%token	<num>		T_EGP 

/* AS Paths */    
%token	<num>		T_ASPATH T_ORIGIN T_INCOMPLETE T_ANY
%type	<asp_match>	aspath_match
%type	<asp_stack>	aspath_regex aspath_sum aspath_symbol aspath_term
%type	<asp_range>	aspath_range
%type	<flag>		aspath_origin_list aspath_origins aspath_origin
%type	<proto>		aspath_prop_init    
/* Tags */
%token	<num>		T_TAG
%type	<metric>	tag_option tag_as tag_as_option tag
%type	<proto>		tag_prop_init

/* RIP and HELLO */
%token	<num>		T_TRUSTEDGATEWAYS T_SOURCEGATEWAYS



/*
#ifndef PROTO_ISIS -- jgs -- fix this later
@BEGIN: PROTO_IDRP -- jgs
%token  <num>           T_IP T_DUAL T_SNPA	
@END:   PROTO_IDRP -- jgs
#endif	PROTO_ISIS -- jgs
*/

%token	<num>		T_IP T_DUAL T_SNPA
%token	<num>		T_ISIS T_CIRCUIT T_SYSTEMID
%token	<num>		T_LEVEL T_IPREACH T_TROLL T_SET T_PREFIX
%token	<num>		T_INTDOMINFO T_MODE
%type	<num>		isis_option isis_mode isis_level ipreach_type prefix_type
%type	<flag>		isis_trace_options isis_trace_option 

/* OSPF and ISIS */
%token	<num>		T_AREA T_SPF

/* OSPF and SLSP */
%token <num>		T_HELLOINTERVAL T_ROUTERDEADINTERVAL

/* OSPF and ISIS and SLSP */
%token	<num>		T_PRIORITY




/* IDRP */
%token	<num>	T_IDRP T_RDI T_OSISTATIC T_NSAP T_NET T_NSAPPREFIX T_OSILOCAL T_MULTIEXIT T_PROTOSOCK
%token	<num> 	T_SIP T_PIP T_UDP T_LOCALNODE
%token	<num>	T_PROTOSUPPORT T_AUTHCODE T_INTRAIS T_RDC T_ASRDI
%token	<num>	T_OUTPDUS T_MAXPDUSIZE 
%token	<num> 	T_REXMIT T_CLOSEWAIT T_RIBCHECK T_MINADV T_MINADVRD
%token	<num>	T_PROHIBITDIST T_ALLOWDIST  T_ROUTESERVER
%token	<num>	T_LISTENOPEN

%type	<flag>	idrp_trace_options
%type	<hex>	osi_rdi osi_address osi_nsap osi_prefix
 

/* SNMP */


/* Redirects */
%token	<num>		T_REDIRECT T_NOREDIRECTS T_ICMP
%type	<config>	redirect_interface_option redirect_interface_options
%type	<flag>		redirect_trace_options

/* Authentication */
%token	<num>		T_AUTHTYPE T_AUTHKEY T_NONE T_SIMPLE

/* Interface */
%token	<num>		T_PASSIVE T_SIMPLEX T_STRICTIFS T_SCANINTERVAL

/* Control */
%token	<num>		T_STATIC T_MARTIANS T_RETAIN
%token	<num>		T_EXPORT T_IMPORT
%token	<num>		T_HOST T_MASK T_MASKLEN T_REJECT T_BLACKHOLE T_RESTRICT
%type	<dm>		route_dest
%type	<flag>		rts_options rts_option   

/* AS control */
%token	<num>		T_AS T_ROUTERID T_IGP

/* Tracing */
%token			T_OPTIONS T_NOINSTALL T_NOSEND T_NORESOLV
%token	<num>		T_TRACEOPTIONS T_EXCEPT T_TRACEFILE T_REPLACE T_SIZE T_FILES T_K T_M
%token	<num>		T_ALL T_GENERAL T_INTERNAL T_EXTERNAL T_ROUTE T_UPDATE T_KERNEL
%token	<num>		T_TASK T_TIMER T_NOSTAMP T_MARK
/* additional tracing for IS-IS */
%token  <num>   	TISIS_IIH TISIS_DUMPLSP TISIS_EVENTS TISIS_LANADJ
%token  <num>   	TISIS_FLOODING TISIS_BUILDLSP TISIS_CSNP TISIS_PSNP
%token  <num>   	TISIS_LSPINPUT TISIS_P2PADJ TISIS_LSPDB 
%token	<num>		TISIS_PATHS TISIS_LSPCONTENT TISIS_SUMMARY

%type	<num>		number				/* A decimal, hex, or octet string number */
%type	<hex>		byte_string			/* A L_HEXSTRING or L_OCTETSTRING */
%type	<metric>	metric metric_option metric_restrict_option
%type	<time>		time
%type	<num>		onoff_option
%type	<flag>		trace_options trace_option trace_option_global trace_options_none
%type	<flag>		trace_trace_options_none trace_trace_options trace_trace_option
%type	<str>		trace_file
%type	<num>		trace_size    
%type	<config>	interface_define_options interface_define_option
%type	<num>		trace_replace
%type	<proto>		proto_exterior
%type	<proto_lists>	control_exterior import_interior export_interior proto_interior


%type	<proto> 	prop_default prop_direct prop_static prop_kernel
%type	<as>		as
%type	<adv>		as_list as_list_option
%type	<pref>		preference
%type	<metric>	preference_option preference_restrict_option
%type	<ifae>		interface_local    
%type	<sockaddr>	addr mask host network interface_addr
%token	<num>		T_INET    
%type	<sockaddr>	ip_addr
%type	<in_addr>	in_addr
%token	<num>		T_ISO
%type	<sockaddr>	iso_addr
%type	<gwp>		gateway
%type	<dm>		dest_mask host_mask network_mask all_mask default_mask martian
%type	<adv>		interface_all interface_all_adv interface_list interface_list_option interface_all_list
%type	<adv>		interface_gateway_list gateway_list
%type	<adv>		martian_list
%type	<adv>		import_list import_listen
%type	<metric_adv>	import_option
%type	<adv>		prop_source prop_source_list
%type	<adv>		prop_restrict prop_restrict_list prop_restrict_list_option
%type	<metric_adv>	prop_source_option prop_metric_option
%type	<str>		string host_name
%type	<num>		port

%%

config		: /* Empty */
		| statements
		;

statements	: statement
		| statements statement
		;

statement	: parse_statement
		| trace_statement
		| option_order option_statement
		| interface_order interface_statement
		| define_order define_statement
		| proto_order proto_statement
		| route_order route_statement
		| control_order control_statement
		| error L_EOS
			{
				yyerrok;
			}
		| L_EOS
		;

/*  */

parse_statement	: T_YYDEBUG onoff_option L_EOS
			{
#if	YYDEBUG != 0
			    if ($2 == T_OFF) {
				yydebug = 0;
			    } else {
				yydebug = 1;
			    }
#endif	/* YYDEBUG */
			}
		| T_YYSTATE L_NUMBER L_EOS
			{
#if	YYDEBUG != 0
			    if ($2 < 0 || $2 > PS_MAX) {
				(void) sprintf(parse_error, "invalid yystate value: %d",
					       $2);
				PARSE_ERROR;
			    }
			    parse_state = $2;
#endif	/* YYDEBUG */
			}
		| T_YYQUIT L_EOS
			{
#if	YYDEBUG != 0
			    task_quit(0);
#endif	/* YYDEBUG */
			}
		;

/*  */

trace_statement	: T_TRACEOPTIONS trace_trace_options_none L_EOS
			{
			    if (!BIT_TEST(task_state, TASKS_TEST)) {
				if ($2) {
				    if (trace_flags) {
					trace_flags = trace_flags_save = $2;
					trace_display(trace_flags, TR_PARSE);
				    } else if (trace_opened) {
					trace_flags = trace_flags_save = $2;
					trace_display(trace_flags, TR_PARSE);
				    }
				} else {
				    if (trace_flags) {
					trace_flags = trace_flags_save = $2;
					trace_display(trace_flags, TR_PARSE);
					trace_off();
				    }
				}
			    }
			}
		| T_TRACEFILE trace_file trace_limits L_EOS
			{
			    if (!BIT_TEST(task_state, TASKS_TEST)) {
				if ($2.len) {
				    int append = (int) $2.type;

				    if (trace_flags) {
					trace_flags_save = trace_flags;
					trace_off();
				    }
				    trace_on($2.ptr, append);
				}
			    }
			}
		;

trace_file	: /* Empty */
			{
			    $$.len = 0;
			}
    		| string trace_replace
			{
			    $$ = $1;
			    $$.type = $2;
			}
    		;

trace_size	: number
		| number T_K
			{
			    $$ = $1 * 1024;
			}
		| number T_M
			{
			    $$ = $1 * 1024 * 1024;
			}
		;

trace_replace	: /* Empty */	{ $$ = TRUE; }
		| T_REPLACE	{ $$ = FALSE; }
		;

trace_limits	: /* Empty */
    		| T_SIZE trace_size T_FILES number
			{
			    if (parse_limit_check("file-size",
						  (u_int) $2,
						  TRACE_LIMIT_FILE_SIZE)) {
				PARSE_ERROR;
			    }
			    trace_limit_size = $2;

			    if (parse_limit_check("files",
						  (u_int) $4,
						  TRACE_LIMIT_FILE_COUNT)) {
				PARSE_ERROR;
			    }
			    trace_limit_files = $4;
			}
    		;

trace_trace_options_none
    		: T_NONE
			{
			    $$ = (flag_t) 0;
			}
		| trace_trace_options
			{
			    $$ = $$;
			}
		;
		
trace_trace_options
		: trace_option
		| trace_trace_options trace_trace_option
			{
			    $$ = $1 | $2;
			}
		| trace_trace_options T_EXCEPT trace_trace_option
			{
			    $$ = $1 & ~$3;
			}
		;

trace_trace_option
		: trace_option
		| trace_option_global
    		;

trace_options_none
		: T_NONE
			{
			    $$ = (flag_t) 0;
			}
		| trace_options
		;

trace_options	: trace_option
		| trace_options trace_option
			{
			    $$ = $1 | $2;
			}
		;

/* Trace options that can only be specified globally */
trace_option_global
    		: T_NOSTAMP
			{
			    $$ = TR_NOSTAMP;
			}
		| T_MARK
			{
			    $$ = TR_MARK;
			}
		| T_PARSE
			{
			    $$ = TR_PARSE;
			}
		| T_ADV
			{
#ifdef	TR_ADV
			    $$ = TR_ADV;
#else	/* TR_ADV */
			    $$ = 0;
#endif	/* TR_ADV */
			}
		;

/* Trace options that can be specified anywhere */
trace_option	: T_ALL
			{
			    $$ = TR_ALL;
			}
		| T_GENERAL
			{
			    $$ = TR_INT|TR_EXT|TR_RT;
			}
		| T_INTERNAL
			{
			    $$ = TR_INT;
			}
		| T_EXTERNAL
			{
			    $$ = TR_EXT;
			}
		| T_ROUTE
			{
			    $$ = TR_RT;
			}
		| T_UPDATE
			{
			    $$ = TR_UPDATE;
			}
		| T_TASK
			{
			    $$ = TR_TASK;
			}
		| T_TIMER
			{
			    $$ = TR_TIMER;
			}
		| T_PROTO
			{
			    $$ = TR_PROTOCOL;
			}
		| T_KERNEL
			{
			    $$ = TR_KRT;
			}
		| T_BGP
			{
			    $$ = TR_BGP;
			}
		| T_ISIS
			{
			    $$ = TR_ISIS;
			}    
		| T_IDRP	{ $$ = TR_IDRP; }
		;

/*  */

define_order	: /*Empty */
			{
				if (parse_new_state(PS_DEFINE)) {
					PARSE_ERROR;
				}
			}
		;

define_statement
		: T_MARTIANS
			{
			    parse_contig_masks = FALSE;
			}
			'{' martian_list '}' L_EOS
			{
			    parse_contig_masks = TRUE;
			    if (parse_adv_append(&martian_list, $4, TRUE)) {
				PARSE_ERROR;
			    }
			}


                | T_RDI osi_rdi L_EOS
                        {
				/* RDI indicates the Routing Domain IDRP peer in
				 * there for both IP and CLNP 
				 * - Kludge for now to save as AS_RDI until
				 * - more parsing code is written for IP 
				 */
 
                                bcopy($2.ptr,&rt_rdi.isoa_genaddr, $2.len);
                                rt_rdi.isoa_len = $2.len;
				trace (TR_PARSE,0,"rdi xxxx configured for idrp %A ", 
					sockbuild_iso(rt_rdi.isoa_genaddr, rt_rdi.isoa_len));
				rt_as_rdi.isoa_len = $2.len - 2;
				bcopy($2.ptr,&rt_as_rdi.isoa_genaddr,rt_as_rdi.isoa_len);
				rt_as_rdi.isoa_family = AF_ISO;
				trace (TR_PARSE,0, "idrp finished call to rdi in parser.y");
                        }
                | osi_address_name osi_nsap L_EOS
                        { 
				/* need an NET for this node 
				 * if we configure multiple ISO host addresses, 
				 * we need to nominate one as NET.
				 * Eventually maybe this could be discovered
				 */

                                bcopy($2.ptr, rt_net.isoa_genaddr, $2.len);
                                rt_net.isoa_len = $2.len;
                        }

		| T_AS as L_EOS
			{
			    if (inet_autonomous_system != $2) {
				if (inet_autonomous_system) {
				    inet_as_unlink(inet_autonomous_system);
				}
				inet_autonomous_system = $2;
				inet_as_link(inet_autonomous_system);
			    }
			}
    		| T_ROUTERID ip_addr L_EOS
			{
			    if (inet_parse_routerid($2, parse_error)) {
				PARSE_ERROR;
			    }
			}
		;


martian_list	: /* Empty */
			{
			    $$ = (adv_entry *) 0;
			}
		| martian_list martian L_EOS
			{
			    $$ = adv_alloc(ADVFT_DM, (proto_t) 0);
			    $$->adv_dm = $2;
			    $$ = parse_adv_dm_append($1, $$);
			    if (!$$) {
				PARSE_ERROR;
			    }
			}
		| martian_list error L_EOS
			{
			    yyerrok;
			}
		;


martian		: host_mask
		| network_mask
		| default_mask
		;

/**/

option_order	: /*Empty */
			{
				if (parse_new_state(PS_OPTIONS)) {
					PARSE_ERROR;
				}
			}
		;

option_statement
    		: T_OPTIONS option_list L_EOS
    		;

option_list	: option
		| option_list option
		;


option		: T_NOINSTALL
			{
			    krt_install = FALSE;
			}
		| T_GENDEFAULT preference_option
			{
			    rt_default_needed = TRUE;
			    if (PARSE_METRIC_ISSET(&$2)) {
				rt_default_rtparms.rtp_preference = $2.metric;
			    }
			}
		| T_NOSEND
			{
			    task_newstate(TASKS_NOSEND, 0);
			}
		| T_NORESOLV
			{
			    task_newstate(TASKS_NORESOLV, 0);
			}
		| T_SYSLOG syslog_mask
			{
#ifdef	LOG_DAEMON
			    (void) setlogmask($2);
#else	/* LOG_DAEMON */
			    (void) sprintf(parse_error, "system does not support setlogmask()");
			    PARSE_ERROR;
#endif	/* LOG_DAEMON */
			}
		;


syslog_mask	:	syslog_level
			{
#ifdef	LOG_DAEMON
			    $$ = LOG_MASK($1);
#endif	/* LOG_DAEMON */
			}
		|	T_UPTO syslog_level
			{
#ifdef	LOG_DAEMON
			    $$ = LOG_UPTO($2);
#endif	/* LOG_DAEMON */
			}
		|	syslog_mask syslog_level
			{
#ifdef	LOG_DAEMON
			    $$ = $1 | LOG_MASK($2);
#endif	/* LOG_DAEMON */
			}
		;

syslog_level	: T_EMERG
			{
			    $$ = LOG_EMERG;
			}
		| T_ALERT
			{
			    $$ = LOG_ALERT;
			}
		| T_CRIT
			{
			    $$ = LOG_CRIT;
			}
		| T_ERR
			{
			    $$ = LOG_ERR;
			}
		| T_WARNING
			{
			    $$ = LOG_WARNING;
			}
		| T_NOTICE
			{
			    $$ = LOG_NOTICE;
			}
		| T_INFO
			{
			    $$ = LOG_INFO;
			}
		| T_DEBUG
			{
			    $$ = LOG_EMERG;
			}
    		;

/*  */

interface_order	: /*Empty */
			{
			    if (parse_new_state(PS_INTERFACE)) {
				PARSE_ERROR;
			    }
			}
		;

interface_statement	
		: T_INTERFACES
			{
			    PROTO_SEEN(RTPROTO_DIRECT);
			}
			'{' interface_stmts '}' L_EOS
		;

interface_stmts
		: /* Empty */
    		| interface_stmts interface_stmt L_EOS
    		| interface_stmts error L_EOS
			{
			    yyerrok;
			}
    		;

interface_stmt
		: T_OPTIONS interface_options
		| T_INTERFACE interface_all_list interface_define_options
			{
			    config_list *list = config_list_alloc($3, 0);

			    parse_adv_propagate_config($2, list, (proto_t) 0);

			    if (parse_adv_append(&int_policy, $2, TRUE)) {
				PARSE_ERROR;
			    }

			    config_list_free(list);
			}
		| T_DEFINE host
			{
			    bzero((caddr_t) &parse_ifi, sizeof (parse_ifi));

			    parse_ifi.ifi_addr = $2;
			}
			interface_type_options interface_common_options
			{
			    if (if_parse_add(&parse_ifi, parse_error)) {
				PARSE_ERROR;
			    }
			}
		;

interface_options
		: interface_option
		| interface_options interface_option
		;

interface_option
		: T_STRICTIFS
			{
			    task_newstate(TASKS_STRICTIFS, 0);
			}
		| T_SCANINTERVAL time
			{
			    if (parse_limit_check("scan-interval",
						  (u_int) $2,
						  KRT_LIMIT_SCANTIMER)) {
				PARSE_ERROR;
			    }
			    
			    timer_set(krt_task->task_timer[KRT_TIMER_IFCHECK],
				      $2,
				      (time_t) 0);
			}
		;

interface_type_options
		: T_BROADCAST host
			{
			    parse_ifi.ifi_addr_broadcast = $2;
			    BIT_SET(parse_ifi.ifi_state, IFS_BROADCAST);
			}
		| T_POINTOPOINT host
			{
			    parse_ifi.ifi_addr_local = $2;
			    BIT_SET(parse_ifi.ifi_state, IFS_POINTOPOINT);
			}
		| /* Empty */
			{
			}
		;

interface_common_options
		: /* Empty */
		| interface_common_options interface_common_option
		;

interface_common_option
		: T_NETMASK mask
			{
			    parse_ifi.ifi_subnetmask = $2;
			}
		| T_MULTICAST
			{
			    BIT_SET(parse_ifi.ifi_state, IFS_MULTICAST);
			}
		;

interface_define_options
		: interface_define_option
		| interface_define_options interface_define_option
			{
			    $$ = config_append($1, $2);
			    if (!$$) {
				PARSE_ERROR;
			    }
			}
		;

interface_define_option
		: T_PREFERENCE preference
			{
			    $$ = config_alloc(IF_CONFIG_PREFERENCE_UP, (void_t) $2);
			}
		| T_DOWN T_PREFERENCE preference
			{
			    $$ = config_alloc(IF_CONFIG_PREFERENCE_DOWN, (void_t) $3);
			}
		| T_PASSIVE
			{
			    $$ = config_alloc(IF_CONFIG_PASSIVE, (void_t) TRUE);
			}
		| T_SIMPLEX
			{
			    $$ = config_alloc(IF_CONFIG_SIMPLEX, (void_t) TRUE);
			}
		| T_REJECT
			{
			    $$ = config_alloc(IF_CONFIG_REJECT, (void_t) TRUE);
			}
		| T_BLACKHOLE
			{
			    $$ = config_alloc(IF_CONFIG_BLACKHOLE, (void_t) TRUE);
			}
		;

interface_list_option
		: /* Empty */
			{
			    $$ = adv_alloc(ADVFT_ANY|ADVF_FIRST, (proto_t) 0);
			}
		| T_INTERFACE interface_list
			{
			    $$ = $2;
			}
		;

interface_gateway_list
		: interface_list_option
		| T_GATEWAY gateway_list
			{
			    $$ = $2;
			}
		;

interface_all_list
		: T_ALL
			{
			    $$ = adv_alloc(ADVFT_ANY|ADVF_FIRST, (proto_t) 0);
			}
		| interface_list
		;

interface_list
		: interface_all
			{
			    BIT_SET($1->adv_flag, ADVF_FIRST);
			    $$ = $1;
			}
		| interface_list interface_all
			{
			    $$ = $1;
			    if (parse_adv_append(&$$, $2, TRUE)) {
				PARSE_ERROR;
			    }
			}
		;

interface_all	: interface_all_adv
			{
			    switch ($1->adv_flag & ADVF_TYPE) {

			    case ADVFT_IFN:
				trace(TR_PARSE, 0, "parse: %s INTERFACE: %A*",
				      parse_where(),
				      $$->adv_ifn->ifae_addr);
				break;
				
			    case ADVFT_IFAE:
				trace(TR_PARSE, 0, "parse: %s INTERFACE: %A",
				      parse_where(),
				      $$->adv_ifae->ifae_addr);
			    }
			}
		;

interface_all_adv
		: host_name
			{
			    int valid_ifname = $1.type == L_KEYWORD && $1.len <= IFNAMSIZ;

			    if (valid_ifname) {
				/* First try to look up as an interface name */

				$$ = if_parse_name($1.ptr, FALSE);
			    } else {
				/* Doesn't meet the criteria to be a name */

				$$ = (adv_entry *) 0;
			    }
			    
			    if (!$$ && !BIT_TEST(task_state, TASKS_NORESOLV)) {
				/* Try to resolve as a host name */

				sockaddr_un *addr = parse_addr_hostname($1.ptr);

				if (addr) {
				    $$ = if_parse_address(addr);
				}

			    }

			    if (!$$) {
				if (valid_ifname && !BIT_TEST(task_state, TASKS_STRICTIFS)) {
				    /* Assume it is a reference to an interface that does not exist */

				    $$ = if_parse_name($1.ptr, TRUE);
				} else {
				    /* Time to give up */
				    
				    (void) sprintf(parse_error, "Interface not found at '%s'",
						   $1.ptr);
				    PARSE_ERROR;
				}
			    }
				
			    free_charstr(&$1);
			}
		| addr
			{
			    $$ = if_parse_address($1);
			    if (!$$) {
				(void) sprintf(parse_error, "Interface not found at '%A'",
					       $1);
				PARSE_ERROR;
			    }

			    sockfree($1);
			}
		;

interface_local	: interface_addr
			{
			    if_addr *ifap;
			    
			    ifap = if_withlcladdr($1, FALSE);
			    if (ifap) {
				$$ = ifae_alloc(ifap->ifa_addrent_local);
			    } else {
				if (BIT_TEST(task_state, TASKS_STRICTIFS)) {
				    (void) sprintf(parse_error, "Invalid interface at '%A'",
						   $1);
				    PARSE_ERROR;
				} else {
				    $$ = ifae_locate($1, &if_local_list);
				}
			    }
			    sockfree($1);
			    trace(TR_PARSE, 0, "parse: %s LOCAL ADDRESS: %A",
				  parse_where(),
				  $$->ifae_addr);
			}
		;

interface_addr	: addr
		| host_name
			{
			    sockaddr_un *addr;

			    if (BIT_TEST(task_state, TASKS_NORESOLV)
				|| !(addr = parse_addr_hostname($1.ptr))) {
				PARSE_ERROR;
			    }
			    $$ = sockdup(addr);
			    free_charstr(&$1);
			}
		;

/*  */

proto_order	: /* Empty */
			{
			    if (parse_new_state(PS_PROTO)) {
				PARSE_ERROR;
			    }
			}
		;

proto_statement	: redirect_statement L_EOS
		| bgp_statement L_EOS
    		| isis_statement L_EOS
		| idrp_statement
		;

/*  */


/*  */


    
/*  */

    
/*  */


/* ISIS protocol statements & etc, in the format of other protocols */

isis_statement  : T_ISIS
                        {
			    PROTO_SEEN(RTPROTO_ISIS);

#ifdef	notyet
			    parse_gwlist = &isis.gw_list;

			    isis_trace_flags = trace_flags;
#endif	/* notyet */

			    /* XXX - should be done by above */
			    isis_plevel = 1;
                        }
		  isis_option isis_group 
			{

			    ip_supported = 0;
			    clnp_supported = 0;
			    switch ($3) {
			    case T_OFF:
				break;
			    /* XXX - Check validity of configuration for the following */
			    case T_IP:
				ip_supported = 1;
				doing_isis = TRUE;
				break;
			    case T_ISO:
				clnp_supported = 1;
				doing_isis = TRUE;
				break;
			    case T_ON:
			    case T_DUAL:
				ip_supported = clnp_supported = 1;
				doing_isis = TRUE;
			    }
			}
		;

isis_option	: onoff_option
		| T_IP		{ $$ = T_IP; }
		| T_ISO		{ $$ = T_ISO; }
		| T_DUAL	{ $$ = T_DUAL; }
		;

isis_group      : /* Empty */
                | '{' isis_group_stmts '}'
		;

isis_group_stmts	: /* Empty */
		| isis_group_stmts isis_group_stmt L_EOS
		| isis_group_stmts error L_EOS
			{
				yyerrok;
			}
		;

isis_group_stmt	: isis_level
			{
			    isis_plevel = $1;
			}
		| T_CIRCUIT string
			{
			    if (BIT_TEST(task_state, TASKS_RECONFIG)) {
				if ((isis_ci = isis_GetIspcByName($2)) == 0) {
				    /* this is a config of a new circuit */    
				    isis_ci = ++isis_circuits_parsed;
				    bzero(&isis_ca[isis_ci], sizeof(struct is_pc));
				    strcpy(isis_ca[isis_ci].name, $2.ptr);
			    	    /* isis_init_ca(&isis_ca[isis_ci], $2); */
				    trace(TR_ISIS, LOG_INFO, "IS-IS reconfig parsed new circuit #%d -> %s", isis_ci, $2);
				} else {
					trace(TR_ISIS, LOG_INFO, "IS-IS reconfig parsed existing circuit #%d -> %s", isis_ci, $2);
				} 
			    } else {
				/* remember new circuit */
			    	isis_ci = ++isis_circuits_parsed;
				trace(TR_ISIS, LOG_INFO, "IS-IS config parsed new circuit #%d -> %s", isis_ci, $2);
				bzero(&isis_ca[isis_ci], sizeof(struct is_pc));
				strcpy(isis_ca[isis_ci].name, $2.ptr);
			    	/* isis_init_ca(&isis_ca[isis_ci], $2); */
			    }
			}
		isis_coptions
			{
			}
		| T_SYSTEMID string
			{
			    isis_set_sysid($2);
			}
		| T_TRACEOPTIONS isis_trace_options
			{
			    isis_save_trace($2);
			}
		| T_SNPA number
			{
			    /* we need an index in lieu of the NSAP address; used to use getpid */
			    isis_snpa = $2;
			}
		| T_AREA string
			{
			    isis_saveArea($2);
			}
		| T_IPREACH isis_level ipreach_type network mask T_METRIC number 
			{
			    save_ipreach($2,$3,&$4,&$5,$7);
			}
		| T_PREFIX prefix_type T_METRIC number string
			{
			    isis_savePrefix($5.ptr, $4, $2);
			}
		| T_INTDOMINFO string
			{
			    isis_saveInterDomInfo($2);
			}
		| T_SET L_HNAME number
			{
			    isis_set_parm($2,$3);
			}
    		/* nothing else for now */
		;

isis_level	: T_LEVEL number
			{
			    $$ = $2;
			}
		| /* empty */
			{
			    $$ = 1;
			}
		;

isis_mode	: T_IP
			{
			    $$ = T_IP;
			}
                | T_ISO
			{
			    $$ = T_ISO;
			}
                | T_DUAL
			{
			    $$ = T_DUAL;
			}
                ;

isis_coptions	: /* Empty */
		| isis_coptions isis_coption
		;

isis_coption	: T_METRIC isis_level number
			{
			    if ($2 == 2) {
				isis_ca[isis_ci].l2metric = $3;
			    } else {
				isis_ca[isis_ci].metric = $3;
			    }
			}
		| T_PRIORITY isis_level number
			{
			    if ($2 == 2) {
				isis_ca[isis_ci].l2prior = $3;
			    } else {
				isis_ca[isis_ci].prior = $3;
			    }
			}
		;

ipreach_type	: T_INTERNAL
			{
			    $$ = ISIS_INTERNAL;
			}
		| T_EXTERNAL
			{
			    $$ = ISIS_EXTERNAL;
			}
		| TISIS_SUMMARY
			{
			    $$ = ISIS_SUMMARY;
			}
		;

prefix_type	: T_INTERNAL
			{
			    $$ = ISIS_INTERNAL;
			}
		| T_EXTERNAL
			{
			    $$ = ISIS_EXTERNAL;
			}
		;

isis_trace_options	: isis_trace_option
			| isis_trace_options isis_trace_option
				{
				    $$ = $1 | $2;
				}
			| isis_trace_options T_EXCEPT isis_trace_option
				{
				    $$ = $1 & ~$2;
				}
			;

isis_trace_option
    		: T_ALL         { $$ = T_ISISALL; }
		| TISIS_IIH		{ $$ = T_DUMPIIH; }
		| TISIS_DUMPLSP         { $$ = T_DUMPLSP; }
		| TISIS_EVENTS          { $$ = T_EVENTS; }
		| TISIS_LANADJ          { $$ = T_LANADJ; }
		| TISIS_FLOODING        { $$ = T_FLOODING; }
		| TISIS_BUILDLSP        { $$ = T_BUILDLSP; }
		| TISIS_CSNP            { $$ = T_CSNP; }
		| TISIS_PSNP            { $$ = T_PSNP; }
		| TISIS_LSPINPUT        { $$ = T_LSPINPUT; }
		| TISIS_P2PADJ          { $$ = T_P2PADJ; }
		| T_SPF             	{ $$ = T_TRSPF; }
		| TISIS_LSPDB           { $$ = T_DUMPDB; }
		| T_ROUTE		{ $$ = T_ROUTE2; }
		| T_UPDATE		{ $$ = T_ROUTE1; }
		| TISIS_PATHS		{ $$ = T_ROUTE3; }
		| TISIS_LSPCONTENT	{ $$ = T_LSPCONTENT; }
		;



/*  */


/*  */


/*  */

    
/*  */


bgp_statement	: T_BGP
			{
			    PROTO_SEEN(RTPROTO_BGP);

			    parse_gwlist = &parse_gwp;

			    bgp_default_trace_flags = trace_flags;
			}
		onoff_option bgp_group
			{
			    switch ($3) {
			    case T_ON:
				doing_bgp = TRUE;

				if (!bgp_conf_check(parse_error)) {
				    PARSE_ERROR;
				}
				break;

			    case T_OFF:
				doing_bgp = FALSE;
			    }

			    parse_gwlist = (gw_entry **) 0;
			}
		;

bgp_group	: /* Empty */
		| '{' bgp_group_stmts '}'
		;

bgp_group_stmts	: /* Empty */
		| bgp_group_stmts bgp_group_stmt L_EOS
		| bgp_group_stmts error L_EOS
			{
			    yyerrok;
			}
		;

bgp_group_stmt	: T_PREFERENCE preference
			{
			    bgp_default_preference = $2;
			}
		| T_DEFAULTMETRIC metric
			{
			    if (parse_metric_check(RTPROTO_BGP, &$2)) {
				PARSE_ERROR;
			    }
			    bgp_default_metric = $2.metric;
			}
		| T_TRACEOPTIONS bgp_trace_options
			{
			    bgp_default_trace_flags = $2;
			}
		| T_GROUP T_TYPE bgp_linktype T_PEERAS as
			{
			    /* Allocate a Group structure */
			    bgp = bgp_conf_group_alloc();

			    /* Set the type and AS */
			    bgp->bgpg_type = $3;
			    bgp->bgpg_peer_as = $5;

			    /* Point at this configuration information */
			    bcp = &bgp->bgpg_conf;

			    /* Inherit global trace flags */
			    bgp->bgpg_trace_flags = bgp_default_trace_flags;
			}
		bgp_peer_options
			{
			    /* Add this group */
			    bgp = bgp_conf_group_add(bgp, parse_error);
			    if (!bgp) {
				PARSE_ERROR;
			    }
			}
		'{' bgp_peer_stmts '}'
			{
			    /* Check group */
			    if (!bgp_conf_group_check(bgp, parse_error)) {
				PARSE_ERROR;
			    }
			}
		;

bgp_peer_options
		: /* Empty */
		| bgp_peer_options bgp_peer_option
		;

bgp_peer_stmts	: /* Empty */
		| bgp_peer_stmts bgp_peer_stmt L_EOS
		| bgp_peer_stmts error L_EOS
			{
				yyerrok;
			}
		;

bgp_peer_stmt	: T_NEIGHBOR
    			{
			    /* Allocate this peer */
			    bnp = bgp_conf_peer_alloc(bgp);

			    /* Point at this configuration information */
			    bcp = &bnp->bgp_conf;
			}
		host bgp_peer_options
			{
			    /* Set peer address */
			    bnp->bgp_addr = $3;

			    bnp = bgp_conf_peer_add(bgp, bnp, parse_error);
			    if (!bnp) {
				PARSE_ERROR;
			    }
			}
		| T_ALLOW
			{
			    parse_contig_masks = FALSE;
			}
			'{' bgp_allow_list '}'
			{
			    parse_contig_masks = TRUE;

			    if (parse_adv_append(&bgp->bgpg_allow, $4, TRUE)) {
				PARSE_ERROR;
			    }
			}
		;

bgp_peer_option	: T_METRICOUT metric
			{
			    if (parse_metric_check(RTPROTO_BGP, &$2)) {
				PARSE_ERROR;
			    }
			    BIT_SET(bcp->bgpc_options, BGPO_METRIC_OUT);
			    bcp->bgpc_metric_out = $2.metric;
			}
		| T_LOCALAS as
			{
			    BIT_SET(bcp->bgpc_options, BGPO_LOCAL_AS);
			    bcp->bgpc_local_as = $2;
			}
		| T_NOGENDEFAULT
			{
			    BIT_SET(bcp->bgpc_options, BGPO_NOGENDEFAULT);
			}
		| T_GATEWAY gateway
			{
			    if (bcp->bgpc_gateway) {
				/* Free the old one */
				sockfree(bcp->bgpc_gateway);
			    }
			    BIT_SET(bcp->bgpc_options, BGPO_GATEWAY);
			    bcp->bgpc_gateway = sockdup($2->gw_addr);
			    gw_freelist(*parse_gwlist);
			    *parse_gwlist = parse_gwp = (gw_entry *) 0;
			}
		| T_PREFERENCE preference
			{
			    BIT_SET(bcp->bgpc_options, BGPO_PREFERENCE);
			    bcp->bgpc_preference = $2;
			}
		| T_LCLADDR interface_local
			{
			    if (bcp->bgpc_lcladdr) {
				/* Free the old one */
				ifae_free(bcp->bgpc_lcladdr);
			    }
			    BIT_SET(bcp->bgpc_options, BGPO_LCLADDR);
			    bcp->bgpc_lcladdr = $2;
			}
		| T_HOLDTIME time
			{
			    if (parse_limit_check("hold time",
						  (u_int) $2,
						  BGP_LIMIT_HOLDTIME)) {
				PARSE_ERROR;
			    }
			    BIT_SET(bcp->bgpc_options, BGPO_HOLDTIME);
			    bcp->bgpc_holdtime_out = $2;
			}
		| T_TRACEOPTIONS bgp_trace_options
			{
			    bcp->bgpc_trace_flags = $2;
			}
		| T_VERSION L_NUMBER
			{
			    if (!BGP_KNOWN_VERSION($2)) {
				(void) sprintf(parse_error, "unsupported BGP version: %d",
					       $2);
				PARSE_ERROR;
			    }
			    BIT_SET(bcp->bgpc_options, BGPO_VERSION);
			    bcp->bgpc_conf_version = $2;
			}
		| T_PASSIVE
			{
			    BIT_SET(bcp->bgpc_options, BGPO_PASSIVE);
			}
		| T_DEFAULTIN
			{
			    BIT_SET(bcp->bgpc_options, BGPO_DEFAULTIN);
			}
		| T_DEFAULTOUT
			{
			    BIT_SET(bcp->bgpc_options, BGPO_DEFAULTOUT);
			}
		| T_KEEPALL
			{
			    BIT_SET(bcp->bgpc_options, BGPO_KEEPALL);
			}
		| T_SENDBUF bgp_kbufsize
			{
			    bcp->bgpc_send_bufsize = $2;
			}
		| T_RECVBUF bgp_kbufsize
			{
			    bcp->bgpc_recv_bufsize = $2;
			}
		| T_SPOOLBUF number
			{
			    if (parse_limit_check("spool buffer size",
						  (u_int) $2,
						  BGP_LIMIT_SBUF)) {
				PARSE_ERROR;
			    }
			    bcp->bgpc_spool_bufsize = $2;
			}
		;

bgp_kbufsize	: number
			{
			    if (parse_limit_check("kernel buffer size",
						  (u_int) $1,
						  BGP_LIMIT_KBUF)) {
				PARSE_ERROR;
			    }
			    $$ = $1;
			}
		;

bgp_linktype	: T_INTERNAL
			{
			    $$ = BGPG_INTERNAL;
			}
		| T_EXTERNAL
			{
			    $$ = BGPG_EXTERNAL;
			}
		| T_IGP
			{
			    $$ = BGPG_INTERNAL_IGP;
			}
		| T_TEST
			{
			    $$ = BGPG_TEST;
			}
		;

bgp_trace_options
		: trace_options
		;

bgp_prop_init
		: T_BGP
			{
			    $$ = parse_proto = RTPROTO_BGP;
			    parse_gwlist = (gw_entry **) 0;
			    trace(TR_PARSE, 0, "parse: %s PROTO: %s",
				  parse_where(),
				  gd_lower(trace_state(rt_proto_bits, $$)));
			}
		;


bgp_allow_list	: /* Empty */
			{
			    $$ = (adv_entry *) 0;
			}
		| bgp_allow_list bgp_allow L_EOS
			{
			    $$ = adv_alloc(ADVFT_DM, (proto_t) 0);
			    $$->adv_dm = $2;
			    $$ = parse_adv_dm_append($1, $$);
			    if (!$$) {
				PARSE_ERROR;
			    }
			}
		| bgp_allow_list error L_EOS
			{
			    yyerrok;
			}
		;
		
	
/* Destination and mask pair used for control lists */
bgp_allow	: all_mask
		| host_mask
		| network_mask
		;


/*  */


idrp_statement          : T_IDRP
                                {
					/* global parameters for protocol */
	
                                        idrp_trace_flags = trace_flags;
                                        parse_proto = RTPROTO_IDRP;
                                        idrp_parse_group_index = 0;
                                }
                        onoff_option idrp_group L_EOS
                                {
			    		PROTO_SEEN(RTPROTO_IDRP);
                                        if ($3 == T_OFF) {
                                                doing_idrp = FALSE;
                                                trace(TR_PARSE, 0, "parse: %s idrp off ;", parse_where());
                                        } else {
						/* set flag and initialize local nodes idrpPeer struct */
                                                doing_idrp = TRUE;
						idrp_local_peer_init();

						/* check to make sure RDI specified */
                                                if (!rt_rdi.isoa_len) {
                                                        (void) sprintf(parse_error, "RDI not specified");
                                                        PARSE_ERROR;
                                                }
                                                /* %% */
						/* here we need to change for */
						/* idrp for IP */
		
                                                if (!rt_net.isoa_len) {
                                                        (void) sprintf(parse_error, "NET not specified");
                                                        PARSE_ERROR;
                                                }

                                                if (!idrp_neighbors) {
                                                        (void) sprintf(parse_error, "no IDRP neighbors specified");
                                                        PARSE_ERROR;
                                                }
                                                IDRP_LIST(idgp) {
                                                        trace(TR_PARSE, 0, "parse:  %s, IDRP neighbor %s, addr %A",
								parse_where(), idgp->name, 
								sockbuild_iso((byte *)idgp->neighbor.isoa_genaddr,
								idgp->neighbor.isoa_len));
                                                } IDRP_LIST_END;
                                        }
                                }
                        ;


idrp_group_stmt         : T_TRACEOPTIONS idrp_trace_options
                                {
                                        idrp_trace_flags = $2;
                                }

			| T_LOCALNODE
		       '{' idrp_local_node '}'

                        | T_GROUP
                                {
					/* zero the group address functions */
 
                                        bzero((caddr_t) &idrp_group, sizeof(idrp_group));
                                        gr_idrp = &idrp_group;
                                        gr_idrp->trace_flags = idrp_trace_flags;
                                        idrp_parse_group_index++;
                                }
                        idrp_group_options '{' idrp_peer_stmts '}'
                        | T_OSILOCAL
                       '{' idrp_local_routes '}'
                        ;

idrp_peer_stmts         : /* empty */
                        | idrp_peer_stmts idrp_peer_stmt L_EOS
                        | idrp_peer_stmts error L_EOS
                                {
                                        yyerrok;
                                }
                        ;


idrp_local_node		: /* empty */
                        | idrp_local_node_stmts  
                        ;

idrp_local_node_stmts   : /* empty */
                        | idrp_local_node_stmts idrp_local_node_stmt L_EOS
                        | idrp_local_node_stmts error L_EOS
                                {
                                        yyerrok;
                                }
                        ;

idrp_local_node_stmt    : /* empty */   
			| T_PROTOSUPPORT idrp_proto_supp
			| T_ROUTESERVER onoff_option
				{
				idrp_this_node.route_server = FALSE;
				if ($2 == T_ON)
					{
					idrp_this_node.route_server = TRUE;
					}
				}

			| T_MULTIEXIT onoff_option 
				{
				idrp_this_node.multi_exit = FALSE;
				if ($2 == T_ON)
					{
					idrp_this_node.multi_exit = TRUE;
					}
				}

			| idrp_auth 
			| idrp_sock_support 
			| T_OUTPDUS L_NUMBER 
				{
				idrp_this_node.rcv_credit = $2;
				}

			| T_LISTENOPEN onoff_option
				{
				idrp_this_node.listen_for_open = FALSE;
				if ($2 == T_ON)
					{
					idrp_this_node.listen_for_open = TRUE;
					}
				}

			| T_HOLDTIME L_NUMBER
				{
				idrp_this_node.hold_time = $2;
				}

			| T_CLOSEWAIT L_NUMBER 
				{
				idrp_this_node.close_wait = $2;
				}

			| T_REXMIT L_NUMBER 
				{
				idrp_this_node.rexmit_time = $2;
				}
			| T_RIBCHECK L_NUMBER
				{
				idrp_this_node.ribcheck_timer = $2;
				}
			| T_MINADV L_NUMBER
				{
				idrp_this_node.min_adv_time = $2;
				}
			| T_MINADVRD L_NUMBER
				{
				idrp_this_node.min_advRD_time = $2;
				}
			;

idrp_auth		: T_AUTHTYPE L_NUMBER 
				{
				idrp_this_node.AuthType = $2; 
				}
			| T_AUTHTYPE L_NUMBER T_AUTHCODE L_HEXSTRING
				{
				idrp_this_node.AuthType = $2;
				bcopy($4.ptr,&idrp_this_node.AuthCode[0],$4.len);
				}	
			;

idrp_sock_support	: /* empty */
			| T_PROTOSOCK idrp_sock_options 
				{
				idgp = &idrp_this_node;
				}
			;

idrp_group              : /* empty */
                        | '{' idrp_group_stmts '}'
                        ;

idrp_group_stmts        : /* empty */
                        | idrp_group_stmts idrp_group_stmt L_EOS
                        | idrp_group_stmts error L_EOS
                                {
                                        yyerrok;
                                }
                        ;


idrp_peer_stmt          : T_NEIGHBOR
                                {
					/* allocate a peer */

					idgp = idrp_peer_alloc(0);

					/* if group has not been initialized - do it now */
					/* all of idrpPeer structure will be zero       */
	 
       	                               if (!gr_idrp)
						{	
						gr_idrp = idrp_peer_alloc(&idrp_group);
						}
                                }
                        onoff_option osi_address host T_RDI osi_address idrp_intf_option idrp_sock_options idrp_peer_options
                                {
                                        idrpPeer *idgp2;

                                        if ($3 == T_OFF)
						 {
						 idgp->idrp_flags = IDRPF_IDLED;
						}
					else	
						{
						idgp->idrp_flags = IDRPF_TRY_CONNECT;
						}   		

					/* copy in group trace flags 
					 * in this sequence, the RDI is not copied 
					 */
 
					idgp->trace_flags = gr_idrp->trace_flags;
					idgp->type = gr_idrp->type;

					/* copy rdi, NET, IP address for now */
					/* need to put interface someplace */

					bcopy($4.ptr, idgp->neighbor.isoa_genaddr,$4.len);
					idgp->neighbor.isoa_len = $4.len;
					idgp->neighbor.isoa_family = AF_ISO;
                                        bcopy($7.ptr, idgp->rdi.isoa_genaddr,$7.len);
                                        idgp->rdi.isoa_len = $7.len;

				        sprintf(idgp->name, "%A", 
						sockbuild_iso((byte *) idgp->neighbor.isoa_genaddr, 
						idgp->neighbor.isoa_len));
	
					/* is it my rdi - if so, internal neighbor */

					if (idgp->type == IDRP_PEER_INTERNAL)
						{
						if (isoaddrcmp(&idgp->rdi,&rt_rdi))
							{
							/* rdi does not match */
							trace ( TR_IDRP,0,"internal neighbor %s has internal neighbor but rdi is not ours", \
							idgp->name);
							}					 
						}

                                        idgp->ip_neighbor = $5;
                                        idgp->gw.gw_proto = RTPROTO_IDRP;
					idgp->gw.gw_addr->in.sin_port = IDRP_PORT;

					/* if reparsing - check for this peer in
					 * existing list
					 */

					/* check for this peer being on the chain
					 * already
					 */

					idgp2 = idrp_find_peer(idgp);
					if (idgp2)
						{
						/* found a peer structure */
						if (idrp_reparsing)
							{
							/* re-use old structure, update
							 * with the new parameters  so far
							 * and free duplicate peer structure
							 */
 
							idrp_peer_update(idgp2,idgp);
							idrp_big_blk_free(idgp);
							idgp = idgp2;	
							}
						else
							{
							trace (TR_IDRP,0,"duplicate peer %s",idgp->name);
							PARSE_ERROR;
							}	
						}
					else
						{
						/* not a duplicate and not a reparsing re-used
						 * peer structure 
						 */
		
						idrp_link_peer(idgp);
						}
                                }
                        ;

idrp_local_routes       : /* empty */
                        | idrp_local_routes idrp_local_route L_EOS  
                        | idrp_local_routes error L_EOS
                                {
                                        yyerrok;
                                }
                        ;


idrp_local_route       : osi_prefix osi_prefix 
                                {
				/* create a local route with this NLRI information */

				int bits; 
				idrpRoute_options	*p_opts = 0; 

				bits = $1.len * 8; 
				if (!idrp_local_rt($1.ptr,$1.len,bits,AF_ISO,$2.ptr,$2.len,p_opts))
					{ 
					trace (TR_PARSE,0,"maximum number of local routes exceeded, no more added to routing table");
					}
                                }
			   | osi_prefix osi_prefix idrp_route_options 
				{
				int bits; 
				bits = $1.len * 8;
				if (!idrp_local_rt($1.ptr,$1.len,bits,AF_ISO,$2.ptr,$2.len,&idrp_rt_opts))
					{ 
					trace (TR_PARSE,0,"maximum number of local routes exceeded, no more added to routing table");
					}
                                }

			| osi_prefix osi_prefix T_IP idrp_route_options 
				{
				int bits; 
				bits = $1.len * 8;
				if (!idrp_local_rt($1.ptr,$1.len,bits,AF_INET,$2.ptr,$2.len,&idrp_rt_opts))
					{ 
					trace (TR_PARSE,0,"maximum number of local routes exceeded, no more added to routing table");
					}
				}

			| osi_prefix osi_prefix idrp_dist_options
				{
				int bits;
				bits = $1.len * 8;
				if (!idrp_local_rt($1.ptr,$1.len,bits,AF_ISO,$2.ptr,$2.len,&idrp_rt_opts))
					{ 
					trace (TR_PARSE,0,"maximum number of local routes exceeded, no more added to routing table");
					}
				}

			| osi_prefix osi_prefix idrp_route_options idrp_dist_options
				{
				int bits;
				bits = $1.len * 8;
				if (!idrp_local_rt($1.ptr,$1.len,bits,AF_ISO,$2.ptr,$2.len,&idrp_rt_opts))
					{ 
					trace (TR_PARSE,0,"maximum number of local routes exceeded, no more added to routing table");
					}
				}
                        ;

idrp_route_options	: idrp_route_options idrp_route_option L_EOS  
                        | idrp_route_option error L_EOS
                                {
                                        yyerrok;
				}
			;

idrp_route_option	: T_SNPA byte_string 
				{
				/* note I am using byte string and next hop gateway
				 * as the operative thing to insert into
				 * - gateway table, but this may have to be re-examined
				 * - is uses a number and then gets the byte string
				 */ 

				bcopy($2.ptr,&idrp_rt_opts.next_hop_snpas.snpa1.snpa,$2.len);
				idrp_rt_opts.next_hop_snpas.cnt = 1;	
				idrp_rt_opts.next_hop_snpas.pdu_len = $2.len; 
				}
				 
			| T_MULTIEXIT number 
				{
				idrp_rt_opts.multi_exit = $2;
				}
			| T_MULTIEXIT number T_SNPA byte_string  
				{
				idrp_rt_opts.multi_exit = $2;
				bcopy($4.ptr,&idrp_rt_opts.next_hop_snpas.snpa1.snpa,$4.len);
				idrp_rt_opts.next_hop_snpas.cnt = 1;	
				idrp_rt_opts.next_hop_snpas.pdu_len = $4.len; 
				}
			| T_SNPA byte_string T_MULTIEXIT number 
				{
				bcopy($2.ptr,&idrp_rt_opts.next_hop_snpas.snpa1.snpa,$2.len);
				idrp_rt_opts.next_hop_snpas.cnt = 1;	
				idrp_rt_opts.next_hop_snpas.pdu_len = $2.len; 
				idrp_rt_opts.multi_exit = $4;
				}
			| T_ROUTESERVER
				{
				idrp_rt_opts.route_server = IDRP_SERVER_ALLOWED;
				}
			; 

idrp_dist_options	: idrp_dist_options idrp_dist_option L_EOS  
                        | idrp_dist_option error L_EOS
                                {
                                        yyerrok;
				}
			;

idrp_dist_option	: /* empty */
			|  T_PROHIBITDIST idrp_dist_list
				{
				if (idrp_rt_opts.p_DIST_LIST_EXCL)
					{
					p_idrp_excl_rdi_tail->p_next = p_rdi_head;
					p_idrp_excl_rdi_tail = p_rdi_tail;
					}
				else
					{
					p_idrp_excl_rdi_tail = p_idrp_excl_rdi_head = p_rdi_head;
					idrp_rt_opts.p_DIST_LIST_EXCL = p_rdi_head;
					} 
				}

			| T_ALLOWDIST	idrp_dist_list
				{
				if (idrp_rt_opts.p_DIST_LIST_INCL)
					{
					p_idrp_incl_rdi_tail->p_next = p_rdi_head;
					}
				else
					{
					p_idrp_incl_rdi_tail = p_idrp_incl_rdi_head = p_rdi_head;
					idrp_rt_opts.p_DIST_LIST_INCL = p_rdi_head;
					}
				}
			;


idrp_dist_list		: idrp_dist_list idrp_dist_rdi L_EOS  
                        | idrp_dist_rdi error L_EOS
                                {
                                        yyerrok;
				}
			;

idrp_dist_rdi		: osi_prefix
				{
				if (!p_rdi_head)
					{
					p_rdi_head = p_rdi_tail = create_local_rd_path($1.len,$1.ptr);
					}
				else
					{
					p_rdi_tail->p_next = create_local_rd_path($1.len,$1.ptr);
					p_rdi_tail = p_rdi_tail->p_next;
					p_rdi_tail->p_next = (idrp_canon_rdpath *) 0;
					}
				} 	
			;	
			
		
idrp_intf_option        : /* empty */
                        | T_INTERFACE host
                                {
                                        idgp->ip_intf_addr = $2;
					idgp->ip_intf_addr->in.sin_port = IDRP_PORT;
					trace (TR_PARSE, 0, "ip interface address configed"); 
                                }
			| T_INTERFACE osi_address
				{
					idrp_set_iso_sockun((struct sockadd_un *)&idgp->iso_gw_addr,$2.len,$2.ptr);	
					trace (TR_PARSE, 0, "iso interface address configed"); 
				}
			| T_INTERFACE host osi_address
				{
	                                idgp->ip_intf_addr = $2;
					idgp->ip_intf_addr->in.sin_port = IDRP_PORT;
					trace (TR_PARSE, 0, "ip interface address configed"); 
					idrp_set_iso_sockun((struct sockadd_un *)&idgp->iso_gw_addr,$3.len,$3.ptr);	
					trace (TR_PARSE, 0, "iso interface address configured"); 
				}
			| T_INTERFACE osi_address host
				{
	                                idgp->ip_intf_addr = $3;
					idgp->ip_intf_addr->in.sin_port = IDRP_PORT;
					trace (TR_PARSE, 0, "ip interface address configed"); 
					idrp_set_iso_sockun((struct sockadd_un *)&idgp->iso_gw_addr,$2.len,$2.ptr);	
					trace (TR_PARSE, 0, "iso interface address configured"); 
				}
                        ;


idrp_group_options      : idrp_group_option
                        | idrp_group_options idrp_group_option
                        | idrp_peer_option
                        | idrp_peer_options idrp_peer_option
                        ;

idrp_group_option       : T_RDI osi_rdi
                                {
                                        bcopy($2.ptr,&gr_idrp->rdi.isoa_genaddr, $2.len);
                                        gr_idrp->rdi.isoa_len = $2.len;
                                }
                        | T_INTERNAL
                                {
                                        gr_idrp->type = IDRP_PEER_INTERNAL;
                                }
                        | T_EXTERNAL
                                {
                                        gr_idrp->type = IDRP_PEER_EXTERNAL;
                                }
                        ;


idrp_sock_options	: /* empty */ 
			| idrp_sock_options idrp_sock_option
			| idrp_sock_option
			;

idrp_sock_option	:  T_IP
				{
				idgp->proto_sock = IDRP_PDU_PROTO_IP_RAW;
				} 
			| T_UDP
				{
				idgp->proto_sock = IDRP_PDU_PROTO_UDP;
				}

			|  T_ISO		 	 
				{
				/*  raw clnp socket */
				idgp->proto_sock = IDRP_PDU_PROTO_CLNP;
				}
			| T_IDRP
				{
				idgp->proto_sock = IDRP_PDU_PROTO_IDRP;
				}
			;
idrp_proto_supp		: T_ISO 
				{
				idrp_this_node.nlri_supp[IDRP_NLRI_ISO] = TRUE;
				}
			| T_IP
				{
				idrp_this_node.nlri_supp[IDRP_NLRI_IP] = TRUE;
				}
			| T_IP T_ISO
				{
				idrp_this_node.nlri_supp[IDRP_NLRI_IP] = TRUE;
				idrp_this_node.nlri_supp[IDRP_NLRI_ISO] = TRUE;
				idrp_this_node.nlri_supp[IDRP_NLRI_PIP] = TRUE;
				}
			
			| T_IP T_ISO T_PIP T_SIP
				{
				idrp_this_node.nlri_supp[IDRP_NLRI_IP] = TRUE;
				idrp_this_node.nlri_supp[IDRP_NLRI_ISO] = TRUE;
				idrp_this_node.nlri_supp[IDRP_NLRI_PIP] = TRUE;
				idrp_this_node.nlri_supp[IDRP_NLRI_SIP] = TRUE;
				}
			| T_PIP
				{
				idrp_this_node.nlri_supp[IDRP_NLRI_PIP] = TRUE;
				}	
			| T_SIP
				{
				idrp_this_node.nlri_supp[IDRP_NLRI_SIP] = TRUE;
				}	
			;

idrp_peer_options       : /* empty */
                        | idrp_peer_options idrp_peer_option
                        ;

idrp_peer_option        : T_TRACEOPTIONS idrp_trace_options
                                {
                                        idgp->trace_flags = $2;
                                }
                        ;

osi_address             : osi_nsap
                        | osi_address_name osi_nsap
				{
                                        $$ = $2;
                                }
                        ;

osi_nsap                : L_HEXSTRING

osi_rdi                 : L_HEXSTRING
                        ;

osi_prefix              : L_HEXSTRING
                        | L_HEXSTRING_ODD
                        ;


osi_address_name        : T_NET
                        ;
/* 			| T_NSAP */

idrp_trace_options      : trace_options
			;


/*  */

redirect_statement	: T_REDIRECT
			{
#if	defined(PROTO_ICMP) || defined(KRT_RT_SOCK)
			    PROTO_SEEN(RTPROTO_REDIRECT);

			    parse_gwlist = &redirect_gw_list;

			    redirect_trace_flags = trace_flags;
#endif	/* defined(PROTO_ICMP) || defined(KRT_RT_SOCK) */
			}
		onoff_option redirect_group
			{
#if	defined(PROTO_ICMP) || defined(KRT_RT_SOCK)
			    switch ($3) {
			    case T_ON:
				/* Enabled by default */
				break;

			    case T_OFF:
				redirect_disable(RTPROTO_REDIRECT);
				break;
			    }
#endif	/* defined(PROTO_ICMP) || defined(KRT_RT_SOCK) */
			}
		;

redirect_group	: /* Empty */
		| '{' redirect_group_stmts '}'
		;

redirect_group_stmts
		: /* Empty */
		| redirect_group_stmts redirect_group_stmt L_EOS
		| redirect_group_stmts error L_EOS
			{
			    yyerrok;
			}
		;

redirect_group_stmt	: T_PREFERENCE preference
			{
#if	defined(PROTO_ICMP) || defined(KRT_RT_SOCK)
			    redirect_preference = $2;
#endif	/* defined(PROTO_ICMP) || defined(KRT_RT_SOCK) */
			}
		| T_INTERFACE interface_all_list redirect_interface_options
			{
#if	defined(PROTO_ICMP) || defined(KRT_RT_SOCK)
			    config_list *list = config_list_alloc($3, 0);

			    parse_adv_propagate_config($2, list, RTPROTO_REDIRECT);

			    if (parse_adv_append(&redirect_int_policy, $2, TRUE)) {
				PARSE_ERROR;
			    }

			    config_list_free(list);
#endif	/* defined(PROTO_ICMP) || defined(KRT_RT_SOCK) */
			}
		| T_TRUSTEDGATEWAYS gateway_list
			{
#if	defined(PROTO_ICMP) || defined(KRT_RT_SOCK)
			    redirect_n_trusted += parse_gw_flag($2, RTPROTO_REDIRECT, GWF_TRUSTED);
			    if (!redirect_n_trusted) {
				PARSE_ERROR;
			    }
#endif	/* defined(PROTO_ICMP) || defined(KRT_RT_SOCK) */
			}
		| T_TRACEOPTIONS redirect_trace_options
			{
			    redirect_trace_flags = $2;
			}
		;

redirect_interface_options
		: redirect_interface_option
		| redirect_interface_options redirect_interface_option
			{
			    $$ = config_append($1, $2);
			    if (!$$) {
				PARSE_ERROR;
			    }
			}
		;

redirect_interface_option
		: T_NOREDIRECTS
			{
#if	defined(PROTO_ICMP) || defined(KRT_RT_SOCK)
			    $$ = config_alloc(REDIRECT_CONFIG_NOIN, (void_t) TRUE);
#endif	/* defined(PROTO_ICMP) || defined(KRT_RT_SOCK) */
			}
		;

redirect_trace_options
		: trace_options
		;

/*  */



/*  */    


/*  */

route_order	: /* Empty */
			{
			    if (parse_new_state(PS_ROUTE)) {
				PARSE_ERROR;
			    }
			}
		;

route_statement : T_STATIC
			{
			    /* Need to set this for static routes, not used for interface routes */
			    parse_proto = RTPROTO_STATIC;
			    parse_gwlist = &rt_gw_list;
			    rt_open(rt_task);
			}
		'{' route_stmts '}'
			{
			    rt_close(rt_task, (gw_entry *) 0, 0, NULL);
			}
		;

route_stmts	: route_stmt L_EOS
		| route_stmts route_stmt L_EOS
		| error L_EOS
			{
				yyerrok;
			}
		;

route_stmt	: route_dest T_GATEWAY gateway_list interface_list_option preference_option rts_options
			{
			    /* Route to a gateway */

			    if (rt_parse_route($1.dm_dest,
					       $1.dm_mask,
					       $3,
					       $4,
					       PARSE_METRIC_ISSET(&$5) ? (pref_t) $5.metric : RTPREF_STATIC,
					       $6 | RTS_GATEWAY,
					       parse_error)) {
				PARSE_ERROR;
			    }

			    sockfree($1.dm_dest);
			}
		| route_dest T_INTERFACE interface_all preference_option rts_options
			{
			    /* Route to an interface */

			    if (rt_parse_route($1.dm_dest,
					       $1.dm_mask,
					       (adv_entry *) 0,
					       $3,
					       PARSE_METRIC_ISSET(&$4) ? (pref_t) $4.metric : RTPREF_STATIC,
					       $5,
					       parse_error)) {
				PARSE_ERROR;
			    }

			    sockfree($1.dm_dest);
			}
		;

route_dest	: host_mask
    		| network_mask
			{
#ifndef	VARIABLE_MASKS
#ifdef	PROTO_INET
			    if (socktype($1.dm_dest) == AF_INET) {
				sockaddr_un *ifmask = inet_ifsubnetmask(sock2in($1.dm_dest));

				if (!ifmask) {
				    ifmask = inet_mask_natural($1.dm_dest);
				}

				if (ifmask != $1.dm_mask) {
				    sprintf(parse_error, "variable subnet masks not supported on this system");
				    PARSE_ERROR;
				}
			    }
#endif	/* PROTO_INET */
#endif	/* VARIABLE_MASKS */
			    $$ = $1;			/* struct copy */
			}
    		| T_DEFAULT
			{
			    $$.dm_dest = sockdup(inet_addr_default);
			    $$.dm_mask = inet_mask_default;
			    trace(TR_PARSE, 0, "parse: %s DEST: %A MASK: %A",
				  parse_where(),
				  $$.dm_dest,
				  $$.dm_mask);
			}
		| T_ISO T_DEFAULT
			{
			    $$.dm_dest = iso_addr_default;
			    $$.dm_mask = iso_mask_default;
			    trace(TR_PARSE, 0, "parse: %s DEST: %A MASK: %A",
				  parse_where(),
				  $$.dm_dest,
				  $$.dm_mask);
			}
    		;

rts_options	: /* Empty */
			{
			    $$ = 0;
			}
		| rts_options rts_option
			{
			    $$ = $1 | $2;
			}
		;

rts_option	: T_RETAIN
			{
			    $$ = RTS_RETAIN;
			}
		| T_REJECT
			{
			    $$ = RTS_REJECT;
			}
		| T_BLACKHOLE
			{
			    $$ = RTS_BLACKHOLE;
			}
		| T_NOINSTALL
			{
			    $$ = RTS_NOTINSTALL;
			}
		;

/*  */

control_order	: /* Empty */
			{
			    if (parse_new_state(PS_CONTROL)) {
				PARSE_ERROR;
			    }
			}
		;
		
control_statement
		: T_IMPORT T_PROTO control_exterior T_AS as import_option L_EOS
			{
#if	defined(PROTO_EGP) || defined(PROTO_BGP)
			    adv_entry *adv;
			    pmet_t *pp = &$6.metric;

			    /*
			     *	Tack the list of destinations onto the end of the list
			     *	for neighbors with the specified AS.
			     */
			    adv = adv_alloc(ADVFT_AS, $3.proto);
			    adv->adv_as = $5;
			    adv->adv_list = $6.adv;
			    if (PARSE_METRIC_ISSET(pp)) {
				if (PARSE_METRIC_ISRESTRICT(pp)) {
				    BIT_SET(adv->adv_flag, ADVF_NO);
				} else {
				    BIT_SET(adv->adv_flag, ADVFOT_PREFERENCE);
				    adv->adv_result.res_preference = pp->metric;
				}
			    }

			    if (!parse_adv_as($3.import_list, adv)) {
				PARSE_ERROR;
			    }
#endif	/* defined(PROTO_EGP) || defined(PROTO_BGP) */
			}
		| T_IMPORT T_PROTO T_BGP aspath_match import_option L_EOS
			{
			    adv_entry *adv;
			    pmet_t *pp = &$5.metric;

			    /*
			     *	Tack the list of destinations onto the end of the list
			     *	for neighbors with the specified AS.
			     */
			    adv = adv_alloc(ADVFT_ASPATH, RTPROTO_BGP);
			    adv->adv_aspath = (caddr_t) $4;
			    adv->adv_list = $5.adv;
			    if (PARSE_METRIC_ISSET(pp)) {
				if (PARSE_METRIC_ISRESTRICT(pp)) {
				    BIT_SET(adv->adv_flag, ADVF_NO);
				} else {
				    BIT_SET(adv->adv_flag, ADVFOT_PREFERENCE);
				    adv->adv_result.res_preference = pp->metric;
				}
			    }

			    if (parse_adv_append(&bgp_import_aspath_list, adv, TRUE)) {
				PARSE_ERROR;
			    }
			}
		| T_IMPORT T_PROTO import_interior interface_gateway_list import_option L_EOS
			{
			    /*
			     *	Append the dest_mask list to the end of the import list
			     *	for the specified protocol.
			     */
			    if (!parse_adv_propagate_preference($4, $3.proto, &$5.metric, $5.adv)) {
				PARSE_ERROR;
			    }

			    adv_free_list($5.adv);

			    switch ($4->adv_flag & ADVF_TYPE) {
			    case ADVFT_GW:
				/* Gateway list */
			        {
				    adv_entry *adv = $4;
				    adv_entry *advn;

				    do {
					advn = adv->adv_next;
					adv->adv_next = NULL;
					/* Reset the type */
					adv->adv_flag = (adv->adv_flag & ~ADVF_TYPE) | ADVFT_ANY;
					if (parse_adv_append(&adv->adv_gwp->gw_import, adv, TRUE)) {
					    PARSE_ERROR;
					}
				    } while (adv = advn);
				}
				break;

			    case ADVFT_ANY:
				/* Global list */
				if (parse_adv_append($3.import_list, $4, TRUE)) {
				    PARSE_ERROR;
				}
				break;

			    case ADVFT_IFN:
			    case ADVFT_IFAE:
				/* Interface list */
				if (parse_adv_append(&int_import[$3.proto], $4, TRUE)) {
				    PARSE_ERROR;
				}
				break;

			    default:
				assert(FALSE);
			    }
			}
		| T_EXPORT T_PROTO control_exterior T_AS as prop_source_option L_EOS
			{
#if	defined(PROTO_EGP) || defined(PROTO_BGP)
			    adv_entry *adv;
			    adv_entry *adv_prop = $6.adv;
			    pmet_t *mp = &$6.metric;

			    /*
			     *	Tack the list of destinations onto the end of the list
			     *	for neighbors with the specified AS.
			     */
			    adv = adv_alloc(ADVFT_AS, $3.proto);
			    adv->adv_as = $5;
			    adv->adv_list = adv_prop;
			    if (PARSE_METRIC_ISRESTRICT(mp)) {
				BIT_SET(adv->adv_flag, ADVF_NO);
			    } else if (PARSE_METRIC_ISSET(mp)) {
				BIT_SET(adv->adv_flag, ADVFOT_METRIC);
				adv->adv_result.res_metric = mp->metric;
			    }

			    if (!parse_adv_as($3.export_list, adv)) {
				PARSE_ERROR;
			    }
#endif	/* defined(PROTO_EGP) || defined(PROTO_BGP) */
			}
		| T_EXPORT T_PROTO export_interior interface_gateway_list prop_source_option L_EOS
			{
			    /*
			     *	Append the dest_mask list to the end of the export list
			     *	for the specified protocol.
			     */

			    if (!parse_adv_propagate_metric($4, $3.proto, &$5.metric, $5.adv)) {
				PARSE_ERROR;
			    }

			    adv_free_list($5.adv);

			    switch ($4->adv_flag & ADVF_TYPE) {
			    case ADVFT_GW:
				/* Gateway list */
			        {
				    adv_entry *adv = $4;
				    adv_entry *advn;
				
				    do {
					advn = adv->adv_next;
					adv->adv_next = NULL;
					if (parse_adv_append(&adv->adv_gwp->gw_export, adv, TRUE)) {
					    PARSE_ERROR;
					}
				    } while (adv = advn);
				}
				break;

			    case ADVFT_ANY:
				/* Global list for the protocol */
				if (parse_adv_append($3.export_list, $4, TRUE)) {
				    PARSE_ERROR;
				}
				break;

			    case ADVFT_IFN:
			    case ADVFT_IFAE:
				/* Interface list */
				if (parse_adv_append(&int_export[$3.proto], $4, TRUE)) {
				    PARSE_ERROR;
				}
				break;

			    default:
				assert(FALSE);
			    }
			}
		;	

/*  */


/*  */

/* Support for import clauses */

control_exterior	: proto_exterior
			{
			    parse_export_proto = $$.proto = $1;

			    switch ($1) {
#if	defined(PROTO_EGP)
			    case RTPROTO_EGP:
				$$.import_list = &egp_import_list;
				$$.export_list = &egp_export_list;
				break;
#endif

#if	defined(PROTO_BGP)
			    case RTPROTO_BGP:
				$$.import_list = &bgp_import_list;
				$$.export_list = &bgp_export_list;
				break;
#endif

			    default:
				(void) sprintf(parse_error,
					       "unknown import protocol: %s",
					       gd_lower(trace_state(rt_proto_bits, $1)));
				PARSE_ERROR;
			    }
			}
		;


import_interior : proto_interior
			{
			    $$ = $1;
			    
			    parse_proto = $1.proto;
			    parse_gwlist = $1.gw_list;

			    if (!$1.import_list) {
				(void) sprintf(parse_error,
					       "unknown import protocol: %s",
					       gd_lower(trace_state(rt_proto_bits, $1.proto)));
				PARSE_ERROR;
			    }
			    
			    trace(TR_PARSE, 0, "parse: %s PROTO: %s",
				  parse_where(),
				  gd_lower(trace_state(rt_proto_bits, $1.proto)));
			}
		;


import_option	: T_RESTRICT
			{
			    PARSE_METRIC_RESTRICT(&$$.metric);
			    $$.adv = (adv_entry *) 0;
			}
		| preference_option '{' import_list '}'
			{
			    $$.metric = $1;
			    $$.adv = $3;
			}
		;

import_list	: /* Empty */
			{
			    $$ = (adv_entry *) 0;
			}
		| import_list import_listen L_EOS
			{
			    
			    $$ = parse_adv_dm_append($1, $2);
			    if (!$$) {
				PARSE_ERROR;
			    }
			}
		| import_list error L_EOS
			{
			    yyerrok;
			}
		;

import_listen	:
			{
			    parse_contig_masks = FALSE;
			}
			dest_mask preference_restrict_option
			{
			    parse_contig_masks = TRUE;

			    $$ = adv_alloc(ADVFT_DM, parse_proto);
			    $$->adv_dm = $2;
			    if (PARSE_METRIC_ISSET(&$3)) {
				if (PARSE_METRIC_ISRESTRICT(&$3)) {
				    BIT_SET($$->adv_flag, ADVF_NO);
				} else {
				    BIT_SET($$->adv_flag, ADVFOT_PREFERENCE);
				    $$->adv_result.res_preference = $3.metric;
				}
			    }
			}
		;

/*  */

/* Support for Export clauses */

export_interior : proto_interior
			{
			    $$ = $1;
			    
			    parse_export_proto = parse_proto = $1.proto;
			    parse_gwlist = $1.gw_list;

			    if (!$1.export_list) {
				(void) sprintf(parse_error,
					       "unknown export protocol: %s",
					       gd_lower(trace_state(rt_proto_bits, $1.proto)));
				PARSE_ERROR;
			    }
			    
			    trace(TR_PARSE, 0, "parse: %s PROTO: %s",
				  parse_where(),
				  gd_lower(trace_state(rt_proto_bits, $1.proto)));
			}
		;




prop_source_option
    		: T_RESTRICT
			{
			    PARSE_METRIC_RESTRICT(&$$.metric);
			    $$.adv = (adv_entry *) 0;
			}
		| metric_option '{' prop_source_list '}'
			{
			    $$.metric = $1;
			    $$.adv = $3;
			}
		;

prop_source_list
		: /* Empty */
			{
			    $$ = (adv_entry *) 0;
			}
		| prop_source_list prop_source L_EOS
			{
			    $$ = $1;
			    if (parse_adv_append(&$$, $2, TRUE)) {
				PARSE_ERROR;
			    }
			}
		| prop_source_list error L_EOS
			{
			    yyerrok;
			}
		;

prop_source	: T_PROTO proto_exterior T_AS as prop_metric_option
			{
			    $$ = adv_alloc(ADVFT_AS | ADVF_FIRST, (proto_t) 0);
			    $$->adv_as = $4;
			    $$ = parse_adv_propagate_metric($$, $2, &$5.metric, $5.adv);
			}
    /* AS paths */
		| T_PROTO aspath_prop_init aspath_match prop_metric_option
			{
			    $$ = adv_alloc(ADVFT_ASPATH | ADVF_FIRST, (proto_t) 0);
			    $$->adv_aspath = (caddr_t) $3;
			    $$ = parse_adv_propagate_metric($$, $2, &$4.metric, $4.adv);
			}
    /* Tags */
    		| T_PROTO tag_prop_init T_TAG tag prop_metric_option
			{
			    $$ = adv_alloc(ADVFT_TAG|ADVF_FIRST, (proto_t) 0);
			    $$->adv_tag = $4.metric;
			    $$ = parse_adv_propagate_metric($$, $2, &$5.metric, $5.adv);
			}
    /* Interfaces */
		| T_PROTO prop_direct interface_list_option prop_metric_option
			{
			    $$ = parse_adv_propagate_metric($3, $2, &$4.metric, $4.adv);
			}
    /* Static routes */
		| T_PROTO prop_static interface_list_option prop_metric_option
			{
			    $$ = parse_adv_propagate_metric($3, $2, &$4.metric, $4.adv);
			}
    /* Kernel static routes (routing socket) */
		| T_PROTO prop_kernel interface_list_option prop_metric_option
			{
			    $$ = parse_adv_propagate_metric($3, $2, &$4.metric, $4.adv);
			}
    /* Internal default route */
		| T_PROTO prop_default prop_metric_option
			{
			    $$ = adv_alloc(ADVFT_ANY | ADVF_FIRST, (proto_t) 0);
			    $$ = parse_adv_propagate_metric($$, $2, &$3.metric, $3.adv);
			}
		;

prop_direct	: T_DIRECT
			{
			    $$ = parse_proto = RTPROTO_DIRECT;
			    parse_gwlist = (gw_entry **)0;
			    trace(TR_PARSE, 0, "parse: %s PROTO: %s",
				  parse_where(),
				  gd_lower(trace_state(rt_proto_bits, $$)));
			}
		;

prop_static	: T_STATIC
			{
			    $$ = parse_proto = RTPROTO_STATIC;
			    parse_gwlist = &rt_gw_list;
			    trace(TR_PARSE, 0, "parse: %s PROTO: %s",
				  parse_where(),
				  gd_lower(trace_state(rt_proto_bits, $$)));
			}
		;

prop_kernel	: T_KERNEL
			{
			    $$ = parse_proto = RTPROTO_KERNEL;
			    parse_gwlist = &krt_gw_list;
			    trace(TR_PARSE, 0, "parse: %s PROTO: %s",
				  parse_where(),
				  gd_lower(trace_state(rt_proto_bits, $$)));
			}
		;

prop_default	: T_DEFAULT
			{
			    $$ = parse_proto = RTPROTO_DEFAULT;
			    parse_gwlist = (gw_entry **)0;
			    trace(TR_PARSE, 0, "parse: %s PROTO: %s",
				  parse_where(),
				  gd_lower(trace_state(rt_proto_bits, $$)));
			}
		;

prop_metric_option
    		: T_RESTRICT
			{
			    PARSE_METRIC_RESTRICT(&$$.metric);
			    $$.adv = (adv_entry *) 0;
			}
		| metric_option prop_restrict_list_option
			{
			    $$.metric = $1;
			    $$.adv = $2;
			}
		;

prop_restrict_list_option
		:	/* Empty */
			{
			    $$ = (adv_entry *) 0;
			}
		| '{' prop_restrict_list '}'
			{
			    $$ = $2;
			}
		;
    

prop_restrict_list
		: /* Empty */
			{
			    $$ = (adv_entry *) 0;
			}
		| prop_restrict_list prop_restrict L_EOS
			{
			    $$ = parse_adv_dm_append($1, $2);
			    if (!$$) {
				PARSE_ERROR;
			    }
			}
		| prop_restrict_list error L_EOS
			{
			    yyerrok;
			}
		;

prop_restrict	:
			{
			    parse_contig_masks = FALSE;
			}
			dest_mask metric_restrict_option
			{
			    parse_contig_masks = TRUE;

			    $$ = adv_alloc(ADVFT_DM, (proto_t) 0);
			    $$->adv_dm = $2;
			    if (PARSE_METRIC_ISRESTRICT(&$3)) {
				BIT_SET($$->adv_flag, ADVF_NO);
			    } else if (PARSE_METRIC_ISSET(&$3)) {
				$$->adv_result.res_metric = $3.metric;
				BIT_SET($$->adv_flag, ADVFOT_METRIC);
			    }
			}
		;

/*  */

/* Addresses */

addr		: T_NOP
			{
			    assert(FALSE);
			}
		| ip_addr
		| iso_addr
    ;


/* Destination and mask pair used for control lists */
dest_mask	: all_mask
		| default_mask
		| host_mask
		| network_mask
		;

all_mask	: T_ALL
			{
			    $$.dm_dest = $$.dm_mask = (sockaddr_un *) 0;
			    trace(TR_PARSE, 0, "parse: %s DEST: all",
				  parse_where());
			}
		;

default_mask	: T_NOP1
			{
			    assert(FALSE);
			}
		| T_DEFAULT
			{
			    $$.dm_dest = sockdup(inet_addr_default);
			    $$.dm_mask = inet_mask_host;
			    trace(TR_PARSE, 0, "parse: %s DEST: %A",
				  parse_where(),
				  $$.dm_dest);
			}
		| T_ISO T_DEFAULT
			{
#ifdef	XXX
			    $$.dm_dest = sockdup(iso_addr_default);
			    $$.dm_dest = iso_mask_natural(iso_addr_default);
#endif	/* XXX */
			    assert(FALSE);
			}
		;

mask		: T_NOP
			{
			    assert(FALSE);
			}
		| in_addr
			{
			    $$ = inet_mask_locate($1.s_addr);
			    if (!$$) {
				if (parse_contig_masks) {
				    (void) sprintf(parse_error, "mask not contiguous");
				    PARSE_ERROR;
				} else {
				    $$ = mask_locate(sockbuild_in(0, $1.s_addr));
				}
			    }
			    trace(TR_PARSE, 0, "parse: %s MASK: %A",
				  parse_where(),
				  $$);
			}
		| iso_addr
			{
			    $$ = iso_mask_locate($1);
			    sockfree($1);
			    trace(TR_PARSE, 0, "parse: %s MASK: %A",
				  parse_where(),
				  $$);
			}
		;

/* Gateway list */
gateway_list	: gateway
			{
			    $$ = adv_alloc(ADVFT_GW | ADVF_FIRST, (proto_t) 0);
			    $$->adv_gwp = $1;
			}
		| gateway_list gateway
			{
			    adv_entry *adv = adv_alloc(ADVFT_GW, (proto_t) 0);

			    adv->adv_gwp = $2;
			    $$ = $1;
			    if (parse_adv_append(&$$, adv, TRUE)) {
				PARSE_ERROR;
			    }
			}
		;

/* A gateway is a host on an attached network */
gateway		: host
			{
			    /*
			     *	Make sure host is on a locally attached network then
			     *	find or create a gw structure for it.  Requires that
			     *	parse_proto and parse_gwlist are previously set
			     */
			    if (!if_parse_withdst($1)) {
				(void) sprintf(parse_error, "gateway not a host address on an attached network: '%A'",
					       $1);
				if (BIT_TEST(task_state, TASKS_STRICTIFS)) {
				    PARSE_ERROR;
				} else {
				    yyerror(parse_error);
				}
			    }
			    if (!parse_gwlist) {
				(void) sprintf(parse_error, "gateway specification not valid for %s",
					       trace_state(rt_proto_bits, parse_proto));
				PARSE_ERROR;
			    }
			    $$ = gw_locate(parse_gwlist,
					   parse_proto,
					   (task *) 0,
					   (as_t) 0,
					   (as_t) 0,
					   (time_t) 0,
					   $1,
					   (flag_t) 0);
			    trace(TR_PARSE, 0, "parse: %s GATEWAY: %A  PROTO: %s",
				  parse_where(),
				  $$->gw_addr,
				  gd_lower(trace_state(rt_proto_bits, $$->gw_proto)));
			    sockfree($1);
			}
		;

/* A host is a host */
host		: host_name
			{ 
			    sockaddr_un *addr;

			    if (BIT_TEST(task_state, TASKS_NORESOLV)
				|| !(addr = parse_addr_hostname($1.ptr))) {
				PARSE_ERROR;
			    }
			    $$ = sockdup(addr);
			    trace(TR_PARSE, 0, "parse: %s HOST: %A",
				  parse_where(),
				  $$);
			    free_charstr(&$1);
			}
		| addr
			{
			    $$ = $1;
			    trace(TR_PARSE, 0, "parse: %s HOST: %A",
				  parse_where(),
				  $$);
			}
		;

host_mask	: T_HOST host
			{
			    $$.dm_dest = $2;
			    $$.dm_mask = sockhostmask($2);

			    trace(TR_PARSE, 0, "parse: %s DEST: %A",
				  parse_where(),
				  $$.dm_dest);
			}
		;

/* A network is a network */
network		: host_name
			{
			    sockaddr_un *addr;

			    if (BIT_TEST(task_state, TASKS_NORESOLV)
				|| !(addr = parse_addr_netname($1.ptr))) {
				PARSE_ERROR;
			    }
			    $$ = sockdup(addr);
			    trace(TR_PARSE, 0, "parse: %s NETWORK: %A",
				  parse_where(),
				  $$);
			    free_charstr(&$1);
			}
		| addr
			{
			    $$ = $1;
			    trace(TR_PARSE, 0, "parse: %s NETWORK: %A",
				  parse_where(),
				  $$);
			}
		;

network_mask	: network
			{
			    $$.dm_dest = $1;
			    switch (socktype($1)) {
#ifdef	PROTO_INET
			    case AF_INET:
				$$.dm_mask = inet_mask_natural($1);
				if (sock2ip($$.dm_dest) & ~sock2ip($$.dm_mask)) {
				    sprintf(parse_error, "host or mask option needed");
				    PARSE_ERROR;
				}
				break;
#endif	/* PROTO_INET */
#ifdef	PROTO_ISO
			    case AF_ISO:
				$$.dm_mask = iso_mask_natural($1);
				break;
#endif	/* PROTO_ISO */
			    default:
				assert(FALSE);
			    }
			    
			    trace(TR_PARSE, 0, "parse: %s DEST: %A MASK: %A",
				  parse_where(),
				  $$.dm_dest,
				  $$.dm_mask);
			}
		| network T_MASKLEN number
			{
			    $$.dm_dest = $1;
			    switch (socktype($1)) {
#ifdef	PROTO_INET
			    case AF_INET:
				if (parse_limit_check("inet mask bits",
						      (u_int) $3,
						      INET_LIMIT_MASKLEN)) {
				    PARSE_ERROR;
				}
				$$.dm_mask = inet_masks[$3];
				break;
#endif	/* PROTO_INET */
#ifdef	PROTO_ISO
			    case AF_ISO:
				if (parse_limit_check("iso mask bits",
						      (u_int) $3,
						      ISO_LIMIT_MASKLEN)) {
				    PARSE_ERROR;
				}
				$$.dm_mask = iso_masks[$3];
				break;
#endif	/* PROTO_ISO */
			    default:
				assert(FALSE);
			    }
			    trace(TR_PARSE, 0, "parse: %s DEST: %A MASK: %A",
				  parse_where(),
				  $$.dm_dest,
				  $$.dm_mask);
			}
		| network T_MASK mask
			{
			    register byte *ap = $1->a.sa_data;
			    register byte *mp = $3->a.sa_data;
			    register byte *alp = (byte *) $1 + socksize($1);
			    register byte *mlp = (byte *) $3 + socksize($3);

			    if (socktype($1) != socktype($3)) {
				sprintf(parse_error, "Network %A and mask %A are not the same family",
					$1,
					$3);
				PARSE_ERROR;
			    }
			    
			    /* Skip zero fields */
			    while (!*ap && ap < alp) {
				ap++, mp++;
			    }
			    
			    while (ap < alp) {
				if (mp < mlp) {
				    if (*ap++ & ~*mp++) {
					break;
				    }
				} else {
				    if (*ap++) {
					break;
				    }
				}
			    }
			    if (ap < alp) {
				sprintf(parse_error, "Non-masked bits not zero for %A mask %A",
					$1,
					$3);
				PARSE_ERROR;
			    }

			    $$.dm_dest = $1;
			    $$.dm_mask = $3;
			    trace(TR_PARSE, 0, "parse: %s DEST: %A MASK: %A",
				  parse_where(),
				  $$.dm_dest,
				  $$.dm_mask);
			}
		;

/* IP address */
ip_addr		: in_addr
			{
			    $$ = sockdup(sockbuild_in(0, $1.s_addr));

			    trace(TR_PARSE, 0, "parse: %s IP_ADDR: %A",
				  parse_where(),
				  $$);
			}
		;

in_addr		: L_NUMBER
			{
			    u_int32 addr = $1;
			    
			    if (addr) {
				/* Normalize the address */
				while (!(addr & 0xff000000)) {
				    addr <<= 8;
				}
			    }
			    $$.s_addr = htonl(addr);
			}
		| byte_string
			{
			    int i;
			    u_int32 addr = 0;

			    if ($1.len > sizeof(addr)) {
				sprintf(parse_error, "octet or hex string too long to be an IP address");
				PARSE_ERROR;
			    }

			    for (i = 0; i < sizeof(addr); i++) {
				addr <<= 8;
				if (i < $1.len) {
				    BIT_SET(addr, $1.ptr[i]);
				}
			    }
			    $$.s_addr = htonl(addr);
			    free_bytestr(&$1);
			}
		;

/* Host name */
host_name	: L_HNAME
			{
			    $$ = $1;
			    $$.ptr = task_mem_strdup((task *) 0, $1.ptr);
			}
		| L_KEYWORD
			{
			    $$ = $1;
			    $$.ptr = task_mem_strdup((task *) 0, $1.ptr);
			}
		| string
		;

/* Protocols */
proto_interior	: T_REDIRECT
			{
#if	defined(PROTO_ICMP) || defined(KRT_RT_SOCK)
			    $$.proto = RTPROTO_REDIRECT;
			    $$.import_list = &redirect_import_list;
			    $$.export_list = (adv_entry **) 0;
			    $$.gw_list = &redirect_gw_list;
#endif	/* defined(PROTO_ICMP) || defined(KRT_RT_SOCK) */
			}
		;


proto_exterior	: T_NOP
			{
			    assert(FALSE);
			}
		| bgp_prop_init
		;

onoff_option	: T_ON		{ $$ = T_ON; }
		| T_OFF		{ $$ = T_OFF; }
		;

/* Metric */

metric		: number
			{
			    PARSE_METRIC_SET(&$$, $1);
			    trace(TR_PARSE, 0, "parse: %s METRIC: %d",
				  parse_where(),
				  $$.metric);
			}
		| T_INFINITY
			{
			    PARSE_METRIC_INFINITY(&$$);
			    trace(TR_PARSE, 0, "parse: %s METRIC: infinity",
				  parse_where());
			}
		;

metric_option	: /* Empty */
			{
			    PARSE_METRIC_CLEAR(&$$);
			}
		| T_METRIC metric
			{
			    $$ = $2;
			}
		;
		
metric_restrict_option
		: /* Empty */
			{
			    PARSE_METRIC_CLEAR(&$$);
			}
		| T_RESTRICT
			{
			    PARSE_METRIC_RESTRICT(&$$);
			}
		| T_METRIC metric
			{
			    if (parse_metric_check(parse_export_proto, &$2)) {
				PARSE_ERROR;
			    }
			    $$ = $2;
			}
		;
		
/* Preference */
preference_option
		: /* Empty */
			{
			    PARSE_METRIC_CLEAR(&$$);
			}
		| T_PREFERENCE preference
			{
			    PARSE_METRIC_SET(&$$, $2);
			}
		;

preference_restrict_option
		: preference_option
		| T_RESTRICT
			{
			    PARSE_METRIC_RESTRICT(&$$);
			}
		;

preference	: L_NUMBER
			{
			    if (parse_limit_check("preference",
						  (u_int) $1,
						  LIMIT_PREFERENCE)) {
				PARSE_ERROR;
			    }
			    $$ = $1;
			}
		;


/* Numbers and such */
string		: L_STRING
			{
			    /* Remove quotes from the string if present */
			    char *cp;

			    $$ = $1;
				
			    switch (*$$.ptr) {
			    case '<':
				*$$.ptr = '>';
				/* Fall through */

			    case '"':
				cp = &$$.ptr[$$.len-1];
				if (*cp == *$$.ptr) {
				    *cp = (char) 0;
				    $$.len -= 2;
				    cp = $$.ptr + 1;
				    break;
				}
				/* Fall Through */

			    default:
				cp = $$.ptr;
			    }

			    $$.ptr = task_mem_strdup((task *) 0, cp);
			    trace(TR_PARSE, 0, "parse: %s STRING: \"%s\" length: %d",
				  parse_where(),
				  $$.ptr,
				  $$.len);
			}
		;


number		: L_NUMBER
		| byte_string
			{
			    int i = 0;
			    int len = sizeof($$);

			    $$ = 0;
			    if ($1.len > len) {
				sprintf(parse_error, "octet or hex string too long to be a number");
				PARSE_ERROR;
			    } else if ($1.len < len) {
				len = $1.len;
			    }

			    while (i < len) {
				$$ = ($$ << 8) | $1.ptr[i++];
			    }
			    free_bytestr(&$1);
			}
		;


byte_string	: L_HEXSTRING
		| L_OCTETSTRING
		;

time		: L_NUMBER
			{
			    if (parse_limit_check("seconds",
						  (u_int) $1,
						  LIMIT_NATURAL)) {
				PARSE_ERROR;
			    };
			    $$ = $1;
			}
		| L_NUMBER ':' L_NUMBER
			{
			    if (parse_limit_check("minutes",
						  (u_int) $1,
						  LIMIT_NATURAL)) {
				PARSE_ERROR;
			    }
			    if (parse_limit_check("seconds",
						  (u_int) $3,
						  LIMIT_SECONDS)) {
				PARSE_ERROR;
			    }
			    $$ = ($1 * 60) + $3;
			}
		| L_NUMBER ':' L_NUMBER ':' L_NUMBER
			{
			    if (parse_limit_check("hours",
						  (u_int) $1,
						  LIMIT_NATURAL)) {
				PARSE_ERROR;
			    }
			    if (parse_limit_check("minutes",
						  (u_int) $3,
						  LIMIT_MINUTES)) {
				PARSE_ERROR;
			    }
			    if (parse_limit_check("seconds",
						  (u_int) $5,
						  LIMIT_SECONDS)) {
				PARSE_ERROR;
			    }
			    $$ = (($1 * 60) + $3) * 60 + $5;
			}
		;

/* Internet ports */
port		: L_NUMBER
			{
			    if (parse_limit_check("port",
						  (u_int) $1,
						  LIMIT_PORT)) {
				PARSE_ERROR;
			    }
			    $$ = htons($1);
			}
		| L_KEYWORD
			{
			    const struct servent *sp;

			    if (!(sp = getservbyname($1.ptr, parse_serv_proto))) {
				(void) sprintf(parse_error, "unknown protocol %s/%s",
					$1.ptr, parse_serv_proto);
				PARSE_ERROR;
			    }

			    $$ = sp->s_port;

			    trace(TR_PARSE, 0, "parse: %s PORT %s (%d)",
				  parse_where(),
				  $1.ptr,
				  ntohs($$));
			}
		;

/**/
/* Autonomous systems */

as		: L_NUMBER
			{
			    if (parse_limit_check("autonomous system",
						  (u_int) $1,
						  LIMIT_AS)) {
				PARSE_ERROR;
			    }
			    $$ = $1;
			}
		;

as_list		: as
			{
			    $$ = adv_alloc(ADVFT_AS | ADVF_FIRST, (proto_t) 0);
			    $$->adv_as = $1;
			}
		| as_list as
			{
			    adv_entry *adv = adv_alloc(ADVFT_AS, (proto_t) 0);

			    adv->adv_as = $2;
			    $$ = $1;
			    if (parse_adv_append(&$$, adv, TRUE)) {
				PARSE_ERROR;
			    }
			}
		;

as_list_option	: /* Empty */
			{
			    $$ = (adv_entry *) 0;
			}
		| T_AS as_list
			{
			    $$ = $2;
			}
		;

/**/
/* Tags */

/* Route tags */
tag_as_option
		: /* Empty */
			{
			    PARSE_METRIC_CLEAR(&$$);
			}
		| T_TAG tag_as
			{
			    $$ = $2;
			}
		;

tag_as		: tag
		| T_AS
			{
			    PARSE_METRIC_ALTERNATE(&$$, 0);
			}
		| T_AS number
			{
			    if (parse_limit_check("external-route-tag",
						  (u_int) $2,
						  PATH_OSPF_TAG_USR_LIMIT)) {
				PARSE_ERROR;
			    }
			    PARSE_METRIC_ALTERNATE(&$$, $2);
			}
		;

tag_option	: /* Empty */
			{
			    PARSE_METRIC_CLEAR(&$$);
			}
		| T_TAG tag
			{
			    $$ = $2;
			}
		;

tag	: number
			{
			    if (parse_limit_check("external-route-tag",
						  (u_int) $1,
						  PATH_OSPF_TAG_LIMIT)) {
				PARSE_ERROR;
			    }
			    PARSE_METRIC_SET(&$$, $1);
			}
/* XXX - various tag options */
		;

tag_prop_init
		: proto_exterior
		;

/**/
/* AS paths */

aspath_prop_init
		: proto_exterior
		;

aspath_match	: T_ASPATH
			{
			    aspath_state = 0;
			    aspath_current = (asmatch_t *) task_block_alloc(aspath_match_index);
			}
		aspath_regex T_ORIGIN aspath_origin
    			{
			    $$ = aspath_current;
			    $$->origin_mask = $5;
			    trace(TR_PARSE, 0, "%s: path %X origins %s",
				  parse_where(),
				  $$,
				  aspath_adv_origins($$->origin_mask));
			    aspath_current = (asmatch_t *) 0;
			}
    		;

aspath_origin	: T_ANY
			{
			    /* Set all bits on */
			    $$ = -1;
			}
		| aspath_origin_list
		;

aspath_origin_list
		: aspath_origins
		| aspath_origin_list '|' aspath_origins
			{
			    $$ = $1 | $3;
			}
		;

aspath_origins	: T_IGP
			{
			    $$ = (1 << PATH_ORG_IGP);
			}
		| T_EGP
			{
			    $$ = (1 << PATH_ORG_EGP);
			}
		| T_INCOMPLETE
    			{
			    $$ = (1 << PATH_ORG_XX);
			}
		;
    
aspath_regex	: aspath_sum
			{
			    NFA_ASG($$.start, $1.start);
			    NFA_ASG(aspath_current->start, $$.start);
			    trace(TR_PARSE, 0, "%s: REDUCED:	regex : sum",
				  parse_where());
			}
		| aspath_sum '|' aspath_regex
			{
			    NFA_OR($$.start, $1.start, $3.start);
			    NFA_OR($$.state, $1.state, $3.state);
			    NFA_ASG(aspath_current->start, $$.start);
			    trace(TR_PARSE, 0, "%s: REDUCED:	regex : sum | regex",
				  parse_where());
			}
		;

aspath_sum	: aspath_term
			{
			    trace(TR_PARSE, 0, "%s: REDUCED:	sum : term",
				  parse_where());
			}
		| aspath_term aspath_sum
			{
			    int i;
			    
			    NFA_OR($$.state, $1.state, $2.state);

			    for (i = 0; i < ASP_ACC; i++) {
				int nonzero;
				
				if (NFA_ISSET($1.state, i)) {
				    if (NFA_ISSET(aspath_current->nfa[i].next, ASP_ACC)) {
					NFA_CLR(aspath_current->nfa[i].next, ASP_ACC);
					NFA_OR(aspath_current->nfa[i].next, aspath_current->nfa[i].next, $2.start);
				    }
				    NFA_CLR($1.state, i);
				    NFA_NONZERO($1.state, nonzero);
				    if (nonzero) {
					break;
				    }
				}
			    }
			    if (i > ASP_ACC) {
				YYERROR;
			    }
			    if (NFA_ISSET($1.start, ASP_ACC)) {
				NFA_CLR($1.start, ASP_ACC);
				NFA_OR($1.start, $1.start, $2.start);
			    }
			    NFA_ASG($$.start, $1.start);
			    trace(TR_PARSE, 0, "%s: REDUCED:	sum : term sum",
				  parse_where());
			}
		;

aspath_term	: aspath_symbol
			{
			    trace(TR_PARSE, 0, "%s: REDUCED:	term : symbol",
				  parse_where());
			}
		| aspath_symbol '*'
			{
			    int i;
			    
			    for (i = 0; i < ASP_ACC; i++) {
				if (NFA_ISSET($1.state, i)) {
				    int nonzero;
				
				    if (NFA_ISSET(aspath_current->nfa[i].next, ASP_ACC)) {
					NFA_OR(aspath_current->nfa[i].next, aspath_current->nfa[i].next, $1.start);
				    }
				    NFA_CLR($1.state, i);
				    NFA_NONZERO($1.state, nonzero);
				    if (nonzero) {
					break;
				    }
				}
			    }
			    NFA_ASG($$.start, $1.start);
			    NFA_SET($$.start, ASP_ACC);
			    trace(TR_PARSE, 0, "%s: REDUCED:	term : symbol *",
				  parse_where());
			}
		| aspath_symbol '+'
			{
			    int i;
			    
			    for (i = 0; i < ASP_ACC; i++) {
				if (NFA_ISSET($1.state, i)) {
				    int nonzero;
				    
				    if (NFA_ISSET(aspath_current->nfa[i].next, ASP_ACC)) {
					NFA_OR(aspath_current->nfa[i].next, aspath_current->nfa[i].next, $1.start);
				    }
				    NFA_CLR($1.state, i);
				    NFA_NONZERO($1.state, nonzero);
				    if (nonzero) {
					break;
				    }
				}
			    }
			    NFA_ASG($$.start, $1.start);
			    trace(TR_PARSE, 0, "%s: REDUCED:	term : symbol +",
				  parse_where());
			}
		| aspath_symbol '?'
			{
			    NFA_ASG($$.start, $1.start);
			    NFA_SET($$.start, ASP_ACC);
			    trace(TR_PARSE, 0, "%s: REDUCED:	term : symbol ?",
				  parse_where());
			}
		| aspath_symbol '{' aspath_range '}'
			{
			    int copy;
			    flag_t finalstate[1 + ASP_ACC / ASP_WORD];
			    flag_t newstate[1 + ASP_ACC / ASP_WORD];
			    flag_t oldstate[1 + ASP_ACC / ASP_WORD];
			    flag_t newstart[1 + ASP_ACC / ASP_WORD];
			    
			    NFA_ASG(finalstate, $1.state);
			    NFA_ASG(newstate, $1.state);
			    NFA_ASG($$.start, $1.start);
			    NFA_ASG(newstart, $1.start);
			    if (!$3.begin) {
				NFA_SET($$.start, ASP_ACC);
			    }
			    for (copy = 1; copy < $3.begin; copy++) {
				int i;
				
				NFA_ASG(oldstate, $1.state);
				NFA_ZERO(newstart);
				NFA_ZERO(newstate);
				for (i = 0; i < ASP_ACC; i++) {
				    if (NFA_ISSET($1.state, i)) {
					aspath_current->nfa[aspath_state].as = aspath_current->nfa[i].as;
					NFA_ASG(aspath_current->nfa[aspath_state].next, aspath_current->nfa[i].next);
					NFA_SHIFT(aspath_current->nfa[aspath_state].next, aspath_state-i);
					NFA_SET(newstate, aspath_state);
					if (NFA_ISSET($1.start, i)) {
					    int nonzero;
					    
					    NFA_SET(newstart, aspath_state);
					    ASPATH_NEXT_STATE;
					    NFA_CLR($1.state, i);
					    NFA_NONZERO($1.state, nonzero);
					    if (nonzero) {
						break;
					    }
					}
				    }
				    if (NFA_ISSET($$.start, ASP_ACC)) {
					NFA_OR($$.start,$$.start,newstart);
				    }
				    for (i = 0; i < ASP_ACC; i++) {
					if (NFA_ISSET(oldstate, i)) {
					    int nonzero;
					    
					    if (NFA_ISSET(aspath_current->nfa[i].next, ASP_ACC)) {
						if (copy < $3.begin) {
						    NFA_CLR(aspath_current->nfa[i].next, ASP_ACC);
						}
						NFA_OR(aspath_current->nfa[i].next, aspath_current->nfa[i].next, newstart);
					    }
					    NFA_CLR(oldstate, i);
					    NFA_NONZERO(oldstate, nonzero);
					    if (nonzero) {
						break;
					    }
					}
				    }
				    NFA_OR(finalstate, finalstate, newstate);
				    NFA_ASG($1.start, newstart);
				    NFA_ASG($1.state, newstate);
				}
				if ($3.end) {
				    for (i = 0; i < ASP_ACC; i++) {
					if (NFA_ISSET(newstate, i)) {
					    int nonzero;
					    
					    if (NFA_ISSET(aspath_current->nfa[i].next, ASP_ACC)) {
						NFA_OR(aspath_current->nfa[i].next, aspath_current->nfa[i].next, newstart);
					    }
					    NFA_CLR(newstate, i);
					    NFA_NONZERO(newstate, nonzero);
					    if (nonzero) {
						break;
					    }
					}
				    }
				}	
				NFA_ASG($$.state, finalstate);
				trace(TR_PARSE, 0, "%s: REDUCED:	term : symbol { range }",
				      parse_where());
			    }
			}
		;

aspath_symbol	: as
			{
			    NFA_ZERO($$.start);
			    NFA_ZERO($$.state);
			    NFA_SET($$.start, aspath_state);
			    NFA_SET($$.state, aspath_state);
			    aspath_current->nfa[aspath_state].as = htons($1);
			    NFA_SET(aspath_current->nfa[aspath_state].next, ASP_ACC);
			    ASPATH_NEXT_STATE;
			    trace(TR_PARSE, 0, "%s: REDUCED:	symbol : %d",
				  parse_where(),
				  $1);
			}
		| '.'
			{
			    NFA_ZERO($$.start);
			    NFA_ZERO($$.state);
			    NFA_SET($$.start, aspath_state);
			    NFA_SET($$.state, aspath_state);
			    NFA_SET(aspath_current->nfa[aspath_state].next, ASP_ACC);
			    aspath_current->nfa[aspath_state].as = 0;
			    ASPATH_NEXT_STATE;
			    trace(TR_PARSE, 0, "%s: REDUCED:	symbol : .",
				  parse_where());
			}
		| '(' aspath_regex ')'
			{
			    NFA_ASG($$.state, $2.state);
			    NFA_ASG($$.start, $2.start);
			    trace(TR_PARSE, 0, "%s: REDUCED:	symbol : ( regex )",
				  parse_where());
			}

aspath_range	: number
			{
			    $$.begin = $1;
			    $$.end = 0;
			    if (!$1) {
				sprintf(parse_error, "invalid range start: %d",
					$1);
				PARSE_ERROR;
			    }
			    trace(TR_PARSE, 0, "%s: RANGE: {%d}",
				  parse_where(),
				  $$.begin);
			}
		| number ','
			{
			    $$.begin = $1;
			    $$.end = 1;
			    if (!$1) {
				sprintf(parse_error, "invalid range start: %d",
					$1);
				PARSE_ERROR;
			    }
			    trace(TR_PARSE, 0, "%s: RANGE: {%d,}",
				  parse_where(),
				  $$.begin);
			}
		| number ',' number
			{
			    $$.begin = $1;
			    $$.end = $3;
			    if (!$1) {
				sprintf(parse_error, "invalid range start: %d",
					$1);
				PARSE_ERROR;
			    }
			    if ($3 < $1 || !$3) {
				sprintf(parse_error, "invalid range end: %d",
					$3);
				PARSE_ERROR;
			    }
			    trace(TR_PARSE, 0, "%s: RANGE: {%d,%d}",
				  parse_where(),
				  $$.begin,
				  $$.end);
			}
		;
/**/

iso_addr	: L_ISOADDR
			{
			    if ($1.len > ISO_MAXADDRLEN) {
				sprintf(parse_error, "ISO address has too many bytes (%d), %d maximum",
					$1.len,
					ISO_MAXADDRLEN);
				PARSE_ERROR;
			    }
			    
			    $$ = sockdup(sockbuild_iso($1.ptr, $1.len));

			    trace(TR_PARSE, 0, "parse: %s ISO_ADDR: %A",
				  parse_where(),
				  $$);

			    free_bytestr(&$1);
			}
%%

/*
 *	Log any parsing errors
 */
static void
yyerror(s)
const char *s;
{
    byte *cp;

    trace(TR_ALL, 0, NULL);
    tracef("parse: %s ",
	   parse_where());

    switch (yychar) {
    case L_HEXSTRING:
	tracef("%s at '",
	       s);
	for (cp = yylval.hex.ptr; (cp - yylval.hex.ptr) < yylval.hex.len; cp++) {
	    tracef("%02x", *cp);
	}
	tracef("'");
	break;

    case L_OCTETSTRING:
	tracef("%s at '",
	       s);
	for (cp = yylval.hex.ptr; (cp - yylval.hex.ptr) < yylval.hex.len; cp++) {
	    tracef("%s%u",
		   (cp == yylval.hex.ptr) ? "" : ".",
		   *cp);
	}
	tracef("'");
	break;

    case L_NUMBER:
	tracef("%s at '%d'",
	       s,
	       yylval.num);
	break;

    case L_EOS:
	tracef("%s at 'end-of-statement'",
	       s);
	break;

    case L_ERROR:
	tracef("%s",
	       parse_error);
	break;

#ifdef	YYEOF
    case YYEOF:
#else	/* YYEOF */
    case 0:
#endif	/* YYEOF */
	tracef("%s at 'end-of-file'",
		   s);
	break;

    default:
	if (!parse_keyword_lookup(yychar)) {
#if	YYDEBUG != 0
#ifdef	YYTRANSLATE
	    tracef("%s at '%s'",
		   s,
		   yytname[YYTRANSLATE(yychar)]);
#else	/* YYTRANSLATE */
	    tracef("%s at '%s'",
		   s,
		   yyname[yychar]);
#endif	/* YYTRANSLATE */
#else	/* YYDEBUG */
	    tracef("%s",
		   s);
#endif	/* YYDEBUG */
	    break;
	}
	/* Fall Through */

    case L_STRING:
    case L_HNAME:
    case L_KEYWORD:
	tracef("%s at '%s' length %d",
	       s,
	       yylval.str.ptr,
	       yylval.str.len);
	break;

#ifdef	YYEMPTY
    case YYEMPTY:
	tracef("%s",
	       s);
	break;
#endif	/* YYEMPTY */
    }
    trace(TR_ALL, LOG_ERR, NULL);
    trace(TR_ALL, 0, NULL);
}



/*
 * ------------------------------------------------------------------------
 * 
 * 	GateD, Release 3
 * 
 * 	Copyright (c) 1990,1991,1992,1993 by Cornell University
 * 	    All rights reserved.
 * 
 * 	THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT ANY
 * 	EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT
 * 	LIMITATION, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * 	AND FITNESS FOR A PARTICULAR PURPOSE.
 * 
 * 	Royalty-free licenses to redistribute GateD Release
 * 	3 in whole or in part may be obtained by writing to:
 * 
 * 	    GateDaemon Project
 * 	    Information Technologies/Network Resources
 * 	    200 CCC
 * 	    Cornell University
 * 	    Ithaca, NY  14853-2601  USA
 * 
 * 	GateD is based on Kirton's EGP, UC Berkeley's routing
 * 	daemon	 (routed), and DCN's HELLO routing Protocol.
 * 	Development of GateD has been supported in part by the
 * 	National Science Foundation.
 * 
 * 	Please forward bug fixes, enhancements and questions to the
 * 	gated mailing list: gated-people@gated.cornell.edu.
 * 
 * 	Authors:
 * 
 * 		Jeffrey C Honig <jch@gated.cornell.edu>
 * 		Scott W Brim <swb@gated.cornell.edu>
 * 
 * ------------------------------------------------------------------------
 * 
 *       Portions of this software may fall under the following
 *       copyrights:
 * 
 * 	Copyright (c) 1988 Regents of the University of California.
 * 	All rights reserved.
 * 
 * 	Redistribution and use in source and binary forms are
 * 	permitted provided that the above copyright notice and
 * 	this paragraph are duplicated in all such forms and that
 * 	any documentation, advertising materials, and other
 * 	materials related to such distribution and use
 * 	acknowledge that the software was developed by the
 * 	University of California, Berkeley.  The name of the
 * 	University may not be used to endorse or promote
 * 	products derived from this software without specific
 * 	prior written permission.  THIS SOFTWARE IS PROVIDED
 * 	``AS IS'' AND WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES,
 * 	INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
 * 	MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */
