#include "include.h"
#include "idrp.h"
#include <stdio.h>

/* idrp_attr_rec.c - all routines about attribute record 
 *  (only routines in from the idrp_rt.c file
 * once that compiles - we will pull the idrp.c file apart
 * 
 * %% rd path routines  
 * idrp_free_canon_rdpath (p_can_list,peer) - free this canon rd path 
 * idrp_canon_my_rdi_path (p_att,peer) - create canon path from my rdi
 * idrp_my_rdi_rdpath(p_att)
 * idrp_dump_rdi_path (p_rdl,peer) - dump rdi path in pretty print 
 * add_nlri_routeid_chain(p_rte,p_idrp_rt)
 *
 */

/* free a cannonical RD path given a peer structure to free from  */

void
idrp_free_canon_rdpath(p_can_list,peer)
idrp_canon_rdpath       *p_can_list;
idrpPeer                *peer;
{
idrp_canon_rdpath *p_can;
idrp_canon_rdpath *p_can_old;

	p_can = p_can_list;
        while (p_can)
                {
                p_can_old = p_can;
                p_can = p_can->p_next;
                task_mem_free(peer->task,p_can_old);
                }

}
/* create an RDI path entry with just my RDI for this node
 * Does canonization make one a saint... this RDI PATH stuff
 * could use all the blessings it can get !!
 */

int
idrp_canon_my_rdi_path(p_att,peer)
idrp_attribute_record *p_att;
idrpPeer        *peer;
{
idrp_canon_rdpath *p_can;

        p_can = (idrp_canon_rdpath *) task_mem_malloc(peer->task,sizeof(idrp_canon_rdpath));
        p_can->p_next = (idrp_canon_rdpath *) 0;
        bcopy(&rt_rdi,&(p_can->rdi),sizeof(rt_rdi));
        p_att->rd_path = p_can;
	p_att->hopcount = 0;
        return((rt_rdi.isoa_len+1));
}

int
idrp_my_rdi_rdpath(p_att)
idrp_attribute_record *p_att;
{

	/* I think a test for a hopcount of zero is enough.
	 * However - we will bcmp the rt_rdi too.
	 */ 

	if (p_att->hopcount == 0 && (!isoaddrcmp(&p_att->rd_path->rdi,&rt_rdi)))
		{
		return(TRUE);
		}
	return (FALSE);
}

void
idrp_dump_rdi_path(p_rdl,peer)
idrp_canon_rdpath 	*p_rdl;
idrpPeer		*peer;
{
idrp_canon_rdpath	*p_can;
int			i = 0;

	/* someday - we will want to make only one copy of
	 * a canonical RD list - but for now
	 * let's try to focus on just one copy of Attribute 
	 * - refocus as first pdu's flying
	 */
	
	trace (TR_IDRP,0," peer %s RDI path print ",peer->name); 
	RDPATH_LIST(p_rdl,p_can)
		{
		trace(TR_IDRP,0,"RDI path: RDI %d =  %A",i,sockbuild_iso((byte *) p_can->rdi.isoa_genaddr,
			p_can->rdi.isoa_len));
		i++;
		}
}

int
add_new_routeid_chain(p_rte,p_att)
idrpRoute_entry		*p_rte;	/* idrpRoute entry */ 
idrp_attribute_record	*p_att;	/* pointer attribute record */
{
idrpRoute_entry	*p_route_ent;

	/* walk the route_id_list to the end and add the new route_id
	 * record at the end
	 */

	p_route_ent = p_att->route_id_list; 
	if (p_route_ent)
		{
		trace ( TR_IDRP,LOG_ERR,"Attribute record has route list with no entries");
		return (FALSE);
		}
 
	while (p_route_ent->p_next)
		{
		p_route_ent = p_route_ent->p_next;
		}
	p_route_ent->p_next = p_rte;
	p_rte->p_next = (idrpRoute_entry *) 0;

	/* jgs -- should there be a return value? */
}

			

int
add_nlri_routeid_chain(p_rte,p_idrp_rt)
idrpRoute_entry	*p_rte;
idrpRoute	*p_idrp_rt;
{
idrpRoute	*p_irt;
int		nlri_id;	

	nlri_id = family_to_nlri_id(p_idrp_rt->family);

	/* Add a nlri to a route_id
	 * - Right now this is only  valid
	 *   for the idrp configured or
	 *   the EXT_INFO from this node.
	 * - checks are made to make sure this is
	 *   the only valid use
	 * - The rest of the route_id chains
	 *   are added when the UPDATE pdu is received
	 */

	if (p_rte->route_id != IDRP_LOCAL_ROUTE_ID  && 
	    p_rte->route_id != IDRP_EXT_INFO_ROUTE_ID )
		{
		/* not local or EXT_INFO - bogus */
		
		return (FALSE);
		}
 
	p_irt = p_rte->p_route[nlri_id];
	if (!p_irt)
		{
		p_rte->p_route[nlri_id] = p_idrp_rt;
		}
	else
		{ 
		while (p_irt->p_next_nlri)
			{
			/* somethings on the list - find the end of the list */

			p_irt = p_irt->p_next_nlri;
			}
		p_irt->p_next_nlri = p_idrp_rt; 
		}
	p_idrp_rt->p_next_nlri = (idrpRoute *) 0;
	return(TRUE);
}
