#include "include.h"
#include "idrp.h"
#include <stdio.h>

/* idrp_parse_pdu.c
 *  
 * Parsing incoming pdus
 *  (process routines validate pdu, and call state machine
 *   with either pdu w/error or pdu w/out error
 *   parse routines - parse in order validate pdu  
 * 
 *	idrp_parse_open(peer,pdu,length) - parse open pdu and validate
 *	process_open(peer,pdu,length) - process open pdu    
 *	parse_update(peer,pdu,length,results) - parse update and validate
 *	parse_update_cleanup(results) - cleanup from the parse update routine after failure
 *	process_update(peer,pdu,length) - process update pdu
 *	process_keepalive(peer,pdu,length) - process keepalive pdu
 *	process_error(peer,pdu,length) - process error pdu
 *	process_cease(peer,pdu,length) - process cease pdu
 *	parse_rib_refresh(peer, pdu, length) - ???
 *	process_rib_refresh(peer,pdu,length) - process rib refresh pdu
 *	process_echo(peer,pdu,length) - process echo pdu
 *	process_echo_reply(peer,pdu,length) - process echo reply pdu 
 *	process_header_error(peer,pdu,length) - process header error   
 *	process_rcv_seq(peeri,pdu)
 *	in_sequence(peer,seq)
 * 	idrp_recv_pdu  - receive idrp pdu 
 *	idrp_recv - receive for master task for idrp protocol
 */ 
  
 

/*
 * parse_open(peer, pdu, length)
 */
int idrp_parse_open(peer, pdu, length)
idrpPeer *peer;
idrpPdu *pdu;
int length;
{
u_short rcv_holdtime;
u_short rcv_maxpdusize;
struct iso_net_addr	rdi_peer;
idrp_open_pdu	*p_pdu;
idrp_RIB_att	*p_open_rib_att;
idrp_confed_id	*p_open_confed_id;
idrp_authentication_code	*p_open_auth;
u_char		*p_auth_data;
idrp_open_pdu_header *hdr;
u_int	idrp_fixed_header_len;	
int	fixed_header_delta;

	/* calculate the idrp_fixed header portion based on
	 *  the pdu and the protocol header type
	 * - The header length in the task socket is a different
	 *   length depending on the protocol socket this IDRP
	 *   is running over. (UDP/IP/CLNP/IDRP socket
	 */

	hdr = (idrp_open_pdu_header *) &pdu->pdu.open; 
	bcopy((caddr_t) hdr->idrp_maxpdu_size, (caddr_t) &rcv_maxpdusize, sizeof(hdr->idrp_maxpdu_size));
	rcv_maxpdusize = ntohs(rcv_maxpdusize);

	/* a) check the header version */
	if (hdr->version != IDRP_VERSION_1)
		{
		peer->last_error_pdu.error_code = IDRP_ERROR_OPEN_PDU;
		peer->last_error_pdu.error_subcode = IDRP_OPEN_ERROR_UNSUP_VER;
		peer->last_error_pdu.data = (u_char *) 0;
		peer->last_error_pdu.data_len = 0; 
		return(IDRP_ERROR_OPEN_PDU);
		}

	/* b) check the maximum PDU size */

	bcopy((caddr_t) hdr->idrp_maxpdu_size, (caddr_t) &rcv_maxpdusize, sizeof(hdr->idrp_maxpdu_size));
	rcv_maxpdusize = ntohs(rcv_maxpdusize);
	if (rcv_maxpdusize < MinPDULength) 
		{
		peer->last_error_pdu.error_code = IDRP_ERROR_OPEN_PDU;
		peer->last_error_pdu.error_subcode = IDRP_OPEN_ERROR_BAD_MAX_SIZE;
		peer->last_error_pdu.data = (u_char *) &hdr->idrp_maxpdu_size; 
		peer->last_error_pdu.data_len = sizeof(hdr->idrp_maxpdu_size);
		return(IDRP_ERROR_OPEN_PDU);
		}

	p_pdu = (idrp_open_pdu *) hdr; 

	/* c) check the RDI */
	
	rdi_peer.isoa_len = p_pdu->source_rdi.source_RDI_length;
	bcopy ((caddr_t) &p_pdu->source_rdi.source_RDI,&rdi_peer.isoa_genaddr,rdi_peer.isoa_len);
	if (isoaddrcmp(&peer->rdi,&rdi_peer))
		{
		/* addresses are not equal so error */

		trace(TR_IDRP, 0, "IDRP open error - RDI from peer wrong peer %s  rdi %s",
			peer->name, sockbuild_iso((byte *) rdi_peer.isoa_genaddr, rdi_peer.isoa_len));
	
		peer->last_error_pdu.error_code = IDRP_ERROR_OPEN_PDU;
		peer->last_error_pdu.error_subcode = IDRP_OPEN_ERROR_BAD_PEER_RD;
		peer->last_error_pdu.data = (u_char *) 0;
		peer->last_error_pdu.data_len = 0; 
		return(IDRP_ERROR_OPEN_PDU);
		}	
		
	/* d) Rib_Att_Set	
	 *     set up the pointer so that variable field is indexed past 
	 */

	p_open_rib_att = (idrp_RIB_att *) &p_pdu->source_rdi.source_RDI + rdi_peer.isoa_len; 
		
	if  (p_open_rib_att->RIB_att_count != RIB_ATT_DEFAULT_ONLY_COUNT)
		{
		peer->last_error_pdu.error_code = IDRP_ERROR_OPEN_PDU;
		peer->last_error_pdu.error_subcode = IDRP_OPEN_ERROR_BAD_RIB_ATT_SET;
		peer->last_error_pdu.data = (u_char *) 0;
		peer->last_error_pdu.data_len = 0; 
		return(IDRP_ERROR_OPEN_PDU);
		}
	
	/* e) RDC match check */

	p_open_confed_id = (idrp_confed_id *) p_open_rib_att + sizeof(idrp_RIB_att);

	if (p_open_confed_id->confed_id_count != NO_RDC_SUPPORT_COUNT)
		{
		peer->last_error_pdu.error_code = IDRP_ERROR_OPEN_PDU;
		peer->last_error_pdu.error_subcode = IDRP_OPEN_ERROR_RDC_MISMATCH;
		peer->last_error_pdu.data = (u_char *) 0;
		peer->last_error_pdu.data_len = 0; 
		return(IDRP_ERROR_OPEN_PDU);
		}

	/* f) Authentication check */

	p_open_auth = (idrp_authentication_code *) p_open_confed_id + sizeof(idrp_confed_id); 

	switch (p_open_auth->authentication_code)
		{
		case IDRP_AUTH_CODE_TEST:
			return(0);
			break;

		case IDRP_AUTH_CODE_CRYPT:
		case IDRP_AUTH_CODE_CLEAR:
			p_auth_data = (u_char * ) p_open_auth;
			p_auth_data += sizeof (idrp_authentication_code);
			if (valid_auth_data(p_auth_data,peer))
				{
				peer->last_error_pdu.error_code = IDRP_ERROR_OPEN_PDU;
				peer->last_error_pdu.error_subcode = IDRP_OPEN_ERROR_AUTH_FAILURE;
				peer->last_error_pdu.data = (u_char *) 0;
				peer->last_error_pdu.data_len = 0; 
				return(IDRP_ERROR_OPEN_PDU);
				}

		default:		
			peer->last_error_pdu.error_code = IDRP_ERROR_OPEN_PDU;
			peer->last_error_pdu.error_subcode = IDRP_OPEN_ERROR_UNSUP_AUTH_CODE;
			peer->last_error_pdu.data = (u_char *) 0;
			peer->last_error_pdu.data_len = 0; 
			return(IDRP_ERROR_OPEN_PDU);
			break;
			}
return (0);
}

int valid_auth_data(p_auth_data,peer)
u_char		*p_auth_data;
idrpPeer	*peer;
{
	return (TRUE);
}
	
/*
 * Open PDU
 */
static void process_open(peer, pdu, length)
idrpPeer *peer;
idrpPdu *pdu;
int length;
{
u_short rcv_holdtime;
u_short rcv_maxpdusize;
idrp_open_pdu_header *hdr;


	/* set-up new hold time from pdu - whether we use it or not */

	hdr = &pdu->pdu.open.header;	
	bcopy((caddr_t) hdr->idrp_hold_time, (caddr_t) &rcv_holdtime, sizeof(hdr->idrp_hold_time));
	rcv_holdtime = ntohs(rcv_holdtime);
	bcopy((caddr_t) hdr->idrp_maxpdu_size, (caddr_t) &rcv_maxpdusize, sizeof(hdr->idrp_maxpdu_size));
	rcv_maxpdusize = ntohs(rcv_maxpdusize);
	peer->keepalive_time = rcv_holdtime / 3 ? rcv_holdtime / 3 : 1;
	peer->pdu_maxsendsize = rcv_maxpdusize;
	trace(TR_IDRP, 0, "Open hold time %d, max pdu size %d", rcv_holdtime, rcv_maxpdusize);

	/* simply call state machine for valid open */
	idrp_sm(peer,IDRP_OPEN);
}


int parse_update(peer, pdu, length,res)
idrpPeer *peer;
idrpPdu *pdu;
int length;
parse_results *res;
{

/* path attribute variables */

idrp_attribute_record *p_att = 0;
idrp_attribute_record *p_attr = 0;
idrp_path_attr *attrs;
idrpRoute_entry *p_idrp_re = 0; 
short total_attr_len;
short attr_len;
u_int attr_type;
u_int usr_type = 0;
u_int usr_attr_index;
idrpRoute_entry	*p_rid_list;
idrpRoute *p_idrp_rt;
idrpRoute_link rt_chg[IDRP_NLRI_FAMILIES_SUPPORTED];
int route_id_in;

/* nlri variables */ 

u_char *nlri;
nlri_info *nlriptr;
nlri_proto_info *nlri_proto_ptr;
u_short nlrilen, nlriproto, nlriproto_id;
int nlri_cnt = 0;


/* general pointer for copies */ 

u_char *cp;
u_char *cp2;
int	len_left;
int	index;
int 	i;
int	reuse_att=FALSE;
int	family;
int 	nlri_id;
u_short	hold;
u_long	longhold;

/* pdu headers work pointers */
idrp_update_pdu *update;
u_int id;

	/* - set update pointer to header plus protocol header
	 * to send PDU (CLNP header, UDP header, PID header,
	 * IDRP socket header)
	 */
 

	update = (idrp_update_pdu *) &pdu->pdu.update; 
	cp = (u_char *) update;
	
	/* Get Withdrawn route-id count and route_ids */
	/* route_ids are variable length */
	len_left = length - sizeof(idrp_fixed_pdu_header);

	bcopy(update->unfeasible_route_count, (caddr_t) &hold,UNFEASIBLE_RT_CNT_LEN);
	hold = ntohs(hold);
	if (hold >= IDRP_WITHDRAW_MAX_IDs)
		{
		trace (TR_IDRP,0,"Withdraw ids over maximum peer %s update pdu ",peer->name); 
		peer->last_error_pdu.error_code = IDRP_ERROR_UPDATE_PDU;
		peer->last_error_pdu.error_subcode = IDRP_UPDATE_ERROR_MALF_ATTR;
		peer->last_error_pdu.data_len = 0;
		return (IDRP_ERROR_UPDATE_PDU);
		}


	cp +=  UNFEASIBLE_RT_CNT_LEN; 

	
	/* initialize results array */	

	res->withdraw->count = hold; 
	res->route_id_in =  IDRP_INIT_ROUTE_ID;
	res->p_att = (idrp_attribute_record *) 0;
	NLRI_FAMILY_LOOP(i)
		{ 
		res->rt_chg[i] = (idrpRoute *) 0;
		rt_chg[i].tail = rt_chg[i].head = (idrpRoute *) 0;
		} NLRI_FAMILY_LOOP_END

 
	/* Parse the withdrawn route list. 
	 * putting withdraw routes into withdraw array 
	 */

	if (hold)
		{
		/* withdrawing route ids */

		for (i = 0; i < res->withdraw->count; i++)
			{
			bcopy((caddr_t) cp,&id,sizeof(route_id));
			id = ntohl(id);
			cp += sizeof(route_id);
			res->withdraw->withdraw[i] = id;
			trace (TR_IDRP,0,"Withdraw route id %d peer %s",id,peer->name);
			}
		}

	/* copy in the length of the whole attribute 
 	 * - cp = total attribute length
	 *   Check it to make sure the length is sane with whole pdu length 
	 */
   
	/* save pointer to nlri information */

	bcopy((caddr_t) cp, (caddr_t) &total_attr_len, sizeof(update->tot_path_att_len));
	total_attr_len = ntohs(total_attr_len);
	cp += sizeof(update->tot_path_att_len); 
	index = sizeof(update->tot_path_att_len) + (sizeof (route_id) * res->withdraw->count) + total_attr_len + UNFEASIBLE_RT_CNT_LEN;
	len_left -= index;

	if (len_left < 0)
		{
		/* length too short - pdu invalid */
		/* return code of zero - says PDU is bogus pocus */ 
		peer->last_error_pdu.error_code = IDRP_ERROR_UPDATE_PDU;
		peer->last_error_pdu.error_subcode = IDRP_UPDATE_ERROR_ATTR_LEN_ERR; 
		peer->last_error_pdu.data_len = 0;
		parse_update_cleanup(res,p_idrp_re,peer,reuse_att);
		return(IDRP_UPDATE_ERROR);
		}			

	/* Check for only withdraws - zero attribute length */
	
	if (total_attr_len == 0)
		{
		res->pref_rcvd = 0;
		return (0);
		}
		
	
	/* path attributes and nrli ahead - start the parse */
	
	/* 
	 * start reading attributes one by one
	 */ 
	cp2 = (u_char *) cp + total_attr_len;
	nlri_proto_ptr = (nlri_proto_info *) cp2; 
	cp2 += sizeof(nlri_proto_info);
	nlriptr = (nlri_info *) cp2; 

	trace(TR_IDRP | TR_UPDATE, 0, "IDRP update from %s attr length %d", peer->name, total_attr_len);

	/* 1) allocate and fill a temporary attribute record  
	 * 2) allocate and fill memory to save attributes from UPDATE 
	 *    to associate with attribute record 
 	 */ 
 
	p_att = (idrp_attribute_record *) task_mem_malloc(peer->task, sizeof(idrp_attribute_record));
	bzero(p_att, sizeof(idrp_attribute_record));
	p_att->peer_alloc = peer;
	p_att->mem_type = IDRP_TASK_MEM;
	p_att->p_attr_pdu = (u_char *) task_mem_malloc(peer->task,total_attr_len);
	bcopy(cp,p_att->p_attr_pdu,total_attr_len);
	p_att->attr_len = total_attr_len;
	cp = p_att->p_attr_pdu; 

	/* set up the indexes to copy attributes into attribute
	 * - copy attributes in and start to read attributes
	 *   out of the saved copy instead of the pdu
	 *   hopefully, if we are mistaken now, we
	 *   will catch it.  This is better than on later transmission
	 *   of the attributes in a UPDATE pdu 
	 */
 
	usr_attr_index = 0;

	/* set-up results array with attribute record id */

	res->p_att = p_att;

	while (total_attr_len > 0) {
		/* set the attribute header to the next attribute
		 * - make sure the index into the bytes is correct
		 *   Been bit by the structure definitions too many
		 *   times. 
		 */
 
		attrs = (idrp_path_attr *) cp;
		attr_type = attrs->attr_hdr.attr_type;
		usr_type = FALSE;

		/* check the type so saving info in arrays 
		 * will not run off the end of the array
		 */

		if (attr_type >= IDRP_ATTR_USR_MIN && attr_type <= IDRP_ATTR_USR_MAX)
			{
			/* valid user type */
			usr_type = TRUE;
			}
	
		/* check for range within IDRP specified attributes */ 
		if (attr_type < IDRP_ATTR_MIN && 
		    attr_type > IDRP_ATTR_MAX && 
			(usr_type == FALSE))
			{
			/* attribute type outside of range */
			/* error in parsing the pdu */

			if (attrs->attr_hdr.attr_flags == IDRP_ATTR_FLAG_MANDATORY)
				{
				peer->last_error_pdu.error_code = IDRP_UPDATE_ERROR;
				peer->last_error_pdu.error_subcode = IDRP_UPDATE_ERROR_UNREC_WELKN_ATTR;
				peer->last_error_pdu.data_len = 0;
				parse_update_cleanup(res,p_idrp_re,peer,reuse_att);
				trace(TR_IDRP, 0, "IDRP %s got pdu type %d outside range ", peer->name, attr_type);
				return(IDRP_UPDATE_ERROR);
				}
			}
		   
		bcopy(attrs->attr_hdr.attr_length, (caddr_t) &attr_len, sizeof(attrs->attr_hdr.attr_length));
		attr_len = ntohs(attr_len);
		trace(TR_IDRP | TR_UPDATE, 0, "  Attr %d, length %d, flags %x",
			 attrs->attr_hdr.attr_type,attr_len, attrs->attr_hdr.attr_flags);
		cp += sizeof(idrp_path_attr_hdr);
		if (!usr_type)
			{
			/* idrp attribute */
	
			if (p_att->attrib[attr_type].present) 
				{
				/* type appeared twice in PDU */

				peer->last_error_pdu.error_code = IDRP_UPDATE_ERROR;
				peer->last_error_pdu.error_subcode = IDRP_UPDATE_ERROR_UNREC_WELKN_ATTR;
				peer->last_error_pdu.data_len = 0;
				parse_update_cleanup(res,p_idrp_re,peer,reuse_att);
				trace(TR_IDRP, 0, "IDRP %s attribute %d present twice", peer->name, attr_type);
				return(IDRP_UPDATE_ERROR);
				}
		
			p_att->attrib[attr_type].present = TRUE;
			p_att->attrib[attr_type].flags = attrs->attr_hdr.attr_flags;
			p_att->attrib[attr_type].length = attr_len;
			p_att->attrib[attr_type].data = cp;
			}
		else
			{
			/* user type */
			p_att->usr_attrib[usr_attr_index].present = TRUE;
			p_att->usr_attrib[usr_attr_index].flags = attrs->attr_hdr.attr_flags;
			p_att->usr_attrib[usr_attr_index].length = attr_len;
			p_att->usr_attrib[usr_attr_index].data = cp;
			usr_attr_index++;
			}  	
		p_att->idrp_mask = p_att->idrp_mask | idrpAttrBits[attr_type];
		cp += attr_len;
		total_attr_len -= attr_len + sizeof(idrp_path_attr_hdr);
	}


	/* set-up the idrpRoute id list in attribute record 
	 *  - malloc space for the idrpRoute_entry structure 
	 *  - Set upthe idrp Route list by:  
	 *    a.) set the initialize ROUTE_ID so if we get aborted we can clean-up easily 
	 *    b.)  set the peer and clearing the linked list pointer
	 */
 
	p_rid_list = (idrpRoute_entry *) task_mem_malloc(peer->task,sizeof(idrpRoute_entry));	

	/* bzero clears p_rid_list to many standard values quickly
	 * status to IDRP_ATTR_ROUTE_ID_NORM
	 * mem_type - IDRP_TASK_MEM   
	 * p_route - to head and tails of zero
	 * p_next  - to zero
	 *  (leave in so we don't waste time in update sequence
         *   as we begin to cut the waste out of that loop) 
	 *
	 */
 	
	bzero(p_rid_list,sizeof(idrpRoute_entry));
	p_rid_list->route_id = IDRP_INIT_ROUTE_ID;
	p_rid_list->peer = peer;
	p_att->route_id_list = p_rid_list;

	
	/* set internal AdjRib id  */
	/* always zero for default case */
  
	p_att->rib_id = idrp_rib_id(p_att->idrp_mask);
	if (p_att->rib_id < 0)
		{
		trace(TR_IDRP | TR_UPDATE, 0, "Non-Default Attributes no support for this now peer %s",peer->name);
		peer->last_error_pdu.error_code = IDRP_UPDATE_ERROR;
		peer->last_error_pdu.error_subcode = IDRP_UPDATE_ERROR_MALF_ATTR;
		peer->last_error_pdu.data_len = 0;
		parse_update_cleanup(res,p_idrp_re,peer,reuse_att);
		return (IDRP_UPDATE_ERROR);
		}


	/* Now validate the attributes. */

	if (!valid_attr(peer,p_att))
		{
		/* trace error,  release pdu, and exit */
		parse_update_cleanup(res,p_idrp_re,peer,reuse_att);
		trace(TR_IDRP | TR_UPDATE, 0, "Atttributes invalid. code = %x",p_att);
		return(IDRP_UPDATE_ERROR);	
		}

	p_attr = find_attr_rec(peer,p_att,FALSE);
	if (p_attr)
		{
		/* if old attribute matches attributes use it by:
		 *  1.) linking a new entry onto the route_id_list
		 * 	with route_id and loc_pref of value
		 *  2.) incrementing  reference count 
		 */

		route_id_in = p_att->route_id_list->route_id;
		p_idrp_re = idrp_add_idrpRoute_entry(route_id_in,p_attr->route_id_list);
		p_idrp_re->loc_pref = p_att->route_id_list->loc_pref;
		p_idrp_re->peer = peer;

		/* now get rid of the temporary attribute record */
 
		p_att->route_id_list->route_id = IDRP_INIT_ROUTE_ID;
		task_mem_free(peer->task,p_att->p_attr_pdu);
		idrp_free_attr_rec(peer,p_att,route_id_in);
		p_att = p_attr;

		/* reset the results array */

		res->p_att = p_att;
		reuse_att = TRUE;
		}
	else
		{
		/* the new attribute has a route_id list
		 * with route_id and preference of route 
		 * with this attribute set
		 */
 
		p_idrp_re = p_att->route_id_list;
		route_id_in = p_att->route_id_list->route_id;
		reuse_att = FALSE;
		trace(TR_IDRP | TR_UPDATE, 0, "RouteId %d (%d) pref %d peer%s",route_id_in, 
		      p_att->route_id_list->route_id,
		      p_att->route_id_list->loc_pref,peer->name);
		}
		


	if (usr_attr_index) 
		{
		if (!valid_usr_attr(peer,p_att))
			{
			/* trace error, release attribute block and exit */
			trace(TR_IDRP | TR_UPDATE, 0, "Atttributes invalid (valid_usr_attr fails). code = %x",p_att);
			parse_update_cleanup(res,p_idrp_re,peer,reuse_att);
			return(IDRP_UPDATE_ERROR);	
			}
		}


	/* Parse the NLRI */
	/* watch our total count to make sure PDU is valid */
	/* NLRI  = prototype(octet), length(octet), value */
	/* repeat 1-5 for each protocol type */
	/* 1.) get protocol type */
	/* 2.) get length this nlri info */
	/* repeat 3-6 until NLRI done */
	/* 3.) get length of NLRI */
	/* 4.) get NLRI -  */
	/* 5.) create IDRPRoute block and rt_entry */
	/* 6.) do rt_lookup and give to add, change or delete list */

	while (len_left > 0)
		{
		/* copy protocol type, length and value  */ 
		nlrilen = (int) nlri_proto_ptr->proto_len;

 
		/* translate family to gated family */
		nlri_id = proto_to_nlri_id(&nlri_proto_ptr->proto_type,nlrilen,(u_char *) &nlri_proto_ptr->proto_val);
		family = nlri_family[nlri_id].family;

		/* length left = length - length of protocol header 
		 *  protocol header = type, length + protocol field (variable ) 
		 *  nlrilen = protocol header length 
		 *  len_left = len_left - protocol_header length
		 */ 
		
		nlrilen += (sizeof(nlri_proto_ptr->proto_type) + sizeof(nlri_proto_ptr->proto_len)); 
		len_left -= nlrilen; 
		trace(TR_IDRP | TR_UPDATE, 0, "NLRI (PROTO_ID %x, family %d length %d):", \
				 nlri_proto_ptr->proto_type,family, nlrilen);


		/* address length for nlri of this protocol type
		 * copy the length and check length left 
 		 */
  
		bcopy(nlriptr,&hold, sizeof(nlriptr->addr_len));
		hold = ntohs(hold);
		len_left -=(hold + sizeof(nlriptr->addr_len));

		if (len_left < 0)
			{
			trace(TR_IDRP | TR_UPDATE, 0, "len of nrli exceeds pdu length len of nlri = %d",hold);
			peer->last_error_pdu.error_code = IDRP_UPDATE_ERROR;
			peer->last_error_pdu.error_subcode = IDRP_UPDATE_ERROR_ATTR_LEN_ERR;
			peer->last_error_pdu.data_len = 0;

			/* flag no route_id list for cleanup */
			parse_update_cleanup(res,p_idrp_re,peer,reuse_att);	
			return (IDRP_UPDATE_ERROR);
			}
	
		nlrilen = hold;
		
		/* loop till finish this protocol nlri group
		 * set-up the input pointer to start of NLRI 
	 	 *  after the protocol identifiers
		 * indexing over the total NLRI length for
		 * this address family 
		 * 
		 */

		cp = (u_char *) nlriptr; 
		cp += sizeof(nlriptr->addr_len);		

		while (nlrilen > 0) 
			{
			/* allocate an  idrpRoute buffer for
			 * each NLRI that is received.
			 * fill it up with basics as we process
			 */ 

			p_idrp_rt = idrp_alloc_idrpRoute(peer,p_att,family,route_id_in);
			p_idrp_rt->status = IDRP_STATUS_ADJ_RIB;
			p_idrp_rt->pref_rcvd = p_idrp_re->loc_pref;
 
			/* copy in the protocol id nlri into buffer 
			 * - watch here for zero length nlri
			 */

			bcopy(cp,&p_idrp_rt->nlri.pfx_len,sizeof(p_idrp_rt->nlri.pfx_len));
			cp += sizeof(p_idrp_rt->nlri.pfx_len);	
			bcopy(cp,&p_idrp_rt->nlri.pfx,PFX_BYTE_LENGTH(p_idrp_rt->nlri));
			cp += PFX_BYTE_LENGTH(p_idrp_rt->nlri);

			trace(TR_IDRP | TR_UPDATE, 0, " nlri =   %s", iso_ptoa(&p_idrp_rt->nlri));

			/* add it to the linked list of IDRP_routes
			 * use temporary space for header and tail
			 */
	
			if (rt_chg[nlri_id].head)
				{
				rt_chg[nlri_id].tail->p_next_nlri = p_idrp_rt;
				p_idrp_rt->p_next_nlri = (idrpRoute *) 0;
				rt_chg[nlri_id].tail = p_idrp_rt;
				}
			else
				{
				rt_chg[nlri_id].tail = rt_chg[nlri_id].head = p_idrp_rt;
				}				
			nlri_cnt++;

			/* Do we have another prefix
		 	 *  - Is length still positive?
			 */		

			nlrilen -= (PFX_BYTE_LENGTH(p_idrp_rt->nlri) + sizeof(p_idrp_rt->nlri.pfx_len));
			}
		} /* end of this protocol id */

	/* valid update pdu
	 * - trace the fact we got a valid update pdu
	 * - up attribute reference count by number of nlri associated
	 *   with it from peer on this route_id
	 */
		
	trace(TR_IDRP|TR_UPDATE, 0,"valid pdu pdu from peer %s last nlri %s", peer->name,iso_ptoa(&p_idrp_rt->nlri));

	/* here copy into the attribute record route_id lists for
	 * these type of pdus, the linked list
	 * p_idrp_re - for a new attribute record points to
	 *	       route_id_list that is new
 	 * p_idrp_re - for an existing attribute record points
	 *	      to a new idrpRoute_entry on an existing
	 *	      route_id list.  It could have same attribute
	 *	      set received from multiple different peers.
	 *	      it could be the same peer with multiple
	 *	      route_ids received.  It is likely
	 *	      that this will be the majority of the route_id
	 * 	      list entries with one attribute... (skh 2/22/93)
	 *		(debugging will tell for sure ) 
	 *  By setting the rt_chg[i].head pointer to that idrpRoute_entry
	 *  the route_id information is saved.
	 *  (Please verify during debugging that the route_id stuff is all
	 *  set-up)
	 */
	  

	NLRI_FAMILY_LOOP(i)
		{
		p_idrp_re->p_route[i] = rt_chg[i].head;
		res->rt_chg[i] = rt_chg[i].head;
		} NLRI_FAMILY_LOOP_END


	/* increment the attribute record for the number of linked nlri on route_id
	 * Exciting note:  route_lists are separate for ip and osi
	 * - but reference count increment once for each NLRI - osi/ip
	 */
 

	p_att->ref_cnt += nlri_cnt;	/* increment by the number of nlri */

	if (!reuse_att)
		{
		link_att_list(p_att);
		}
	/* save results in result array from UPDATE pdu */
 
	res->route_id_in = route_id_in;
	res->pref_rcvd = p_idrp_re->loc_pref;

	/*  parsed pdu UPDATE pdu */

	return (0);

}
	
void parse_update_cleanup(p_res,p_ent,peer,reuse_att)
parse_results	*p_res;
idrpRoute_entry *p_ent;
idrpPeer	*peer;
int		reuse_att;
{
idrpRoute_entry	*p_last = 0;
idrpRoute_entry	*p_cur = 0;

/* routine to delete the partial results of the attribute
 * record when aborted mid-stream in parse_update 
 */

	/* two different types of parse_update cleanups
	 * resusing an attribute
	 * and no re-use of attribute)
	 */

	peer = p_res->p_att->peer_alloc;
 
	if (reuse_att)
		{
		/* 1) find the id on the route_id list 
		 * 2) free any idrpRoutes associated with it
		 * 3) free the route_id list
		 * 4) exit
		 * - do not decrement reference count 
		 */	 
	
		if (!p_res->p_att)
			{
			trace (TR_IDRP,0,"IDRP parse_cleanup has reuse_att and no attribute record peer %s",peer->name);
			return;
			}	
		
		/* attribute record exists - try to find this route_id */


		if (!p_res->p_att->route_id_list)
			{
			trace (TR_IDRP,0,"IDRP parse_cleanup has reuse_att and no route_id chain for attribute record peer %s",peer->name);
			return;
			}

		/* free the nlris on the found chain and relink around it */

		idrp_free_routeid_chain_id(p_res->p_att,p_ent->route_id,peer);
		}			
		
	else
		{	
		
		/* just first creation of pdu in various stages of hacking 
		 */ 

	
		/* check to see if we have an attribute record at all */ 

		if (!p_res->p_att)
			{
			return;
			}

		/* attribute record exists - now check for the route_id list */

		if (p_ent)
			{
			idrp_free_routeid_chain_id(p_res->p_att,p_ent->route_id,peer);
			}	
		if (p_res->p_att->p_attr_pdu)
			{
			task_mem_free(peer->task,p_res->p_att->p_attr_pdu);
			}

		task_mem_free(peer->task,(void_t)p_res->p_att);
		}
			
}


static void process_update(peer, pdu, length)
idrpPeer *peer;
idrpPdu *pdu;
int length;
{
int family;
parse_results *results;

	results = (parse_results *) task_mem_malloc(peer->task,sizeof(parse_results));
	results->withdraw = (Withdrawl *) task_mem_malloc(peer->task,sizeof(Withdrawl));

	if (parse_update(peer,pdu,length,results))
		{
		/* invalid pdu - %% does gated really handle
		 * returning pdus - I think Dave's code loses pdus 
		 * error code and data are in the last_error_pdu 
		 * in the idrp peer structure
		 */ 

		idrp_sm(peer,IDRP_UPDATE_ERROR);
		task_mem_free(peer->task,results->withdraw);
		task_mem_free(peer->task,results);
		}
	else
		{
		/* valid pdu so let's send it through the state machine */

		if (idrp_sm(peer,IDRP_UPDATE))
			{
			/* state machine says process update pdu */

			if (peer->refresh.seq_start)
				{
				/* mid rib_refresh, hook the rib refresh pdus */
				/* together for processing */

				refresh_pdu_link(peer,results,pdu);
				return;
  				}
			else
				{
				/* away for processing of pdus into phase1 processing
				 *
				 */ 

				idrp_process_pdu_routes(peer,results,IDRP_UPDATE_TYPE);

				/* clean up the results array */

				task_mem_free(peer->task,results->withdraw);
				task_mem_free(peer->task,results);
				}
			} /* if state machine returns process pdu */
		else
			{
			/* state machine error - don't process pdus and release it */ 
			task_mem_free(peer->task,results->withdraw);
			task_mem_free(peer->task,results);
			}
			
		} /* end if parse update valid */

}

/*
 * KEEPALIVE PDU
 */
static void process_keepalive(peer, pdu, length)
idrpPeer *peer;
idrpPdu *pdu;
int length;
{
	idrp_sm(peer,IDRP_KEEPALIVE);
}


static void process_error(peer, pdu, length)
idrpPeer *peer;
idrpPdu *pdu;
int length;
{
int err_code = 0;
int subcode = 0;
int error = 0; 

	/* validate ERROR pdu */

	err_code = pdu->pdu.error.header.error_code;
	subcode = pdu->pdu.error.header.error_subcode;
	
	switch (err_code)
		{
		case IDRP_ERROR_OPEN_PDU:
			if ((subcode < IDRP_OPEN_ERROR_MIN_SUBCODE)
			 && (subcode > IDRP_OPEN_ERROR_MAX_SUBCODE))
				{	 
				error=TRUE;
				}
			break;

		case IDRP_ERROR_UPDATE_PDU:
			if ((subcode < IDRP_UPDATE_ERROR_MIN_SUBCODE)
			 && (subcode > IDRP_UPDATE_ERROR_MAX_SUBCODE))
				{	 
				error = TRUE; 
				}
			break;

		case IDRP_ERROR_HOLDTIME_EXPIRED:
			if (subcode != IDRP_HOLDTIME_EXPIRED_ERROR_SUBCODE)
				{
				error = TRUE;  
				}
			break;

		case IDRP_ERROR_FSM:
			if ((subcode < IDRP_FSM_ERROR_MIN_SUBCODE)
			 && (subcode > IDRP_FSM_ERROR_MAX_SUBCODE))
				{	 
				error = TRUE; 
				}
			break;

		case IDRP_ERROR_RIB_REFRESH:
			if ((subcode < IDRP_RIB_REFRESH_ERROR_MIN_SUBCODE)
			 && (subcode > IDRP_RIB_REFRESH_ERROR_MAX_SUBCODE))
				{
				error = TRUE;
				}
			break;

		default:
			error = TRUE;
			break;
		}

	if (error)
		{
		idrp_sm(peer,IDRP_ERROR_ERROR);
		log_nm_event(peer,IDRP_ERROR_PDU_ERROR,err_code);
		}
	else
		{
		idrp_sm(peer,IDRP_ERROR);
		}

}


static void process_cease(peer, pdu, length)
idrpPeer *peer;
idrpPdu *pdu;
int length;
{
	idrp_sm(peer,IDRP_CEASE);	
}

u_int parse_rib_refresh(peer,pdu,length)
idrpPeer *peer;
idrpPdu *pdu;
int length;
{
	/* range check for the Rib Refresh types */

	if (pdu->pdu.refresh.hdr.opcode < IDRP_RIB_REFRESH_MIN_CODE  ||
	    pdu->pdu.refresh.hdr.opcode > IDRP_RIB_REFRESH_MAX_CODE)
		{
		peer->last_error_pdu.error_code = IDRP_ERROR_RIB_REFRESH;
		peer->last_error_pdu.error_subcode = IDRP_RIB_REFRESH_ERROR_INV_OPCODE;
		peer->last_error_pdu.data_len = 0;
	 	return (IDRP_ERROR_RIB_REFRESH);
		}

	/* check the attribute types */
		
	/* d) Rib_Att_Set - Default only	 */
	
	if  (pdu->pdu.refresh.rib_att[0].RIB_att_count != RIB_ATT_DEFAULT_ONLY_COUNT)
		{
		peer->last_error_pdu.error_code = IDRP_ERROR_RIB_REFRESH;
		peer->last_error_pdu.error_subcode = IDRP_RIB_REFRESH_ERROR_UNSUP_ATTRS;
		peer->last_error_pdu.data_len = 0;
	 	return (IDRP_ERROR_RIB_REFRESH);
		}
	return(0);
}


static void process_rib_refresh(peer, pdu, length)
idrpPeer *peer;
idrpPdu *pdu;
int length;
{

int seq_no = 0;
	
	if (parse_rib_refresh(peer,pdu,length))
		{
		/* error in pdu */

		idrp_sm(peer,IDRP_REFRESH_ERROR);
		}
	else
		{
		/* valid Rib Refresh */
		/* check against state machine */

		if(idrp_sm(peer,IDRP_REFRESH))
			{
	    		switch (pdu->pdu.refresh.hdr.opcode)
				{
				case IDRP_RIB_REFRESH_REQ:
					idrp_send_refresh_all_routes(peer);
					break;	

				case IDRP_RIB_REFRESH_START:
					if (peer->refresh.seq_start !=0)
						{
						rib_refresh_pdu_release(peer);
						}	
					bcopy(&pdu->header.seq,&seq_no,sizeof(pdu->header.seq));

					trace(TR_IDRP, 0, "IDRP Refresh with start Seq no: = %d peer %s",
					seq_no,peer->name);
					peer->refresh.seq_start=seq_no;
					peer->refresh.pdu_list_start = (idrpRefresh * ) 0;
					peer->refresh.pdu_list_end = (idrpRefresh * ) 0;
					break;

				case IDRP_RIB_REFRESH_END:	
					bcopy(&pdu->header.seq,&seq_no,sizeof(pdu->header.seq));
					trace(TR_IDRP, 0, "IDRP Refresh with END Seq no = %d peer %s",
					seq_no,peer->name);
					peer->refresh.seq_stop=seq_no;
					rib_refresh_update_process(peer);
					peer->refresh.seq_start = 0;
					peer->refresh.seq_stop = 0;
					break;
				} 
			} /* if for state machine */
		} /* if parse valid */

}


#ifdef ECHOKLUDGE
static void process_echo(peer, pdu, length)
idrpPeer *peer;
idrpPdu *pdu;
int length;
{
	switch (peer->state) {

	case IDRP_ESTABLISHED:
		send_echo_reply(peer);
		break;

	default:
		break;
	}
}


static void process_echo_reply(peer, pdu, length)
idrpPeer *peer;
idrpPdu *pdu;
int length;
{
	switch (peer->state) {

	default:
		break;
	}
}
#endif


static void process_header_error(peer, pdu, length)
idrpPeer *peer;
idrpPdu *pdu;
int length;
{
	log_nm_event(peer, IDRP_PDU_HEADER_ERROR,IDRP_PDU_HEADER_ERR_BAD_TYPE);
	idrp_sm(peer,IDRP_HEADER_ERROR);
}


/*
 * Bump the expected receive sequence number if the packet is sequenced.
 * Send a keepalive as acknowledgement if either the send queue is empty,
 * or the connection is flow blocked, so long as we're in ESTABLISHED state.
 *
 * If the received frame is sequenced, cancel any pending keepalive
 * retransmit (due to the peer being flow blocked).  If the received frame
 * has credit_available nonzero (the peer is not flow blocked), cancel any
 * pending keepalive retransmit.
 */
static void process_rcv_seq(peer, pdu)
idrpPeer *peer;
idrpPdu *pdu;
{
	if (IS_SEQUENCED(pdu->header.pdu_type)) {
		peer->rcv_seq_expected++;
		if (peer->cred_avail_zero_seq_last) {
			peer->cred_avail_zero_seq_last = 0;
			kill_rexmit_timer(peer);
		}
	}

	if (pdu->header.credit_available) {
		if (peer->cred_avail_zero_seq_last) {
			peer->cred_avail_zero_seq_last = 0;
			kill_rexmit_timer(peer);
		}
	}

	/*
 	 * Track the received credit available as part of a mechanism to ensure that
	 * our peer doesn't get flow blocked for a long time because our ack got lost.
	 */

	peer->rcved_credit_avail = pdu->header.credit_available;
	if (peer->rcved_credit_avail == 0) {
		peer->cred_avail_zero_seq = peer->rcv_seq_expected - 1;
	}

	if (IS_SEQUENCED(pdu->header.pdu_type)) {
		if (peer->state == IDRP_ESTABLISHED &&
		    (peer->xmit_queue_head == NULL || (peer->upper_window_edge <= peer->send_next && !acked_something))) {
			send_keepalive(peer);
		}
	}
}

static in_sequence(peer, seq)
idrpPeer *peer;
u_int seq;
{
	if (seq == peer->rcv_seq_expected) {
		trace(TR_IDRP, 0, "In sequence");
		return(TRUE);
	} else {
		trace(TR_IDRP, 0, "Out of sequence");
		return(FALSE);
	}
}


/*
 * A packet for a particular peer task has arrived.  Figure out what kind of packet it
 * is, and do something about it.
 */

static void idrp_recv_pdu(peer, pdu, length)
idrpPeer *peer;
idrpPdu *pdu;
int length;
{
u_char type;
idrp_error_pdu	error_code;	/* error codes */
u_int rcv_seq, rcv_ack;		/* Seq, Ack fields from received PDU */
u_int	idrp_fixed_header_len;	/* calculation of fixed header length depending on protocol */ 


	/* If we're closed, don't even worry about it. */
	if (peer->state == IDRP_CLOSED)
		return;


	/* Validate the checksum. */

	if (!checksum_ok(peer, pdu, length))
		{
                trace(TR_IDRP, 0, "IDRP PDU error - checksum failed on PDU from peer %s", peer->name);
		return;
		}

	/* Validate the header - min and maximum length 
	 * - please notice the header length is only the BISPDU header
	 *   length having been unwrapped from the protocol header
	 *
	 */ 
	
	if (length < MinPDULength)
		{ 
		trace(TR_IDRP,0,"failed minimum size check");
		idrp_sm(peer,IDRP_HEADER_ERROR);
		return;
		}
	if (!valid_header(peer,pdu,length))
		{
		trace(TR_IDRP,0,"failed valid header check");
		idrp_sm(peer,IDRP_HEADER_ERROR);	
		return;
		}
 
	/* Extract sequence information. */

	bcopy((caddr_t) pdu->header.seq, &rcv_seq, sizeof(pdu->header.seq));
	rcv_seq = ntohl(rcv_seq);
	bcopy((caddr_t) pdu->header.ack, &rcv_ack, sizeof(pdu->header.ack));
	rcv_ack = ntohl(rcv_ack);


	/* Update the hold timer */

	timer_set(TASK_TIMER(peer->task, IDRPTIMER_HOLDTIME), peer->hold_time, (time_t) 0);

	/* Process CEASE and ERROR PDUs before anything else. */
	/* CEASE and ERROR pdus need to be processed w/ack */ 
	/* - processed first because bad news needs */
	/* to travel fast */

	type = pdu->header.pdu_type;

	trace(TR_IDRP, 0, "IDRP (%s) RCV %s seq %d ack %d offer %d avail %d len %d from %s", 
	      trace_state(peer->state, idrpStates),
	      (type > IDRP_PDUTYPE_MAX ? "Bogus" : idrpTypes[type]), 
	       rcv_seq, rcv_ack, pdu->header.credit_offered, pdu->header.credit_available,
		length, peer->name);

	switch (type) {

	case IDRP_OPEN_TYPE:
		if (idrp_parse_open(peer, pdu, length))  
			{
			/* bogus open - don't process acks */
			idrp_sm(peer,IDRP_OPEN_ERROR);
			pdu = NULL;
			}
		break;

	case IDRP_ERROR_TYPE:
		process_error(peer, pdu, length);
		pdu = NULL;
		break;

	case IDRP_CEASE_TYPE:
		process_cease(peer, pdu, length);
		pdu = NULL;
		break;
	}
	
	if (pdu == NULL)
		return;		/* Nothing else to do. */

	/* Process received acknowledgements. */
	/* the above code has streamed out    */
	/* - close state 		*/
	/* ERROR pdus (with or without errors) */  
	/* CEASE pdus (with or without errors) */
	/* OPEN pdus with errors */
	/* - ack_pdu handles ack by type and state */  

	acked_something = FALSE;
	switch (peer->state) {
	/* Open_Sent state: for OPEN PDUs only, process ack */
	
	case IDRP_OPEN_SENT:
		if (type == IDRP_OPEN_TYPE)
			{ 	
			ack_pdu(peer, rcv_ack, pdu->header.credit_offered,type);
			}
		break;

	default:
		ack_pdu(peer,rcv_ack,pdu->header.credit_offered,type);
		break;
	}

	/* Now process the send sequence number. */

	switch (peer->state) {

	/*
	 * Open_sent state:  treat everything as in-sequence except an out-of-sequence OPEN.
	 * This allows us to bitch about FSM errors.
	 */
	case IDRP_OPEN_SENT:
		peer->rcv_seq_expected = rcv_seq;  /* Establish initial sequence number */
		if (type == IDRP_OPEN_TYPE) {
			if (in_sequence(peer, rcv_seq)) {
				process_rcv_seq(peer, pdu);
			} else {
				pdu = NULL;
			}
		}
		break;

	/*
	 * Everything else--received packets must be in-sequence or we drop them.
	 */
	default:
		if (in_sequence(peer, rcv_seq)) {
			process_rcv_seq(peer, pdu);
		} else {
			pdu = NULL;
		}
		break;
	}

	if (pdu == NULL)
		return;		/* Nothing else to do. */

	/*
	 * Now process the received PDUs.
	 */
	switch (type) {

	case IDRP_OPEN_TYPE:
		process_open(peer, pdu, length);
		break;

	case IDRP_UPDATE_TYPE:
		process_update(peer, pdu, length);
		break;

	case IDRP_KEEPALIVE_TYPE:
		process_keepalive(peer, pdu, length);
		break;

	case IDRP_RIB_REFRESH_TYPE:
		process_rib_refresh(peer, pdu, length);
		break;

#ifdef ECHOKLUDGE
	case IDRP_ECHO_TYPE:
		process_echo(peer, pdu, length);
		break;

	case IDRP_ECHO_REPLY_TYPE:
		process_echo_reply(peer, pdu, length);
		break;
#endif

	default:
		process_header_error(peer, pdu, length);
		break;

	}
}


/*
 * A packet has arrived on the common socket.  Fetch it and demux.
 * - here we have to change radically the idrp code 
 * - currently it does only IP based CLNP -  
 * - note received pdu is returned by task???  
 */

void idrp_recv(tp)
task *tp;
{

	idrpPeer *peer;
	idrpPdu *pdu;
	idrp_proto_header *pdu_proto = (idrp_proto_header *) task_recv_buffer;
	size_t length;
	int i,len;

	if (task_receive_packet(tp, &length)) {
		return;
	}

	switch (idrp_this_node.proto_sock)
		{
		case IDRP_PDU_PROTO_IP_RAW:
			peer = idrp_ip_raw_sock_recv(pdu_proto);
			if (!peer)
				{
				/* trace this and return */
				return;
				}
			pdu = (idrpPdu *) &pdu_proto->hdr.ip.pdu;
			len = length - IDRP_IP_HEADER_LEN;
			break;



		case IDRP_PDU_PROTO_UDP:
			peer = idrp_udp_sock_recv(pdu_proto);
			if (!peer)
				{
				/* trace this as below and return */
				return;
				}
			pdu = (idrpPdu *) &pdu_proto->hdr.udp.pdu;
			len = length - IDRP_UDP_HEADER_LEN;
			break;

		case IDRP_PDU_PROTO_CLNP:
			peer = idrp_clnp_sock_recv(pdu_proto);
			if (!peer)
				{
				/* trace this and exit */

				return;
				}
			pdu = (idrpPdu *) &pdu_proto->hdr.clnp.pdu;
			len = length - (peer->neighbor.isoa_len + rt_net.isoa_len);
			break;

		case IDRP_PDU_PROTO_IDRP:
			peer = idrp_idrp_sock_recv(pdu_proto);
			if (!peer)
				{
				/* just exit */
				return;
				}
			pdu = (idrpPdu *) &pdu_proto->hdr.idrp.pdu;
			len = length - IDRP_IDRP_HEADER_LEN;   
			break;
		}


	/* Ensure that it's an IDRP packet. */

	if (pdu->header.pid != IDRP_PID) {
		trace(TR_IDRP, 0, "PDU type 0x%x not IDRP, from ", peer->name);
		return;
	}

	/* Deliver the packet to the appropriate task. */
	
	idrp_recv_pdu(peer, pdu, len);
}
