#define DROP_PACKET (drop_packet())
/* #define DROP_PACKET (FALSE) */

#include "include.h"
#include "idrp.h"
#include <stdio.h> 
#include "md4global.h"	/* RAREF globals and constants */
#include "md4.h"	/* md4-specific includes */

/* 
 * idrp_pdus.c 
 *
 * pdu sending 
 *	idrp_post(peer,buf) - post a message, and put on re-xmit if seq
 *	post_enqueued_pdus(peer) - post as many enqueued PDUs as possible
 * 	idrp_send_pdu(peer,pdu,type,length) - send pdu to peer 
 *	send_open(peer) - send open BISPDU to peers
 *	send_update_pdu(peer,route,withdrawl,unreachable)
 *	idrp_send_error(peer,code)
 *      send_cease(peer)
 * 	send_keepalive(peer)
 *	send Refresh_pdu
 */ 
  
 
/*
 * Post a buffer to the socket for transmission.  Move it to the
 * retransmission queue if it is sequenced.
 *
 * The buffer must not be linked into any queue.
 */
static void idrp_post(peer, buf)
idrpPeer *peer;
idrpBuffer *buf;
{
	sockaddr_un *p_dst;
	u_int longhold;
	idrp_fixed_pdu_header *header = &(buf->pdu->header);
	u_char	digest[16];	

	/* Fill in the flow control fields.  The sequence number was assigned
	 * earlier; all other fields come from the current state. */

	longhold = htonl(buf->seq);
	bcopy((caddr_t) &longhold, (caddr_t) header->seq, sizeof(header->seq));

	longhold = htonl(peer->rcv_seq_expected - 1);
	bcopy((caddr_t) &longhold, (caddr_t) header->ack, sizeof(header->ack));

	header->credit_offered = peer->rcv_credit;
	header->credit_available = peer->upper_window_edge - peer->send_next;

	bzero(&digest,16);
	bzero((caddr_t) header->validation, sizeof(header->validation)); /* zero validation now */
	switch  (peer->AuthType)
		{
		case IDRP_AUTH_CODE_TEST:
			break;
		case IDRP_AUTH_CODE_CLEAR:
			md4((u_char *)buf->pdu,buf->length,digest); 	 /* validation scheme */
			break;

		case IDRP_AUTH_CODE_CRYPT:
			/* with zeros in buffer should fail reception */ 
			trace(TR_IDRP,0, "No Authentication code crypt support, set on this peer %s",
				peer->name);
			break;   		
		}

	/* Copy the buffer to the send buffer and post it. */

	bcopy ((caddr_t) &digest,&header->validation,sizeof(header->validation));
	bcopy((caddr_t) buf->pdu, task_send_buffer, buf->proto_len);
	p_dst = idrp_send_dest_set(peer);

	/* test on the type of socket, and then call appropriate send
	 * routine
	 */

	switch (idrp_this_node.proto_sock)
		{
		case IDRP_PDU_PROTO_CLNP:
			idrp_send_clnp_pdu(&idrp_this_node,task_send_buffer,buf->proto_len,p_dst); 
			break;

		case IDRP_PDU_PROTO_UDP:
			(void) task_send_packet(idrp_this_node.task, task_send_buffer, buf->proto_len,0, p_dst);
			break;

		case IDRP_PDU_PROTO_IP_RAW:
			if (task_send_packet(idrp_this_node.task,task_send_buffer,buf->proto_len,0,p_dst) < 0)
				{
				trace (TR_IDRP,LOG_ERR,"error sending IDRP PID pdu peer %s",peer->name);
				}
			break;
		case IDRP_PDU_PROTO_IDRP:
			idrp_send_idrp_pdu(peer,task_send_buffer,buf->proto_len,p_dst);
			break;
		}	

					
	trace(TR_IDRP, 0, "IDRP (%s) XMIT %s (%s) seq %d ack %d offer %d avail %d len %d to %s", 
		trace_state(peer->state, idrpStates),
		idrpTypes[header->pdu_type], 
		buf->sequenced ? "seq" : "unseq",
		buf->seq,
		peer->rcv_seq_expected - 1,
		header->credit_offered,
		header->credit_available,
		buf->length,
		peer->name);

	if (peer->trace_flags)
		{
		trace(TR_IDRP, 0, "Sending:");
			{
			int i;
			char *cp = (char *) buf->pdu;
			for (i=0; i<buf->length; i++) {
				tracef("%2.2x  ", *cp++);
			}
			trace(TR_IDRP, 0, NULL);
			}
		}
		
	/* Put the buffer on the retransmission queue if it was sequenced. */

	if (buf->sequenced) {
		*peer->rexmit_queue_tail = buf;
		peer->rexmit_queue_tail = &buf->next;
		start_rexmit_timer(peer);
	}
}

/*
 * Post as many enqueued PDUs as possible.
 */
void 
post_enqueued_pdus(peer)
idrpPeer *peer;
{
	idrpBuffer *buf;

	/* Process the transmit queue. */

	while (peer->xmit_queue_head) {

		/* If we're flow blocked, give up. */

		if ((peer->state == IDRP_ESTABLISHED)
			&& (peer->upper_window_edge <= peer->send_next )) {
			trace(TR_IDRP, 0, "IDRP %s flow blocked", peer->name);
			break;
		}

		/* Not flow blocked.  Pop the top of the queue, assign it a sequence number, and post it. */

		buf = peer->xmit_queue_head;
		peer->xmit_queue_head = buf->next;
		if (buf->next == NULL)
			peer->xmit_queue_tail = &peer->xmit_queue_head;
		buf->next = NULL;
		buf->seq = peer->send_next++;

		idrp_post(peer, buf);
	}
}


/*
 * Build and attempt to send a PDU.
 */

void idrp_send_pdu(peer,pdu,pdu_proto,type,length,proto_length)
idrpPeer *peer;
idrpPdu *pdu;
idrp_proto_header	*pdu_proto;
u_int type;
u_int length;			/* length of total buffer size */
u_int proto_length;		/* length of protocol header */
{
	int i;
	int len;
	int waitack;
	u_short hold;
	u_int longhold;
	char *cp = (char *) pdu;
	idrp_fixed_pdu_header *header = &(pdu->header);
	idrpBuffer *buf;

	/* Fill in most of the fixed header. */
	/* PID, length, type, source, destination */
	/* copy source and destination from the Peer structure */

	header->pid = IDRP_PID;
	hold = htons(length);
	bcopy((caddr_t) &hold, (caddr_t) header->pdu_length, sizeof(header->pdu_length));
	header->pdu_type = type;

	/* fill in the UDP header if proto_sock = UDP */
	if (peer->proto_sock == IDRP_PDU_PROTO_UDP)
		{
		bcopy((caddr_t) &peer->neighbor.isoa_genaddr, (caddr_t) &pdu_proto->hdr.udp.hdr.dest.isoa_genaddr,peer->neighbor.isoa_len);
		pdu_proto->hdr.udp.hdr.dest.isoa_len = peer->neighbor.isoa_len;
		bcopy((caddr_t) &rt_net.isoa_genaddr, (caddr_t) &pdu_proto->hdr.udp.hdr.src.isoa_genaddr,rt_net.isoa_len);
		pdu_proto->hdr.udp.hdr.src.isoa_len = rt_net.isoa_len;
		}

	/* Build the buffer structure. */

	buf = (idrpBuffer *) task_mem_malloc(peer->task, sizeof(idrpBuffer));
	buf->next = NULL;
	buf->pdu_proto = pdu_proto;
	buf->proto_len = length;
	buf->pdu = pdu;
	buf->length = length - proto_length;
	buf->sequenced = IS_SEQUENCED(type);

	/* If not sequenced, post the buffer immediately.  Otherwise, enqueue it. */

	if (buf->sequenced) {
		*peer->xmit_queue_tail = buf;
		peer->xmit_queue_tail = &buf->next;
		post_enqueued_pdus(peer);
	} else {
		buf->seq = peer->send_next;
		idrp_post(peer, buf);
	}
}

/*
 * Send an OPEN PDU to (peer).
 */
void send_open(peer)
idrpPeer *peer;
{
	u_short hold;
	idrp_source_RDI *openrdi;
	idrp_RIB_att *openribatt;
	idrp_confed_id *openconfed;
	idrp_open_pdu_header *openhead;
	idrp_authentication_code *openauthcode;
	int i,len,header_len;
	u_char *cp;
	idrp_open_pdu	*open;
	u_char	*basecp;
	idrpPdu	*pdu;
	idrp_proto_header	*pdu_proto;

	/* allocate the pdu */

	pdu_proto = (idrp_proto_header *) task_mem_malloc(peer->task, peer->pdu_maxsendsize);

	/* offset the pdu pointer for the protocol socket
	 * and record the extra header length 
	 */

	header_len = proto_offset(&pdu,pdu_proto,peer->proto_sock);
	open = (idrp_open_pdu *) &pdu->pdu.open;
	cp = (u_char *) open;
	basecp = (u_char *) open;

	/* Fixed header */


	bzero(pdu,peer->pdu_maxsendsize);
 	openhead = &(open->header);
	openhead->version = IDRP_VERSION_1;

	hold = htons(peer->hold_time);
	bcopy((caddr_t) &hold, (caddr_t) openhead->idrp_hold_time, sizeof(openhead->idrp_hold_time));

	hold = htons(peer->pdu_maxrcvsize);
	bcopy((caddr_t) &hold, (caddr_t) openhead->idrp_maxpdu_size, sizeof(openhead->idrp_maxpdu_size));

	openhead->idrp_initial_credit = peer->rcv_credit;
	cp += sizeof(idrp_open_pdu_header);

	/* Source RDI field */

	openrdi = (idrp_source_RDI *) cp;
	openrdi->source_RDI_length = rt_rdi.isoa_len;
	bcopy((caddr_t) rt_rdi.isoa_genaddr, (caddr_t) openrdi->source_RDI, openrdi->source_RDI_length);
	cp += openrdi->source_RDI_length + 1;

	/* RIB attributes */

	openribatt = (idrp_RIB_att *) cp;
	openribatt->RIB_att_count = 0;
	cp += sizeof(openribatt->RIB_att_count);

	/* Confederation IDs */
	/* %%eventually take from config */

	openconfed = (idrp_confed_id *) cp;
	openconfed->confed_id_count = 0;
	cp += sizeof(openconfed->confed_id_count);

	/* Authorization code */

	openauthcode = (idrp_authentication_code *) cp;
	openauthcode->authentication_code = IDRP_AUTH_CODE_TEST;
	cp += sizeof(openauthcode->authentication_code);

	/* Now send it out. */
	len = cp - basecp + sizeof(idrp_fixed_pdu_header) + header_len; 
	idrp_send_pdu(peer, pdu, pdu_proto, IDRP_OPEN_TYPE, len,header_len);
}


/*
 * Send an update PDU
 */

extern void
send_update_pdu(peer,p_send_list,unreachable)
idrpPeer *peer;
idrp_send_list *p_send_list;
u_int unreachable;
{
/* send list provides the following
 * - pointer to attribute record 
 * - pointer head of idrpRoute list for nlri
 *   nlri is linked idrpRoute->route_out[peer->id] structure  
 * - pointer to tail of idrpRoute list for nlri 
 * - Witdrawl structure with IDs 

/* idrpRoute pointers */
idrpRoute	*p_idrp_rt;		/* generic route pointer */
idrpRoute	*p_irt_head;		/* list of idrpRoute structures - head */
idrpRoute	*p_irt_tail;		/* list of idrpRoute structures - tail */
idrp_attribute_record	*p_att;		/* attribute record */ 
int		next_hop_off;		/* next hop attribute off */

/* pdu pointers */

idrp_proto_header	*pdu_proto;
idrpPdu *pdu;
idrp_update_pdu *hdr;
nlri_info *nlriptr;
u_char	*nlristart;

/* generic character pointers */

int i;
int cnt;
int len;
int seglen;	
int tot_len,header_len;
u_char *segout;
u_char *cpout;
u_char *p_out;
u_char *cpin;
u_char *cpbase;
u_short hold;
u_int longhold;


	/* set up pointers for the task */
	/* pdu and maximum pdu size */

	pdu_proto = (idrp_proto_header *) task_mem_malloc(peer->task, peer->pdu_maxsendsize);
	bzero(pdu_proto,peer->pdu_maxsendsize);

	/* set-up the pdu pointer to point to idrp fixed header */

	header_len = proto_offset(&pdu,pdu_proto,peer->proto_sock); 

	/* fill in the unfeasible route info. 
	 * make sure routes won't overload the UPDATE 
	 * maximum size 
	 * return with complaint
	 */

	tot_len = p_send_list->withdraw->count * sizeof (route_id);
	tot_len += MinPDULength + IDRP_NO_NLRI_UPDATE; 

	if (tot_len > peer->pdu_maxsendsize)
		{
		trace (TR_IDRP,0,"bad withdraw pdu size at send_update_pdu peer %s cnt %s",
				peer->name,p_send_list->withdraw->count);
		trace (TR_IDRP,0,"p_send_list pointer = %x",p_send_list);
		return;
		}

	/* drop the count of the withdraws into the pdu */

	hold = cnt = p_send_list->withdraw->count;
	hold = htons(hold);

	cpout = (u_char *) &pdu->pdu.update;

	bcopy(&hold,&pdu->pdu.update.unfeasible_route_count,UNFEASIBLE_RT_CNT_LEN);

	/* set the pointer to the first byte beyond the unfeasible route count */

	cpout += UNFEASIBLE_RT_CNT_LEN;

	if (cnt)
		{
		/* unfeasible route ids - so stuff in update pdu */ 
		for (i = 0;  i < cnt; i++)
			{
			/* 4 bytes of route_id out */

			longhold = htonl(p_send_list->withdraw->withdraw[i]);
			bcopy(&longhold, cpout, sizeof(route_id));
			cpout += sizeof(route_id);
			}
		}

	/* is it just unreachables - */

	if (unreachable)
		{ 
		/* - Just unreachables so put 
		   - put out fixed header for path attribute length */

		longhold = 0;
		bcopy((caddr_t) &longhold,cpout,sizeof(hdr->tot_path_att_len));
		cpout += sizeof(hdr->tot_path_att_len);
	
		/* trace this update and send with just unreachables */
	
		trace(TR_UPDATE, 0, "IDRP XMIT Update to %s", peer->name);
		len = cpout - (u_char *) pdu + header_len;
		idrp_send_pdu(peer, pdu, pdu_proto, IDRP_UPDATE_TYPE, len, header_len);
		return;
		}

	/* set up the attribute record */

	p_att = (idrp_attribute_record *) p_send_list->p_attr; 
	
	next_hop_off = send_next_hop(p_att,peer);
  
	/* set the cpbase (output counter) to the first byte of the path attributes 
	 * set the cpin to the path attribute - stuff that in there first  
	 */
		
	cpin =  p_att->attrib[IDRP_ATTR_RD_PATH].data;  
	cpout += sizeof(hdr->tot_path_att_len);
	cpbase = cpout;
	
	/* Now add the route ID. */

	*cpout++ = IDRP_ATTRFLG_ROUTE_SEPARATOR;  /* Well-known, mandatory */
	*cpout++ = IDRP_ATTR_ROUTE_SEPARATOR;
	*cpout++ = 0;
	*cpout++ = IDRP_ROUTE_SEPARATOR_LENGTH;

	/* route id out depends on the neighbor */ 

	longhold = htonl(peer->route_id_out);
	bcopy((caddr_t) &longhold, cpout, IDRP_ROUTE_ID_LENGTH);
	cpout += IDRP_ROUTE_ID_LENGTH;
	bcopy((caddr_t) &p_send_list->loc_pref,cpout,IDRP_ROUTE_PREF_LENGTH);
	cpout += IDRP_ROUTE_PREF_LENGTH;

	/* path attributes new routes - so fill in attributes 
	 * - from the attribute record 
	 * Mandatory RD_PATH, HOPCOUNT and ROUTE_ID
	 *
	 */  

	if (p_att->attrib[IDRP_ATTR_RD_PATH].present == 0)
		{
		trace (TR_IDRP,0,"update has no rd_path peer = %A", 
			sockbuild_iso((byte *) peer->neighbor.isoa_genaddr, peer->neighbor.isoa_len));
			return;
		}
	
	/* First encode the path information. 
	 * In, array but at first lets hard code it
	 */ 
	
	*cpout++ = IDRP_ATTRFLG_RD_PATH; /* Well-known, mandatory */
	*cpout++ = IDRP_ATTR_RD_PATH;
	
	/* our RD is in local part of array
	 * if neighbor is external or local Route 
	 * this part of array is already set-up 
	 * in the attribute - attr_rec->attrib[IDRP_LOCAL_RD_PATH];
	 */
	
	len = hold = p_att->attrib[IDRP_ATTR_RD_PATH].length; 
	hold = htons(hold);
	bcopy((caddr_t) &hold, cpout, 2);	/* Attribute length */
	cpout += 2;

	/* check to make sure length is not zero */

	if (!p_att->attrib[IDRP_ATTR_RD_PATH].length) {
		trace(TR_ALL, LOG_ERR, "null path in send_update_pdu, %x peer %s",p_idrp_rt,p_idrp_rt->peer->name);
		kill(task_pid, SIGINT);
		}

	/* HOPCOUNT is mandatory */
	/* Add it to the pdu */

	/* if a local RD path has been added to this RD path */
	/* then add the local in from the local path */ 

	if (p_att->attrib[IDRP_ATTR_LOCAL_RD].present)
		{
		/* take off the length of the local RD */
		seglen = p_att->attrib[IDRP_ATTR_LOCAL_RD].length;	
		len -= seglen; 
		segout = cpout + len;	
		bcopy ((caddr_t) p_att->attrib[IDRP_ATTR_LOCAL_RD].data,(caddr_t) segout,seglen); 
		}
		
	/* copy part that came from another router 
	 * for a local router this may be zero 
	 */

	if (len)
		{
		bcopy ((caddr_t) p_att->attrib[IDRP_ATTR_RD_PATH].data,cpout,len);
		}

	/* reset the length for the path attribute length to be after path attribute */

	cpout += p_att->attrib[IDRP_ATTR_RD_PATH].length; 

	if (!p_att->attrib[IDRP_ATTR_HOP_COUNT].present)
		{	
		cpout = p_att->attrib[IDRP_ATTR_ROUTE_SEPARATOR].data; 
		trace(TR_ALL, LOG_ERR, "NO HOPCOUNT in route %x peer %s",cpout,peer->name);
		kill(task_pid, SIGINT);
		}

	*cpout++ = IDRP_ATTRFLG_HOP_COUNT;  /* Well-known, mandatory */
	*cpout++ = IDRP_ATTR_HOP_COUNT;
	*cpout++ = 0;
	*cpout++ = IDRP_RD_HOP_COUNT_LENGTH;
	*cpout++ = p_att->hopcount + 1;


	/* past the mandatory attributes 
	 * here's where we add the following copies 
 	 * of attribute record 
	 */

	if (p_att->idrp_mask == IDRP_ATTBT_MANDATORY)
		{
		trace (TR_IDRP,0,"update send to peer %s with just mandatory attributes",peer->name);
		}
	else
		{
		trace (TR_IDRP,0,"peer %s got UPDATE with attribute mask %x",peer->name,p_att->idrp_mask); 
		/* more attributes to go in the pdu */
		
		for (i = IDRP_ATTR_EXT_INFO; i < IDRP_ATTR_MAX; i++)
			{
			/* loop copying in the attributes verbatum
			 * flags(1 byte), type(1 byte), length(2 bytes) 
			 * probably should have this above too
			 */
	
			if (p_att->attrib[i].present)
				{
				/* present and non-mandatory - stuff in the update pdu */
 
				if (p_att->attrib[i].flags != IDRP_ATTR_FLAG_MANDATORY)
					{
					/* must change value 2 to symoblic reference (skh 3/26/93) */
					*cpout++ = p_att->attrib[i].flags;
					*cpout++ = (u_char) i;
					hold = htons(p_att->attrib[i].length);
					bcopy((caddr_t) &hold,cpout,2);  
					cpout +=2;
   					bcopy((caddr_t) p_att->attrib[i].data,cpout,p_att->attrib[i].length);
					cpout +=p_att->attrib[i].length;
					}
				}
			}   
		}
	 
		
	/* That was fun.  Now go back and update the total attribute length. */

	hold = htons(cpout - cpbase);
	p_out = cpbase - sizeof(hdr->tot_path_att_len);
	bcopy((caddr_t) &hold, p_out,sizeof(hdr->tot_path_att_len));

	/* Lastly, add the NLRI */
	/* walk list of idrp Routes putting out routes */

	/* send list will provide linked list per ISO families */
	NLRI_FAMILY_LOOP(i)
		{
		if (!p_send_list->ann_nlri[i].head)
			{
			continue;
			}

		/* copy in protocol id, length and value */
	
		*cpout = idrp_nlri_proto_id[i].proto_type;
		cpout++;
		*cpout = hold = idrp_nlri_proto_id[i].proto_len;
		cpout ++;
		bcopy(idrp_nlri_proto_id[i].proto_val,cpout,hold);
		cpout +=hold;
	 	
		/* skip over total type length */	

		cpout += 2; 
		nlristart = cpout;	

		/* set up linked list for array headers */
	
		p_idrp_rt  = p_send_list->ann_nlri[i].head;
		p_irt_tail = p_send_list->ann_nlri[i].tail;

		/* start copying in the nlris */

		while (p_idrp_rt)  
			{		
			/* prefix length */

			hold = PFX_BYTE_LENGTH(p_idrp_rt->nlri);
			*cpout = p_idrp_rt->nlri.pfx_len;
			cpout++;
		
			/* prefix */

			bcopy(&p_idrp_rt->nlri.pfx,cpout,hold);
			cpout += hold;

			/* set up route_out array and get next nlri if
			 * it is there
			 */

			if (p_idrp_rt == p_irt_tail)
				{
				/* last one - so close the outbound list for
				 * this peer and zero p_idrp_rt to stop loop
				 */

				p_idrp_rt->route_out[peer->id].p_next = p_send_list->ann_nlri[i].head;
				p_idrp_rt->route_out[peer->id].route_id = peer->route_id_out;
				p_idrp_rt = (idrpRoute *) 0;
				}
			else
				{
				/* still more on the loop - so link route_out list
				 * to next peer
				 * get next route
				 */
 
				p_idrp_rt = p_idrp_rt->route_out[peer->id].p_next;
				}
			}  /* done with nlri per family */

		/* copy in length */

		hold = htons(cpout - nlristart);
		p_out = nlristart - sizeof(nlriptr->addr_len); 
		bcopy(&hold,p_out,sizeof(nlriptr->addr_len));
 
		} /* looping for packing nlri for all families */ 

	/* Finally, send it out. */

	tot_len = cpout - (u_char *)pdu + header_len;
	if (tot_len > peer->pdu_maxsendsize)
		{
		trace (TR_IDRP | TR_UPDATE,0,"IDRP XMIT peer %s update has too big pdu %d max size %d", peer->name,tot_len,peer->pdu_maxsendsize);
		return;
		}

	/* well - it's going out as best as we can tell update route_id
	 * to be ready for the next UPDATE BISPDU
	 */ 
	peer->route_id_out++;
	if (next_hop_off)
		{
		/* next hop attribute turned off because
		 * it is redundant - but turn it back on later
		 */
		p_att->attrib[IDRP_ATTR_NEXT_HOP].present = TRUE;
		}
	trace(TR_UPDATE, 0, "IDRP XMIT Update to %s", peer->name);
	idrp_send_pdu(peer, pdu, pdu_proto, IDRP_UPDATE_TYPE, tot_len,header_len);
}




/*
 * Send an error PDU
*/
void
idrp_send_error(peer, code)
idrpPeer	*peer;
u_int		code;
{
u_int subcode;
caddr_t *data;
u_int datalen;
u_int datalen2;
u_int header_len;
u_int packetlen;
idrp_proto_header	*pdu_proto;
idrpPdu *pdu;
u_char *p_errpdu;

	if (code != peer->last_error_pdu.error_code)
		{
		trace (TR_IDRP,0,"error with send_error routine peer %s last_error_pdu code = %c call = %c", 
		peer->name,code,peer->last_error_pdu.error_code);
		}

	header_len = MinPDULength + proto_header_len[peer->proto_sock]; 
	#ifdef ADN_PATCH
	/* just for compilation */
	{
		u_int	tmp	= 64;

		packetlen = header_len + tmp;
	}
#else
	packetlen = header_len +  IDRP_ERROR_PDU_HEADER_LEN + datalen + datalen2 + sizeof(idrp_error_header);
#endif /* ADN_PATCH */
	pdu_proto = (idrp_proto_header *) task_mem_malloc(peer->task,packetlen); 

	/* set up pointer to error pdu */
	header_len = proto_offset(&pdu,pdu_proto,peer->proto_sock);	
	p_errpdu = (u_char *) &pdu->pdu.error;

	/* put out the error code, then the subcode */

	*p_errpdu = peer->last_error_pdu.error_code;
	p_errpdu++;	
	*p_errpdu = peer->last_error_pdu.error_subcode;
	p_errpdu++;	
	datalen = peer->last_error_pdu.data_len;
	datalen2 = peer->last_error_pdu.data2_len;
	if (datalen)
		{
		bcopy(peer->last_error_pdu.data, p_errpdu,datalen);
		p_errpdu += datalen;
		}
	if (datalen2)
		{
		bcopy(peer->last_error_pdu.data2,p_errpdu,datalen2);
		}
	idrp_send_pdu(peer, pdu,pdu_proto, IDRP_ERROR_TYPE, packetlen,header_len);
	begin_close(peer);
}


/*
 * Send a cease PDU
 */

void send_cease(peer)
idrpPeer *peer;
{
int	header_len,len;
idrpPdu	*pdu;
idrp_proto_header	*pdu_proto;

	/* get enough for IDRP_CEASE plus the udp header */
 	
	pdu_proto = (idrp_proto_header *) task_mem_malloc(peer->task, IDRP_CEASE_SIZE + sizeof(idrp_fixed_udp_header));
	header_len = proto_offset(&pdu,pdu_proto,peer->proto_sock);
	len = IDRP_CEASE_SIZE + header_len;
	idrp_send_pdu(peer, pdu,pdu_proto,IDRP_CEASE_TYPE,len ,header_len);
}


/*
 * Send a keepalive PDU.  Ensure that this keepalive will be retransmitted soon if
 * the peer thinks he is flow blocked.
 */

void send_keepalive(peer)
idrpPeer *peer;
{
idrp_proto_header	*pdu_proto;
idrpPdu	*pdu;
int	header_len;
int	len;

	pdu_proto = (idrp_proto_header *) task_mem_malloc(peer->task, IDRP_KEEPALIVE_SIZE+ sizeof(idrp_fixed_udp_header));
	header_len = proto_offset(&pdu,pdu_proto,peer->proto_sock);
	len = IDRP_KEEPALIVE_SIZE + header_len;
	idrp_send_pdu(peer, pdu,pdu_proto,IDRP_KEEPALIVE_TYPE,len ,header_len);

	if (peer->cred_avail_zero_seq_last == 0) {	/* If we aren't currently retransmitting keepalives */
		if (peer->rcved_credit_avail == 0) {	/* If the other guy is flow blocked */
			peer->cred_avail_zero_seq_last = peer->cred_avail_zero_seq;
			start_rexmit_timer(peer);
		}
	}
}


#ifdef ECHOKLUDGE
static void send_echo(peer)
idrpPeer *peer;
{
idrp_proto_header	*pdu_proto;
idrpPdu			*pdu;
int	len,header_len;

	pdu_proto = (idrp_proto_header *) task_mem_malloc(peer->task, (IDRP_ECHO_SIZE+ sizeof(idrp_fixed_udp_header));
	header_len = proto_offset(&pdu,pdu_proto,peer->proto_sock);
	len = IDRP_ECHO_SIZE+ header_len;
	idrp_send_pdu(peer, pdu,pdu_proto,IDRP_ECHO_TYPE,len ,header_len);
}

static void send_echo_reply(peer)
idrpPeer *peer;
{
idrp_proto_header	*pdu_proto;
idrpPdu			*pdu;
int	len,header_len;

	pdu_proto = (idrp_proto_header *) task_mem_malloc(peer->task, (IDRP_ECHO_REPLY_SIZE+ sizeof(idrp_fixed_udp_header));
	header_len = proto_offset(&pdu,pdu_proto,peer->proto_sock);
	len = IDRP_ECHO_REPLY_SIZE+ header_len;
	idrp_send_pdu(peer, pdu,pdu_proto,IDRP_ECHO_TYPE,len ,header_len);
}

#endif /* ECHOKLUDGE */

/* send a Rib Refresh PDU */


void
send_Refresh_pdu(peer,type)
idrpPeer	*peer;
int		type;
{
idrp_proto_header	*pdu_proto;
idrpPdu	*pdu;
char	*cp;
int	len,header_len;
idrp_rib_refresh_pdu	*p_rrib;

	switch (type)
		{
		case IDRP_RIB_REFRESH_REQ:
		case IDRP_RIB_REFRESH_START:
		case IDRP_RIB_REFRESH_END:
			break;

		default:
			trace(TR_IDRP,0," Logic error, IDRP tried to send Rib Refresh with bad type = %d peer %s",  
			type,peer->name); 
			return;
		}
				
	pdu_proto = (idrp_proto_header *) task_mem_malloc(peer->task, IDRP_RIB_REFRESH_SIZE+sizeof(idrp_fixed_udp_header));

	/* change pdu to jump over different header lengths points */


	header_len = proto_offset(&pdu,pdu_proto,peer->proto_sock);
	cp = (u_char *) pdu;
	len = MinPDULength + header_len + IDRP_RIB_REFRESH_ADD_SIZE;

	/* set up for rib refresh pdu */

	p_rrib = (idrp_rib_refresh_pdu *) &pdu->pdu.refresh; 

	/* copy in 1 byte type */

	p_rrib->hdr.opcode = (u_char) type;
	p_rrib->rib_att[0].RIB_att_count =  RIB_ATT_DEFAULT_ONLY_COUNT;
	idrp_send_pdu(peer, pdu, pdu_proto,IDRP_RIB_REFRESH_TYPE, len, header_len);
}	
	
#if 0 /* under construction */
/*
 * Trace a PDU
 */
idrp_trace_pdu(idrpPdu *pdu)
{
	char *i = pdu;
	int  j = 0;
	int  type = 0;
	int  length = 0;
	iso_net_addr theAddr;

	trace(TR_IDRP, LOG_ERR, "IDRP PDU:");
	
	/* now log the fixed header */
	trace(TR_IDRP, LOG_ERR, " Identifier: %d", *i++);
	trace(TR_IDRP, LOG_ERR, " Length    : %d", (short)*i++);
	trace(TR_IDRP, LOG_ERR, " Type      : %s", idrpTypes[type = *i++]);
	trace(TR_IDRP, LOG_ERR, " Sequence  : %d", (int)*i++);
	trace(TR_IDRP, LOG_ERR, " Ack       : %d", (int)*i++);
	trace(TR_IDRP, LOG_ERR, " Cred Offer: %d", *i++);
	trace(TR_IDRP, LOG_ERR, " Cred Avail: %d", *i++);
	trace(TR_IDRP, LOG_ERR, " Validation: 0x%x%x%x%x", (int)*i++,(int)*i++,(int)*i++,(int)*i++);

	/* now log the rest */
	switch(type) 
		{
		case IDRP_OPEN_TYPE:
			{
			trace(TR_IDRP, LOG_ERR, " Version   : %d", *i++);
			trace(TR_IDRP, LOG_ERR, " Holdtime  : %d", (short)*i++);
			trace(TR_IDRP, LOG_ERR, " MaxPDUSize: %d", (short)*i++);

			theAddr.isoa_len = *i++;
			theAddr.isoa_family = *i++;
			bcopy(i, theAddr.isoa_genaddr, theAddr.isoa_len);
			i += theAddr.isoa_len;

			trace(TR_IDRP, LOG_ERR, " Source RDI: %A", (sockaddr_un)theAddr);
			UNDER CONSTRUCTION -- jgs

			
	trace(
#endif /* under construction */
