#ifndef _idrp_prototypes_h
#define _idrp_prototypes_h

/* prototypes for routines */ 

/* initialization routines */

/* routines set-up in task.c  */

/* routines set-up in task initialization for master task */
/* see idrp_init.c for details */
/*INIT_SEQ - 	idrp_cleanup, 
 *		idrp_var_init,
 * 		parse,
 *		idrp_init,
 *		idrp_flash,
 *		idrp_newpolicy 
 *
 * tasks - idrp_recv - receives all BISPDUs from
 * 			UDP or PID or CLNP or IDRP socket
 *	- idrp_reinit - reinitialize protocol
 * 	- idrp_new_policy - reinitialize policy  
 *	- idrp_delete - delete master task of idrp protocol turned off 
 */ 

PROTOTYPE(idrp_cleanup, extern void ,(task *tp)); /* task cleanup on re-config */
PROTOTYPE(idrp_var_init, extern void , (void));   /* task Initialization prior to config */
PROTOTYPE(idrp_init, extern void , (void));       /* task init after config */
PROTOTYPE(idrp_find_interface,sockaddr_un *,());	/* get a local interface for local node's peer structure */ 


PROTOTYPE(idrp_init_this_node, extern void, (void));	/* initialize master task & nodes peer structure */ 

/* master task routines */

PROTOTYPE(idrp_recv, extern void, (task * tp));  /* Read data from socket */
PROTOTYPE(idrp_reinit, extern void, (task *tp));  /* Protocol re-initialization */
PROTOTYPE(idrp_newpolicy, extern void, (task *tp, rt_list *rtl));  /* Protocol re-initialization */
PROTOTYPE(idrp_terminate,extern void, (task *tp)); /* idrp task terminate */
PROTOTYPE(idrp_master_dump, extern void, (task *tp, FILE *)); /* Dump IDRP master task state */

/* initialize local routes from IDRP - associated with master task */

PROTOTYPE(idrp_init_local_rt, void, ());

/* routines set-up in peer task initialization 
 * idrp_parse_init - routines for idrp parsing
 */

PROTOTYPE(idrp_find_peer, extern idrpPeer * , (idrpPeer *p_new_peer));	/* find a peer on the peer list */
PROTOTYPE(idrp_link_peer, void, (idrpPeer *p_peer));	/* link a new peer to the peer list */
PROTOTYPE(idrp_peer_alloc, extern idrpPeer *, (idrpPeer * peer)); /* allocate a peer */
PROTOTYPE(idrp_small_blk_alloc,extern u_char *, ()); /* allocate small task block */
PROTOTYPE(idrp_small_blk_free,extern void,(u_char *p)); /* allocate small task block */
PROTOTYPE(idrp_med_blk_alloc,extern u_char *, ()); /* allocate medium task block */
PROTOTYPE(idrp_med_blk_free,void, (u_char *p)); /* free medium task block */
PROTOTYPE(idrp_big_blk_free,void, (idrpPeer *peer)); /* free big peer task block */
PROTOTYPE(idrp_local_peer_init,extern void,());		/* fill in local node's peer structure */ 
PROTOTYPE(idrp_delete, extern void, (idrpPeer *peer)); /* idrp peer structure deletion  and list relinked */
PROTOTYPE(idrp_peer_config_init, extern void, (idrpPeer *peer)); /* configuration init a peer structure */
PROTOTYPE(idrp_update_peer,extern void, (idrpPeer *peer_old, idrpPeer *peer)); /* update old peer from new */



/* routines to set-up the sockets for the master task 
 * file: idrp_init_sock 
 * - idrp_ip_pdi_sock_init - initialize the ip pid socket for intialization
 * - idrp_udp_sock_init - initialize the UDP socket
 * - idrp_clnp_raw_sock_init - intialize the CLNP socket
 * - idrp_idrp_sock_init - intialize the idrp socket
 * - idrp_send_clnp_pdu - send the IDRP packet via the CLNP pdu
 * - idrp_idrp_send_pdu - send IDRP packte via the CLNP in idrp socket
 */ 

PROTOTYPE(idrp_ip_pid_sock_init,void, (task *p_task, sockaddr_un *p_addr));
PROTOTYPE(idrp_udp_sock_init,void, (task *p_task, sockaddr_un *p_addr));
PROTOTYPE(idrp_clnp_raw_sock_init,void , (task *p_task, sockaddr_un *p_addr));
PROTOTYPE(idrp_idrp_sock_init,void, (task *p_task, sockaddr_un *p_addr));


/* routines for socket interaction recv and transmit
 * (transmit and recieve also found in idrp_recv and
 * idrp_post
 */

PROTOTYPE(idrp_ip_raw_sock_recv,idrpPeer *,(idrp_proto_header *pdu_proto));
PROTOTYPE(idrp_udp_sock_recv,idrpPeer *,(idrp_proto_header *pdu_proto));
PROTOTYPE(idrp_clnp_sock_recv,idrpPeer *,(idrp_proto_header *pdu_proto));
PROTOTYPE(idrp_idrp_sock_recv,idrpPeer *,(idrp_proto_header *pdu_proto));
PROTOTYPE(idrp_send_clnp_pdu,void,(idrpPeer *peer, u_char *p_buf, int len, sockaddr_un *where_to));
PROTOTYPE(idrp_send_idrp_pdu,void,(idrpPeer *peer, u_char *p_buf, int len, sockaddr_un *where_to));


/* here's the peer routines
 * idrp_peer_dump - (not done)
 * idrp_peer_reinit (not done) - reinitialize for policy
 * idrp_peer_cleanup (not done) - cleanup after terminate or on re-config
 * idrp_peer_terminate - needs to be done prior to 1st release (peer down)
 *			configure peer down
 *
 *
 * idrp_peer_ete_init	- initialize ETE stuff in peer (reinit stuff)
 */ 
   
 
PROTOTYPE(idrp_peer_dump, extern void, (task *tp, FILE *)); /* Dump IDRP state */
PROTOTYPE(idrp_peer_reinit, extern void, (task *tp)); /* reinitialize a peer */  
PROTOTYPE(idrp_peer_cleanup, extern void, (task *tp)); /* idrp cleanup on re-init */
PROTOTYPE(idrp_peer_terminate, extern void, (task *tp)); /* termination of peer task */
PROTOTYPE(idrp_peer_ete_init, extern void, (idrpPeer *peer)); /* ete variable init in a peer structure */
PROTOTYPE(idrp_peer_ete_reinit, extern void, (idrpPeer *peer)); /* ete variable init in a peer structure */
PROTOTYPE(idrp_reinit_ete_timers, extern void, (idrpPeer *peer)); /* reinit timers on working peer */


/* %% routines for iso address pretty-printing */

PROTOTYPE(iso_ptoa, extern char * , (struct iso_prefix *prefix));   /* pretty print ISO prefix */

/* comparison of 2 iso address - yes or no */

PROTOTYPE(isoaddrcmp, extern int, (struct iso_net_addr *addr1, struct iso_net_addr *addr2)); 
PROTOTYPE(isoaddr_netiso_cmp, extern int, (struct iso_addr *addr1, struct iso_net_addr *addr2)); 
	
/* compare 2 iso addresses - return <,  =, > */
	
PROTOTYPE(compare_iso_addr, extern int, (struct iso_net_addr *addr1, struct iso_net_addr *addr2)); 

/* compare iso prefixes */

PROTOTYPE(isopfxcompare, extern int, (struct iso_prefix *pfx1, struct iso_prefix *pfx2)); /* Compare two prefixes */

/* idrp.c routines */
/* %% general utility routines */

PROTOTYPE (drop_packet, int , ());
PROTOTYPE(idrp_trace, extern void, (idrpPeer *peer, \
		char *comment, \
		int  send_flag, \
		idrpPdu *pdu, \
		int	length));

PROTOTYPE(log_nm_event, extern void, (idrpPeer *peer, \
		int type, \
		int value));

PROTOTYPE(free_buffer,extern void, (idrpPeer *peer, \
		idrpBuffer *buf));

/* %% timer routines */

PROTOTYPE(start_rexmit_timer, extern void, (idrpPeer *peer));
PROTOTYPE(kill_rexmit_timer,extern void, (idrpPeer *peer));
PROTOTYPE(start_keepalive_timer, extern void, (idrpPeer *peer));
PROTOTYPE(start_opensent_timer, extern void, (idrpPeer *peer));

#ifdef ECHOKLUDGE
PROTOTYPE(start_echo_timer, extern void, (idrpPeer *peer));
#endif


/* %% pdu sending routines  */

/* post idrp pdu, and enqueued pdus */

PROTOTYPE(idrp_post, void, (idrpPeer *peer, idrpBuffer *buf)); 
PROTOTYPE(post_enqueued_pdus, void, (idrpPeer *peer));

/* send a pdu - format stuff in to pdu  */

PROTOTYPE(idrp_send_pdu, void, (idrpPeer *peer, idrpPdu *pdu,idrp_proto_header *pdu_proto, \
				u_int type, u_int length,u_int proto_len));

PROTOTYPE(send_open, void, (idrpPeer *peer));
PROTOTYPE(send_update_pdu, void, (idrpPeer *peer, idrp_send_list *p_send_list, u_int unreachable));

PROTOTYPE(idrp_send_error, extern void, (idrpPeer *peer, u_int code));  

PROTOTYPE(send_cease, extern void, (idrpPeer *peer)); /* send cease pdu */
PROTOTYPE(send_keepalive, extern void, (idrpPeer *peer)); /* send keepalive */

#ifdef ECHOKLUDGE
PROTOTYPE(send_echo, extern void,(idrpPeer *peer)); /* send echo pdu */
PROTOTYPE(send_echo_reply,extern void, (idrpPeer *peer)); /* send echo reply pdu */
#endif
PROTOTYPE(send_refresh_pdu,void, (idrpPeer *peer, int type));	/* send Rib Refresh */


/* idrp_rt_peer.c routines - actions for a peer 
 */ 
   
PROTOTYPE(flush_xmit_queue, extern void,(idrpPeer *peer)); /* flush transmit queue */
PROTOTYPE(path_down,extern void, (idrpPeer *peer));
PROTOTYPE(free_acked_pdu, void,(idrpPeer *peer)); 
PROTOTYPE(free_rexmit_queue,extern void,(idrpPeer *peer)); /* flush re-xmit queue */
PROTOTYPE(begin_close,extern void,(idrpPeer *peer)); /* begin close sequence */
PROTOTYPE(ack_pdu,void,(idrpPeer *peer,u_int ack, u_int credit,u_int type));		/* process acks */
PROTOTYPE(path_up,void,(idrpPeer *peer));		/* path up BIS-BIS established */
PROTOTYPE(idrp_refresh_all_peers, void, (idrp_ann_list *p_ann_list));	/* send Rib Refresh to all connected peers */
PROTOTYPE(idrp_req_refresh, void, (idrpPeer *peer));		/* request a Rib Refresh from a Peer */ 

/* parsing routines for pdus */

PROTOTYPE(idrp_parse_open,int,(idrpPeer *peer, idrpPdu *pdu, int length));

/* put valid_auth_data here for now - move either to idrp.c or idrp_auth.c */

PROTOTYPE(valid_auth_data,int,(u_char *p_auth_data,idrpPeer *peer));

PROTOTYPE(process_open,void,(idrpPeer *peer, idrpPdu *pdu, int length));

PROTOTYPE(parse_update,int,(idrpPeer *peer, idrpPdu *pdu, int length,parse_results *res));
PROTOTYPE(parse_update_cleanup,void,(parse_results *res,idrpRoute_entry *p_ent,idrpPeer *peer,int reuse_att));

PROTOTYPE(process_update,void,(idrpPeer *peer, idrpPdu *pdu, int length));

PROTOTYPE(process_keepalive,void,(idrpPeer *peer,idrpPdu *pdu, int length));

PROTOTYPE(process_error,void,(idrpPeer *peer,idrpPdu *pdu,int length));

PROTOTYPE(process_cease,void,(idrpPeer *peer,idrpPdu *pdu,int length));

PROTOTYPE(parse_rib_refresh,u_int,(idrpPeer *peer, \
				   idrpPdu  *pdu, \
				   int	    length));

PROTOTYPE(process_rib_refresh,void,(idrpPeer *peer, \
				    idrpPdu  *pdu, \
				    int      length));

#ifdef ECHOKLUDGE
PROTOTYPE(process_echo,void,(idrpPeer *peer, idrpPdu  *pdu, int length));
PROTOTYPE(process_echo_reply,void,(idrpPeer *peer, idrpPdu  *pdu, int length));
#endif

PROTOTYPE(process_header_error,void,(idrpPeer *peer,idrpPdu *pdu,int length));

PROTOTYPE(process_rcv_seq,void,(idrpPeer *peer, idrpPdu *pdu));

PROTOTYPE(in_sequence,static int,(idrpPeer *peer, u_int seq));
PROTOTYPE(idrp_recv_pdu,static void,(idrpPeer *peer, \
				     idrpPdu *pdu, \
				     int length));

PROTOTYPE(idrp_recv,extern void,(task *tp));
					

/* %% idrp events */
/* task timer events - also idrp state machine events */

PROTOTYPE(idrp_start_event, extern void, (idrpPeer *peer)); 	/* start this peer */
PROTOTYPE(idrp_stop_event, extern void, (idrpPeer *peer));	/* stop this peer */
PROTOTYPE(idrp_event_closetimer, extern void, (timer * tip, \
		                   time_t interval));                   /* Connect timer fired */
PROTOTYPE(idrp_event_holdtimetimer, extern void, (timer * tip, \
			                   time_t interval));            /* Abort timer fired */

/* task timer events - not idrp state machine events */
 
PROTOTYPE(idrp_event_keepalivetimer, extern void, (timer * tip, \
			                    time_t interval));                  /* KeepAlive Timer */
PROTOTYPE(idrp_event_rexmittimer, extern void, (timer * tip, \
			                    time_t interval));                  /* Retransmission Timer */
PROTOTYPE(idrp_event_starttimer, extern void, (timer * tip, \
		                    time_t interval));                  /* Startup Timer */
PROTOTYPE(idrp_event_opensenttimer, extern void, (timer *tip, \
				    time_t interval));			/* open sent timer */
#ifdef ECHOKLUDGE
PROTOTYPE(idrp_event_echotimer, extern void, (timer * tip, \
                    time_t interval));                  /* Echo Timer %% */
#endif

/* %% pdu parsing routines */

PROTOTYPE(checksum_ok,int,(idrpPeer *peer, \
			   idrpPdu *pdu, \
			   int length));

PROTOTYPE(valid_header,int,(idrpPeer *peer, idrpPdu *pdu, int length));  /* validate header */
PROTOTYPE(idrp_rib_id,int,(int att_mask));

/* md4 checksum pdu */ 


/* %% validate attributes in attribute record */

PROTOTYPE(valid_attr,int,(idrpPeer *peer, \
			  idrp_attribute_record *att_rec));
PROTOTYPE(valid_usr_attr,int,(idrpPeer *peer, \
			  idrp_attribute_record *att_rec));

/* rd pathwy checks */
 
PROTOTYPE(valid_seg_type,int,(int type, idrpPeer *peer));  /* validate path segment type */
PROTOTYPE(valid_rd_path,int,(idrpPeer *peer, \
			     idrp_attribute_record *att_rec, \
			     u_char *p_rdpath, \
			     int    rdpath_length));  /* validate rd path by linking to cannon list */


/* compare attributes - 1 if match, 0 if don't */
 
PROTOTYPE(compare_atts,u_int,(idrp_attribute_record *p_att1, idrp_attribute_record *p_att2));
PROTOTYPE(find_attr_ann_list, idrp_ann_list *,(idrp_ann_list *p_ann_list, idrp_attribute_record *attr));
PROTOTYPE(find_attr_rec, idrp_attribute_record *,(idrpPeer *peer, idrp_attribute_record *attr,int local));
PROTOTYPE(idrp_dif_path, extern int,(idrp_attribute_record *p_att1,idrp_attribute_record *p_att2));
PROTOTYPE(idrp_free_attr_block, extern void,(idrpPeer *peer, idrp_attribute_record *attr));
PROTOTYPE(ATTS_REC_ZERO,void,(idrp_attribute_record *p_atts));
PROTOTYPE(idrp_free_routeid_chain_id,void,(idrp_attribute_record *p_att, int id, idrpPeer *Peer)); 
PROTOTYPE(idrp_find_routeid_chain,idrpRoute_entry *,(idrpRoute *p_rt,idrpRoute_entry **p_rte_last));
PROTOTYPE(idrp_create_routeid_chain_ent,idrpRoute_entry *,(idrpRoute *p_rt));
PROTOTYPE(idrp_free_nlri_att_rec, extern void, (idrpRoute *p_rt, int with));
PROTOTYPE(idrp_free_attr_rec,extern void,(idrpPeer *peer, idrp_attribute_record *attr,int id));
PROTOTYPE(idrp_path_dump,void,(idrp_canon_rdpath *p_rdl,idrpPeer *peer));
PROTOTYPE(relink_free_att,void,(idrp_attribute_record *p_atts));
PROTOTYPE(link_att_list,void,(idrp_attribute_record *p_atts));
PROTOTYPE(idrp_free_nlri_chain,void,(idrpRoute_entry *p_rte));


/* %% - idrp_attr_rec - routines to handle attribute record
 * -idrp_free_canon_rdpath - free the canonical rdi path list
 * -idrp_canon_my_rdi_path - make canonical path from rdi list
 * -idrp_my_rdi_rdpath - is this pathway only my rdi (returns true/false)
 * -idrp_dump_rdi_path - dump the canonical rdi path in pretty print
 */
 
PROTOTYPE(idrp_free_canon_rdpath, void, (idrp_canon_rdpath *p_can_list, idrpPeer *peer));
PROTOTYPE(idrp_canon_my_rdi_path, extern int ,(idrp_attribute_record *p_att, idrpPeer *peer));
PROTOTYPE(idrp_my_rdi_rdpath,int, (idrp_attribute_record *p_att));
PROTOTYPE(idrp_dump_rdi_path, void,(idrp_canon_rdpath *p_rdl,idrpPeer *peer));
PROTOTYPE(add_new_routeid_chain,int,(idrpRoute_entry *p_rte,idrp_attribute_record *p_att));
PROTOTYPE(add_nlri_routeid_chain,int,(idrpRoute_entry *p_rte,idrpRoute *p_idrp_rt));

PROTOTYPE(link_rd_canon, int,(idrp_canon_rdpath *p_rd_new,idrp_attribute_record *p_att));
PROTOTYPE(add_local_rd_to_path,int,(idrpPeer *peer, idrp_attribute_record *attr_rec));

/* protocol to family transfer routines -
 * can these be done in macros ?? 
*/

PROTOTYPE(proto_to_nlri_id, extern u_int,(char *type, int len, u_char *proto));
PROTOTYPE(family_to_nlri_id, u_int,(int	family));

/* %% rib refresh routines */


PROTOTYPE(rib_refresh_pdu_release,void,(idrpPeer *peer)); /* refresh aborted, release pdus */ 
							  /* %% check for parse results release */
PROTOTYPE(refresh_pdu_link,void,(idrpPeer *peer, parse_results *results, idrpPdu *pdu));

PROTOTYPE(rib_refresh_update_process,void,(idrpPeer *peer));  /* process all updates for Rib refresh */
PROTOTYPE(rib_refresh_remove_dup, void, (idrpPeer *peer, idrp_ann_list *p_w_list, idrp_ann_list *p_ann_list));
								/* remove duplicate from rib refresh lists */

PROTOTYPE(idrp_sm,extern int,(idrpPeer *peer, int event));

PROTOTYPE(idrp_send_dest_set,sockaddr_un *,(idrpPeer *peer));	/* set send destination sockaddr_un */
 
PROTOTYPE(my_bis_peer,int,(gw_entry *gw));	/* Is it my bis peer */ 


PROTOTYPE(md4, u_char *, (u_char *, u_int, u_char *));	/* md4 master routine */
							/* rest of code in md4.c */ 


/* idrp_init.c */

/* iso utilities */

PROTOTYPE(idrp_trace,extern void, (idrpPeer *peer, char *comment, int send_flag, idrpPdu *pdu, int length));
PROTOTYPE(isopfxcompare, extern int, (struct iso_prefix *pfx1, struct iso_prefix *pfx2)); /* Compare two prefixes */



/* %%% idrp_rt_util.c routines  */
/* %% hash routines      */

PROTOTYPE(idrp_rid_hash_get,extern idrp_rid_hash *,(idrpPeer *peer,u_int id));	/* get hash table entry */
PROTOTYPE(idrp_add_hash_entry,extern void, (idrpPeer *peer, u_int id, idrpRoute *p_route[])); /* add entry */ 
PROTOTYPE(idrp_free_hash_entry,extern void, (idrpPeer *peer,u_int id));  /* free hash table entry */
PROTOTYPE(idrp_release_hash_tbl,extern void, (idrpPeer *peer));	 /* free memory from peer hash table */

PROTOTYPE(idrp_init_hash_tbl,extern void, (idrpPeer *peer));	 /* re-initialize memory from peer hash table */

PROTOTYPE(idrp_reinit_hash_tbl,extern void, (idrpPeer *peer));	 /* reinit memory from peer hash table */

/* %% idrpRoute structure release routines */

PROTOTYPE(idrp_alloc_idrpRoute, extern idrpRoute *, (idrpPeer *peer, idrp_attribute_record * attr_rec, int family, \
							int route_id_in));
/* Allocate route block */

PROTOTYPE(idrp_free_idrpRoute, void, (idrpRoute *p_rt));

PROTOTYPE(idrp_free_idrpRoute_entry, extern idrpRoute_entry *, (int id, idrpPeer *peer, idrpRoute_entry *p_rt_list, int circular));

PROTOTYPE(free_idrpRoute_ent,  idrpRoute_entry *, (int id, idrpPeer *peer, idrpRoute_entry *p_rt_list));	
/* free entry of idrpRoute list */

PROTOTYPE(circular_free_idrpRoute, idrpRoute_entry *, (int id, idrpRoute_entry *p_rt_list));

PROTOTYPE(idrp_add_idrpRoute_entry, idrpRoute_entry *, (u_int id,idrpRoute_entry *p_rtl));
PROTOTYPE(idrp_free_outlist, void, (idrpRoute *p_idrp_rt)); 



/* %% dump routines for now/
 * 	- idrp_master_dump	- dump master task
 */
PROTOTYPE(idrp_master_dump,void,(task *tp, FILE *fd));	/* dump master task */
PROTOTYPE(idrp_peer_dump,void,(task *tp,  FILE *fd));	/* dump peer task */
PROTOTYPE(idrp_dump_peer_mib, extern void, (FILE *fd, idrpPeer *peer));
PROTOTYPE(idrp_dump_peer_local_mib, extern void, (FILE *fd));
PROTOTYPE(idrp_dump_peer_extra, extern void, (FILE *fd, idrpPeer *peer));
PROTOTYPE(idrp_dump_attr_list, extern void, (FILE *fd, idrpPeer *peer));
PROTOTYPE(idrp_dump_attr_entry, extern void, (FILE *fd, idrp_attribute_record *att));
PROTOTYPE(idrp_peer_rtbit_dump, extern void,(rt_head *, void_t data, char *));
PROTOTYPE(idrp_dump_local_rib,extern void,(FILE *fd, idrpPeer *peer));
PROTOTYPE(idrp_dump_adj_rib,extern void,(FILE *fd, idrpPeer *peer));
PROTOTYPE(idrp_dump_pref_list,extern void, (FILE *fd, idrpPeer *peer));
PROTOTYPE(idrp_dump_aggr_list,extern void, (FILE *fd, idrpPeer *peer));
PROTOTYPE(idrp_dump_dist_list,extern void, (FILE *fd, idrpPeer *peer));
PROTOTYPE(idrp_dump_route_mib,extern void, (FILE *fd, idrpRoute *idrp_rt));
PROTOTYPE(idrp_dump_internal_systems, extern void, (FILE *fd));
PROTOTYPE(idrp_dump_intra_is_systems, extern void, (FILE *fd));
PROTOTYPE(idrp_dump_snpa_list, extern void, (FILE *fd, snpa_list snpal));
PROTOTYPE(idrp_dump_snpa_entry, extern void, (FILE *fd, snpa_entry *snpa));
PROTOTYPE(idrp_get_byte_value, extern long, (char *data, int n));


/* idrp_rt_peer.c - routines 
 * %% AdjRib checks
 */ 

PROTOTYPE(consistency_check, extern void, (int rib_no));  /* consistency check one rib or all */
PROTOTYPE(validate_AdjRib, extern void, (int rib_no, idrpPeer *peer)); /* validate AdjRib */
PROTOTYPE(validate_allRibs, extern void,(int rib_no)); 		/* validate all AdjRib rib_no */

/* %% idrp_rt_policy routines
 * PREF - calculate preference for routine
 *  
 * %% Distribution policy routines for sending 
 * DIST - can this route be distributed to this peer
 * DIST_ATTR - modify or keep attribute record
 *
 * idrp_proto_pref -  given equal gated pref which protocol preferred 
 */
 
PROTOTYPE(idrp_proto_pref, extern int, (u_int proto_rcv, idrpPeer *peer));
PROTOTYPE(PREF,int,(rt_entry *p_rt, idrpRoute *p_idrp_rt));
PROTOTYPE(DIST,int, (idrpRoute *p_idrp_rt, idrpPeer *peer,int type));
PROTOTYPE(DIST_ATTR, idrp_attribute_record *, (idrp_attribute_record *p_att, idrpPeer *peer));
PROTOTYPE(AGGR, void, (idrpPeer *peer,  idrp_ann_list *p_ann_list));

/* %% Descision processes */
/* phase 1 */

PROTOTYPE(idrp_pref,extern void,(idrpRoute *p_idrp_rt));  /* idrp preference calculation */
PROTOTYPE(idrp_ext_info_pref, int,(rt_entry *p_rt));

/* phase 2 */

PROTOTYPE(idrp_to_gated_pref, extern int, (idrpRoute *p_rt));	/* transmuting idrp preference to gated */ 

PROTOTYPE(tie_break, int,(idrpRoute *p_idrp_rt1, idrpRoute *p_idrp_rt2));   /* break the tie */
PROTOTYPE(tie_break_iso, int,(idrpRoute *p_idrp_rt1, idrpRoute *p_idrp_rt2));	/* tie break iso routes */
PROTOTYPE(tie_break_ip, int,(idrpRoute *p_idrp_rt1, idrpRoute *p_idrp_rt2)); /* tie break ip routes */

/* iso specific - phase2 comparisons */

PROTOTYPE(iso_next_hop_compare, int, \
		 (idrp_attribute_record *p_att1, \
		  idrp_attribute_record *p_att2));  /* compare next hop */

PROTOTYPE(multi_exit_compare, int, \
		 (idrp_attribute_record *p_att1, \
		  idrp_attribute_record *p_att2));  /* multi_exit_disc */

/* %% phase3 - processing */


PROTOTYPE(idrp_flash, extern void, (task *tp, rt_list *rtl)); /* idrp flash update */
PROTOTYPE(phase3_process, extern void, (rt_list *rtl));		/* phase3_process */
PROTOTYPE(idrp_phase3_dump, int,  (idrpPeer *peer, rt_list *rtl[]));	/* send phase 3 routes full dump */

PROTOTYPE(phase3_newpolicy , idrp_ann_list *, (rt_list *rtl, idrp_ann_list *p_ann_list));
PROTOTYPE(phase3_status_change, void,  (rt_list *rtl,idrp_ann_list *));
PROTOTYPE(ph3_status_case1, void, (rt_head *rth,idrp_ann_list *p_ann_list)); /* new idrp route */
PROTOTYPE(ph3_status_case2, int, (rt_head *rth,idrp_ann_list *p_ann_list)); /* new ext_info route */
PROTOTYPE(ph3_status_case3, void, (rt_head *rth,idrp_ann_list *p_ann_list)); /* old active idrp, new active idrp */
PROTOTYPE(ph3_status_case4, int, (rt_head *rth,idrp_ann_list *p_ann_list)); /* old active ext_info -> new ext_info */
PROTOTYPE(ph3_status_case5, void, (rt_head *rth,idrp_ann_list *p_ann_list)); /* old ext_info to new idrp */
PROTOTYPE(ph3_status_case6, int, (rt_head *rth,idrp_ann_list *p_ann_list)); /* old idrp to new ext_info */
PROTOTYPE(ph3_status_case7, void, (rt_head *rth,idrp_ann_list *p_ann_list)); /* old idrp to delete route */
PROTOTYPE(ph3_status_case8, void, (rt_head *rth,idrp_ann_list *p_ann_list)); /* old ext_info to new route */

PROTOTYPE(idrp_send_phase3_routes, void, (idrp_ann_list *p_idrp_ann));
PROTOTYPE(idrp_del_ph3_routes, void, (idrp_ann_list *p_idrp_ann));

/* phase 1 processing */
/* processing update routes routines */

PROTOTYPE(idrp_route_mod, extern int, (idrpRoute *p_idrp_rt, idrpPeer *peer, idrpRoute_link rt_chg));
PROTOTYPE(idrp_process_pdu_routes, idrp_ann_list *, (idrpPeer *peer, parse_results *res,int type)); /* process update pdu */

/* phase one processing */

PROTOTYPE(idrp_phase1_process,void,(idrp_ann_list *ann_list));  /* phase1 handing */
PROTOTYPE(phase1_internal,int,(idrp_ann_list *ann_list));	/* phase1 handling internal routes */
PROTOTYPE(phase1_external,int,(idrp_ann_list *ann_list));	/* phase1 handling external routes */
PROTOTYPE(ph1_with_int_route,void,(idrpRoute *p_idrp_rt));	/* phase1 withdraw internal route */
PROTOTYPE(ph1_with_rep_int_route,void,(idrpRoute *p_idrp_rt));  /* phase1 withdraw/replaced internal route */
PROTOTYPE(ph1_add_int_route,void,(idrpRoute *p_idrp_rt));	/* phase1 add internal route */

PROTOTYPE(ph1_with_ext_route,void,(idrpRoute *p_idrp_rt,idrp_ann_list *p_ext_ann_list));
					/* Phase 1 withdrawl of external route */

PROTOTYPE(ph1_with_rep_ext_route,void,(idrpRoute *p_idrp_rt,idrp_ann_list *p_ext_ann_list));
PROTOTYPE(ph1_add_ext_route,void,(idrpRoute *p_idrp_rt,idrp_ann_list *p_ext_ann_list));

PROTOTYPE(idrp_replace_ext, void, (idrpRoute *p_idrp_rt,rt_entry *p_rt, idrp_ann_list *p_ext_ann_list, \
					idrpRoute *p_best_rt,int pref));
  
PROTOTYPE(send_best_ext,void,(idrp_ann_list *p_ext_ann_list));
PROTOTYPE(del_routes_best_ext,void,(idrp_ann_list *p_ext_ann_list));
PROTOTYPE(remove_ann_dup,idrpRoute *,(idrpRoute *p_withdraw,idrpRoute *p_announce));
								/* remove announce duplicates */
								/* for the withdraw 10, replace 8 case */

PROTOTYPE(verify_extinfo_idrp_rt,int,(idrpRoute *p_idrp_rt,rt_entry *p_rt));
PROTOTYPE(ph3_create_ext_route,idrpRoute *,(rt_entry *p_rt, rt_head *rth, int pref));

/* utility routines shared by both phase 1 and phase 3 processing */

/* %% routines for general sending */

PROTOTYPE(link_ann_list, int , (idrpRoute *p_idrp_rt, idrp_ann_list *p_ann_list, int type)); /* link to announce list */
PROTOTYPE(create_ann_list_entry, idrp_ann_list *, (idrpPeer *peer, idrp_attribute_record *p_att));
PROTOTYPE(link_ann_list_ann_list,void, (idrp_ann_list *p_ann_list, idrpPeer *peer, \
					idrp_ann_list *p_add_ann_list));

PROTOTYPE(free_ann_list, void , (idrp_ann_list *p_ann_list));
PROTOTYPE(find_ann_list_entry, idrp_ann_list *, (idrpPeer *peer, idrp_attribute_record *p_att, \
				idrp_ann_list *p_ann_list)); 
PROTOTYPE(link_send_list,int , (idrp_send_list *p_send_list,idrp_send_list *p_send_new));
PROTOTYPE(send_with_attr, idrp_send_list *, (idrpRoute *p_idrp_rt,idrp_attribute_record *p_att_dist, \
				 idrp_send_list *p_send_list,idrpPeer *peer));
PROTOTYPE(send_update_reset_send_list, void, (idrpPeer *peer, idrp_send_list *p_send_list, idrp_send_list *p_tmp_send));

PROTOTYPE(send_nlri_attr, idrp_send_list *, (idrpRoute *p_idrp_rt,idrp_attribute_record *p_att_dist,  \
				 idrp_send_list *p_send_list,idrpPeer *peer)); 

PROTOTYPE(create_send_list, idrp_send_list *, (idrp_attribute_record *p_att_dist, idrpPeer *peer));
PROTOTYPE(free_send_list, void, (idrp_send_list *p_send_list, idrpPeer *peer)); 
PROTOTYPE(flush_att_send_list,void,(idrpPeer *peer,idrp_attribute_record *p_att_dist, \
				idrp_send_list *p_send_list));
PROTOTYPE(idrp_send_list_room,int, (idrp_send_list *p_list, idrp_send_list *p_new));
PROTOTYPE(idrp_del_sent_routes,void,(idrp_ann_list *list)); /* delete routes given type */  
PROTOTYPE(best_ext_route,int,(idrpRoute	*p_idrp_rt,idrpRoute *p_best_ext)); /* give best ext of pair */ 

/* - mixture of phase1 and phase3 utilites and gated utilities
 *   ( a true no-man's land of code - filled with dragons)
 * 
 *  idrp_add_route_locate - do gated checking for routes to be added to gated tables,  
 *			  - look up best ext, gateway and preference matches
 *  idrp_with_route_locate - do gated checking for routes to be withdrawn from gated tables
 */


PROTOTYPE(idrp_add_route_locate,extern int, (sockaddr_un *p_dest, \
						 idrpRoute *p_idrp_rt, \
						 idrp_rt_chain_walk *p_gate, \
						 idrp_rt_chain_walk *p_pref, \
						 idrpRoute **p_best_ext));

PROTOTYPE(idrp_with_route_locate,extern int, (sockaddr_un *dest, \
						 idrpRoute *p_idrp_rt, \
						 idrp_rt_chain_walk *p_gate, \
						 idrp_rt_chain_walk *p_pref, \
						 idrpRoute **p_best_ext));

/* group processing */

PROTOTYPE(idrp_rt_send_init, void, (idrpPeer *peer));	/* send all routes initially */
PROTOTYPE(idrp_peer_down, extern void, (idrpPeer *peer)); /* Process a down peer */
PROTOTYPE(idrp_peer_route_pull, extern int, (rt_list *rtl, idrp_ann_list *p_ann_list)); /* process down peer */

PROTOTYPE(idrp_rt_reinit,extern void,(idrpPeer *peer));	/* re-init peer after policy change */ 
PROTOTYPE(idrp_send_refresh_all_routes,extern void,(idrpPeer *peer));	/* send all routes for refresh */

/* generic processing routines */

/* iso sockaddr structure set from iso net address */

PROTOTYPE(idrp_iso_sockaddr_mask,extern void,(sockaddr_un *mask, int cnt)); /* create bit mask for iso address */
PROTOTYPE(idrp_set_iso_sockaddr,extern void,(struct sockaddr_iso *spt, int iso_len,u_char *iso_net_addr));

/* %%  idrp_rt_minadv.c routines -
 *     routines dealing with the Minimum Route Advertisement timer
 */ 

PROTOTYPE(idrp_minadv_annlist,void,(idrp_ann_list *p_ann));   /* set minadv for all of announce list */
PROTOTYPE(idrp_add_min_route_adv,void,(idrpRoute *p_idrp_rt));  	/* add route to min_rte adv list */
PROTOTYPE(advmin_interval_calc,int,(struct timeval *p_tp, idrpPeer *peer,int type));
	/* calculate interval for rte list */ 
PROTOTYPE(idrp_set_min_route_timer,int,());				/* set min_route_adv timer */
PROTOTYPE(idrp_process_minadv,extern void,(timer *tip, time_t interval));  /* all routes process loop */
PROTOTYPE(idrp_min_adv_rt, int, (idrpRoute *p_idrp_rt,idrp_ann_list *p_ann_list));	/* one route process handling */ 

/* process minimum advertisement for this RD */

PROTOTYPE(idrp_add_rt_min_advRD,void,(idrpRoute *p_idrp_rt));
PROTOTYPE(idrp_set_min_advRD_time,int,());
PROTOTYPE(idrp_process_minadvRD,extern void,(timer *tip, time_t interval));  /* all routes process loop */
PROTOTYPE(idrp_min_advRD_rt, int, (idrpRoute *p_idrp_rt, idrp_ann_list *p_ann_list)); /* one route process handling */ 


/*  %% gated related routines 
 * 
 *  idrp_rt_change - has this idrp route really changed 
 *  free_rt_chain_walk - free idrp_rt_chain_walk structure - build in idrp_add_route_locate
 */  

PROTOTYPE(idrp_rt_change,extern int, (rt_entry *p_rt, idrpRoute *p_idrp_rt2));  /* idrp route change tested */
PROTOTYPE(free_rt_chain_walk,void,(idrp_rt_chain_walk *p_ch,idrpPeer *peer));

/* %% local RD routines */

PROTOTYPE(idrp_add_local_rt_gated,extern int,(idrpRoute *p_idrp_rt));
PROTOTYPE(idrp_local_rt, extern int, (u_char *p_nlri_str, int byte_cnt, int bit_cnt, \
			 int family, u_char *p_gw_str, int gw_byte_cnt, idrpRoute_options  *p_opts)); /* add a local route to rib */

PROTOTYPE(idrp_find_local_dest, idrpRoute *, (u_char *p_nlri_str, int byte_cnt, int bit_cnt, \
			 int family, u_char *p_gw_str, int gw_byte_cnt, idrpRoute_options  *p_opts)); /* add a local route to rib */
PROTOTYPE(local_gw_match, int, (idrpRoute *p_irt, u_char *p_gw_str, int byte_cnt));
PROTOTYPE(local_options_match, int, (idrpRoute *p_irt, idrpRoute_options *p_opts));

PROTOTYPE(create_local_attr_record, idrp_attribute_record *,(sockaddr_un *p_next_hop, \
				idrpRoute_options *p_opts)); /* create local */
PROTOTYPE(find_local_att_rec, idrp_attribute_record *, (sockaddr_un *p_next_gw,idrpRoute_options *p_opts));
PROTOTYPE(create_task_att_rec, idrp_attribute_record *, (idrpPeer *peer));
PROTOTYPE(find_local_ext_attr, idrp_attribute_record *, (sockaddr_un *p_next_gw,idrpRoute_options *p_opts));
PROTOTYPE(create_local_ext_info_att_rec, idrp_attribute_record * ,(sockaddr_un *p_next_hop, idrpRoute_options *p_opts)); /* create local ext_info record  */
PROTOTYPE(idrp_fill_local_man_attrib, void, (idrp_attribute_record *p_atts));  /* fill local attribute array */ 
PROTOTYPE(idrp_fill_local_opt_attrib, void, (idrp_attribute_record *p_atts,idrpRoute_options *p_opts));  /* fill local attribute array */ 

PROTOTYPE(fill_next_hop, int,(idrp_attribute_record *p_atts, idrpRoute_options *p_opts));
PROTOTYPE(fill_DIST_LIST,int, (idrp_attribute_record *p_atts, idrp_canon_rdpath *p_rd, int type));
PROTOTYPE(fill_MULTI_EXIT,int, (idrp_attribute_record *p_atts, idrpRoute_options *p_opts)); 
PROTOTYPE(link_local_attr_list,int, (idrp_attribute_record *p_atts));	/* link local attribute */
PROTOTYPE(create_local_rd_path,idrp_canon_rdpath *, (int len, u_char *ptr));	/* create a canonical rd path */
PROTOTYPE(idrp_create_local_gw, sockaddr_un *,(u_char *p_gw_str, int gw_byte_cnt, int family)); /* create gw struture */ 

PROTOTYPE(idrp_local_route_reset,void,());	/* flag all local routes for reparsing */
PROTOTYPE(idrp_local_route_clean,int,());	/* clean up all local routes and attribute records */
PROTOTYPE(idrp_free_local_attrib_task,void ,(idrp_attribute_record *p_att)); 
PROTOTYPE(idrp_free_local_att, void ,(idrp_attribute_record *p_att)); 
PROTOTYPE(idrp_flag_local_att,void, (idrp_attribute_record *p_att));
PROTOTYPE(idrp_clean_local_att, int, (idrp_attribute_record *p_att));
PROTOTYPE(free_local_attr_record, void, (idrp_attribute_record *p_att));
PROTOTYPE(clean_local_route_chain, int, (idrpRoute_entry *p_rte));

PROTOTYPE(idrp_free_local_idrpRoute_ent,void,(idrpRoute_entry *p_rte));	/* free local route entry */

PROTOTYPE(idrp_free_local_rt,void,(idrpRoute *p_idrp_rt));	/* free one local route on chain */
PROTOTYPE(idrp_free_local_route_chain,void, (idrpRoute_entry *p_idrp_rtl));	/* release nrlis associated */
										/* with local routes before task */
										/* initialization */  

PROTOTYPE(idrp_local_mem_fit,u_char *, (int len));	/* fit memory in length */
PROTOTYPE(idrp_mem_fit_free,void, (idrp_attribute_entry *));	/* free memory from attribute entry */
PROTOTYPE(rdi_canon_length, int,(idrp_canon_rdpath *p_rdi, idrp_rdi_mem *p_req)); /* memory required DIST_LIST */


/* %% ext_info routines for the attribute record (gated related) */

PROTOTYPE(find_ext_info_attr_rec, extern idrp_attribute_record *, (rt_head *p_rth, rt_entry *p_rt));
PROTOTYPE(create_ext_info_attr_rec, extern idrp_attribute_record *, (rt_head *p_rth, rt_entry *p_rt));


#ifdef 	IDRP_ASPATH
PROTOTYPE(idrp_aspath_to_canon_rdpath, int, (as_path *p_as,idrp_attribute_record *p_att, \
			 idrpPeer *peer, int type, u_char *p_bgp));
PROTOTYPE(idrp_aspath_ext_info, int, (idrp_canon_rdpath *p_can_list, idrp_attribute_record *p_att, \
			 idrpPeer *peer, int size));
PROTOTYPE(idrp_aspath_bgp_xfr, int, (idrp_canon_rdpath *p_can_list, \
	idrp_attribute_record *p_att, idrpPeer *peer, int size, u_char *p_bgp));

PROTOTYPE(idrp_as_rdi_map,struct iso_net_addr *, (int as));
PROTOTYPE(idrp_bgp_within_rd,int, (idrpRoute *p_idrp_rt));
#endif

PROTOTYPE(idrp_ext_info_peer,idrpPeer *,(rt_entry *p_rt, idrpRoute *p_idrp_rt));
/* idrp_rt.c routines */
PROTOTYPE(sock_supported, int, (idrpPeer *peer, sockaddr_un *addr));
PROTOTYPE(idrp_sock_dup, sockaddr_un *, (idrpPeer *peer, sockaddr_un *addr));

/* %% gated interaction routines
 * - mediate_pref_match - modify the gated preference match as phase2 desires
 * - idrp_add_rt_to_gated - add a route to gated tables
 * - idrp_repl_gated_rt - replace an idrpRoute structure in gated table
 *    			- call the rt_change routine to handle changes if needed 
 * - idrp_mod_rt_gated - modify gated routing table 
 */

PROTOTYPE(mediate_pref_match, int , (idrp_rt_chain_walk *p_pref, idrpRoute *p_irt));
PROTOTYPE(idrp_add_rt_to_gated, extern int, (idrpRoute *p_idrp_rt, sockaddr_un *p_dest));   /* add a local route to rib */
PROTOTYPE(idrp_repl_gated_rt, idrpRoute *, (idrpRoute *p_idrp_rt, rt_entry *p_rt));  /* replace rtd portion of gated rt */
PROTOTYPE(idrp_mod_rt_gated, int, (idrpRoute *p_idrp_rt,int pref)); 
PROTOTYPE(idrp_del_rt_gated, int, (idrpRoute *p_idrp_rt));
PROTOTYPE(idrp_repl_rt_gated, idrpRoute *, (idrpRoute *p_idrp_rt, rt_entry *p_rt));
PROTOTYPE(idrp_find_ext_info_gated_rt, idrpRoute *, (rt_head *rth, rt_entry *p_rt));

#endif /* not define _idrp_prototypes_h */

/* %jgs% -- junk not in any rational order */
PROTOTYPE(insert_in_pref_order, void, (idrpRoute *p_rt, idrpRoute *p_best));
PROTOTYPE(find_next_best, idrpRoute *, (idrpRoute *p_idrp_rt));
PROTOTYPE(idrp_ann_list_empty, int, (idrp_ann_list *p_ann_list));
PROTOTYPE(find_nlri_ann_list, int, (idrpRoute *p_idrp_rt, idrp_ann_list *p_ann_list));
