#include "include.h"
#include "iso.h"
#include "inet.h"
#include "idrp.h"

/* idrp_rt_phase_util.c
 * - utilities for idrp phases 
 * 
 * %% utility routines used by phase 1 and phase 3
 *
 * link_ann_list 	- link route to announce list
 * create_ann_list_entry - create an announce list entry
 * find_ann_list_entry	- find an announce list entry
 * link_ann_list_ann_list  - link announce list to existing announce list 
 *				
 * free_ann_list - 	free an announce lsit
 *  
 * link_send_list	- link new send_list to old send list 
 * send_with_attr	- send withdrawn from announce list  
 * send_update_reset_send - send update pdu and reset send list
 * 
 * send_nlri_attr	- send nlri from from announce list 
 * create_send_list	- create a send list from task memory and
 *		          fill in with attributes
 * free_send_list	- free the send linst  
 *
 * flush_att_send_list  - flush the send list at end of
 *			  send list entry for an attribute record
 * idrp_send_list_room  - is there room in the UPDATE PDU for
 *			  this peer to include the additions
 *			  in the send list additions 
 *  
 * idrp_del_sent_routes  - delete routes after sending withdrawls
 * idrp_pref_compare(p_idrp_rt,p_best_ext) - preference comparison -
 *	returns TRUE if first argument is better than the second. 
 * %% parsing update help routines
 *   
 * idrp_add_route_locate
 *	- locate information about route addition  
 *
 * idrp_with_route_locate 
 *	- locate information about route withdrawl
 *  
 * idrp_rt_change(p_idrp_rt1,p_idrp_rt2)
 *	- check that these two idrpRoutes are different
 * 
 * free_rt_chain_walk - free idrp_rt_chain_walk structure
 *			create in idrp_add_route_locate or 
 *			idrp_with_route_locate
 *
 * find_next_best(p_idrp_rt) - find the next-best external route
 *	to the same NLRI as the external route p_idrp_rt
 *	points to.  Return a pointer to it.
 *  
 * %% - dumping of tables for debugging routines
 * 
 * idrp_peer_tsi_dump  
 * dump_local_rib
 * find_nlri_ann_list - find nlri in the announce list
 */

		
/* %% routine for general sending
 * 
 * link_ann_list  - link route to ann list for type withdraw or announce  
 * create_ann_list_entry - create an announce list entry
 * find_ann_list_entry - find an announce list entry
 * link_send_list - link route to send list 
 * send_with_attr - build send list for withdrawls 
 * send_nlri_attr - build send list for nlri 
 */

int
link_ann_list(p_idrp_rt,p_ann_list,type)
idrpRoute	*p_idrp_rt;
idrp_ann_list	*p_ann_list;
int		type;
{
idrp_ann_list	*p_atl;
int		nlri_id;
int	found = 0;
int	i;


	nlri_id = family_to_nlri_id(p_idrp_rt->family);

	/* is the announce list empty
	 * if so skip search
	 */

	if (!p_ann_list)
		{
		trace (TR_IDRP,LOG_ERR,"IDRP Fatal:link_ann_list called with no announce list peer %s",p_idrp_rt->peer->name);
		task_quit(0);
		/* note - should we change to allow to create an entry, we
		 * need to go back through all the code and pull this out.
		 * - I think this is a better way, but it is too
	 	 * - baked into the code to fix right now.
		 * --- skh (3/28/93)
		 */
		}
	else
		{
		p_atl = find_ann_list_entry(p_idrp_rt->peer,p_idrp_rt->p_attr,p_ann_list);
		}

	/* entry for attribute list has been found or created.
	 * link the new idrp_route to it with the P_ANN_NLRI
	 * link in the idrp route
	 */
 
	switch (type)
		{
		case NLRI_LINKED_P_NEXT:
			trace (TR_IDRP,0," link_ann_list - only list through ann or withdraw route %s tried to link through p_next",iso_ptoa(&p_idrp_rt->nlri)); 
			return(FALSE);
			break;			

		case NLRI_LINKED_P_ANN:
			if (p_atl->ann_nlri[nlri_id].tail)
				{
				p_atl->ann_nlri[nlri_id].tail->p_ann_nlri = p_idrp_rt;
				p_atl->ann_nlri[nlri_id].tail = p_idrp_rt;
				}
			else
				{
				p_atl->ann_nlri[nlri_id].head = p_atl->ann_nlri[nlri_id].tail = p_idrp_rt;
				}
			p_idrp_rt->p_ann_nlri = (idrpRoute *) 0; 
			break;

		case NLRI_LINKED_P_WITH:
			if (p_atl->with_nlri[nlri_id].tail)
				{
				p_atl->with_nlri[nlri_id].tail->p_with = p_idrp_rt;
				p_atl->with_nlri[nlri_id].tail = p_idrp_rt;
				}
			else
				{
				p_atl->with_nlri[nlri_id].tail = p_atl->with_nlri[nlri_id].head = p_idrp_rt;
				}
			p_idrp_rt->p_with  = (idrpRoute *) 0;	

			break;
		}
	return(TRUE);
}

idrp_ann_list *
create_ann_list_entry(peer,p_att)
idrpPeer *peer;
idrp_attribute_record *p_att;
{
idrp_ann_list 	*p_anl;

	p_anl = (idrp_ann_list *) task_mem_malloc(peer->task,sizeof(idrp_ann_list));
	/* zeroing the structure will
	 * now allow default rib to be set and all pointers zeroed 
	 */

	bzero(p_anl,sizeof(idrp_ann_list));
	p_anl->peer = peer;
	p_anl->p_attr = p_att;
	p_anl->rib_id = p_att->rib_id;
	return(p_anl);
}

idrp_ann_list *
find_ann_list_entry(peer,p_att,p_ann_list)
idrpPeer	*peer;
idrp_attribute_record	*p_att;
idrp_ann_list		*p_ann_list;
{
idrp_ann_list	*p_atl;
idrp_ann_list	*p_last_atl;

	/* search for announce list entry  
	 * that matches
	 */

	if (!p_ann_list->p_attr)
		{
		/* list entry is there but empty - return this
		 * one after  initialization
		 */

		p_ann_list->peer = peer;
		p_ann_list->p_attr = p_att;
		p_ann_list->rib_id = p_att->rib_id;
		return (p_ann_list);
		}
		

	ANN_LIST(p_atl,p_ann_list) 
		{
		if (p_atl->p_attr == p_att)
			{
			/* attribute record match */
			return(p_atl);
			}
		p_last_atl = p_atl;
		} ANN_LIST_END 

	p_atl = create_ann_list_entry(peer,p_att);
	p_last_atl->p_next = p_atl;
	return(p_atl);
}

void
link_ann_list_ann_list(p_ann_list,peer,p_add_ann_list)
idrp_ann_list	*p_ann_list;
idrpPeer	*peer;
idrp_ann_list	*p_add_ann_list;
{
idrp_ann_list	*p_anl;
int		i;


	p_anl = find_ann_list_entry(peer,p_add_ann_list->p_attr,p_ann_list);
	NLRI_FAMILY_LOOP(i)
		{
		if (p_anl->ann_nlri[i].head)
			{
			p_anl->ann_nlri[i].tail->p_next_nlri = p_add_ann_list->ann_nlri[i].head;
			p_anl->ann_nlri[i].tail = p_add_ann_list->ann_nlri[i].tail;
			}
		else
			{
			p_anl->ann_nlri[i].head = p_add_ann_list->ann_nlri[i].head;
			p_anl->ann_nlri[i].tail = p_add_ann_list->ann_nlri[i].tail;
			}
		if (p_anl->with_nlri[i].head)
			{
			p_anl->with_nlri[i].tail->p_with = p_add_ann_list->ann_nlri[i].head;
			p_anl->with_nlri[i].tail = p_add_ann_list->ann_nlri[i].tail;
			}
		else
			{
			p_anl->with_nlri[i].head = p_add_ann_list->with_nlri[i].head;
			p_anl->with_nlri[i].tail = p_add_ann_list->with_nlri[i].tail;
			}

	      } NLRI_FAMILY_LOOP_END;

}

void
free_ann_list(p_ann_list)
idrp_ann_list	*p_ann_list;
{
idrp_ann_list *p_atl,*p_atl2;

	p_atl = p_ann_list->p_next;
	while (p_atl)
		{
		p_atl2 = p_atl;
		p_atl = p_atl->p_next;
		task_mem_free(p_atl2->peer->task,p_atl2);
		}
}

/* link_send_list - link a new send list on original
 * 
 */

int
link_send_list(p_send1,p_send_new)
idrp_send_list	*p_send1;
idrp_send_list	*p_send_new;
{
idrp_send_list	*p_send;
idrp_send_list	*p_send_last;

	p_send_last = p_send1;
	p_send = p_send1->p_next;

	/* loop to find the end of the send list */

	while(p_send)
		{
		p_send_last  = p_send;
		p_send = p_send->p_next;
		} 

	p_send_last->p_next = p_send_new;
	p_send_new->p_next = (idrp_send_list *) 0;
}


	
/* 
 * send_with_attr - 
 */
idrp_send_list *
send_with_attr (p_idrp_rt,p_att_dist,p_send_list,peer)
idrpRoute		*p_idrp_rt;	/* idrpRoute with NLRI to be announced */ 
idrp_attribute_record	*p_att_dist;	/* attribute record distributed to this peer */

idrp_send_list		*p_send_list;	/* send list */
idrpPeer		*peer;		/* peer to send withdraw sequence to */
{
#ifdef ADN_PATCH
	int	route_id_adn;
#else
int		route_id;
#endif /* ADN_PATCH */

idrp_send_list	tmp_send;
Withdrawl	withdraw;
idrpRoute	*p_irt;
idrpRoute	*p_irt_first;
idrpRoute	*p_irt_last;
int		pdu_bytes;
int		nlri_cnt = 0;
int		nlri_ann = 0;
int		nlri_early = 0;
int		i;
int		nlri_id;

	bzero(&tmp_send,sizeof(idrp_send_list));
	bzero(&withdraw,sizeof(Withdrawl));	
	tmp_send.withdraw = &withdraw;
	
	if (p_idrp_rt->status & IDRP_STATUS_WITH_EARLY_PROC)
		{	
		return(p_send_list);
		}


	if (!p_idrp_rt->route_out[peer->id].p_next)
		{
		/* no route announced out to this peer */
		
		return(p_send_list);
		}


	/* this route was announced out so add the withdraw to
	 * to our outbound list
	 *
	 * build temporary send list for this node
	 * see if this withdraw list can be added to the 
	 * send list withdraw sequence 
 	 *   
	 *  
	 * 1) save this route_id in  
	 *
	 * 2) walk the outbound list chain until we reach this
	 *    this address again,
	 * 	2.1)  build send list of those NLRI which will still need to be added   
	 * 	2.2) after list is built, check list against the withdraw processing 
	 *	2.3) flag any routes found for this route_id with early processing
	 *	 	so they will be ignored in the loop
	 *		Inefficent, but stable. 
	 *
	 *     2.4) Force out the withdraw pdus with one sequence initially
	 *
	 * later - calculate pdus and see if can stuff max in 
	 * 
	 * (later) 
	 * 3) calculate the number of bytes that will need
	 *    to be used to add the nlri sequence to 
	 *
	 * 4) see if we can stuff this addition to the
	 *	the UPDATE pdu
 	 *
	 * 5) if can stuff in UPDATE PDU, then add withdraw ID
	 *	and nlri to send list and exit   
	 * 
	 * 6) if we cannot stuff this withdraw sequence
	 * 	then give this pdu off to the send_update_pdu
	 *	routine to send the pdu
	 */ 

#ifdef ADN_PATCH
        route_id_adn = p_idrp_rt->route_out[peer->id].route_id;
#else
	route_id = p_idrp_rt->route_out[peer->id].route_id;
#endif /*  ADN_PATCH */
	p_irt_first = p_idrp_rt; 
	p_irt_last = p_idrp_rt;
	p_irt = p_idrp_rt->route_out[peer->id].p_next;

	nlri_cnt++;
	/* set-up the count for the first nlri in withdraw list */
 
	tmp_send.pdu_bytes = PFX_BYTE_LENGTH(p_idrp_rt->nlri);
	while (p_irt != p_irt_first)
		{
		nlri_cnt++;
		/* go through the list until you get to
	 	* the original node 
	 	* - only link out nlris without withdraw set  
		*/

		if ((p_irt->status & IDRP_STATUS_WITH) == 0)
			{
			nlri_ann++;	
			nlri_id = family_to_nlri_id(p_irt->family);
			if (!tmp_send.ann_nlri[nlri_id].head)
				{
				/* set-up the linked list head */

				tmp_send.ann_nlri[nlri_id].head = p_irt;
				tmp_send.ann_nlri[nlri_id].tail = p_irt;
				}
			else
				{
				/* list already set-up - so just add to tail of list */

				tmp_send.ann_nlri[nlri_id].tail->p_ann_nlri = p_irt;
				}
			p_irt->p_ann_nlri = (idrpRoute *) 0;  
			}
		else
			{
			if (p_idrp_rt->route_out[peer->id].route_id ==  p_irt->route_out[peer->id].route_id)
				{
				/* early processing of another route on the chain */
				
				IDRP_STATUS_BIT_SET(p_irt,IDRP_STATUS_WITH_EARLY_PROC);
				nlri_early++;
				}
			}
		
		/* 1) set this idrpRoute to the last idrproute
	 	 *   pointer so can check for circular list
		 * 2) set to next link in chain
		 * 3) make sure route_id is equal to current
		 * 4) add pdu bytes to end
		 */ 
	 
		p_irt_last = p_irt;
		p_irt = p_irt->route_out[peer->id].p_next;

		/* check to make sure route_id is equal or links screwed up
		 */
	
#ifdef ADN_PATCH
		if (route_id_adn != p_irt->route_out[peer->id].route_id)
#else
		if (route_id != p_irt->route_out[peer->id].route_id)
#endif /* ADN_PATCH */
			{
#ifdef ADN_PATCH
			trace(TR_IDRP,LOG_ERR,"outbound route linked list screwed up peer %s route_id %d new id %d", 
				peer->name,
				route_id_adn,   /* !!! */
				p_irt->route_out[peer->id].route_id);
#else
			trace(TR_IDRP,LOG_ERR,"outbound route linked list screwed up peer %s route_id %d new id %d", \
				peer->name,route_id,p_irt->route_out[peer->id].route_id);
#endif /* ADN_PATCH */
			return(p_send_list);
			}
		tmp_send.pdu_bytes = tmp_send.pdu_bytes + PFX_BYTE_LENGTH(p_irt->nlri); 
		}
#ifdef ADN_PATCH
	trace (TR_IDRP,0,
		"%d nlri processed for withdraw route_id %d peer %s.  ONLY %d re-ann early proc %d routes", 
		nlri_cnt,
		route_id_adn,   /* !!! */
		peer->name, nlri_ann,nlri_early);
#else
	trace (TR_IDRP,0,"%d nlri processed for withdraw route_id %d peer %s.  ONLY %d re-ann early proc %d routes", \
			nlri_cnt, route_id, peer->name, nlri_ann,nlri_early);
#endif /* ADN_PATCH */
 	
	/* add in bytes for withdraw ID */

	tmp_send.pdu_bytes = tmp_send.pdu_bytes  + (sizeof(route_id));
   
	/* finished the output loop 
	 * 1.) Does send list have pdu_bytes_attr set?
	 * 	-no - set attribute record stuff in send list
	 *  
	 * 2.) do we have enough room in the pdu
	 *	- if send list is new, increment send list
	 *	- for attribute byte list 
	 * 
	 * 3.) if no more room (idrp_send_list_room_pdu)
	 *		(idrp_send_list_room_pdu - governs how much
	 *		 gets stuffed in an IDRP BISPDU UPDATE)
	 * 
	 *		a.) send pdu with current list
	 *		b.) put temporary send list into running send list
	 * 
	 * 4.) if enough room (idrp_send_list_room)
	 *		a.) add temporary send list to permanent send list
	 *
	 */

	if (!p_send_list)
		{
		/* no send list exists yet - so set it up
		 * with attributes record, rib id, attribute length
		 */

		p_send_list = create_send_list(p_att_dist,peer);
		}	
	
	/* now ready to stufrf the route id 
	 * in temp - it is redudant now, but left
	 * so that we will rexamine the packing of pdus - skh 4/18/93 */
	
	tmp_send.withdraw->count++;
#ifdef ADN_PATCH
	tmp_send.withdraw->withdraw[1] = route_id_adn;
#else
	tmp_send.withdraw->withdraw[1] = route_id;
#endif /* ADN_PATCH */

	if (idrp_send_list_room_pdu(p_send_list,&tmp_send))
		{
		/* enough room in pdu - so stuff in the withdraw id
		 * and nlri routes
		 */

#ifdef ADN_PATCH
		p_send_list->withdraw->withdraw[p_send_list->withdraw->count] = route_id_adn;		/* !!! */
#else
		p_send_list->withdraw->withdraw[p_send_list->withdraw->count] = route_id;
#endif /* ADN_PATCH */
		p_send_list->withdraw->count++;
		p_send_list->pdu_bytes = p_send_list->pdu_bytes + tmp_send.pdu_bytes;
		if (nlri_ann)
			{
			/* loop adding all NLRI per family */

			NLRI_FAMILY_LOOP(i)
				{
				/* are the linked lists set-up ? */
				if (p_send_list->ann_nlri[i].head)
					{
					/* link list for routes set-up 
					 * so just add to the tail 
					 */

					p_send_list->ann_nlri[i].tail->p_ann_nlri = tmp_send.ann_nlri[i].head;
					}
				else
					{
					p_send_list->ann_nlri[i].head = tmp_send.ann_nlri[i].head;
					}	
				p_send_list->ann_nlri[i].tail = tmp_send.ann_nlri[i].tail;
				}
			}
		
		/* kludge to send 1 withdraw with associated announces per PDU 
		 * UPDATE pdu and go on
		 */

		tmp_send.withdraw->count = 0;
		send_update_reset_send_list(peer,p_send_list,&tmp_send);	
		}
	else
		{
		/* send out the pdu */
		trace (TR_IDRP,0,"sending out 2nd withdraw in pdu send_with_nlri peer %s ",peer->name); 

		send_update_reset_send_list(peer,p_send_list,&tmp_send);
		}
	return(p_send_list);
}

void
send_update_reset_send_list(peer,p_send_list,p_tmp_send)
idrpPeer        *peer;
idrp_send_list  *p_send_list;
idrp_send_list  *p_tmp_send;
{
int     i;
int     only_unreachables;

	/* not enough room in this pdu - so send the
	* UPDATE pdu and go on
	* Note that every time the upper routines
	* change  attribute records , the
	* rest of the send list will be forced out
	*/


	only_unreachables = TRUE;
	NLRI_FAMILY_LOOP(i)
		{
		if (p_send_list->ann_nlri[i].head != (idrpRoute *) 0)
			{
			/* empty list */
			only_unreachables = FALSE;
			break;
			}
		} NLRI_FAMILY_LOOP_END

	send_update_pdu(peer,p_send_list,only_unreachables);

	/* clear out linked list of send lists */

	if (!p_tmp_send)
		{
		bzero(p_send_list,sizeof(idrp_send_list));
		bzero(p_send_list->withdraw,sizeof(Withdrawl));
		return;
		}

	p_send_list->p_next = (idrp_send_list *) 0;
	p_send_list->p_attr = p_tmp_send->p_attr;
	p_send_list->rib_id = p_tmp_send->rib_id;
	p_send_list->pdu_bytes = p_tmp_send->pdu_bytes;
	NLRI_FAMILY_LOOP(i)
		{
		p_send_list->ann_nlri[i].head = p_tmp_send->ann_nlri[i].head;
		p_send_list->ann_nlri[i].tail = p_tmp_send->ann_nlri[i].tail;
		} NLRI_FAMILY_LOOP_END

	p_send_list->withdraw->count = p_tmp_send->withdraw->count;
	for (i = 0; i < p_send_list->withdraw->count; i++)
		{
		p_send_list->withdraw->withdraw[i] = p_tmp_send->withdraw->withdraw[i];
		}		 	
}




idrp_send_list *
send_nlri_attr (p_idrp_rt,p_att_dist,p_send_list,peer)
idrpRoute		*p_idrp_rt;	/* idrpRoute with NLRI to be announced */ 
idrp_attribute_record 	*p_att_dist;	/* attribute record to be distributed */ 
idrp_send_list		*p_send_list;	/* send list */
idrpPeer		*peer;		/* peer to send UPDATE to */
{
int	i;
int	id = 0;
idrp_send_list		tmp_send;	/* send list */
Withdrawl		withdraw;	/* withdraw array */


	/* The send_nlri_attr and send_with_attr routines are
 	 *   threaded to a single attribute
	 *   during  upgrade of code to IS ballot this send
	 *   routines will be changed to allow multiple attribute  
	 */ 


	id = family_to_nlri_id(p_idrp_rt->family);

	/* set-up temporary send list  */

	bzero(&tmp_send,sizeof(idrp_send_list));
	tmp_send.p_attr = p_idrp_rt->p_attr;
	tmp_send.rib_id = p_idrp_rt->p_attr->rib_id;
	tmp_send.pdu_bytes_attr = p_idrp_rt->p_attr->attr_len;	
	tmp_send.pdu_bytes = PFX_BYTE_LENGTH(p_idrp_rt->nlri);
	tmp_send.ann_nlri[id].head = tmp_send.ann_nlri[id].tail = p_idrp_rt;
	tmp_send.withdraw = &withdraw;
	tmp_send.withdraw->count = 0;	
	if (p_idrp_rt->pref_cal > 255)
		{
		tmp_send.loc_pref = (byte) 0xff;
		}
	else
		{ 
		tmp_send.loc_pref = (byte) p_idrp_rt->pref_cal;
		}

	/* if no send list - allocate it */

	if (!p_send_list)
		{
		/* no existing send list allocate  a send list structure 
		 * - and fill it with  
		 */

		p_send_list = create_send_list(p_att_dist,peer);
		}

	/* see if this nlri overflows the pdu - if so send pdu */

	if (idrp_send_list_room_pdu(p_send_list,&tmp_send))
		{
		/* room to stuff in this nlri
		 * - so add it to announce tail 
		 * (here's where the code needs to
		 *  change for multiple attributes per
		 *  the send list.  The code here
		 *  needs to search for right attribute in pdu 
		 *  idrp_send_list will be multiple records 
		 *  with the combination being sent out)
	 	 * 
		 */
		p_send_list->pdu_bytes = p_send_list->pdu_bytes + tmp_send.pdu_bytes;
		p_send_list->p_attr = tmp_send.p_attr;
		p_send_list->rib_id = tmp_send.rib_id; 
		p_send_list->loc_pref = tmp_send.loc_pref;

		if (p_send_list->ann_nlri[id].head)
			{
			p_send_list->ann_nlri[id].tail->route_out[peer->id].p_next = p_idrp_rt;
			p_send_list->ann_nlri[id].tail = p_idrp_rt;
			}
		else	
			{
			p_send_list->ann_nlri[id].head = p_idrp_rt;
			p_send_list->ann_nlri[id].tail = p_idrp_rt;
			}
		p_send_list->ann_nlri[id].tail->route_out[peer->id].route_id = peer->route_id_out;
		p_idrp_rt->route_out[peer->id].p_next = (idrpRoute *) 0;
		}
	else
		{
		
		/* send update and reset send list */

		send_update_reset_send_list(peer,p_send_list,&tmp_send);
		}

	return (p_send_list);
}


idrp_send_list *
create_send_list(p_att_dist,peer)
idrp_attribute_record 	*p_att_dist;
idrpPeer		*peer;
{
idrp_send_list *p_send_list;
int		i;

	p_send_list = (idrp_send_list *) task_mem_malloc(peer->task,sizeof(idrp_send_list));
	bzero(p_send_list, sizeof(idrp_send_list));
	p_send_list->p_attr = p_att_dist;
	p_send_list->pdu_bytes_attr = p_att_dist->attr_len;
	p_send_list->rib_id = p_att_dist->rib_id;
	p_send_list->withdraw = (Withdrawl *) task_mem_malloc(peer->task,sizeof(Withdrawl));
	bzero(p_send_list->withdraw, sizeof(Withdrawl));
	NLRI_FAMILY_LOOP(i)
		{
		p_send_list->ann_nlri[i].head = p_send_list->ann_nlri[i].tail = (idrpRoute *) 0;
		}
	return(p_send_list);
}

void
free_send_list(p_send_list,peer)
idrp_send_list	*p_send_list;
idrpPeer	*peer;
{

	/* this is a simple task mem free
	 *  - if send list becomes a linked list
	 *  - then this call must be changed to
	 *  - free the linked list
	 *  -- skh 3/10/93
	 */

	if (p_send_list)
	{
		if (p_send_list->withdraw)
		{
		task_mem_free(peer->task,p_send_list->withdraw);
		}
		task_mem_free(peer->task,p_send_list);
	}
}
	

void
flush_att_send_list(peer,p_att_dist,p_send_list)
idrpPeer		*peer;
idrp_attribute_record	*p_att_dist;
idrp_send_list		*p_send_list;
{
u_int	only_unreachables = TRUE;
int	i;


	/* if no send list - don't flush it or it will flush you */

	if (!p_send_list)
		{
		return;
		}

	NLRI_FAMILY_LOOP(i)
		{
		if (p_send_list->ann_nlri[i].head != (idrpRoute *) 0 )
			{
			only_unreachables  = FALSE;
			break;
			}
		}


	if (only_unreachables && (p_send_list->withdraw->count == 0))
		{
		return;
		} 
	trace (TR_IDRP,0,"flush attribute send list routine entered ");
	send_update_pdu(peer,p_send_list,only_unreachables);
	p_send_list->p_next = (idrp_send_list *) 0;
	p_send_list->p_attr = (idrp_attribute_record *) 0;
	p_send_list->pdu_bytes = 0;
	p_send_list->pdu_bytes_attr = 0;
	NLRI_FAMILY_LOOP(i)
		{
		p_send_list->ann_nlri[i].head = p_send_list->ann_nlri[i].tail = (idrpRoute *) 0;
		}	
}


int
idrp_send_list_room_pdu(p_list,p_new)
idrp_send_list	*p_list;
idrp_send_list	*p_new;
{

	/* this is where the logic lies to pack in
	 *
	 * - in the future it will try to package as
	 * - many sequences as possible
	 */

	if (p_new->withdraw->count)
		{
		if (p_new->pdu_bytes < (IDRP_PKTSIZE - sizeof(idrp_fixed_pdu_header)))
			{
			return(TRUE);
			}
		}
	
	if (p_list->pdu_bytes)
		{
		return (FALSE);
		}
			
	
	return (TRUE);
} 


void
idrp_del_sent_routes(p_ann_list)
idrp_ann_list *p_ann_list;
{
int		changes = 0;	/* changes to route table */
int			i;
idrp_ann_list		*p_atl;
idrpRoute		*p_irt;

	/* walk through the announce list building building
	 * the send list   
	 */

	rt_open(idrp_this_node.task);
	ANN_LIST(p_atl,p_ann_list)
		{
		NLRI_FAMILY_LOOP(i)
			{
			ANN_WITH_LIST_WALK(p_irt,i,p_atl)
				{
				/* delete the routes with IDRP_STATUS_DEL_SEND 
				 *
				 */

				if (p_irt->status & IDRP_STATUS_DEL_SEND)
					{
					idrp_del_rt_gated(p_irt);
					idrp_free_nlri_att_rec(p_irt,NLRI_LINKED_P_WITH);
					idrp_free_idrpRoute(p_irt);
					changes++;
					}
				} ANN_WITH_LIST_WALK_END;
			} NLRI_FAMILY_LOOP_END;
		} ANN_LIST_END;

	rt_close(idrp_this_node.task,&idrp_this_node.gw,changes,NULL);

	
	/* trace the deleting withdrawn routes */
 
	trace (TR_IDRP,0,"%d Withdrawn routes idrp routes deleted ",changes);

}




/* idrp_add_locate -
 * 	calculate idrp preference
 *	calculate gated preferences
 *
 * 	walk nlri list to 	
 *	Set the gated preference and add to tables
 * 		- as change or add 
 *	so idrp decision process 2 needs to be run  	
 */

int
idrp_add_route_locate(p_dest,p_idrp_rt,p_gate,p_pref,p_best_ext)
sockaddr_un		*p_dest;	/* destination structure */
idrpRoute		*p_idrp_rt;	/* route */
idrp_rt_chain_walk	*p_gate;	/* return chain for matched gateways */
idrp_rt_chain_walk	*p_pref;	/* return chain for routes with same IDRP pref */ 
idrpRoute		**p_best_ext;	/* best external route  */
{
rt_entry	*p_rt,*p_rt_first;	/* rt_entry for gated route table entry */
idrpRoute	*p_irt;		/* idrpRoute */ 
idrpPeer	*peer;		/* peer idrpRoute received from */
int		type;		/* type of peer route came from */
				/* internal, external, local (this node) */
				/* %%%% future - test to mirror bgp */
#ifndef ADN_PATCH
rt_parms	rt_parms;	/* route parameters for idrp route */
#endif /* ADN_PATCH */
gw_entry	*rt_gwp;	/* gateway for gated */
int		gated_pref;	/* gated preference (phase 2) for route */ 
int		pfx_len;	/* gated preference (phase 2) for route */ 
int		length;		/* length of iso prefix */
sockaddr_un	*p_mask;	/* mask for gated route hacking &*%# */
			
	
	/* Note you MUST USE THE GATED MASK - Why?
	 * - the radix compare does a pointer compare
	 * - not a bit by bit compare 
	 * It's really not a bad way to do it. But I'm just frosted from
	 * having to learn it during the debugging of the code.
	 * I guess it is too much to expect the 'old boys
	 * to tell me about it.  It must be some test
	 * of sorts to enter the club to have to figure it
	 * out from the code. 
	 *   
	 */   
 

	/* logic of adding route is:
	 *  1.) calculate IDRP preference and gated preference for new route
	 *  2.) determine type of peer route came from: 
	 *      - external or internal, local (this node), test
	 *  3.) try to get any match on the nlri and protocol
	 *	via rt_locate call to gated 
	 *  4.) if no for idrp, set flags
	 *	  no gateway match
	 *	  no preference match
	 *	  no old best_ext (nulled out if the p_ext chain pointer is null )
	 *
	 *  5.) if return, walk the rest of the chain
	 *	to find:
	 *	 if gateway match
	 *	 if preference match, link matched preference
	 *	   on an idrpRoute list - using preference as route_id
	 *	 if new route is external, 	
	 *		link any external route to idrpRoute_list
	 *		(nulled out if the p_ext chain pointer is null 
  	 * 
 	 * chains to the caller:
	 */ 

	/* initialize the idrp chain walking structures */

	p_gate->p_next = (idrp_rt_chain_walk *) 0;	
	p_gate->p_rt = (rt_entry *) 0;
	p_pref->p_next = (idrp_rt_chain_walk *) 0;	
	p_pref->p_rt = (rt_entry *) 0;
	*p_best_ext = (idrpRoute *) 0;	/* no best external best external route  */


	/* calculate IDRP preference for route */
	/* and put it in pref_cal part of idrpRoute structure */
 
	idrp_pref(p_idrp_rt);

	/* calculate gated preference for route */
	/* this is a phase 2 calculation, but not action */

	gated_pref = idrp_to_gated_pref(p_idrp_rt);
	peer = p_idrp_rt->peer;
		
	/* set-up masks to either first locate, then add the route
	 * - recalculate the byte length to force check on copy
	 */ 

	/* set the family using the idrpRoute  family value */
 
	pfx_len = p_idrp_rt->nlri.pfx_len;		
	switch (p_idrp_rt->family)
		{
		case AF_ISO:
			length = PFX_BYTE_LENGTH(p_idrp_rt->nlri); 
			p_dest->iso.siso_len = length + 2;
			p_dest->iso.siso_family = p_idrp_rt->family;
			bcopy((caddr_t) &p_idrp_rt->nlri.pfx,(caddr_t) &p_dest->iso.siso_addr,length);
			p_mask = iso_mask_prefix(pfx_len);
			break;

		case AF_INET:
			/* copy the prefix over as byte */
 
			bcopy((caddr_t) &p_idrp_rt->nlri.pfx,(caddr_t) &p_dest->in.sin_addr,sizeof(struct in_addr));
			p_dest->in.sin_family = AF_INET; 
			p_dest->in.sin_len = 8; 
			p_dest->in.sin_port = 0;
			p_mask = inet_mask_prefix(pfx_len);	
			break;
		}
		
	/* here add in the look-up in the rt_table to find match for gateway */
	/* to find exact match and best match */

	p_rt = rt_locate(RTS_NETROUTE,p_dest,p_mask,RTPROTO_BIT(RTPROTO_IDRP));
	
	/* if nothing was return - no nlri with IDRP protocol */

	if (!p_rt)
		{
		/* set rt_entry pointers to zero for gate, preference 
		* and the external if it is called
		*/ 

		p_gate->p_rt = (rt_entry *) 0;
		p_pref->p_rt = (rt_entry *) 0;
		return(FALSE);
		}

	/* walk the rest of nlri structure 
	 * looking for gateway match, best_ext 
	 * and preference matches 
	 */

	p_rt_first  = p_rt;
	while (p_rt) 
		{
		/* %jgs% -- rip this out!!! */
		/* if (p_rt->rt_gwp) -- the original */
		if ((p_rt->rt_gwp) && (p_rt->rt_gwp != (gw_entry*)1))
			{	
			if (p_rt->rt_gwp->gw_proto == RTPROTO_IDRP)
				{
				/* idrp route */
	
				if (p_rt->rt_gwp->gw_addr == peer->gw.gw_addr || p_rt->rt_gwp->gw_addr == peer->iso_gw.gw_addr)
					{
					/* matching gateway, because pointers match save on list */
	
					p_gate->p_rt = p_rt;
					p_gate->p_next = (idrp_rt_chain_walk *) task_mem_malloc(peer->task,sizeof(idrp_rt_chain_walk));
					p_gate->p_next->p_next = 0;
					p_gate->p_next->p_rt = 0;
					p_gate = p_gate->p_next; 
					}
				else
					{
					/* if this route is not from the same gateway, check preference and */
					/* external route */

					/* compare pref on existing route to pref on our new route
				 	 * top of idrpRoute is rt_data structure so cast as idrp_route
					 */
					p_irt = (idrpRoute *) p_rt->rt_data; 
					if (p_irt->pref_cal == p_idrp_rt->pref_cal)
						{
						/* equal gated preferences - link to list of equiv pref rts */
	
						p_pref->p_rt = p_rt;
						p_pref->p_next = (idrp_rt_chain_walk *) task_mem_malloc(peer->task,sizeof(idrp_rt_chain_walk));
						p_pref->p_next->p_next = 0;
						p_pref->p_next->p_rt = 0;
						p_pref = p_pref->p_next;
						}
			
					if (p_irt->peer->type == IDRP_PEER_EXTERNAL)
						{
						*p_best_ext = p_idrp_rt->p_p_best_ext->p_best_ext;
						} /* end test for external  */
					} /* end test for matching gateway */ 
				} /* end bit test for the idrp route - */
			} /* does a gateway */ 

		/* walk to the next route */
		p_rt = p_rt->rt_forw;
		if (p_rt == p_rt_first)
			{
			/* circular loop - so watch for first route */

			p_rt = (rt_entry *) 0;
			}

		}  /* end loop to walk forward in rt_entry chain */

	return(TRUE);
}

/* idrp_with_route_locate-
 * 	calculate idrp preference
 *	calculate gated preferences
 *
 * 	walk nlri list using p_with parameter 
 *	Set the gated preference and add to tables
 * 		- as change or add 
 *	so idrp decision process 2 needs to be run  	
 */

int
idrp_with_route_locate(p_dest,p_idrp_rt,p_gate,p_pref,p_best_ext)
sockaddr_un		*p_dest;	/* destination structure */
idrpRoute		*p_idrp_rt;	/* route */
idrp_rt_chain_walk	*p_gate;	/* return chain for matched gateways */
idrp_rt_chain_walk	*p_pref;	/* return chain for matched idrp preferences */ 
idrpRoute		**p_best_ext;	/* best external route  */
{
int	found;

	found =  idrp_add_route_locate(p_dest,p_idrp_rt,p_gate,p_pref,p_best_ext);
	if (!found)
		{
		trace (TR_IDRP, 0, "idrp_with_route_locate:  No IDRP routes for NLRI %A", p_dest);
		return(FALSE);
		}
	
	/* This is the real code to find the best external route.  Dead easy.
   	 */

	*p_best_ext = p_idrp_rt->p_p_best_ext->p_best_ext;
	return (TRUE);	/* ---skh return value - true if route found */
}


int idrp_rt_change(p_rt,p_idrp_rt)
rt_entry *p_rt;
idrpRoute *p_idrp_rt;
{
idrpRoute *p_idrp_rt2;
int	change = FALSE;

	/* copy over changes to the rt_entry data structure */

	p_idrp_rt2 = (idrpRoute *)p_rt->rt_data;

	/* walk down and find out what is different between the two routes */

	if (p_idrp_rt2->peer != p_idrp_rt->peer)
		{
		trace (TR_IDRP,0,"match for idrp_rt_change peer %s to peer %s - ERROR!!",
			p_idrp_rt->peer->name,p_idrp_rt2->peer->name);
		return(change);
		}

	/* check on attribute record change */

	if (p_idrp_rt2->p_attr != p_idrp_rt->p_attr)
		{
		change = TRUE;	
		}

	if (p_idrp_rt2->route_id_in != p_idrp_rt->route_id_in ||
		p_idrp_rt2->route_id_in == IDRP_LOCAL_ROUTE_ID)
		{
		change = TRUE;
		}
	return (change);

}

/* free_rt_chain_walk
 *
 */

void
free_rt_chain_walk(p_ch,peer)
idrp_rt_chain_walk	*p_ch;
idrpPeer		*peer;
{
idrp_rt_chain_walk	*p_free;
idrp_rt_chain_walk	*p_next;


	for (p_next = p_ch,p_free = p_ch; p_next; p_free = p_next)
		{
		p_next = p_free->p_next;
		task_mem_free(peer->task,p_free);
		}
return;
}

/*
 * find_next_best(p_idrp_rt) - find the next-best external route
 *	to the same NLRI as the external route p_idrp_rt
 *	points to.  Return a pointer to it.
 */
idrpRoute *
find_next_best(p_idrp_rt)
idrpRoute *p_idrp_rt;
{
	if (!(p_idrp_rt->status & IDRP_STATUS_BEST_EXT))
		{
		trace (TR_IDRP, LOG_WARNING, "IDRP (find_next_best) route is not best external");
		return NULL;
		}
		
	return p_idrp_rt->p_next_best;
}
			
/*
 * find_best_ext(p_idrp_rt) - find the best external route
 *	to the same NLRI as the external route p_idrp_rt
 *	points to.  Return a pointer to it.
 */
idrpRoute *
find_best_ext(p_idrp_rt)
idrpRoute *p_idrp_rt;
{
	return p_idrp_rt->p_p_best_ext->p_best_ext;
}

/*
 * idrp_pref_compare(idrpRoute *p_idrp_rt, idrpRoute *p_best_ext)
 *
 * Description:  Return TRUE if p_idrp_rt is the best external route
 *  and FALSE otherwise.  Do this by comparing to p_best_ext,
 *  the current best external (or NULL if there is no best
 *  external to this destination).
 *
 * Discussion:  The algorithm used (just compare to current BER) 
 *  assumes that the tie-break algorithm imposes a complete ordering
 *  on routes to the same destination.  I can't imagine how it could
 *  just be a partial ordering, so I think we're okay.
 * 
 *  We assume that whoever called idrp_pref_compare is planning to 
 *  actually set the best route pointer to the (new) best ext route,
 *  set the IDRP_STATUS_BEST_EXT bit in the status flag, and so on.
 */
int
idrp_pref_compare(p_idrp_rt, p_best_ext)
idrpRoute	*p_idrp_rt;	/* the route we're checking */
idrpRoute	*p_best_ext;	/* Current best ext route */
{
	/* 
	 * 7.17.1.b says to pick this route as a new BER if:
	 *  1. This route has the highest degree of pref,
	 *  2. This route is the only route to the destination, or
	 *  3. This route is selected as the result of a tie-break.
	 */

	/* Check clause 2 || 1 : */
	if ((p_best_ext == NULL) || (p_idrp_rt->pref_cal > p_best_ext->pref_cal))
		return (TRUE);
	/* Else check clause 3 */
	else if (tie_break(p_idrp_rt, p_best_ext) == ISO_ADDR_LESS_THAN)
		return (TRUE);
	else
		return (FALSE);
}

/*
 * insert_in_pref_order(idrpRoute *p_rt, idrpRoute *p_best)
 *
 * Insert p_rt into the sorted-by-preference list of routes to the same NLRI.
 */
void
insert_in_pref_order(p_rt, p_best)
idrpRoute *p_rt;	/* route to be inserted */
idrpRoute *p_best;	/* best route in chain of idrpRoutes to NLRI */
{
	while (p_best != NULL)
		{
		if (idrp_pref_compare(p_rt, p_best->p_next_best))
			/* insert between scratch and p_best->p_next_best */
			{
			p_rt->p_next_best = p_best->p_next_best;
			p_rt->p_better = p_best;
			p_best->p_next_best = p_rt;
			if (p_rt->p_next_best != NULL) /* NULL if we inserted at end of list */
				{
				p_rt->p_next_best->p_better = p_rt;
				}
			break;
			}
		p_best = p_best->p_next_best;
		}
}

int idrp_ann_list_empty(p_ann_list)
idrp_ann_list 	*p_ann_list;
{
int	i;
/* check to see that the neither the ann_nlri or with_nrli
 * has anything on it - heads of lists should be non-zero
 * in first attribute
 */

	NLRI_FAMILY_LOOP(i)
		{
		if (p_ann_list->ann_nlri[i].head)
			{
			return(FALSE);
			}
		if (p_ann_list->with_nlri[i].head)
			{
			return(FALSE);
			}
		} NLRI_FAMILY_LOOP_END;
	return(TRUE);
} 

int
find_nlri_in_ann_nlri(p_idrp_rt,p_ann_list)
idrpRoute	*p_idrp_rt;
idrp_ann_list	*p_ann_list;
{
idrp_ann_list	*p_atl;
idrpRoute	*p_irt;
int		i;

	i = family_to_nlri_id(p_idrp_rt->family); 
	ANN_LIST(p_atl,p_ann_list)
		{
		ANN_ANN_NEXT_NLRI(p_irt,i,p_atl)
			{
			if (!isopfxcompare(&p_irt->nlri,&p_idrp_rt->nlri))
				{
				return(TRUE);
				}
			} ANN_ANN_NEXT_NLRI_END;
		} ANN_LIST_END;
	return(FALSE);
}
