#include "include.h"
#include "idrp.h"
#include <stdio.h>


/* idrp_attrib.c
 *  
 *  attribute validation and handling routines
 *  (rd path handling and validation takes up lots of the module)
 *  there also seems to be a bunch of route-id chain handling here.
 *  perhaps there should be another chain for route-id handling?
 * 	valid_attr(peer,attr) - validate attribute record 
 *	valid_usr_attr(peer,attr) - user attributes
 *	add_local_rd_to_path(peer,attr_rec) - add our rd to the rd path
 *		also detects routing loops
 *	valid_seg_type(type,peer) - ??? something to do w/ rd path
 *	valid_rd_path(peer,att_rec,p_rdpath,rdpath_length) - ??? validate rd
 *		path, I guess
 *	link_rd_canon(p_rd_new,p_att) - link rds in canonical order
 * 	find_attr_rec - search for duplicate attribute record in attribute list  
 *     	compare_atts  - compare attributes
 *	idrp_dif_path - see if different path attribute- return
 *			mask of attributes that are different 
 *	ATTS_REC_ZERO(p_atts) - zero out path atts.  Maybe this should be a 
 *		macro and not a func.
 *	idrp_free_routeid_chain_id(p_att,id,peer) - free routeid by id, att and peer
 *	idrp_find_routeid_chain(p_rt,p_last_rte) - find routeid chain pointer for this idrpRoute
 *	idrp_free_nlri_att_rec(p_rt) - free nlri from chain
 *      idrp_free_attr_rec - free attribute record
 * 	link_att_list - link an attribute record to the list 
 * 	relink_free_att - relink attribute list to exclude attribute record and free
 * 	idrp_free_nlri_chain - free nlris off route_id chain off attribute list
 */ 
  
 

/*		
 * valid_attr - check for valid attributes 
 */
int valid_attr(peer,att_rec)
idrpPeer *peer;
idrp_attribute_record *att_rec;
{
int attr_type;
int attr_len;
u_int	longhold;
u_char *dataptr;
u_char *rdpath;
int	rdpath_length;
short seglen, seglen_local;
struct	iso_net_addr *addr;
int	attr_present = 0;
idrp_canon_rdpath *p_rd_canon_list = (idrp_canon_rdpath *) 0;
idrp_canon_rdpath *p_rd = (idrp_canon_rdpath *) 0;

	/* now validate the attribute sections of UPDATE */

	/* (7.21.3 - d) check for well known attributes  */

	if ((att_rec->idrp_mask & IDRP_ATTBT_MANDATORY) != IDRP_ATTBT_MANDATORY)
		{
		peer->last_error_pdu.error_code = IDRP_UPDATE_ERROR;
		peer->last_error_pdu.error_subcode = IDRP_UPDATE_ERROR_MISS_WELKN_ATTR;
		if (att_rec->idrp_mask & IDRP_ATTBT_ROUTE_SEPARATOR == 0 ) 
			{
			peer->last_error_pdu.data = (u_char *) &idrpAttrType[IDRP_ATTR_ROUTE_SEPARATOR]; 
			}
		else	{
			if (att_rec->idrp_mask & IDRP_ATTBT_RD_PATH == 0)
				{
				peer->last_error_pdu.data = (u_char *) &idrpAttrType[IDRP_ATTR_RD_PATH];
				}
			else
				{
				peer->last_error_pdu.data = (u_char *) &idrpAttrType[IDRP_ATTR_HOP_COUNT];
				}
			}	
		peer->last_error_pdu.data_len = 1;
		trace(TR_IDRP , 0, "IDRP valid_attr error - update missing well known attributes.  Mandatory attribute mask = %x, provided attribute mask = %x", IDRP_ATTBT_MANDATORY, att_rec->idrp_mask);
		return(FALSE);
		}

	/* 7.21.3 i) check that DIST_LIST_INCL and DIST_LIST_EXCL  */
	/*           are not both active */

	if ((att_rec->idrp_mask & IDRP_ATTBT_DIST_LIST_INCL) && 
	    (att_rec->idrp_mask & IDRP_ATTBT_DIST_LIST_EXCL))
		{
		/* both IDRP DIST_LIST_INCL and IDRP_DIST_LIST_EXCL */
		/* exist - malformed attribute list error */
		
		peer->last_error_pdu.error_code = IDRP_UPDATE_ERROR;
		peer->last_error_pdu.error_subcode = IDRP_UPDATE_ERROR_MALF_ATTR;
		peer->last_error_pdu.data = NULL;
		peer->last_error_pdu.data_len = 0;
		trace(TR_IDRP , 0, "IDRP valid_attr error - both DIST_LIST_INCL and DIST_LIST_EXCL are present");
		return (FALSE);
		} 

	/* all mandatory attribute there - check flag codes and each attribute */

	for (attr_type = 1; attr_type < IDRP_ATTR_MAX + 1; attr_type++) {

		/* if  attribute not present skip all the tests */

		if (!att_rec->attrib[attr_type].present)
			continue;

		dataptr = att_rec->attrib[attr_type].data;
		attr_len = att_rec->attrib[attr_type].length;

		trace(TR_IDRP | TR_UPDATE, 0, "  Attr %d, length %d, flags %x", attr_type,
			att_rec->attrib[attr_type].length, att_rec->attrib[attr_type].flags);

		/* validate attribute flags  (idrp 7.xx.x - x)  */
		if (att_rec->attrib[attr_type].flags != idrpAttrFlags[attr_type])
			{
			peer->last_error_pdu.error_code=IDRP_UPDATE_ERROR;
			peer->last_error_pdu.error_subcode = IDRP_UPDATE_ERROR_ATTR_FLAGS_ERR;
			peer->last_error_pdu.data = att_rec->attrib[attr_type].data;
			peer->last_error_pdu.data_len = att_rec->attrib[attr_type].length;
			trace(TR_IDRP , 0, "IDRP valid_attr error - problems with attribute type %d with flags 0x%x",
				attr_type, att_rec->attrib[attr_type].flags);
			return(FALSE);
			}


		switch(attr_type) {

		/* ROUTE_SEPARATOR */
		case IDRP_ATTR_ROUTE_SEPARATOR:
			if (attr_len != IDRP_ROUTE_SEPARATOR_LENGTH)
				{ 
				peer->last_error_pdu.error_code=IDRP_UPDATE_ERROR;
				peer->last_error_pdu.error_subcode = IDRP_UPDATE_ERROR_ATTR_LEN_ERR;
				peer->last_error_pdu.data = att_rec->attrib[attr_type].data;
				peer->last_error_pdu.data_len = att_rec->attrib[attr_type].length;
				trace(TR_IDRP , 0, "IDRP valid_attr error - route separator has invalid length.  Desired length = %d, provided length = %d", IDRP_ROUTE_SEPARATOR_LENGTH, attr_len);
				return(FALSE);
				}
			bcopy(dataptr, (caddr_t) &longhold,sizeof(route_id));
			longhold = htonl(longhold);
			trace(TR_IDRP | TR_UPDATE, 0, "   ROUTE_ID 0x%x",longhold );
			att_rec->route_id_list->route_id=longhold;
			trace(TR_IDRP | TR_UPDATE, 0, "   ROUTE_ID 0x%x",att_rec->route_id_list->route_id);
			dataptr += sizeof(route_id);
			att_rec->route_id_list->loc_pref = *dataptr;	
			break;			

		case IDRP_ATTR_EXT_INFO:
			if (attr_len != IDRP_EXT_INFO_LENGTH)
				{ 
				peer->last_error_pdu.error_code=IDRP_UPDATE_ERROR;
				peer->last_error_pdu.error_subcode = IDRP_UPDATE_ERROR_ATTR_FLAGS_ERR;
				peer->last_error_pdu.data = att_rec->attrib[attr_type].data;
				peer->last_error_pdu.data_len= att_rec->attrib[attr_type].length;
				trace(TR_IDRP , 0, "IDRP valid_attr error - ext info has invalid length.  Desired length = %d, provided length = %d", IDRP_EXT_INFO_LENGTH, attr_len);
				return(FALSE);
				}
			break;

		/* RD Path.  Note its position so that we can copy it into our route. */

		case IDRP_ATTR_RD_PATH:
			if (valid_rd_path(peer,att_rec,dataptr,attr_len))
				{
				/* add in the local rd path if peer
				 * received from is an external peer 
			   	 */
				if (peer->type == IDRP_PEER_EXTERNAL)
					{
					if (!add_local_rd_to_path(peer,att_rec))
						{
						trace(TR_IDRP , 0, "IDRP valid_attr error - received update from external peer, failed to successfully add our RD to the path");
						return (FALSE);
						}
					}
				}
			else
				{
				trace(TR_IDRP , 0, "IDRP valid_attr error - failed to validate RD path (valid_rd_path() failed)");
				return (FALSE);
				}	
			
			break;

		case IDRP_ATTR_HOP_COUNT:
			trace(TR_IDRP | TR_UPDATE, 0, "   HOP_COUNT %d", *dataptr);
			att_rec->hopcount = *dataptr;
			break;

		} /* end of switch statement */
	}	/* end of for statement checking all attributes */

return(TRUE);
}

int valid_usr_attr(peer,att_rec)
idrpPeer *peer;
idrp_attribute_record *att_rec;
{
	/* null for now */
	return(1);

}

/* 
 *  add_local_rd_to_path(peer,attr_rec))
 * 
 * put the local RD in the rd_path by adding it into the 
 * locally added attributes portion of the attribute field
 * LOCAL_RD_PATH attribute of attributes 
 * 
 * link the attribute to the cannonical attribute list
 * if it fails to link, the we have a loop
 * 
 * return code = 0 failed
 * return code = 1 worked  
 */
int  add_local_rd_to_path(peer,attr_rec)
idrpPeer *peer;
idrp_attribute_record *attr_rec;
{
u_int16 hold;
idrp_canon_rdpath *p_rd;


	/* add in our RDI length + (1) RDI length byte + segement header length  */
	/*  to the total RD path length */

	hold = sizeof(idrp_path_segment_header) + rt_rdi.isoa_len + 1;
	attr_rec->attrib[IDRP_ATTR_LOCAL_RD].length = hold;
 	hold += attr_rec->attrib[IDRP_ATTR_RD_PATH].length;
        hold = htons(hold);

	/* modify Attribute length */

        bcopy((caddr_t) &hold,&attr_rec->attrib[IDRP_ATTR_RD_PATH].length, 2); 
        attr_rec->attrib[IDRP_ATTR_LOCAL_RD].present = TRUE;
	attr_rec->attrib[IDRP_ATTR_LOCAL_RD].data = (u_char *) &local_rd_path;

	p_rd = (idrp_canon_rdpath *) task_mem_malloc(peer->task,sizeof(idrp_canon_rdpath));  
	bcopy(&rt_rdi,&p_rd->rdi,sizeof(struct iso_net_addr));
	if (!link_rd_canon(p_rd,attr_rec)) 
		{
		peer->last_error_pdu.error_code = IDRP_UPDATE_ERROR;
		peer->last_error_pdu.error_subcode = IDRP_UPDATE_ERROR_RD_ROUTING_LOOP;
		peer->last_error_pdu.data = (u_char *) &rt_rdi.isoa_genaddr; 
		peer->last_error_pdu.data_len = rt_rdi.isoa_len;      		
		return(FALSE);
		}
	return(TRUE);
}
	
int valid_seg_type(type,peer)
int type;
idrpPeer *peer;
{
	switch (type)
		{
		case IDRP_PATH_SEGMENT_SET:
		case IDRP_PATH_SEGMENT_SEQ:
			return (TRUE);
			break;

		case IDRP_PATH_SEGMENT_ENTRY_SEQ:
		case IDRP_PATH_SEGMENT_ENTRY_SET:
			trace (TR_IDRP | TR_UPDATE,0,"IDRP peer %s got RD_PATH with RDI SET/SEQ",
				peer->name);
			return (TRUE);
			break;	

		}
	return (FALSE);
}


int valid_rd_path(peer,att_rec,p_rdpath,rdpath_length)
idrpPeer *peer;
idrp_attribute_record *att_rec;
u_char	*p_rdpath;
int	rdpath_length;
{
u_char	*cp;
int	pathlen;
u_short hold;
int	seglen;
int	seglen_local;
struct	iso_net_addr	*addr;
idrp_path_segment_header 	*seghead;
idrp_canon_rdpath	*p_rd;
 
	/* clear rd path list pointer, 
	 * set up path length for loop
	 * set-up pointer to pathway sequence
	 */ 
 
	att_rec->rd_path = (idrp_canon_rdpath *) 0 ; 
	pathlen = rdpath_length;
	cp = (u_char *) p_rdpath;
	
	/* parse the rd path information */


	while (pathlen > 0)  {

	/* walk through segment type, length value */ 

		seghead = (idrp_path_segment_header *) cp;		
		cp += sizeof(idrp_path_segment_header);
		bcopy(&seghead->path_seg_len, (caddr_t) &hold, sizeof(seghead->path_seg_len));
		hold = ntohs(hold);
		seglen_local = seglen = (int) hold;
		trace(TR_IDRP | TR_UPDATE, 0, "   RD_PATH type %d, length %d",
			seghead->path_seg_type,seglen);

		if (!valid_seg_type(seghead->path_seg_type, peer))
			{ 
			/* illegal type in the segement type */
			trace(TR_IDRP | TR_UPDATE, 0, "RD PATH type %d illegal peer %s",
			seghead->path_seg_type,peer->name);
			peer->last_error_pdu.error_code = IDRP_UPDATE_ERROR;
			peer->last_error_pdu.error_subcode = IDRP_UPDATE_ERROR_INV_SEG_TYPE;
			peer->last_error_pdu.data = &seghead->path_seg_type;
			peer->last_error_pdu.data_len = sizeof(seghead->path_seg_type);	 
			return(FALSE);
			}
	
		/* valid segment type,  walk thorugh the RDI list */
 
		while (seglen_local > 0) {
			/* build canonical rd path */
			p_rd = (idrp_canon_rdpath *) task_mem_malloc(peer->task,sizeof(idrp_canon_rdpath));  

			/* copy in length byte
			 * - this will have to change if
		 	 * - changes off a byte 
			 */ 

			p_rd->rdi.isoa_len = hold = *cp;		/* one byte length for RDI */
			cp++;

			
			/* copy in RDI */

			bcopy(cp,&p_rd->rdi.isoa_genaddr,hold);
			trace(TR_IDRP | TR_UPDATE, 0, "     %A", 
			      sockbuild_iso((byte *) &p_rd->rdi.isoa_genaddr, p_rd->rdi.isoa_len));


			if (!link_rd_canon(p_rd,att_rec)) 
				{
				peer->last_error_pdu.error_code =IDRP_UPDATE_ERROR;
				peer->last_error_pdu.error_subcode = IDRP_UPDATE_ERROR_RD_ROUTING_LOOP;
				peer->last_error_pdu.data = (u_char *) &p_rd->rdi.isoa_genaddr; 
				peer->last_error_pdu.data_len = p_rd->rdi.isoa_len;      		
				return(FALSE);	
				}
			seglen_local = seglen_local - (hold +1);
			cp += hold;
			}
		/* set up for the next segment */

		pathlen -= seglen + sizeof(idrp_path_segment_header);

		} /* end of loop for the full rd_path */
	return(TRUE);
}

/* 
 * link_rd_canon(p_rd,att)
 *  - link RDIs in cannonical order 
 */
int link_rd_canon(p_rd_new,p_att)
idrp_canon_rdpath	*p_rd_new;
idrp_attribute_record	*p_att;
{
/* local variables */

idrp_canon_rdpath	*p_rd;		/* pointer to rd path entry */
idrp_canon_rdpath	*p_cur;		/* pointer to rd path entry */
idrp_canon_rdpath	*p_next;	/* pointer to rd path entry */
u_int			len;		/* length in bits of match */
int			res;		/* result */

	/* list empty */

	if (p_att->rd_path == (idrp_canon_rdpath *)0 )
		{
		p_rd_new->p_next = (idrp_canon_rdpath *) 0;
		p_att->rd_path = p_rd_new;
		return(TRUE);
		}

	if (p_att->rd_path->p_next == (idrp_canon_rdpath *) 0)
		{
		/* just one on the list link this one in before or after */
		
		res = compare_iso_addr(&p_rd_new->rdi,&p_att->rd_path->rdi);


		if (res == ISO_ADDR_EQUAL)
			{
			/* iso address equal - RD twice in pathway */
			return(FALSE);
			}
		else 	{
			if (res == ISO_ADDR_LESS_THAN)
				{
				/* this iso address less than one on list */
				/* this address is new head of list */

				p_rd_new->p_next = p_att->rd_path;
 				p_att->rd_path->p_next = (idrp_canon_rdpath *) 0;
				p_att->rd_path = p_rd_new;
				}	
			else
				{
				/* this iso address greater than one on list */
				/* link to end of list */

				p_rd_new->p_next = 0;
				p_att->rd_path->p_next = p_rd_new;
				}
			}
		return(TRUE);
		}

	/* Case 3 - more than one node on list 
	* simple insertion search 
	*/

	p_next = p_att->rd_path->p_next;
	p_cur = p_att->rd_path;

	while (p_next)
		{
		res = compare_iso_addr(&p_rd_new->rdi,&p_next->rdi);
		if (res == ISO_ADDR_EQUAL)
			{
			/* equal - rd twice in path */ 
			return (FALSE);
			}
		else 
			{
			if (res == ISO_ADDR_LESS_THAN)
				{
				/* new iso address less than next ISO 
				 * compare against current address 
				 * should it fit between current and next?
				 */ 

				res = compare_iso_addr(&p_rd_new->rdi,&p_cur->rdi);
				if ( res == ISO_ADDR_EQUAL)
					{
					/* equal rd -twice in path */

					return(FALSE);
					}

				if (res == ISO_ADDR_LESS_THAN)
					{
					/* less than current address */

					if (p_cur == p_att->rd_path)
						{
						/* current begins list 
						 * link to front of list
						 */  
						p_rd_new->p_next = p_cur;
						p_att->rd_path = p_rd_new;
						return(TRUE); 
						}
					else
						{
						/* not front of list */
						trace(TR_IDRP, 0, "link-rd_path routine broken");
						return(FALSE);
						}
					}
				if (res == ISO_ADDR_GREATER_THAN)
					{
					p_rd_new->p_next = p_next;
					p_cur->p_next = p_rd_new;
					return(TRUE);
					}
				} /* end of if (res == ISO_ADDR_LESS_THAN) next */
			}	/* end if else */

		p_cur = p_next;
		p_next = p_next->p_next;
		}  /* end of while loop to find next */

	/* link to end of the chain */

	p_rd_new->p_next = (idrp_canon_rdpath *) 0;
	p_att->rd_path->p_next = p_rd_new;
	return (TRUE);

}

/* find_attr_ann_list
 * - find the attribute in the announce list entry
 */

idrp_ann_list *
find_attr_ann_list(p_ann_list,p_att)
idrp_ann_list	*p_ann_list;
idrp_attribute_record 	*p_att;
{
idrp_ann_list	*p_atl;

	ANN_LIST(p_atl,p_ann_list)
		{
		if (p_atl->p_attr == p_att)
			return(p_atl);
		} ANN_LIST_END;

	return (0);

}		


/* find_attr_rec (idrpPeer *peer, idrp_attribute_record *attr)
 *  
 * look down link list of attributes for attribute that matches
 *	Match : 1) attribute mask 
 *		2) canonical represent of RDs
 *		3) full Attributes
 *
 */

idrp_attribute_record *
find_attr_rec(peer, attr,local)
idrpPeer *peer;
idrp_attribute_record *attr;
int	local;	
{
/* walk down chain of attributes */
/* linked in match of attribute mask */

idrp_attribute_record 	*p_attr;	/* attribute record */
int			found;		/* found the attribute record */

	for (p_attr = idrp_attr_list; p_attr; p_attr = p_attr->next)
		{
		if (p_attr->idrp_mask == attr->idrp_mask)
			{
			/* compare the next_hop address */

			/* add in here later a compare on the
			 * canonical RD path -
			 * skh --- (3/25/93) - we need to
			 * to add in the effieciency of
			 * a list of canonical rd_paths
			 *-- I am leaving this space
			 * efficiency out until
			 * the base protocol is working
			 */

			found = FALSE; 
			if (!attr->next_hop)
				{
				found = compare_atts(p_attr,attr);
				}
			else if (sockaddrcmp(p_attr->next_hop,attr->next_hop))
				{
				found = compare_atts(p_attr,attr);	
				}

			if (found)
				{ 
				if (local && p_attr->local_mask)
					{
					/* local attribute match */ 
					return(p_attr);
					}
				if (!local && p_attr->local_mask == 0)
					{
					/* return match at external attribute */
					return(p_attr);
					}
				}
			}
		}
	return (0);
}

/*
 * compare_atts(p_att1,p_att2)
 *
 * 	- compare two attribute records
 *        attribute by attribute 
 * 	  returns 1 if match 
 *	  returns 0 if do not match 
 */ 
u_int compare_atts(p_att1,p_att2)
idrp_attribute_record *p_att1;
idrp_attribute_record *p_att2;
{
idrp_attribute_record *p_att;
u_int att_type;


	/* walk through all types */
		
	for (att_type = 1; att_type < IDRP_ATTR_MAX + 1; att_type++)
		{
		if (p_att1->attrib[att_type].present)
			{
			switch (att_type)
				{
				/* see that attribute data matches */
						
				case IDRP_ATTR_RD_PATH:
				case IDRP_ATTR_NEXT_HOP:
				case IDRP_ATTR_DIST_LIST_INCL:
				case IDRP_ATTR_DIST_LIST_EXCL:
				case IDRP_ATTR_MULTI_EXIT_DISC:
					{
					if (p_att1->attrib[att_type].length  !=
					    p_att2->attrib[att_type].length)
						{
						return (0);
						}
					if (bcmp(p_att1->attrib[att_type].data, 
					     p_att2->attrib[att_type].data, 
					     p_att1->attrib[att_type].length))
						{
						return (0);
						}
					}
					break;
			
				/* distinquishing attributes cases here */ 	
				}  /* end switch */
			} /* end if */
		}   /* end for loop */

	return (1);
}

int idrp_dif_path(p_att1,p_att2)
idrp_attribute_record *p_att1;
idrp_attribute_record *p_att2;
{
int	mask = 0;		/* mask for return what doesn't match */ 
int	att_type;		/* attribute type for loop */
	
	for (att_type = 1; att_type < IDRP_ATTR_MAX + 1; att_type++)
		{
		if (p_att1->attrib[att_type].present)
			{
			switch (att_type)
				{
				/* see that attribute data matches */
						
				case IDRP_ATTR_RD_PATH:
				case IDRP_ATTR_NEXT_HOP:
				case IDRP_ATTR_DIST_LIST_INCL:
				case IDRP_ATTR_DIST_LIST_EXCL:
				case IDRP_ATTR_MULTI_EXIT_DISC:
					{
					if (p_att1->attrib[att_type].length  !=
					    p_att2->attrib[att_type].length)
						{
						mask = mask | idrpAttrBits[att_type];	
						}
					if (bcmp(p_att1->attrib[att_type].data, 
					     p_att2->attrib[att_type].data, 
					     p_att1->attrib[att_type].length))
						{
						mask = mask | idrpAttrBits[att_type];
						}
					}
					break;
			
				/* distinquishing attributes cases here */ 	
				}  /* end switch */
			} /* end if */
		}   /* end for loop */

	return (mask);
}


/* zero attributes */
/* jgs -- why the use of all CAPS?  This conventionally indicates
 * jgs -- a macro rather than a func.
 * skh -- need to pull ATT_REC_ZERO into macro -(2/26/93)
 * skh -- caps indicate this should be done after debugging
 * skh -- or decide that this call is OK and
 * skh -- change to small letters 
  */
void ATTS_REC_ZERO(p_atts)
idrp_attribute_record *p_atts;
{
int i;

	for (i = 0; i < IDRP_ATTR_MAX; i++) 
		{ 
                p_atts->attrib[i].present = FALSE;
 		p_atts->attrib[i].flags = (u_char) 0; 
 		p_atts->attrib[i].length = 0; 
 		p_atts->attrib[i].data = (u_char *) 0;  
 		} 
        p_atts->idrp_mask = 0;
	return;
}


/* idrp_find_routeid_chain_id(p_att,id,peer)
 * - find routeid chain pointer based on id
 * - free entry along with nlris if
 * the memory is tasks
 * local routeid_chains should go through
 * the local memory freeing
 * - NOTE this route must only
 *   be called at parse_cleanup - no initial
 *   output lists
 *   OR after the output list have been relinked 
 * 
 */

void
idrp_free_routeid_chain_id(p_att,id,peer)
idrp_attribute_record *p_att;
int 	id;
idrpPeer	*peer;
{
idrpRoute_entry		*p_rte;
idrpRoute_entry		*p_rte_last = 0;
idrpRoute_entry		*p_found = 0;
int			i;

	if (p_rte->mem_type == IDRP_LOC_MEM)
		{
		trace (TR_IDRP,0,"idrp_free_routeid_chain_id called with local memory, not freed - peer %",peer->name);
		return;
		}

	p_rte = p_att->route_id_list;
	while (p_rte)
		{
		/* look for route_id match */

		if (id == p_rte->route_id && (peer == p_rte->peer))
			{
			/* got a match - free the associated nlris
		 	 * then relink route_id and free
			 */ 

			idrp_free_nlri_chain(p_rte);
			if (p_rte_last)
				{
				/* relink the route_id list */
				p_rte_last->p_next = p_rte->p_next;
				}
			else
				{
				p_att->route_id_list = p_rte->p_next;	
				}
			task_mem_free(peer->task,p_rte);
			return;
			}
		p_rte_last = p_rte;
		p_rte = p_rte->p_next;
		}
	
	return;
}



/*  idrp_find_routeid_chain(p_rt,p_last_rte)
 *  - find routeid chain pointer for this idrpRoute
 *
 */
idrpRoute_entry *
idrp_find_routeid_chain(p_rt,p_last_rte)
idrpRoute *p_rt;	/* pointer to route */
idrpRoute_entry	**p_last_rte;
{
idrpRoute_entry		*p_rte;
idrpRoute_entry		*p_rte_last = 0;
idrpRoute_entry		*p_found = 0;
int			nlri_id;

	/* set the nlri id value from AF.ISO value */
	nlri_id = family_to_nlri_id(p_rt->family);

	/* walk down list of route_ids looking for match */
	p_rte = p_rt->p_attr->route_id_list; 
	while (p_rte)
		{
		/* look for route_id match */

		if (p_rt->route_id_in == p_rte->route_id && 
			(p_rt->peer == p_rte->peer))
			{
			p_found = p_rte;
			*p_last_rte = p_rte_last;
			break;
			}
		p_rte_last = p_rte;
		p_rte = p_rte->p_next;
		}

return (p_found);
}

/* idrp_free_nlri_att_rec(p_rt)
 * - free nlri from chain
 *   (with indicates whether chain is linked through
 *	p_with or p_nlri_next
 */ 
void
idrp_free_nlri_att_rec(p_rt,link)
idrpRoute	*p_rt;
int		link;	  /* how linked */ 
{
idrpRoute_entry	*p_rte,*p_last_rte;
idrpRoute	*p_route;
idrpRoute	*p_last = 0;
idrp_attribute_record 	*p_att;
idrp_canon_rdpath	*p_rd,*p_rd2;

int	found = FALSE;
int	list_empty = TRUE;
int	i;
int	nlri_id;

	/* found a route_id match, walk down nlri */
	/* list looking for the route pointer */

	nlri_id = family_to_nlri_id(p_rt->family);
	p_rte = idrp_find_routeid_chain(p_rt,&p_last_rte); 

	/* cannot find id */
	if (!p_rte)
		{
		trace(TR_IDRP, LOG_ERR, "IDRP idrp_free_nlri_att_rec error -- can't find route ID chain for NLRI %A",
			p_rt->p_rt->rt_dest);
		return;
		}

	/* look for the nlri inside the route_id chain */

	p_route = p_rte->p_route[nlri_id];
	while (p_route) 
		{
		if (p_route == p_rt)
			{
			found = TRUE;
			break;
			}
		p_last = p_route;
		switch (link)
			{
			case NLRI_LINKED_P_NEXT:
				p_route = p_route->p_next_nlri;
				break;

			case NLRI_LINKED_P_WITH:
				p_route = p_route->p_with;
				break;

			case NLRI_LINKED_P_ANN:
				p_route = p_route->p_ann_nlri;
				break;

			case NLRI_LINKED_P_MINADV:
				p_route = p_route->p_min_adv;
				break;

			case NLRI_LINKED_P_MIN_ADV_RD:
				p_route = p_route->p_min_advRD;
				break;
			}	
		}

	/* check to see if the route was found */

	if (!found)
		{
		return;
		}	

	/* found - so let's see if we can relink the attribute
	 * record's route id list around it
	 */

	if (p_last)
		{
		/* more on route_id chain */ 
		switch (link)
			{
			case NLRI_LINKED_P_NEXT:
				p_last->p_next_nlri = p_route->p_next_nlri;
				break;

			case NLRI_LINKED_P_WITH:
				p_last->p_with = p_route->p_with;
				break;

			case NLRI_LINKED_P_ANN:
				p_rte->p_route[nlri_id] = p_route->p_ann_nlri;
				break;

			case NLRI_LINKED_P_MINADV:
				p_rte->p_route[nlri_id] = p_route->p_min_adv;
				break;

			case NLRI_LINKED_P_MIN_ADV_RD:
				p_rte->p_route[nlri_id] = p_route->p_min_advRD;
				break;
			}
		}
	else	
		{
		/* first entry in the chain
		 * - re-link 2nd entry (which may be null)
		 */

		switch (link)
			{
			case NLRI_LINKED_P_NEXT:
				p_rte->p_route[nlri_id] = p_route->p_next_nlri;
				break;

			case NLRI_LINKED_P_WITH:
				p_rte->p_route[nlri_id] = p_route->p_with;
				break;

			case NLRI_LINKED_P_ANN:
				p_rte->p_route[nlri_id] = p_route->p_ann_nlri;
				break;

			case NLRI_LINKED_P_MINADV:
				p_rte->p_route[nlri_id] = p_route->p_min_adv;
				break;

			case NLRI_LINKED_P_MIN_ADV_RD:
				p_rte->p_route[nlri_id] = p_route->p_min_advRD;
				break;
			}
		}

	/* check for empty chains and remove route_id */

	NLRI_FAMILY_LOOP(i)
		{
		if (p_rte->p_route[nlri_id])
			{
			list_empty = FALSE;
			break;
			}
		}	

	if (list_empty)
		{
		/* relink around empty list */

		if (p_last_rte)
			{
			p_last_rte->p_next = p_rte->p_next;
			}


		if (p_rte->mem_type == IDRP_TASK_MEM)
			{
			/* free memory out of task */
			task_mem_free(p_rte->peer->task,p_rte);
			}
		else
			{	
			/* free out of local memory */

			idrp_free_local_idrpRoute_ent(p_rte);
			}
		}

	if (found)
		{
		/* decrement reference count on attribute block */ 
		p_rt->p_attr->ref_cnt--;
		if (p_rt->p_attr->ref_cnt == 0)
			{
			p_att = p_rt->p_attr;
			if (p_att->mem_type == IDRP_TASK_MEM && p_att->local_mask == IDRP_LOCAL_NO_PDU)
				{
				idrp_free_local_attrib_task(p_att);
				}
			else	
				{
				task_mem_free(p_att->peer_alloc->task,p_att->p_attr_pdu);
				}

			task_mem_free(p_att->peer_alloc->task,p_att->next_hop);

			/* FREE rd canonical list */

			for (p_rd = p_att->rd_path;p_rd; p_rd = p_rd2)
				{
				p_rd2 = p_rd->p_next;
				task_mem_free(p_att->peer_alloc->task,p_rd);  
				}

			/* finally let's free the attribute record
			 * - after we remember to relink around it
			 */	
	

			relink_free_att(p_att);
			}
		}
						
}
								 
/*
 * free_attr_attr_record
 *
 * 	free attribute record from this use 
 */  
void idrp_free_attr_rec(peer,attr,id)
idrpPeer *peer;
idrp_attribute_record *attr;
int	id;
{
idrp_attribute_record	*p_attr;
idrp_canon_rdpath	*p_rd;

	/* this routine is only called with an unlinked attribute record
	 * with no nlris attached 
	 * - if you get here without - big problems
	 * refcnt should be zero 
	 */

	if (id != IDRP_INIT_ROUTE_ID)
		{
		if (attr->ref_cnt)
			{
			trace (TR_IDRP | TR_UPDATE, 0, "Big time error peer %s has tried to free attribute record %s with route_id to init and ref_cnt = %d", 
			peer->name,attr,attr->ref_cnt);
			return;
			}
		task_mem_free(peer->task,attr->route_id_list);
		task_mem_free(peer->task,attr);
		return;
		}

 
	if (attr->ref_cnt)
		{
		trace(TR_IDRP,0,"attribute ref cnt = 0 peer %s",peer->name);
		return;
		}

	/* free the next hop memory if it is there */
	if (attr->next_hop)
		{
		task_mem_free(peer->task,attr->next_hop);
		}

	/* free cannonical rd_path */

	p_rd = attr->rd_path;

	while (p_rd)	
  		{
		task_mem_free(peer->task,p_rd);	
		p_rd = p_rd->p_next;
		}

	/* free the pdu */

	if (attr->p_attr_pdu)
		{
		task_mem_free(peer->task,attr->p_attr_pdu);
		}

	/* at last free the memory block */	
					
	task_mem_free(peer->task,attr);

}

/* relink around this attribute and free it */ 

void
relink_free_att(p_att)
idrp_attribute_record 	*p_att;
{
idrp_attribute_record 	*p_attr;

	if (idrp_attr_list == p_att)
		{
		/* attribute at begining of list */	

		idrp_attr_list = p_att->next;
		idrp_attr_list->prev = (idrp_attribute_record *) 0;
		}
	else
		{
		/* attribute mid or end list */

		p_attr = p_att->prev;
		p_attr->next = p_att->next;
		if (p_att->next)
			{
			p_att->next->prev = p_attr;
			}
		}
	if (p_att->mem_type == IDRP_LOC_MEM)
		{
		free_local_attr_record(p_att);
		}
	else
		{
		task_mem_free(p_att->peer_alloc->task,p_att);
		}
}

void
link_att_list(p_atts)
idrp_attribute_record	*p_atts;
{
idrp_attribute_record	*p_att,*p_last_att;


	if (idrp_attr_list)
		{
		/* walk through the attribute list look for the end */
		ATTR_LIST(p_att)
			{
			p_last_att= p_att;
			}
		
		p_last_att->next = p_atts;
		p_atts->prev = p_last_att;
		}
	else
		{
		idrp_attr_list = p_atts;
		idrp_attr_list->prev = (idrp_attribute_record *) 0;
		} 
	p_atts->next = (idrp_attribute_record *) 0;
}	

 /* free the nlris on the found chain and relink around it */

void
idrp_free_nlri_chain(p_ire)
idrpRoute_entry	*p_ire;
{
idrpRoute	*p_idrp_rt,*p_idrp_rt2;
int		i;
idrpPeer	*peer;

	peer = p_ire->peer;

	if (p_ire->mem_type == IDRP_LOC_MEM)
		{
		trace (TR_IDRP,0,"idrp_free_nlri_chain: peer %s IDRP local routes should disappear via reconfig not mid stream",peer->name);
		return;
		}	

	NLRI_FAMILY_LOOP(i)
		{
		/* Is there any idrp routes */

		p_idrp_rt = p_ire->p_route[i];
		while (p_idrp_rt)
			{
			/* free the idrp routes
			 * - this should be linked by p_next_nlri
			 * - all list pointers should have
			 * already been taken care of 
			 * - warning if you do not,
			 * - things will screw up big
			 * time.
			 * --skh(3/27/93) my thought is
			 * -- it is too pendatic to 
			 * --- check each time we use this
			 * -- to watch lists.  But I may be
			 * -- wrong. 
			 */
			p_idrp_rt2 = p_idrp_rt;
			p_idrp_rt = p_idrp_rt->p_next_nlri;
			task_mem_free(peer->task,p_idrp_rt);
			}
		} NLRI_FAMILY_LOOP_END;
}
