#ifdef	PROTO_IDRP

#include "include.h"
#include "idrp.h"
#include <stdio.h>
#include "md4global.h"	/* RSAREF globals and constants */
#include "md4.h"	/* md4-specific includes */

/* idrp.c 
 *
 * miscellaneous functions.  It may be appropriate to move some to other
 * locations.
 *	drop_packet - drop transmit packet for testing
 *	idrp_trace  - idrp trace function
 *	log_nm_event - log network management event 
 *	free_buffer - only for buffers allocated from task
 *		not for receive buffer, but I'm not sure about this
 * 	path_up(peer) - send all routes to new neighbor
 *	proto_to_family(proto) - ???
 *	family_to_nlri_id(family) - ???
 *	idrp_send_dest_set(peer) - ??? I have NO idea!
 */ 
  
 
/*
 * Decide whether to drop a transmitted packet for fun. %%
 */
static drop_packet()
{
	int dropit = ((rand() % 10) == 0);
	if (dropit)
		trace(TR_IDRP, 0, "IDRP xmit packet dropped");
	return(dropit);
}


/*
 * Trace an IDRP packet
 */
static void
idrp_trace(peer, comment, send_flag, pdu, length)
idrpPeer *peer;
char *comment;
int send_flag;
idrpPdu *pdu;
int length;
{
	trace(TR_IDRP, 0, "idrp_trace called");
}



static void log_nm_event(peer, type, value)
idrpPeer *peer;
int type;
int value;
{
	trace(TR_IDRP, 0, "IDRP mgmt event %s, peer %s, value %d",
		idrpManagementEvents[type], peer->name, value); /* %% */
};


/*
 * Free a buffer
 */
static void free_buffer(peer, buf)
idrpPeer *peer;
idrpBuffer *buf;
{
	task_mem_free(peer->task, buf->pdu);
	task_mem_free(peer->task, buf);
}


/*
 * Bring up the path to our neighbor
 * jgs -- where to put this?
 */
static void path_up(peer)
idrpPeer *peer;
{
	trace(TR_ALL, LOG_ERR, "IDRP path to %s is up", peer->name);
	send_keepalive(peer);
	start_keepalive_timer(peer);
#ifdef ECHOKLUDGE
	start_echo_timer(peer);
#endif
	idrp_rt_send_init(peer);
}


int proto_to_family(proto)
int proto;
{
	/* ISO only until IP gets encoded */

	return (AF_ISO);
}


u_int family_to_nlri_id(family)
int family;
{
	/* ISO only until IP gets encoded */

	return (IDRP_NLRI_ISO);

}


sockaddr_un *idrp_send_dest_set(peer)
idrpPeer *peer;
{
sockaddr_un	*p_dst;
	switch(peer->proto_sock)
		{
		case IDRP_PDU_PROTO_UDP:
			p_dst = peer->gw.gw_addr;
			break;
		case IDRP_PDU_PROTO_IP_RAW:
			p_dst = peer->ip_intf_addr;
			break;

		case IDRP_PDU_PROTO_IDRP:
		case IDRP_PDU_PROTO_CLNP:
			p_dst = (sockaddr_un *) &peer->iso_gw_addr;
			break;
		}
	return(p_dst);
}
			

/* 
 * Function: md4
 * Summary: Compute the 16-byte md4 digest of the buffer.
 * Status: Thought to work; not finalized.
 * Author: jgs
 * Date: Mon Dec 07 12:47:00 1992
 * Notes:
 *   The md4 code can probably be tuned to be faster.  The code as included
 * is exactly the RFC 1320 code, which is said to sacrifice speed for
 * portability.  I have made no attempt to speed it up.
 */

u_char * 
md4 __PF3(buffer, u_char *,	/* buffer to be md4-digested */
	  length, u_int,	/* length of buffer in bytes */
	  digest, u_char *)	/* 16 bytes for md4 digest */
{
	MD4_CTX context;

	/* These all call functions in md4c.c, straight from RFC 1320 */
	MD4Init (&context);
	MD4Update(&context, buffer, length);
	MD4Final(digest, &context);

	return(digest);
}

#endif /* PROTO_IDRP */
