#ifdef	PROTO_IDRP

#include "include.h"
#include "idrp.h"
#include <stdio.h>



/* idrp_attrib.c
 *  
 *  attribute validation and handling routines
 *  (rd path handling and validation takes up lots of the module)
 *  there also seems to be a bunch of route-id chain handling here.
 *  perhaps there should be another chain for route-id handling?
 * 	valid_attr(peer,attr) - validate attribute record 
 *	valid_usr_attr(peer,attr) - user attributes
 *	add_local_rd_to_path(peer,attr_rec) - add our rd to the rd path
 *		also detects routing loops
 *	valid_seg_type(type,peer) - ??? something to do w/ rd path
 *	valid_rd_path(peer,att_rec,p_rdpath,rdpath_length) - ??? validate rd
 *		path, I guess
 *	link_rd_canon(p_rd_new,p_att) - link rds in canonical order
 * 	find_attr_rec - search for duplicate attribute record in attribute list  
 *     	compare_atts  - compare attributes
 *	idrp_dif_path - see if different path attribute- return
 *			mask of attributes that are different 
 *	ATTS_REC_ZERO(p_atts) - zero out path atts.  Maybe this should be a 
 *		macro and not a func.
 *	idrp_find_routeid_chain(p_rt) - find routeid chain pointer for this idrpRoute
 *	idrp_free_nlri_att_rec(p_rt) - free nlri from chain
 *      idrp_free_attr_rec - free attribute record
 */ 
  
 

/*		
 * valid_attr - check for valid attributes 
 */
int valid_attr(peer,att_rec)
idrpPeer *peer;
idrp_attribute_record *att_rec;
{
int attr_type;
int attr_len;
u_int	longhold;
u_char *dataptr;
u_char *rdpath;
int	rdpath_length;
short seglen, seglen_local;
struct	iso_net_addr *addr;
int	attr_present = 0;
idrp_canon_rdpath *p_rd_canon_list = (idrp_canon_rdpath *) 0;
idrp_canon_rdpath *p_rd = (idrp_canon_rdpath *) 0;

	/* now validate the attribute sections of UPDATE */

	/* (7.21.3 - d) check for well know attributes  */

	if ((att_rec->idrp_mask & IDRP_ATTBT_MANDATORY) != IDRP_ATTBT_MANDATORY)
		{
		peer->last_error_pdu.error_code = IDRP_UPDATE_ERROR;
		peer->last_error_pdu.error_subcode = IDRP_UPDATE_ERROR_MISS_WELKN_ATTR;
		if (att_rec->idrp_mask & IDRP_ATTBT_ROUTE_ID == 0 ) 
			{
			peer->last_error_pdu.data = (u_char *) &idrpAttrType[IDRP_ATTR_ROUTE_ID]; 
			}
		else	{
			if (att_rec->idrp_mask & IDRP_ATTBT_RD_PATH == 0)
				{
				peer->last_error_pdu.data = (u_char *) &idrpAttrType[IDRP_ATTR_RD_PATH];
				}
			else
				{
				peer->last_error_pdu.data = (u_char *) &idrpAttrType[IDRP_ATTR_HOP_COUNT];
				}
			}	
		peer->last_error_pdu.data_len = 1;
		return(FALSE);
		}

	/* 7.21.3 i) check that DIST_LIST_INCL and DIST_LIST_EXCL  */
	/*           are not both active */

	if ((att_rec->idrp_mask & IDRP_ATTBT_DIST_LIST_INCL) && 
	    (att_rec->idrp_mask & IDRP_ATTBT_DIST_LIST_EXCL))
		{
		/* both IDRP DIST_LIST_INCL and IDRP_DIST_LIST_EXCL */
		/* exist - malformed attribute list error */
		
		peer->last_error_pdu.error_code = IDRP_UPDATE_ERROR;
		peer->last_error_pdu.error_subcode = IDRP_UPDATE_ERROR_MALF_ATTR;
		peer->last_error_pdu.data = NULL;
		peer->last_error_pdu.data_len = 0;
		return (FALSE);
		} 

	/* all mandatory attribute there - check flag codes and each attribute */

	for (attr_type = 1; attr_type < IDRP_ATTR_MAX + 1; attr_type++) {
		dataptr = att_rec->attrib[attr_type].data;
		attr_len = att_rec->attrib[attr_type].length;

		/* if  attribute not present skip all the tests */

		if (!att_rec->attrib[attr_type].present)
			continue;

		trace(TR_IDRP | TR_UPDATE, 0, "  Attr %d, length %d, flags %x", attr_type,
			att_rec->attrib[attr_type].length, att_rec->attrib[attr_type].flags);

		/* validate attribute flags  (idrp 7.21.3 - b)  */
		if (att_rec->attrib[attr_type].flags != idrpAttrFlags[attr_type])
			{
			peer->last_error_pdu.error_code=IDRP_UPDATE_ERROR;
			peer->last_error_pdu.error_subcode = IDRP_UPDATE_ERROR_ATTR_FLAGS_ERR;
			peer->last_error_pdu.data = att_rec->attrib[attr_type].data;
			peer->last_error_pdu.data_len = att_rec->attrib[attr_type].length;
			return(FALSE);
			}

		att_rec->idrp_mask += idrpAttrBits[attr_type];  

		switch(attr_type) {

		/* ROUTE_ID */
		case IDRP_ATTR_ROUTE_ID:
			if (attr_len != sizeof(route_id))
				{ 
				peer->last_error_pdu.error_code=IDRP_UPDATE_ERROR;
				peer->last_error_pdu.error_subcode = IDRP_UPDATE_ERROR_ATTR_LEN_ERR;
				peer->last_error_pdu.data = att_rec->attrib[attr_type].data;
				peer->last_error_pdu.data_len = att_rec->attrib[attr_type].length;
				return(FALSE);
				}
			bcopy(dataptr, (caddr_t) &longhold,sizeof(route_id));
			htonl(longhold);
			trace(TR_IDRP | TR_UPDATE, 0, "   ROUTE_ID 0x%x",longhold );
			att_rec->route_id_list->route_id=longhold;
			trace(TR_IDRP | TR_UPDATE, 0, "   ROUTE_ID 0x%x",att_rec->route_id_list->route_id);
			break;			

		case IDRP_ATTR_EXT_INFO:
			if (attr_len != IDRP_EXT_INFO_LENGTH)
				{ 
				peer->last_error_pdu.error_code=IDRP_UPDATE_ERROR;
				peer->last_error_pdu.error_subcode = IDRP_UPDATE_ERROR_ATTR_FLAGS_ERR;
				peer->last_error_pdu.data = att_rec->attrib[attr_type].data;
				peer->last_error_pdu.data_len= att_rec->attrib[attr_type].length;
				return(FALSE);
				}
			break;

		/* RD Path.  Note its position so that we can copy it into our route. */

		case IDRP_ATTR_RD_PATH:
			if (valid_rd_path(peer,att_rec,dataptr,attr_len))
				{
				/* add in the local rd path */

				if (!add_local_rd_to_path(peer,att_rec))
					{
					return (FALSE);
					}
				}
			else
				{
				return (FALSE);
				}	
			
			break;

		case IDRP_ATTR_HOP_COUNT:
			trace(TR_IDRP | TR_UPDATE, 0, "   HOP_COUNT %d", *dataptr);
			att_rec->hopcount = *dataptr;
			break;

		} /* end of switch statement */
	}	/* end of for statement checking all attributes */

return(TRUE);
}

int valid_usr_attr(peer,att_rec)
idrpPeer *peer;
idrp_attribute_record *att_rec;
{
	/* null for now */
	return(1);

}

/* 
 *  add_local_rd_to_path(peer,attr_rec))
 * 
 * put the local RD in the rd_path by adding it into the 
 * locally added attributes portion of the attribute field
 * LOCAL_RD_PATH attribute of attributes 
 * 
 * link the attribute to the cannonical attribute list
 * if it fails to link, the we have a loop
 * 
 * return code = 0 failed
 * return code = 1 worked  
 */
int  add_local_rd_to_path(peer,attr_rec)
idrpPeer *peer;
idrp_attribute_record *attr_rec;
{
u_int	hold;
idrp_canon_rdpath *p_rd;


	/* add in our RDI length + (1) RDI length byte + segement header length  */
	/*  to the total RD path length */

	hold = sizeof(idrp_path_segment_header) + rt_rdi.isoa_len + 1;
	attr_rec->attrib[IDRP_ATTR_LOCAL_RD].length = hold;
 	hold += attr_rec->attrib[IDRP_ATTR_RD_PATH].length;
        htons(hold);

	/* modify Attribute length */

        bcopy((caddr_t) &hold,&attr_rec->attrib[IDRP_ATTR_RD_PATH].length, 2); 
        attr_rec->attrib[IDRP_ATTR_LOCAL_RD].present = TRUE;
	attr_rec->attrib[IDRP_ATTR_LOCAL_RD].data = (u_char *) &local_rd_path;

	p_rd = (idrp_canon_rdpath *) task_mem_malloc(peer->task,sizeof(idrp_canon_rdpath));  
	bcopy(&rt_rdi,&p_rd->rdi,sizeof(struct iso_net_addr));
	if (!link_rd_canon(p_rd,attr_rec)) 
		{
		peer->last_error_pdu.error_code = IDRP_UPDATE_ERROR;
		peer->last_error_pdu.error_subcode = IDRP_UPDATE_ERROR_RD_ROUTING_LOOP;
		peer->last_error_pdu.data = (u_char *) &rt_rdi.isoa_genaddr; 
		peer->last_error_pdu.data_len = rt_rdi.isoa_len;      		
		return(FALSE);
		}
}
	
int valid_seg_type(type,peer)
int type;
idrpPeer *peer;
{
	switch (type)
		{
		case IDRP_PATH_SEGMENT_SET:
		case IDRP_PATH_SEGMENT_SEQ:
			return (TRUE);
			break;

		case IDRP_PATH_SEGMENT_ENTRY_SEQ:
		case IDRP_PATH_SEGMENT_ENTRY_SET:
			trace (TR_IDRP | TR_UPDATE,0,"IDRP peer %s got RD_PATH with RDI SET/SEQ",
				peer->name);
			return (TRUE);
			break;	

		}
	return (FALSE);
}


int valid_rd_path(peer,att_rec,p_rdpath,rdpath_length)
idrpPeer *peer;
idrp_attribute_record *att_rec;
u_char	*p_rdpath;
int	rdpath_length;
{
int	pathlen;
int	seglen;
int	seglen_local;
struct	iso_net_addr	*addr;
idrp_path_segment_header 	*seghead;
idrp_canon_rdpath	*p_rd;
 
	/* clear rd path list pointer */
 
	att_rec->rd_path = (idrp_canon_rdpath *) 0 ; 
	pathlen = rdpath_length;
	

	/* parse the rd path information */

	seghead = (idrp_path_segment_header *) p_rdpath;

	while (rdpath_length > 0)  {

	/* walk through segment type, length value */ 

		addr = (struct iso_net_addr *) (seghead + sizeof(idrp_path_segment_header));
		bcopy(seghead->path_seg_len, (caddr_t) &seglen, sizeof(seghead->path_seg_len));
		ntohs(seglen);
		seglen_local = seglen;
		trace(TR_IDRP | TR_UPDATE, 0, "   RD_PATH type %d, length %d",
			seghead->path_seg_type,seglen);

		if (!valid_seg_type(seghead->path_seg_type, peer))
			{ 
			/* illegal type in the segement type */
			trace(TR_IDRP | TR_UPDATE, 0, "RD PATH type %d illegal peer %s",
			seghead->path_seg_type,peer->name);
			/* jgs -- why do we set error_code and then immediately reset it??? */
			peer->last_error_pdu.error_code = IDRP_UPDATE_ERROR;
			peer->last_error_pdu.error_subcode = IDRP_UPDATE_ERROR_INV_SEG_TYPE;
			peer->last_error_pdu.data = &seghead->path_seg_type;
			peer->last_error_pdu.data_len = sizeof(seghead->path_seg_type);	 
			return(FALSE);
			}
	
		/* valid segment type,  walk thorugh the RDI list */
 
		while (seglen_local > 0) {
			trace(TR_IDRP | TR_UPDATE, 0, "     %s", iso_netiso_ntoa(addr));

			/* build canonical rd path */
			p_rd = (idrp_canon_rdpath *) task_mem_malloc(peer->task,sizeof(idrp_canon_rdpath));  
			bcopy(addr,&p_rd->rdi,sizeof(struct iso_net_addr));
			if (!link_rd_canon(p_rd,att_rec)) 
				{
				peer->last_error_pdu.error_code =IDRP_UPDATE_ERROR;
				peer->last_error_pdu.error_subcode = IDRP_UPDATE_ERROR_RD_ROUTING_LOOP;
				peer->last_error_pdu.data = (u_char *) &addr->isoa_genaddr; 
				peer->last_error_pdu.data_len = addr->isoa_len;      		
				return(FALSE);	
				}
			seglen_local -= addr->isoa_len + 1;
			addr += addr->isoa_len + 1;
			}
		/* set up for the next segment */

		pathlen -= seglen + sizeof(idrp_path_segment_header);
		seghead = (idrp_path_segment_header *) addr;

		} /* end of loop for the full rd_path */
	return(TRUE);
}

/* 
 * link_rd_canon(p_rd,att)
 *  - link RDIs in cannonical order 
 */
int link_rd_canon(p_rd_new,p_att)
idrp_canon_rdpath	*p_rd_new;
idrp_attribute_record	*p_att;
{
/* local variables */

idrp_canon_rdpath	*p_rd;		/* pointer to rd path entry */
idrp_canon_rdpath	*p_cur;		/* pointer to rd path entry */
idrp_canon_rdpath	*p_next;	/* pointer to rd path entry */
u_int			len;		/* length in bits of match */
int			res;		/* result */

	/* list empty */

	if (p_att->rd_path == (idrp_canon_rdpath *)0 )
		{
		p_rd_new->p_next = (idrp_canon_rdpath *) 0;
		p_att->rd_path = p_rd_new;
		return(TRUE);
		}

	if (p_att->rd_path->p_next == (idrp_canon_rdpath *) 0)
		{
		/* just one on the list link this one in before or after */
		
		res = compare_iso_addr(&p_rd_new->rdi,&p_att->rd_path->rdi);


		if (res == ISO_ADDR_EQUAL)
			{
			/* iso address equal - RD twice in pathway */
			return(FALSE);
			}
		else 	{
			if (res == ISO_ADDR_LESS_THAN)
				{
				/* this iso address less than one on list */
				/* this address is new head of list */

				p_rd_new->p_next = p_att->rd_path;
 				p_att->rd_path->p_next = (idrp_canon_rdpath *) 0;
				p_att->rd_path = p_rd_new;
				}	
			else
				{
				/* this iso address greater than one on list */
				/* link to end of list */

				p_rd_new->p_next = 0;
				p_att->rd_path->p_next = p_rd_new;
				}
			}
		}
	else
		{
		/* Case 3 - more than one node on list */
		/* simple insertion search */

		p_next = p_att->rd_path->p_next;
		p_cur = p_att->rd_path;

		while (p_next)
			{
			res = compare_iso_addr(&p_next->rdi,&p_rd_new->rdi);
			if (res == ISO_ADDR_EQUAL)
				{
				/* equal - rd twice in path */ 
				return (FALSE);
				}
			else 
				{
				if (res == ISO_ADDR_LESS_THAN)
					{
					/* next iso address less than next ISO */
					/* compare against current address */
					/* should it fit between current and next? */ 

					res = compare_iso_addr(&p_cur->rdi,&p_rd_new->rdi);
					if ( res == ISO_ADDR_EQUAL)
						{
						/* equal rd -twice in path */

						return(FALSE);
						}
					if (res == ISO_ADDR_LESS_THAN)
						{
						/* less than current address */

						if (p_cur ==  p_att->rd_path)
							{
							/* current begins list */
							/* link to front of list */  
							p_rd_new->p_next = p_cur;
							p_att->rd_path = p_rd_new;
							return(TRUE); 
							}
						else
							{
							/* not front of list */
							trace(TR_IDRP, 0, "link-rd_path routine broken");
							return(FALSE);
							}
						}
					if (res == ISO_ADDR_GREATER_THAN)
						{
						p_rd_new->p_next = p_next;
						p_cur->p_next = p_rd_new;
						return(TRUE);
						}
					} /* end of if (res == ISO_ADDR_LESS_THAN) next */

				p_cur = p_next;
				p_next = p_next->p_next;
				}  /* end of while loop to find next */
		}	/* end of case 3 */

		/* link to end of the chain */

		p_rd_new->p_next = (idrp_canon_rdpath *) 0;
		p_rd->p_next = p_rd_new;
		}
return (TRUE);

}


/* find_attr_rec (idrpPeer *peer, idrp_attribute_record *attr)
 *  
 * look down link list of attributes for attribute that matches
 *	Match : 1) attribute mask 
 *		2) canonical represent of RDs
 *		3) full Attributes
 *
 */
idrp_attribute_record *
find_attr_rec(peer, attr)
idrpPeer *peer;
idrp_attribute_record *attr;
{
/* walk down chain of attributes */
/* linked in match of attribute mask */

idrp_attribute_record 	*p_attr;	/* attribute record */

	for (p_attr = idrp_attr_list; p_attr !=0; p_attr = p_attr->next)
		{
		if (p_attr->idrp_mask == attr->idrp_mask)
			{
			/* mask flags are the same */
			/* test cannonical RD path */ 
		
			if (p_attr->rd_path == attr->rd_path)
				{
				/* pointers match */
				/* compare all the attributes */

				if (compare_atts(p_attr,attr))
					{
					return(p_attr);
					}
				}
			}
		}
	return (0);
}

/*
 * compare_atts(p_att1,p_att2)
 *
 * 	- compare two attribute records
 *        attribute by attribute 
 * 	  returns 1 if match 
 *	  returns 0 if do not match 
 */ 
u_int compare_atts(p_att1,p_att2)
idrp_attribute_record *p_att1;
idrp_attribute_record *p_att2;
{
idrp_attribute_record *p_att;
u_int att_type;


	/* for now duplicate attribute paths */
 
	return (0);
 
	/* walk through all types */
		
	for (att_type = 1; att_type < IDRP_ATTR_MAX + 1; att_type++)
		{
		if (p_att1->attrib[att_type].present)
			{
			switch (att_type)
				{
				/* see that attribute data matches */
						
				case IDRP_ATTR_RD_PATH:
				case IDRP_ATTR_NEXT_HOP:
				case IDRP_ATTR_DIST_LIST_INCL:
				case IDRP_ATTR_DIST_LIST_EXCL:
				case IDRP_ATTR_MULTI_EXIT_DISC:
					{
					if (p_att1->attrib[att_type].length  !=
					    p_att2->attrib[att_type].length)
						{
						return (0);
						}
					if (bcmp(p_att1->attrib[att_type].data, 
					     p_att2->attrib[att_type].data, 
					     p_att1->attrib[att_type].length))
						{
						return (0);
						}
					}
					break;
			
				/* distinquishing attributes cases here */ 	
				}  /* end switch */
			} /* end if */
		}   /* end for loop */

	return (1);
}

int idrp_dif_path(p_att1,p_att2)
idrp_attribute_record *p_att1;
idrp_attribute_record *p_att2;
{
int	mask = 0;		/* mask for return what doesn't match */ 
int	att_type;		/* attribute type for loop */
	
	for (att_type = 1; att_type < IDRP_ATTR_MAX + 1; att_type++)
		{
		if (p_att1->attrib[att_type].present)
			{
			switch (att_type)
				{
				/* see that attribute data matches */
						
				case IDRP_ATTR_RD_PATH:
				case IDRP_ATTR_NEXT_HOP:
				case IDRP_ATTR_DIST_LIST_INCL:
				case IDRP_ATTR_DIST_LIST_EXCL:
				case IDRP_ATTR_MULTI_EXIT_DISC:
					{
					if (p_att1->attrib[att_type].length  !=
					    p_att2->attrib[att_type].length)
						{
						mask = mask | idrpAttrBits[att_type];	
						}
					if (bcmp(p_att1->attrib[att_type].data, 
					     p_att2->attrib[att_type].data, 
					     p_att1->attrib[att_type].length))
						{
						mask = mask | idrpAttrBits[att_type];
						}
					}
					break;
			
				/* distinquishing attributes cases here */ 	
				}  /* end switch */
			} /* end if */
		}   /* end for loop */

	return (mask);
}


/* zero attributes */
/* jgs -- why the use of all CAPS?  This conventionally indicates
 * jgs -- a macro rather than a func. */
void ATTS_REC_ZERO(p_atts)
idrp_attribute_record *p_atts;
{
int i;

	for (i = 0; i < IDRP_ATTR_MAX; i++) 
		{ 
                p_atts->attrib[i].present = FALSE;
 		p_atts->attrib[i].flags = (u_char) 0; 
 		p_atts->attrib[i].length = 0; 
 		p_atts->attrib[i].data = (u_char *) 0;  
 		} 
        p_atts->idrp_mask = 0;
	return;
}



/*  idrp_find_routeid_chain(p_rt)
 *  - find routeid chain pointer for this idrpRoute
 *
 */
idrpRoute_entry *
idrp_find_routeid_chain(p_rt)
idrpRoute *p_rt;	/* pointer to route */
{
idrpRoute_entry		*p_rte;	
idrpRoute_entry		*p_found = 0;
int			nlri_id;

	/* set the nlri id value from AF.ISO value */
	nlri_id = family_to_nlri_id(p_rt->family);

	/* walk down list of route_ids looking for match */
	p_rte = p_rt->p_attr->route_id_list; 

	while (p_rte)
		{
		/* look for route_id match */

		if (p_rt->route_id_in == p_rte->route_id && 
			(p_rt->peer == p_rte->peer))
			{
			p_found = p_rte;
			break;
			}
		p_rte = p_rte->p_next;
		}

return (p_found);
}

/* idrp_free_nlri_att_rec(p_rt)
 * - free nlri from chain
 *   (with indicates whether chain is linked through
 *	p_with or p_nlri_next
 */ 
void
idrp_free_nlri_att_rec(p_rt,with)
idrpRoute	*p_rt;
int		with;	  /* if true, link through withdraw - other p_next-nlri */ 
{
idrpRoute_entry	*p_rte;
idrpRoute	*p_route;
idrpRoute	*p_last = 0;

int	found = FALSE;
int	list_empty = FALSE;
int	i;
int	nlri_id;

	/* found a route_id match, walk down nlri */
	/* list looking for the route pointer */

	nlri_id = family_to_nlri_id(p_rt->family);
	p_rte = idrp_find_routeid_chain(p_rt); 

	/* cannot find id */
	if (!p_rte)
		{
		return;
		}

	/* look for the nlri inside the route_id chain */

	p_route = p_rte->p_route[nlri_id];
	while (p_route ) 
		{
		if (p_route == p_rt)
			{
			found = TRUE;
			/* found re-link */
			if (p_last)
				{
				/* mid list - determine link withdraw/next nlri */
				if (with)
					{
					/* withdraw link */
					p_last->p_with = p_rt->p_with;
					}
				else
					{
					/* next nlri link */
					p_last->p_next_nlri = p_rt->p_next_nlri;
					}
				}
			else
				{
				/* top of list */
				if (with)
					{
					/* withdraw link  */
					p_rte->p_route[nlri_id] = p_rt->p_with;
					}
				else
					{
					/* next nlri link */
					p_rte->p_route[nlri_id] = p_rt->p_next_nlri;
					}
				} 
			}
		if (!found)
			{
			if (with)
				{
				p_route = p_route->p_with;
				}
			else
				{
				p_route = p_route->p_next_nlri;
				}
			p_last = p_route;	
			}
		} 	

	/* check for empty chains and remove route_id */

	list_empty = TRUE;
	NLRI_FAMILY_LOOP(i)
		{
		if (p_rte->p_route[nlri_id])
			{
			list_empty = FALSE;
			}
		}	

	if (list_empty)
		{
		if (p_rte->mem_type == IDRP_TASK_MEM)
			{
			idrp_free_local_idrpRoute_ent(p_rte);
			}
		else
			{	
			/* free memory out of task */

			task_mem_free(p_rt->peer,p_rte);	
			}
		}
	if (found)
		{
		/* decrement reference count on attribute block */ 
		p_rt->p_attr->ref_cnt--;
		if (p_rt->p_attr->ref_cnt == 0)
			{
			if (p_rt->p_attr->mem_type == IDRP_LOC_MEM)
				{
				idrp_free_local_rt(p_rt->p_attr);
				}
			else
				{
				task_mem_free(p_rt->peer->task,p_rt->p_attr);
				}
			}
		}
						
}
								 
/*
 * free_attr_attr_record
 *
 * 	free attribute record from this use 
 */  
void idrp_free_attr_rec(peer,attr,id)
idrpPeer *peer;
idrp_attribute_record *attr;
int	id;
{
idrp_attribute_record	*p_attr;
idrp_canon_rdpath	*p_rd;

	/* decrement reference count */

	if (id == IDRP_INIT_ROUTE_ID)
		{
		if (attr->ref_cnt)
			{
			trace (TR_IDRP | TR_UPDATE, 0, "Big time error peer %s has tried to free attribute record %s with route_id to init and ref_cnt = %d", 
			peer->name,attr,attr->ref_cnt);
			return;
			}
		task_mem_free(peer->task,attr->route_id_list);
		task_mem_free(peer->task,attr);
		return;
		}

	/* local routes, EXT_INFO and received IDRP routes */
	/* should have route_id structure */

 
	if (!attr->route_id_list)
		{
		trace (TR_IDRP | TR_UPDATE,0,"IDRP error peer %s has no route_id list and more than initialized ", 
			peer->name);
		return;
		}

	if (id  == IDRP_LOCAL_ROUTE_ID)
		{
		/* freeing the local route - this better be re-init time */
		if (attr != local_route_attr_rec || peer != &idrp_this_node)
			{
			trace (TR_IDRP | TR_UPDATE,0,"IDRP idrp_free_attr_rec local routes logic error peer %s - freeing local routes erroneously",peer->name);
			return;
			} 
		idrp_free_local_route_chain(attr->route_id_list);
		return;
		}	

	if (id == IDRP_EXT_INFO_ROUTE_ID)
		{
		/* freeing external ext_info route - should be 
		 * done by a different routine
		 * -warn and exit
		 */

		trace (TR_IDRP,0, "IDRP idrp_attr_free tries to remove ext_info attr record peer %s",peer->name);
		return;
		}  	


	/* not external info or local info - decrement ref_cnt
	 * and free up the route_id list 
	 */
 
	attr->ref_cnt--;
	idrp_free_idrpRoute_entry(id,peer,attr->route_id_list,IDRP_ROUTE_LIST_LINKED);

	if (attr->ref_cnt)
		{
		return;
		}
	if (attr->ref_cnt < 0)
		{
		trace(TR_IDRP,0,"attribute ref cnt = 0 peer %s",	
			peer->name);
		return;
		}

	if (attr->route_id_list)
		{
		trace(TR_IDRP,0,"attribute ref cnt = 0 and route_list not empty peer %s",
			peer->name);
		return;
		}

	/* unlink it from attribute list */
	
	if (idrp_attr_list == attr)
		{
		/* attribute at begining of list */	

		idrp_attr_list = attr->next;
		idrp_attr_list->prev = (idrp_attribute_record *) 0;
		}
	else
		{
		/* attribute mid or end list */

		p_attr = attr->prev;
		p_attr->next = attr->next;
		attr->next->prev = p_attr;
		}

	/* free cannonical rd_path */

	p_rd = attr->rd_path;

	while (p_rd)	
  		{
		task_mem_free(peer->task,p_rd);	
		p_rd = p_rd->p_next;
		}

	/* at last free the memory block */	
					
	task_mem_free(peer->task,attr);

}

#endif /* PROTO_IDRP */
