#ifdef	PROTO_IDRP

#include "include.h"
#include "idrp.h"
#include <stdio.h>


/* idrp_parse_pdu.c
 *  
 * Parsing incoming pdus
 *  (process routines validate pdu, and call state machine
 *   with either pdu w/error or pdu w/out error
 *   parse routines - parse in order validate pdu  
 * 
 *	idrp_parse_open(peer,pdu,length) - parse open pdu and validate
 *	process_open(peer,pdu,length) - process open pdu    
 *	parse_update(peer,pdu,length,results) - parse update and validate
 *	parse_update_cleanup(results) - cleanup from the parse update routine after failure
 *	process_update(peer,pdu,length) - process update pdu
 *	process_keepalive(peer,pdu,length) - process keepalive pdu
 *	process_error(peer,pdu,length) - process error pdu
 *	process_cease(peer,pdu,length) - process cease pdu
 *	parse_rib_refresh(peer, pdu, length) - ???
 *	process_rib_refresh(peer,pdu,length) - process rib refresh pdu
 *	process_echo(peer,pdu,length) - process echo pdu
 *	process_echo_reply(peer,pdu,length) - process echo reply pdu 
 *	process_header_error(peer,pdu,length) - process header error   
 *	process_rcv_seq(peeri,pdu)
 *	in_sequence(peer,seq)
 * 	idrp_recv_pdu  - receive idrp pdu 
 *	idrp_recv - receive for master task for idrp protocol
 */ 
  
 

int acked_something;			/* True if last received packet acknowledged something */
					/* jgs -- need to patch this up */


/*
 * parse_open(peer, pdu, length)
 */
int idrp_parse_open(peer, pdu, length)
idrpPeer *peer;
idrpPdu *pdu;
int length;
{
u_short rcv_holdtime;
u_short rcv_maxpdusize;
struct iso_net_addr	rdi_peer;
idrp_open_pdu	*p_pdu;
idrp_open_pdu_header *hdr = (idrp_open_pdu_header *) &(pdu->pdu.open);

	bcopy((caddr_t) hdr->idrp_maxpdu_size, (caddr_t) &rcv_maxpdusize, sizeof(hdr->idrp_maxpdu_size));
	ntohs(rcv_maxpdusize);

	/* a) check the header version */
	if (hdr->version != IDRP_VERSION_1)
		{
		peer->last_error_pdu.error_code = IDRP_ERROR_OPEN_PDU;
		peer->last_error_pdu.error_subcode = IDRP_OPEN_ERROR_UNSUP_VER;
		peer->last_error_pdu.data = (u_char *) 0;
		peer->last_error_pdu.data_len = 0; 
		return(IDRP_ERROR_OPEN_PDU);
		}

	/* b) check the maximum PDU size */

	bcopy((caddr_t) hdr->idrp_maxpdu_size, (caddr_t) &rcv_maxpdusize, sizeof(hdr->idrp_maxpdu_size));
	ntohs(rcv_maxpdusize);
	if (rcv_maxpdusize < MinPDULength) 
		{
		peer->last_error_pdu.error_code = IDRP_ERROR_OPEN_PDU;
		peer->last_error_pdu.error_subcode = IDRP_OPEN_ERROR_BAD_MAX_SIZE;
		peer->last_error_pdu.data = (u_char *) &hdr->idrp_maxpdu_size; 
		peer->last_error_pdu.data_len = sizeof(hdr->idrp_maxpdu_size);
		return(IDRP_ERROR_OPEN_PDU);
		}

	p_pdu = (idrp_open_pdu *) (&pdu->pdu.open);

	/* c) check the RDI */
	
	rdi_peer.isoa_len = p_pdu->source_rdi.source_RDI_length;
	bcopy ((caddr_t) &p_pdu->source_rdi.source_RDI,&rdi_peer.isoa_genaddr,rdi_peer.isoa_len);
	if (isoaddrcmp(&peer->rdi,&rdi_peer))
		{
		/* addresses are not equal so error */

		trace(TR_IDRP, 0, "IDRP open error - RDI from peer wrong peer %s  rdi %s",
			peer->name,iso_netiso_ntoa(&rdi_peer));
	
		peer->last_error_pdu.error_code = IDRP_ERROR_OPEN_PDU;
		peer->last_error_pdu.error_subcode = IDRP_OPEN_ERROR_BAD_PEER_RD;
		peer->last_error_pdu.data = (u_char *) 0;
		peer->last_error_pdu.data_len = 0; 
		return(IDRP_ERROR_OPEN_PDU);
		}	
		
	/* d) Rib_Att_Set	 */

	if  (p_pdu->rib_att.RIB_att_count != RIB_ATT_DEFAULT_ONLY_COUNT)
		{
		peer->last_error_pdu.error_code = IDRP_ERROR_OPEN_PDU;
		peer->last_error_pdu.error_subcode = IDRP_OPEN_ERROR_BAD_RIB_ATT_SET;
		peer->last_error_pdu.data = (u_char *) 0;
		peer->last_error_pdu.data_len = 0; 
		return(IDRP_ERROR_OPEN_PDU);
		}
	
	/* e) RDC match check */

	if (p_pdu->confed_id.confed_id_count != NO_RDC_SUPPORT_COUNT)
		{
		peer->last_error_pdu.error_code = IDRP_ERROR_OPEN_PDU;
		peer->last_error_pdu.error_subcode = IDRP_OPEN_ERROR_RDC_MISMATCH;
		peer->last_error_pdu.data = (u_char *) 0;
		peer->last_error_pdu.data_len = 0; 
		return(IDRP_ERROR_OPEN_PDU);
		}

	/* f) Authentication check */

	if (p_pdu->auth_code.authentication_code == IDRP_AUTH_CODE_CLEAR || 
	  (p_pdu->auth_code.authentication_code == IDRP_AUTH_CODE_CRYPT) || 
	  (p_pdu->auth_code.authentication_code == IDRP_AUTH_CODE_CLEAR)) 
		{
		peer->last_error_pdu.error_code = IDRP_ERROR_OPEN_PDU;
		peer->last_error_pdu.error_subcode = IDRP_OPEN_ERROR_AUTH_FAILURE;
		peer->last_error_pdu.data = (u_char *) 0;
		peer->last_error_pdu.data_len = 0; 
		return(IDRP_ERROR_OPEN_PDU);
		}
	else
		{
		peer->last_error_pdu.error_code = IDRP_ERROR_OPEN_PDU;
		peer->last_error_pdu.error_subcode = IDRP_OPEN_ERROR_UNSUP_AUTH_CODE;
		peer->last_error_pdu.data = (u_char *) 0;
		peer->last_error_pdu.data_len = 0; 
		return(IDRP_ERROR_OPEN_PDU);
		}
return (0);
}
	
/*
 * Open PDU
 */
static void process_open(peer, pdu, length)
idrpPeer *peer;
idrpPdu *pdu;
int length;
{
u_short rcv_holdtime;
u_short rcv_maxpdusize;
idrp_open_pdu_header *hdr;


	/* set-up new hold time from pdu - whether we use it or not */

	hdr = &pdu->pdu.open.header;	
	bcopy((caddr_t) hdr->idrp_hold_time, (caddr_t) &rcv_holdtime, sizeof(hdr->idrp_hold_time));
	ntohs(rcv_holdtime);
	bcopy((caddr_t) hdr->idrp_maxpdu_size, (caddr_t) &rcv_maxpdusize, sizeof(hdr->idrp_maxpdu_size));
	ntohs(rcv_maxpdusize);
	peer->keepalive_time = rcv_holdtime / 3 ? rcv_holdtime / 3 : 1;
	peer->pdu_maxsendsize = rcv_maxpdusize;
	trace(TR_IDRP, 0, "Open hold time %d, max pdu size %d", rcv_holdtime, rcv_maxpdusize);

	/* simply call state machine for valid open */
	idrp_sm(peer,IDRP_OPEN);
}


int parse_update(peer, pdu, length,res)
idrpPeer *peer;
idrpPdu *pdu;
int length;
parse_results *res;
{

/* preset return code to be valid */

int return_code= 0; 

/* path attribute variables */

idrp_attribute_record *att_rec = 0;
idrp_attribute_record *p_attr = 0;
idrp_path_attr *attrs;
idrpRoute_entry *p_idrp_re = 0; 
short total_attr_len;
short attr_len;
u_int attr_type;
u_int usr_type = 0;
u_int usr_attr_index;
idrpRoute_entry	*p_rid_list;
idrpRoute *p_idrp_rt;
idrpRoute_link rt_chg[IDRP_NLRI_FAMILIES_SUPPORTED];
int route_id_in;

/* nlri variables */ 
nlri_info *nlriptr;
nlri_proto_info *nlri_proto_ptr;
struct iso_prefix *nlri;
u_short nlrilen, nlrilen_local, nlriproto, nlriproto_id;
int nlri_cnt = 0;


/* general pointer for copies */ 
u_char *cp;
int	len_left;
int	index;
int 	i;
int	family;
int 	nlri_id;
u_short	hold;
u_long	longhold;

/* pdu headers work pointers */
idrp_update_pdu *update;
u_int id;

	/* set up initial pointers into pdu */

	update = &pdu->pdu.update;
	
	len_left = length - sizeof(idrp_fixed_header);
	
	/* Get Withdrawn route-id count and route_ids */
	/* route_ids are variable length */

	bcopy(update->unfeasible_route_count, (caddr_t) &hold, sizeof(update->unfeasible_route_count));
	ntohs(hold);
	if (hold >= IDRP_WITHDRAW_MAX_IDs)
		{
		trace (TR_IDRP,0,"Withdraw ids over maximum peer %s update pdu ",peer->name); 
		peer->last_error_pdu.error_code = IDRP_ERROR_UPDATE_PDU;
		peer->last_error_pdu.error_subcode = IDRP_UPDATE_ERROR_MALF_ATTR;
		peer->last_error_pdu.data_len = 0;
		return (IDRP_ERROR_UPDATE_PDU);
		}
	
	cp =  (u_char *) &update->unfeasible_routes[1];

	res->withdraw->count = hold; 

	/* initialize results array */	

	res->route_id_in =  IDRP_INIT_ROUTE_ID;
	res->p_att = (idrp_attribute_record *) 0;
	NLRI_FAMILY_LOOP(i)
		{ 
		res->rt_chg[i] = (idrpRoute *) 0;
		rt_chg[i].tail = rt_chg[i].head = (idrpRoute *) 0;
		}
	res->free_pdu = 0; 
 
	
	/* Parse the withdrawn route list. */
	/* putting withdraw routes into withdraw array */

	for (i = 0; i < res->withdraw->count; i++)
		{
		bcopy((caddr_t) cp,&id,sizeof(route_id));
		ntohl(id);
		cp += sizeof(route_id);
		res->withdraw->withdraw[i] = id;
		}

	/* Parse the update into an attribute record. */
	/* skip over the withdrawn route length, routes, */
	/* attributes length and attributes */
	/* save pointer to nlri information */

	bcopy((caddr_t) cp, (caddr_t) &total_attr_len, sizeof(update->tot_path_att_len));
	ntohs(total_attr_len);
	index = sizeof(update->tot_path_att_len) + (sizeof (route_id) * res->withdraw->count);

	/* update the length left to make sure pdu is valid */
	/* bytes to end of pdu should be		   */
	/*  -  total withdrawn routes bytes (variable)     */
	/* -  total_path_att_len bytes (2) */
	/* -  total path attribute bytes (variable) */
	len_left -= index;

	if (len_left < 0)
		{
		/* length too short - pdu invalid */
		/* return code of zero - says PDU is bogus pocus */ 
		peer->last_error_pdu.error_code = IDRP_ERROR_UPDATE_PDU;
		peer->last_error_pdu.error_subcode = IDRP_UPDATE_ERROR_ATTR_LEN_ERR; 
		peer->last_error_pdu.data_len = 0;
		parse_update_cleanup(res,p_idrp_re,peer);
		return(IDRP_UPDATE_ERROR);
		}			

	/* only withdraws in this pdu */
	
	if (total_attr_len == 0)
		{
		res->free_pdu =  TRUE;
		return (0);
		}
		
	
	/* path attributes and nrli ahead - start the parse */
	
	/* skip over path attribute length */
	/* start reading attributes one by one */

	nlriptr = (nlri_info *) ((caddr_t) cp + index + total_attr_len+ sizeof(nlri_proto_info)); 
	nlri_proto_ptr = (nlri_proto_info *) nlriptr;
	cp += sizeof(update->tot_path_att_len); 

	trace(TR_IDRP | TR_UPDATE, 0, "IDRP update from %s attr length %d", peer->name, total_attr_len);

	/* allocate and fill a temporary attribute record  */
	/* later we will see if this can like AS PATHS */
 
	att_rec = (idrp_attribute_record *) task_mem_malloc(peer->task, sizeof(idrp_attribute_record));
	att_rec->p_attr_pdu = (u_char *) task_mem_malloc(peer->task,total_attr_len);
	usr_attr_index = 0;
	bzero(att_rec, sizeof(idrp_attribute_record));
	bcopy(cp,att_rec->p_attr_pdu,total_attr_len);
	att_rec->attr_len = total_attr_len;
	cp = att_rec->p_attr_pdu; 

	/* set-up results array with attribute record id */
	res->p_att = att_rec;

	while (total_attr_len > 0) {
		attrs = (idrp_path_attr *) cp;
		attr_type = attrs->attr_hdr.attr_type;
		usr_type = FALSE;

		/* check the type so saving info in arrays */
		/* will not run off the end of the array */
		if (attr_type >= IDRP_ATTR_USR_MIN || attr_type <= IDRP_ATTR_USR_MAX)
			{
			/* valid user type */
			usr_type = TRUE;
			}
	
		/* check for range within IDRP specified attributes */ 
		if (attr_type < IDRP_ATTR_MIN ||
		    attr_type > IDRP_ATTR_MAX ||
			(usr_type == FALSE))
			{
			/* attribute type outside of range */
			/* error in parsing the pdu */

			if (attrs->attr_hdr.attr_flags == IDRP_ATTR_FLAG_MANDATORY)
				{
				peer->last_error_pdu.error_code = IDRP_UPDATE_ERROR;
				peer->last_error_pdu.error_subcode = IDRP_UPDATE_ERROR_UNREC_WELKN_ATTR;
				peer->last_error_pdu.data_len = 0;
				parse_update_cleanup(res,p_idrp_re,peer);
				trace(TR_IDRP, 0, "IDRP %s got pdu type %d outside range ", peer->name, attr_type);
				return(IDRP_UPDATE_ERROR);
				}
			}
		   
		bcopy(attrs->attr_hdr.attr_length, (caddr_t) &attr_len, sizeof(attrs->attr_hdr.attr_length));
		ntohs(attr_len);
		trace(TR_IDRP | TR_UPDATE, 0, "  Attr %d, length %d, flags %x",
			 attrs->attr_hdr.attr_type,attr_len, attrs->attr_hdr.attr_flags);
		cp += sizeof(idrp_path_attr_hdr);
		if (!usr_type)
			{
			/* idrp attribute */
	
			if (att_rec->attrib[attr_type].present) 
				{
				/* type appeared twice in PDU */

				peer->last_error_pdu.error_code = IDRP_UPDATE_ERROR;
				peer->last_error_pdu.error_subcode = IDRP_UPDATE_ERROR_UNREC_WELKN_ATTR;
				peer->last_error_pdu.data_len = 0;
				parse_update_cleanup(res,p_idrp_re,peer);
				trace(TR_IDRP, 0, "IDRP %s attribute %d present twice", peer->name, attr_type);
				return(IDRP_UPDATE_ERROR);
				}
		
			att_rec->attrib[attr_type].present = TRUE;
			att_rec->attrib[attr_type].flags = attrs->attr_hdr.attr_flags;
			att_rec->attrib[attr_type].length = attr_len;
			att_rec->attrib[attr_type].data = cp;
			}
		else
			{
			/* user type */
			att_rec->usr_attrib[usr_attr_index].present = TRUE;
			att_rec->usr_attrib[usr_attr_index].flags = attrs->attr_hdr.attr_flags;
			att_rec->usr_attrib[usr_attr_index].length = attr_len;
			att_rec->usr_attrib[usr_attr_index].data = cp;
			usr_attr_index++;
			}  	
		cp += attr_len;
		total_attr_len -= attr_len + sizeof(idrp_path_attr_hdr);
	}
	/* set-up the idrpRoute id list */ 

	p_rid_list = (idrpRoute_entry *) task_mem_malloc(peer->task,sizeof(idrpRoute_entry));	
	att_rec->route_id_list = p_rid_list;
	p_rid_list->route_id = IDRP_INIT_ROUTE_ID;
	p_rid_list->peer = peer;
	p_rid_list->p_next = (idrpRoute_entry *) 0;

	NLRI_FAMILY_LOOP(i)
		{
		p_rid_list->p_route[i] = (idrpRoute *) 0;
		}
	
	/* set internal AdjRib id  */
	/* always zero for default case */
  
	att_rec->rib_id = idrp_rib_id(att_rec->idrp_mask);
	if (att_rec->rib_id < 0)
		{
		trace(TR_IDRP | TR_UPDATE, 0, "Non-Default Attributes no support for this now peer %s",peer->name);
		peer->last_error_pdu.error_code = IDRP_UPDATE_ERROR;
		peer->last_error_pdu.error_subcode = IDRP_UPDATE_ERROR_MALF_ATTR;
		peer->last_error_pdu.data_len = 0;
		parse_update_cleanup(res,p_idrp_re,peer);
		return (IDRP_UPDATE_ERROR);
		}


	/* Now validate the attributes. */

	if (!valid_attr(peer,att_rec))
		{
		/* trace error,  release pdu, and exit */
		parse_update_cleanup(res,p_idrp_re,peer);
		trace(TR_IDRP | TR_UPDATE, 0, "Atttributes invalid. code = %x",att_rec);
		return(IDRP_UPDATE_ERROR);	
		}

	p_attr = find_attr_rec(peer,att_rec);
	if (p_attr)
		{
		/* if old attribute matches attributes use it by:
		 *  1.) linking a new entry onto the route_id_list
		 *  2.) incrementing  reference count 
		 *  3.) set flag to free pdu after we're all done 
		 */

		route_id_in = att_rec->route_id_list->route_id;
		p_idrp_re = idrp_add_idrpRoute_entry(route_id_in,p_attr->route_id_list);
		att_rec->route_id_list->route_id = IDRP_INIT_ROUTE_ID;
		idrp_free_attr_rec(peer,att_rec,route_id_in);
		res->free_pdu = 1;
		att_rec = p_attr;
		}
	else
		{
		/* get route id */
		bcopy(&route_id_in,att_rec->attrib[IDRP_ATTR_ROUTE_ID].data,sizeof(route_id));
		ntohl(route_id_in);
		trace(TR_IDRP | TR_UPDATE, 0, "RouteId %d (%d) peer%s",route_id_in, \
			att_rec->route_id_list->route_id,peer->name);
		att_rec->route_id_list->route_id = route_id_in;
		p_idrp_re = att_rec->route_id_list;
		}
		

	if (usr_attr_index) 
		{
		if (!valid_usr_attr(peer,att_rec))
			{
			/* trace error, release attribute block and exit */
			trace(TR_IDRP | TR_UPDATE, 0, "Atttributes invalid. code = %x",att_rec);
			parse_update_cleanup(res,p_idrp_re,peer);
			return(IDRP_UPDATE_ERROR);	
			}
		}


	/* Parse the NLRI */
	/* watch our total count to make sure PDU is valid */
	/* NLRI  = prototype(octet), length(octet), value */
	/* repeat 1-5 for each protocol type */
	/* 1.) get protocol type */
	/* 2.) get length this nlri info */
	/* repeat 3-6 until NLRI done */
	/* 3.) get length of NLRI */
	/* 4.) get NLRI -  */
	/* 5.) create IDRPRoute block and rt_entry */
	/* 6.) do rt_lookup and give to add, change or delete list */

	while (len_left > 0)
		{
		/* copy protocol type, length and value  */ 
		bcopy(&nlri_proto_ptr->proto_type, &nlriproto_id, sizeof(nlri_proto_ptr->proto_type));
		bcopy(&nlri_proto_ptr->proto_len,&nlrilen, sizeof(nlri_proto_ptr->proto_len));
		ntohs(nlrilen);
		bcopy(&nlri_proto_ptr->proto_val,&nlriproto,sizeof(nlri_proto_ptr->proto_val));
		ntohs(nlriproto);

		/* translate family to gated family */
		family = proto_to_family(nlriproto);
		nlri_id = family_to_nlri_id(family);

		/* recalculate remaining length */
		/* taking out the length of this protocol group */ 

		nlrilen += (sizeof(nlri_proto_ptr->proto_type) + sizeof(nlri_proto_ptr->proto_len)); 
		len_left -= nlrilen;
		trace(TR_IDRP | TR_UPDATE, 0, "NLRI (PROTO_ID %x, length %d):", nlriproto, nlrilen);

		bcopy(nlriptr,&nlrilen, sizeof(nlriptr->addr_len));
		ntohs(nlrilen);
		nlri = nlriptr->prefixes;		
		len_left -=nlrilen;

		if (len_left < 0)
			{
			trace(TR_IDRP | TR_UPDATE, 0, "len of nrli exceeds pdu length len of nlri = %d",nlrilen);
			peer->last_error_pdu.error_code = IDRP_UPDATE_ERROR;
			peer->last_error_pdu.error_subcode = IDRP_UPDATE_ERROR_ATTR_LEN_ERR;
			peer->last_error_pdu.data_len = 0;

			/* flag no route_id list for cleanup */
			parse_update_cleanup(res,p_idrp_re,peer);	
			return (IDRP_UPDATE_ERROR);
			}
	
		
		/* loop till finish this protocol nlri group */

		while (nlrilen > 0) {
			/* allocate buffer */
			trace(TR_IDRP | TR_UPDATE, 0, " nlri =   %s", iso_ptoa(nlri));
			p_idrp_rt = idrp_alloc_idrpRoute(peer,att_rec,family);

			/* copy in the nlri into buffer */
			bcopy((caddr_t) &p_idrp_rt->nlri,&nlri,PFX_BYTE_LENGTH(*nlri));

			/* add it to the linked list of IDRP_routes */
			
			idrp_route_mod(p_idrp_rt,peer,rt_chg[nlri_id]);
			nlri_cnt++;
			p_idrp_rt->p_attr->ref_cnt++;	

			/* advance to the next prefix */		
			nlrilen -= nlri->pfx_len;
			nlri = (struct iso_prefix *) ((caddr_t) nlri + PFX_BYTE_LENGTH(*nlri));
			}
		} /* end of this protocol id */

	/* valid update pdu */
		
	trace(TR_IDRP | TR_UPDATE,0,"valid pdu pdu from peer %s last nlri %s", peer->name,iso_ptoa(nlri));

	/* increment the attribute record and link nlri on route_id */
	/* note - that the array of pointers is set for ip and osi */

	att_rec->ref_cnt += nlri_cnt;  /* increment by the number of nlri */

	NLRI_FAMILY_LOOP(i)
		{
		p_idrp_re->p_route[i] = rt_chg[i].head;
		res->rt_chg[i] = rt_chg[i].head;
		}

	/* save results in result array from UPDATE pdu */
 
	res->route_id_in = route_id_in;

	/*  parsed pdu UPDATE pdu */

	return (0);

}
	
void parse_update_cleanup(p_res,p_ent,peer)
parse_results	*p_res;
idrpRoute_entry *p_ent;
idrpPeer	*peer;
{
idrpRoute_entry	*p_last = 0;
idrpRoute_entry	*p_cur = 0;
int i;
int found;

/* routine to delete the partial results of the attribute
 * record when aborted mid-stream in parse_update 
 */

	/* check to see if we have an attribute record at all */ 

	if (p_res->p_att == (idrp_attribute_record *) 0 )
		{
		return;
		}


	/* attribute record exists - now check for the route_id list */

	if (p_res->p_att->route_id_list == 0)
		{
		if (p_res->p_att->ref_cnt)
			{
			/* warn error and then exit */
			trace (TR_IDRP | TR_UPDATE,0,"peer %s has tried to free attribute record with no route_id_list",peer->name);
			return;
			}
		else
			{
			if (p_res->p_att->p_attr_pdu)
				{
				task_mem_free(peer->task,p_res->p_att->p_attr_pdu);
				}
			task_mem_free(peer->task,(void_t)p_res->p_att);
			return;
			}
		}


			
	/* route_list exists - now check on route requested to delete */

		if (p_ent == p_res->p_att->route_id_list)
			{
			/* first route_id in list - is there anything after it? */

			p_res->p_att->route_id_list = p_ent->p_next;
				
			}
		else
			{
			/* walk the loop searching for the route_id list */

			p_last = p_res->p_att->route_id_list;
			p_cur = p_res->p_att->route_id_list->p_next;
	
			while (p_cur)
				{
				if (p_cur == p_ent)
					{
					found = TRUE;
					}
				else
					{
					p_last = p_cur;
					p_cur = p_cur->p_next;
					}
				}
			if (found)
				{
				p_last->p_next = p_cur->p_next;
				}
			}	

		/* free the idrpRoute_entry */

		NLRI_FAMILY_LOOP(i)
			{
			if (p_ent->p_route[i])
				{
				/* warn - no routes should be in the list when we free it here */

				trace (TR_IDRP,0,"parse_update_cleanup error peer %s has called with bad routeEntry", \
					peer->name);
				}
			}
		
		task_mem_free(peer->task,p_ent);

		if (p_res->p_att->route_id_list)
			{
			/* more on this attribute record */ 
			return;
			}				
		
		/* no more on this route_id list */

		if (p_res->p_att->ref_cnt)
			{
			/* problem - reference count is greater than zero attribute record */
			/* but no route_ids are on the list */

			trace (TR_IDRP | TR_UPDATE,0,"parse_update_cleanup error peer %s has attribute record no route_list and ref_cnt != 0",peer->name);
			return;
			}

		/* zero reference count - free the attribute block */
		/* free copy of attributes from pdu */
	
		if (p_res->p_att->p_attr_pdu)
			{
			task_mem_free(peer->task,p_res->p_att->p_attr_pdu);
			}
		task_mem_free(peer->task,p_res->p_att);			
		return;
			
}


static void process_update(peer, pdu, length)
idrpPeer *peer;
idrpPdu *pdu;
int length;
{
int family;
parse_results *results;

	results = (parse_results *) task_mem_malloc(peer->task,sizeof(parse_results));
	results->withdraw = (Withdrawl *) task_mem_malloc(peer->task,IDRP_WITHDRAW_MAX_IDs);

	if (parse_update(peer,pdu,length,results))
		{
		/* invalid pdu - %% does gated really handle
		 * returning pdus - I think Dave's code loses pdus 
		 * error code and data are in the last_error_pdu 
		 * in the idrp peer structure
		 */ 

		idrp_sm(peer,IDRP_UPDATE_ERROR);
		if (results->free_pdu)
			{
			/* find out how to free the pdu here */
			}
		task_mem_free(peer->task,results->withdraw);
		task_mem_free(peer->task,results);
		}
	else
		{
		/* valid pdu so let's send it through the state machine */

		if (idrp_sm(peer,IDRP_UPDATE))
			{
			/* state machine says process update pdu */

			if (peer->refresh.seq_start)
				{
				/* mid rib_refresh, hook the rib refresh pdus */
				/* together for processing */

				refresh_pdu_link(peer,results,pdu);
				return;
  				}
			else
				{
				/* away for processing of pdus into phase1 processing */ 
				idrp_process_pdu_routes(peer,results);

				/* clean up the results array */
				if (results->free_pdu)
					{
					/* free idrp pdu - since attribute record duplicate */
					}	
				task_mem_free(peer->task,results->withdraw);
				task_mem_free(peer->task,results);
				}
			} /* if state machine returns process pdu */
		else
			{
			/* state machine error - don't process pdus and release it */ 
			if (results->free_pdu);
				{
				/* free idrp pdu - since error */
				}
			task_mem_free(peer->task,results->withdraw);
			task_mem_free(peer->task,results);
	
			}
			
		} /* end if parse update valid */

}

/*
 * KEEPALIVE PDU
 */
static void process_keepalive(peer, pdu, length)
idrpPeer *peer;
idrpPdu *pdu;
int length;
{
	idrp_sm(peer,IDRP_KEEPALIVE);
}


static void process_error(peer, pdu, length)
idrpPeer *peer;
idrpPdu *pdu;
int length;
{
int err_code = 0;
int subcode = 0;
int error = 0; 

	/* validate ERROR pdu */

	err_code = pdu->pdu.error.header.error_code;
	subcode = pdu->pdu.error.header.error_subcode;
	
	switch (err_code)
		{
		case IDRP_ERROR_OPEN_PDU:
			if ((subcode < IDRP_OPEN_ERROR_MIN_SUBCODE)
			 && (subcode > IDRP_OPEN_ERROR_MAX_SUBCODE))
				{	 
				error=TRUE;
				}
			break;

		case IDRP_ERROR_UPDATE_PDU:
			if ((subcode < IDRP_UPDATE_ERROR_MIN_SUBCODE)
			 && (subcode > IDRP_UPDATE_ERROR_MAX_SUBCODE))
				{	 
				error = TRUE; 
				}
			break;

		case IDRP_ERROR_HOLDTIME_EXPIRED:
			if (subcode != IDRP_HOLDTIME_EXPIRED_ERROR_SUBCODE)
				{
				error = TRUE;  
				}
			break;

		case IDRP_ERROR_FSM:
			if ((subcode < IDRP_FSM_ERROR_MIN_SUBCODE)
			 && (subcode > IDRP_FSM_ERROR_MAX_SUBCODE))
				{	 
				error = TRUE; 
				}
			break;

		case IDRP_ERROR_RIB_REFRESH:
			if ((subcode < IDRP_RIB_REFRESH_ERROR_MIN_SUBCODE)
			 && (subcode > IDRP_RIB_REFRESH_ERROR_MAX_SUBCODE))
				{
				error = TRUE;
				}
			break;

		default:
			error = TRUE;
			break;
		}

	if (error)
		{
		idrp_sm(peer,IDRP_ERROR_ERROR);
		log_nm_event(peer,IDRP_ERROR_PDU_ERROR,err_code);
		}
	else
		{
		idrp_sm(peer,IDRP_ERROR);
		}

}


static void process_cease(peer, pdu, length)
idrpPeer *peer;
idrpPdu *pdu;
int length;
{
	idrp_sm(peer,IDRP_CEASE);	
}

u_int parse_rib_refresh(peer,pdu,length)
idrpPeer *peer;
idrpPdu *pdu;
int length;
{
	/* range check for the Rib Refresh types */

	if (pdu->pdu.refresh.hdr.opcode < IDRP_RIB_REFRESH_MIN_CODE  ||
	    pdu->pdu.refresh.hdr.opcode > IDRP_RIB_REFRESH_MAX_CODE)
		{
		peer->last_error_pdu.error_code = IDRP_ERROR_RIB_REFRESH;
		peer->last_error_pdu.error_subcode = IDRP_RIB_REFRESH_ERROR_INV_OPCODE;
		peer->last_error_pdu.data_len = 0;
	 	return (IDRP_ERROR_RIB_REFRESH);
		}

	/* check the attribute types */
		
	/* d) Rib_Att_Set - Default only	 */
	
	if  (pdu->pdu.refresh.rib_att[0].RIB_att_count != RIB_ATT_DEFAULT_ONLY_COUNT)
		{
		peer->last_error_pdu.error_code = IDRP_ERROR_RIB_REFRESH;
		peer->last_error_pdu.error_subcode = IDRP_RIB_REFRESH_ERROR_UNSUP_ATTRS;
		peer->last_error_pdu.data_len = 0;
	 	return (IDRP_ERROR_RIB_REFRESH);
		}
	return(0);
}


static void process_rib_refresh(peer, pdu, length)
idrpPeer *peer;
idrpPdu *pdu;
int length;
{

int seq_no = 0;
	
	if (parse_rib_refresh(peer,pdu,length))
		{
		/* error in pdu */

		idrp_sm(peer,IDRP_REFRESH_ERROR);
		}
	else
		{
		/* valid Rib Refresh */
		/* check against state machine */

		if(idrp_sm(peer,IDRP_REFRESH))
			{
	    		switch (pdu->pdu.refresh.hdr.opcode)
				{
				case IDRP_RIB_REFRESH_REQ:
					idrp_send_refresh_all_routes(peer);
					break;	

				case IDRP_RIB_REFRESH_START:
					if (peer->refresh.seq_start !=0)
						{
						rib_refresh_pdu_release(peer);
						}	
					bcopy(&pdu->header.seq,&seq_no,sizeof(pdu->header.seq));

					trace(TR_IDRP, 0, "IDRP Refresh with start Seq =%d eer %s",
					seq_no,peer->name);
					peer->refresh.seq_start=seq_no;
					break;

				case IDRP_RIB_REFRESH_END:	
					bcopy(&pdu->header.seq,&seq_no,sizeof(pdu->header.seq));
					trace(TR_IDRP, 0, "IDRP Refresh with start Seq =%d eer %s",
					seq_no,peer->name);
					peer->refresh.seq_stop=seq_no;
					rib_refresh_update_process(peer);
					peer->refresh.seq_start = 0;
					peer->refresh.seq_stop = 0;
					break;
				} 
			} /* if for state machine */
		} /* if parse valid */

}


#ifdef ECHOKLUDGE
static void process_echo(peer, pdu, length)
idrpPeer *peer;
idrpPdu *pdu;
int length;
{
	switch (peer->state) {

	case IDRP_ESTABLISHED:
		send_echo_reply(peer);
		break;

	default:
		break;
	}
}


static void process_echo_reply(peer, pdu, length)
idrpPeer *peer;
idrpPdu *pdu;
int length;
{
	switch (peer->state) {

	default:
		break;
	}
}
#endif


static void process_header_error(peer, pdu, length)
idrpPeer *peer;
idrpPdu *pdu;
int length;
{
	log_nm_event(peer, IDRP_PDU_HEADER_ERROR,IDRP_PDU_HEADER_ERR_BAD_TYPE);
	idrp_sm(peer,IDRP_HEADER_ERROR);
}


/*
 * Bump the expected receive sequence number if the packet is sequenced.
 * Send a keepalive as acknowledgement if either the send queue is empty,
 * or the connection is flow blocked, so long as we're in ESTABLISHED state.
 *
 * If the received frame is sequenced, cancel any pending keepalive
 * retransmit (due to the peer being flow blocked).  If the received frame
 * has credit_available nonzero (the peer is not flow blocked), cancel any
 * pending keepalive retransmit.
 */
static void process_rcv_seq(peer, pdu)
idrpPeer *peer;
idrpPdu *pdu;
{
	if (IS_SEQUENCED(pdu->header.pdu_type)) {
		peer->rcv_seq_expected++;
		if (peer->cred_avail_zero_seq_last) {
			peer->cred_avail_zero_seq_last = 0;
			kill_rexmit_timer(peer);
		}
	}

	if (pdu->header.credit_available) {
		if (peer->cred_avail_zero_seq_last) {
			peer->cred_avail_zero_seq_last = 0;
			kill_rexmit_timer(peer);
		}
	}

	/*
 	 * Track the received credit available as part of a mechanism to ensure that
	 * our peer doesn't get flow blocked for a long time because our ack got lost.
	 */

	peer->rcved_credit_avail = pdu->header.credit_available;
	if (peer->rcved_credit_avail == 0) {
		peer->cred_avail_zero_seq = peer->rcv_seq_expected - 1;
	}

	if (IS_SEQUENCED(pdu->header.pdu_type)) {
		if (peer->state == IDRP_ESTABLISHED &&
		    (peer->xmit_queue_head == NULL || (peer->upper_window_edge <= peer->send_next && !acked_something))) {
			send_keepalive(peer);
		}
	}
}

static in_sequence(peer, seq)
idrpPeer *peer;
u_int seq;
{
	if (seq == peer->rcv_seq_expected) {
		trace(TR_IDRP, 0, "In sequence");
		return(TRUE);
	} else {
		trace(TR_IDRP, 0, "Out of sequence");
		return(FALSE);
	}
}


/*
 * A packet for a particular peer task has arrived.  Figure out what kind of packet it
 * is, and do something about it.
 * jgs -- where to put this?
 */
static void idrp_recv_pdu(peer, pdu, length,family)
idrpPeer *peer;
idrpPdu *pdu;
int length;
int family;
{
u_char type;
idrp_error_pdu	error_code;	/* error codes */
u_int rcv_seq, rcv_ack;		/* Seq, Ack fields from received PDU */
				/* jgs -- these were formarly globals (static globals, actually)
				 * jgs -- which were never ref'd except for in this func, as far
				 * jgs -- as I can tell.  Sue, please let me know if they need to
				 * jgs -- be global for some reason I missed. */


	/* If we're closed, don't even worry about it. */
	if (peer->state == IDRP_CLOSED)
		return;

	/* Validate the checksum. */

	if (!checksum_ok(peer, pdu, length))
		return;

	/* Validate the header - min and maximum length */
	
	if (length < sizeof(idrp_fixed_header))
		{ 
		idrp_sm(peer,IDRP_HEADER_ERROR);
		return;
		}
	if (!valid_header(peer,pdu,length,family))
		{
		idrp_sm(peer,IDRP_HEADER_ERROR);	
		return;
		}
 
	/* Extract sequence information. */

	bcopy((caddr_t) pdu->header.seq, &rcv_seq, sizeof(pdu->header.seq));
	ntohl(rcv_seq);
	bcopy((caddr_t) pdu->header.ack, &rcv_ack, sizeof(pdu->header.ack));
	ntohl(rcv_ack);

	trace(TR_IDRP, 0, "IDRP (%s) RCV %s seq %d ack %d offer %d avail %d len %d from %s", 
	      trace_state(peer->state, idrpStates),
	      (type > IDRP_PDUTYPE_MAX ? "Bogus" : idrpTypes[type]), 
	       rcv_seq, rcv_ack, pdu->header.credit_offered, pdu->header.credit_available,
		length, peer->name);

	/* Update the hold timer */

	timer_set(TASK_TIMER(peer->task, IDRPTIMER_HOLDTIME), peer->hold_time, (time_t) 0);

	/* Process CEASE and ERROR PDUs before anything else. */
	/* CEASE and ERROR pdus need to be processed w/ack */ 
	/* - processed first because bad news needs */
	/* to travel fast */

	type = pdu->header.pdu_type;
	switch (type) {

	case IDRP_OPEN_TYPE:
		if (idrp_parse_open(peer, pdu, length))  
			{
			/* bogus open - don't process acks */
			idrp_sm(peer,IDRP_OPEN_ERROR);
			pdu = NULL;
			}
		break;

	case IDRP_ERROR_TYPE:
		process_error(peer, pdu, length);
		pdu = NULL;
		break;

	case IDRP_CEASE_TYPE:
		process_cease(peer, pdu, length);
		pdu = NULL;
		break;
	}
	
	if (pdu == NULL)
		return;		/* Nothing else to do. */

	/* Process received acknowledgements. */
	/* the above code has streamed out    */
	/* - close state 		*/
	/* ERROR pdus (with or without errors) */  
	/* CEASE pdus (with or without errors) */
	/* OPEN pdus with errors */
	/* - ack_pdu handles ack by type and state */  

	acked_something = FALSE;
	switch (peer->state) {
	/* Open_Sent state: for OPEN PDUs only, process ack */
	
	case IDRP_OPEN_SENT:
		if (type == IDRP_OPEN_TYPE)
			{ 	
			ack_pdu(peer, rcv_ack, pdu->header.credit_offered,type);
			}
		break;

	default:
		ack_pdu(peer,rcv_ack,pdu->header.credit_offered,type);
		break;
	}

	/* Now process the send sequence number. */

	switch (peer->state) {

	/*
	 * Open_sent state:  treat everything as in-sequence except an out-of-sequence OPEN.
	 * This allows us to bitch about FSM errors.
	 */
	case IDRP_OPEN_SENT:
		peer->rcv_seq_expected = rcv_seq;  /* Establish initial sequence number */
		if (type == IDRP_OPEN_TYPE) {
			if (in_sequence(peer, rcv_seq)) {
				process_rcv_seq(peer, pdu);
			} else {
				pdu = NULL;
			}
		}
		break;

	/*
	 * Everything else--received packets must be in-sequence or we drop them.
	 */
	default:
		if (in_sequence(peer, rcv_seq)) {
			process_rcv_seq(peer, pdu);
		} else {
			pdu = NULL;
		}
		break;
	}

	if (pdu == NULL)
		return;		/* Nothing else to do. */

	/*
	 * Now process the received PDUs.
	 */
	switch (type) {

	case IDRP_OPEN_TYPE:
		process_open(peer, pdu, length);
		break;

	case IDRP_UPDATE_TYPE:
		process_update(peer, pdu, length);
		break;

	case IDRP_KEEPALIVE_TYPE:
		process_keepalive(peer, pdu, length);
		break;

	case IDRP_RIB_REFRESH_TYPE:
		process_rib_refresh(peer, pdu, length);
		break;

#ifdef ECHOKLUDGE
	case IDRP_ECHO_TYPE:
		process_echo(peer, pdu, length);
		break;

	case IDRP_ECHO_REPLY_TYPE:
		process_echo_reply(peer, pdu, length);
		break;
#endif

	default:
		process_header_error(peer, pdu, length);
		break;

	}
}


/*
 * A packet has arrived on the common socket.  Fetch it and demux.
 * - here we have to change radically the idrp code 
 * - currently it does only IP based CLNP -  
 * - note received pdu is returned by task???  
 * jgs -- where to put this?
 */

void idrp_recv(tp)
task *tp;
{

	idrpPeer *peer;
	idrpPdu *pdu = (idrpPdu *) task_recv_buffer;
	size_t length;
	int i;

	if (task_receive_packet(tp, &length)) {
		return;
	}

	/* Validate the destination address. */

	if (isoaddr_netiso_cmp(&pdu->header.dest, &rt_net)) {
		/* jgs -- note that idrp_fixed_header.src (and .dst) are iso_addrs, while iso_netiso_ntoa wants a */
		/* jgs -- iso_net_addr.  Which should change?  I'd like to better understand the reason to have   */
		/* jgs -- these two similar yet differing types...						  */
		tracef("IDRP PDU with bad destination %s", iso_netiso_ntoa(&pdu->header.dest));
		/* jgs -- note that idrp_fixed_header.src (and .dst) are iso_addrs, while iso_netiso_ntoa wants a */
		/* jgs -- iso_net_addr.  Which should change?  I'd like to better understand the reason to have   */
		/* jgs -- these two similar yet differing types...						  */
		trace(TR_IDRP, 0, " from %s", iso_netiso_ntoa(&pdu->header.src));
		return;
	}

	/* Figure out which subtask the packet is really intended for. */

	IDRP_LIST(peer) {
		if (!isoaddr_netiso_cmp(&pdu->header.src, &peer->neighbor))
		break;
	} IDRP_LIST_END;

	if (!peer) {
		/* jgs -- note that idrp_fixed_header.src (and .dst) are iso_addrs, while iso_netiso_ntoa wants a */
		/* jgs -- iso_net_addr.  Which should change?  I'd like to better understand the reason to have   */
		/* jgs -- these two similar yet differing types...						  */
		trace(TR_IDRP, 0, "IDRP PDU arrived from unconfigured neighbor %s", iso_netiso_ntoa(&pdu->header.src));
		return;
	}

	/* Ensure that it's an IDRP packet. */

	if (pdu->header.pid != IDRP_PID) {
		/* jgs -- note that idrp_fixed_header.src (and .dst) are iso_addrs, while iso_netiso_ntoa wants a */
		/* jgs -- iso_net_addr.  Which should change?  I'd like to better understand the reason to have   */
		/* jgs -- these two similar yet differing types...						  */
		trace(TR_IDRP, 0, "PDU type 0x%x not IDRP, from %s", pdu->header.pid, iso_netiso_ntoa(&pdu->header.src));
		return;
	}

	/* Deliver the packet to the appropriate task. */
	
	idrp_recv_pdu(peer, pdu, length,AF_ISO);
}

#endif /* PROTO_IDRP */
