#ifdef	PROTO_IDRP

#define DROP_PACKET (drop_packet())
/* #define DROP_PACKET (FALSE) */

#include "include.h"
#include "idrp.h"
#include <stdio.h>
#include "md4global.h"	/* RSAREF globals and constants */
#include "md4.h"	/* md4-specific includes */


/* 
 * idrp_pdus.c 
 *
 * pdu sending 
 *	idrp_post(peer,buf) - post a message, and put on re-xmit if seq
 *	post_enqueued_pdus(peer) - post as many enqueued PDUs as possible
 * 	idrp_send_pdu(peer,pdu,type,length) - send pdu to peer 
 *	send_open(peer) - send open BISPDU to peers
 *	send_update_pdu(peer,route,withdrawl,unreachable)
 *	
 * jgs -- what is this stuff (the "distribute" stuff)?  I don't find it in idrp.c.
 * jgs -- is it "to-do" stuff?
 *	all peers distribution for send updates   
 *       	distribute_route(route,withdrawl,unreachble)
 *			 - distribute route to all peers
 *		idrp_distribute_route(route) - distribute new nlri to all peers
 *		idrp_distribute_unreachable(route,withdrawl)
 *		 - distribute unfeasible only to all peers
 *		idrp_distribute_both(route,withdrawl) - both to all peers
 * 
 *	idrp_send_error(peer,code)
 *      send_cease(peer)
 * 	send_keepalive(peer)
 */ 
  
 
/*
 * Post a buffer to the socket for transmission.  Move it to the
 * retransmission queue if it is sequenced.
 *
 * The buffer must not be linked into any queue.
 */
static void idrp_post(peer, buf)
idrpPeer *peer;
idrpBuffer *buf;
{
	sockaddr_un *p_dst;
	u_int longhold;
	idrp_fixed_header *header = &(buf->pdu->header);
	u_char	digest[16];	

	/* Fill in the flow control fields.  The sequence number was assigned
	 * earlier; all other fields come from the current state. */

	longhold = buf->seq;
	longhold = htonl(longhold);
	bcopy((caddr_t) &longhold, (caddr_t) header->seq, sizeof(header->seq));

	longhold = peer->rcv_seq_expected - 1;
	bcopy((caddr_t) &longhold, (caddr_t) header->ack, sizeof(header->ack));

	header->credit_offered = peer->rcv_credit;
	header->credit_available = peer->upper_window_edge - peer->send_next;

	bzero((caddr_t) header->validation, sizeof(header->validation)); /* zero validation now */
	switch  (peer->AuthType)
		{
		case IDRP_AUTH_CODE_TEST:
			break;
		case IDRP_AUTH_CODE_CLEAR:
			md4((u_char *)buf->pdu,buf->length,digest); 	 /* validation scheme */
			break;

		case IDRP_AUTH_CODE_CRYPT:
			/* with zeros in buffer should fail reception */ 
			trace(TR_IDRP,0, "No Authentication code crypt support, set on this peer %s",
				peer->name);
			break;   		
		}

	/* Copy the buffer to the send buffer and post it. */

	bcopy ((caddr_t) &digest,&header->validation,sizeof(header->validation));
	bcopy((caddr_t) buf->pdu, task_send_buffer, buf->length);
	p_dst = idrp_send_dest_set(peer);
	
	if (!DROP_PACKET) /* %% */
		(void) task_send_packet(idrp_this_node.task, task_send_buffer, buf->length, 0, p_dst);
	trace(TR_IDRP, 0, "IDRP (%s) XMIT %s (%s) seq %d ack %d offer %d avail %d len %d to %s", 
		trace_state(peer->state, idrpStates),
		idrpTypes[header->pdu_type], 
		buf->sequenced ? "seq" : "unseq",
		buf->seq,
		peer->rcv_seq_expected - 1,
		header->credit_offered,
		header->credit_available,
		buf->length,
		peer->name);

#ifdef notdef
	trace(TR_IDRP, 0, "Sending:");
	{
		int i;
		char *cp = (char *) buf->pdu;
		for (i=0; i<buf->length; i++) {
			tracef("%2.2x  ", *cp++);
		}
		trace(TR_IDRP, 0, NULL);
	}
#endif

	/* Put the buffer on the retransmission queue if it was sequenced. */

	if (buf->sequenced) {
		*peer->rexmit_queue_tail = buf;
		peer->rexmit_queue_tail = &buf->next;
		start_rexmit_timer(peer);
	}
}

/*
 * Post as many enqueued PDUs as possible.
 */
void 
post_enqueued_pdus(peer)
idrpPeer *peer;
{
	idrpBuffer *buf;

	/* Process the transmit queue. */

	while (peer->xmit_queue_head) {

		/* If we're flow blocked, give up. */

		if (peer->upper_window_edge <= peer->send_next) {
			trace(TR_IDRP, 0, "IDRP %s flow blocked", peer->name);
			break;
		}

		/* Not flow blocked.  Pop the top of the queue, assign it a sequence number, and post it. */

		buf = peer->xmit_queue_head;
		peer->xmit_queue_head = buf->next;
		if (buf->next == NULL)
			peer->xmit_queue_tail = &peer->xmit_queue_head;
		buf->next = NULL;
		buf->seq = peer->send_next++;

		idrp_post(peer, buf);
	}
}


/*
 * Build and attempt to send a PDU.
 */
static void idrp_send_pdu(peer, pdu, type, length)
idrpPeer *peer;
idrpPdu *pdu;
u_int type;
u_int length;
{
	int i;
	int len;
	int waitack;
	u_short hold;
	u_int longhold;
	char *cp = (char *) pdu;
	idrp_fixed_header *header = &(pdu->header);
	idrpBuffer *buf;

	/* Fill in most of the fixed header. */
	/* PID, length, type, source, destination */
	/* copy source and destination from the Peer structure */

	header->pid = IDRP_PID;
	hold = length;
	hold =  htons(hold);
	bcopy((caddr_t) &hold, (caddr_t) header->pdu_length, sizeof(header->pdu_length));
	header->pdu_type = type;

	bcopy((caddr_t) &peer->neighbor.isoa_genaddr, (caddr_t) &header->dest.isoa_genaddr,peer->neighbor.isoa_len);
	header->dest.isoa_len = peer->neighbor.isoa_len;
	bcopy((caddr_t) &rt_net.isoa_genaddr, (caddr_t) &header->src.isoa_genaddr,rt_net.isoa_len);
	header->src.isoa_len = rt_net.isoa_len;

	/* Build the buffer structure. */

	buf = (idrpBuffer *) task_mem_malloc(peer->task, sizeof(idrpBuffer));
	buf->next = NULL;
	buf->pdu = pdu;
	buf->length = length;
	buf->sequenced = IS_SEQUENCED(type);

	/* If not sequenced, post the buffer immediately.  Otherwise, enqueue it. */

	if (buf->sequenced) {
		*peer->xmit_queue_tail = buf;
		peer->xmit_queue_tail = &buf->next;
		post_enqueued_pdus(peer);
	} else {
		buf->seq = peer->send_next;
		idrp_post(peer, buf);
	}
}

/*
 * Send an OPEN PDU to (peer).
 */
static void send_open(peer)
idrpPeer *peer;
{
	u_short hold;
	idrp_source_RDI *openrdi;
	idrp_RIB_att *openribatt;
	idrp_confed_id *openconfed;
	idrp_open_pdu_header *openhead;
	idrp_authentication_code *openauthcode;
	int i;

	idrpPdu *pdu = (idrpPdu *) task_mem_malloc(peer->task, peer->pdu_maxsendsize);
	idrp_open_pdu *open = &(pdu->pdu.open);
	char *cp = (char *) open;
	char *basecp = cp;

	/* Fixed header */


	bzero(pdu,peer->pdu_maxsendsize);
 	openhead = &(open->header);
	openhead->version = IDRP_VERSION_1;

	hold = peer->hold_time;
	hold = htons(hold);
	bcopy((caddr_t) &hold, (caddr_t) openhead->idrp_hold_time, sizeof(openhead->idrp_hold_time));

	hold = peer->pdu_maxrcvsize;
	hold = htons(hold);
	bcopy((caddr_t) &hold, (caddr_t) openhead->idrp_maxpdu_size, sizeof(openhead->idrp_maxpdu_size));

	openhead->idrp_initial_credit = peer->rcv_credit;
	cp += sizeof(idrp_open_pdu_header);

	/* Source RDI field */

	openrdi = (idrp_source_RDI *) cp;
	openrdi->source_RDI_length = rt_rdi.isoa_len;
	for (i=0; i<openrdi->source_RDI_length; i++) {
		openrdi->source_RDI[i] = rt_rdi.isoa_genaddr[i];
	}
	cp += openrdi->source_RDI_length + 1;

	/* RIB attributes */

	openribatt = (idrp_RIB_att *) cp;
	openribatt->RIB_att_count = 0;
	cp += sizeof(openribatt->RIB_att_count);

	/* Confederation IDs */
	/* %%eventually take from config */

	openconfed = (idrp_confed_id *) cp;
	openconfed->confed_id_count = 0;
	cp += sizeof(openconfed->confed_id_count);

	/* Authorization code */

	openauthcode = (idrp_authentication_code *) cp;
	openauthcode->authentication_code = IDRP_AUTH_CODE_CLEAR;
	cp += sizeof(openauthcode->authentication_code);

	/* Now send it out. */

	idrp_send_pdu(peer, pdu, IDRP_OPEN_TYPE, cp - basecp + sizeof(idrp_fixed_header));
}


/*
 * Send an update PDU
 */
extern void
send_update_pdu(peer,p_send_list,unreachable)
idrpPeer *peer;
idrp_send_list *p_send_list;
u_int unreachable;
{
/* send list provides the following
 * - pointer to attribute record 
 * - pointer head of idrpRoute list for nlri
 *   nlri is linked idrpRoute->route_out[peer->id] structure  
 * - pointer to tail of idrpRoute list for nlri 
 * - Withdrawl structure with IDs 

/* idrpRoute pointers */
idrpRoute	*p_idrp_rt;		/* generic route pointer */
idrpRoute	*p_irt_head;		/* list of idrpRoute structures - head */
idrpRoute	*p_irt_tail;		/* list of idrpRoute structures - tail */
idrp_attribute_record	*attr_rec;		/* attribute record */ 


/* pdu pointers */

idrpPdu *pdu;
idrp_update_pdu *hdr;
nlri_info *nlriptr;
u_char	*nlristart;

/* generic character pointers */

int i;
int cnt;
int len;
int seglen;	
int tot_len;
u_char *segout;
u_char *cpout;
u_char *cpin;
u_char *cpbase;
/* jgs -- this function needs to be looked at more carefully */
/* jgs -- hold seems to be used indiscriminately(?) as both  */
/* jgs -- an int and a pointer.  This will not work right!   */
u_short hold;
u_int longhold;

	/* set up pointers for the task */
	/* pdu and maximum pdu size */

	pdu = (idrpPdu *) task_mem_malloc(peer->task, peer->pdu_maxsendsize);
	hdr = (idrp_update_pdu *) &pdu->pdu.update;

	/* fill in the unfeasible route info. */
	/* make pdus won't overload the send_pdu */
	/* return with complaint */

	tot_len = p_send_list->withdraw->count * sizeof (route_id);
	tot_len += sizeof(idrp_fixed_header) + IDRP_NO_NLRI_BODY;

	if (tot_len > peer->pdu_maxsendsize)
		{
		trace (TR_IDRP,0,"bad withdraw pdu size at send_update_pdu peer %s",peer->name);
		return;
		}
	cnt = p_send_list->withdraw->count;
	cpout = (u_char *) &pdu->pdu.update.unfeasible_routes[1];
	for (i = 0;  i < cnt; i++)
		{
		/* route out_id is dependent on peer */
		longhold = p_send_list->withdraw->withdraw[i];
		htonl(longhold);
		bcopy(&longhold, cpout, sizeof(route_id));
		cpout += sizeof(route_id);
		cpbase = cpout;
		}

	/* just unreachables - just fire away */

	if (unreachable)
		{ 
		/* put out fixed header for path attribute length */

		longhold = 0;
		htonl(longhold);
		bcopy((caddr_t) &longhold,cpout,sizeof(hdr->tot_path_att_len));
		cpout += sizeof(hdr->tot_path_att_len);
	
		/* trace this update and send with just unreachables */
	
		trace(TR_UPDATE, 0, "IDRP XMIT Update to %s", peer->name);
		idrp_send_pdu(peer, pdu, IDRP_UPDATE_TYPE, cpout - (u_char *) pdu);
		}

	  
	/* path attributes new routes - so fill in attributes */ 
	/* path attribute stuff */

	attr_rec = (idrp_attribute_record *) p_idrp_rt->p_attr; 
	if (attr_rec->attrib[IDRP_ATTR_RD_PATH].present == 0)
		{
		trace (TR_IDRP,0,"update has no rd_path peer = %s",iso_netiso_ntoa(&peer->neighbor));
			return;
		}	
	cpin =  attr_rec->attrib[IDRP_ATTR_RD_PATH].data;  
	cpbase = cpout;
	
	/* First encode the path information. */
	/* In, array but at first lets hard code it */
	
	*cpout++ = 0; /* Well-known, mandatory */
	*cpout++ = IDRP_ATTR_RD_PATH;
	
	/* our RD is in local part of array */
	/* if neighbor is external or local Route */
	/* this part of array is already set-up */
	/* in the attribute - attr_rec->attrib[IDRP_LOCAL_RD_PATH];
	
	hold = attr_rec->attrib[IDRP_ATTR_RD_PATH].length; 
	htons(hold);
	bcopy((caddr_t) &hold, cpout, 2);	/* Attribute length */
	cpout += 2;

	/* check to make sure length is not zero */

	if (!attr_rec->attrib[IDRP_ATTR_RD_PATH].length) {
		trace(TR_ALL, LOG_ERR, "null path in send_update_pdu, %x peer %s",p_idrp_rt,p_idrp_rt->peer->name);
		kill(task_pid, SIGINT);
		}

	len = attr_rec->attrib[IDRP_ATTR_RD_PATH].length;		

	/* HOPCOUNT is mandatory */
	/* Add it to the pdu */

	/* if a local RD path has been added to this RD path */
	/* then add the local in from the local path */ 

	if (attr_rec->attrib[IDRP_ATTR_LOCAL_RD].present)
		{
		/* take off the length of the local RD */
		seglen = rt_rdi.isoa_len + 1;	
		len -= seglen; 
		segout = cpout + len;	
		bcopy ((caddr_t) attr_rec->attrib[IDRP_ATTR_LOCAL_RD].data,(caddr_t) segout,seglen); 
		}
		
	/* copy part that came from another router */
	/* for a local router this may be zero */

	if (len)
		{
		bcopy ((caddr_t) attr_rec->attrib[IDRP_ATTR_RD_PATH].data,cpout,len);
		}


	if (!attr_rec->attrib[IDRP_ATTR_HOP_COUNT].present)
		{	
		cpout = attr_rec->attrib[IDRP_ATTR_ROUTE_ID].data; 
		trace(TR_ALL, LOG_ERR, "NO HOPCOUNT in route %x peer %s",cpout,peer->name);
		kill(task_pid, SIGINT);
		}

	*cpout++ = IDRP_ATTRFLG_HOP_COUNT;  /* Well-known, mandatory */
	*cpout++ = IDRP_ATTR_HOP_COUNT;
	*cpout++ = 0;
	*cpout++ = IDRP_RD_HOP_COUNT_LENGTH;
	*cpout++ = p_idrp_rt->p_attr->hopcount + 1;

	/* Now add the route ID. */

	*cpout++ = 0;  /* Well-known, mandatory */
	*cpout++ = IDRP_ATTR_ROUTE_ID;
	*cpout++ = 0;
	*cpout++ = IDRP_ROUTE_ID_LENGTH;

	/* route id out depends on the neighbor */ 
	longhold = p_idrp_rt->route_out[peer->id].route_id;  
	bcopy((caddr_t) &longhold, cpout, IDRP_ROUTE_ID_LENGTH);
	cpout += IDRP_ROUTE_ID_LENGTH;

	/* past the mandatory attributes */
	/* here's where we add the following copies */
	/* of attribute record */

	if ((attr_rec->idrp_mask & IDRP_ATTBT_MANDATORY) != IDRP_ATTBT_MANDATORY)
		{
		/* more attributes to go in the pdu */

		for (i = IDRP_ATTR_EXT_INFO; i < IDRP_ATTR_MAX; i++)
			{ 
			/* loop copying in the attributes verbadum */
			/* flags(1 byte), type(1 byte), length(2 bytes) */
			/* probably should have this above too */
	
			if (attr_rec->attrib[i].present)
				{
				*cpout++ = attr_rec->attrib[i].flags;
				*cpout++ = (u_char) i;
				longhold = attr_rec->attrib[i].length;
				bcopy((caddr_t) &longhold,cpout,sizeof(attr_rec->attrib[i].length));  
				cpout +=2;
   				bcopy((caddr_t) attr_rec->attrib[i].data,cpout,longhold);
				cpout +=longhold;
				}
			}   
		}
	 
		
	/* That was easy.  Now go back and update the total attribute length. */

	hold = cpout - cpbase;
	htons(hold);
	bcopy((caddr_t) &hold, hdr->tot_path_att_len, sizeof(hdr->tot_path_att_len));

	/* Lastly, add the NLRI */
	/* walk list of idrp Routes putting out routes */

	/* send list will provide linked list per ISO families */
	NLRI_FAMILY_LOOP(i)
		{
		if (!p_send_list->ann_nlri[i].head)
			{
			continue;
			}

		/* copy in protocol id, length and value */
	
		*cpout = idrp_nlri_proto_id[i].proto_type;
		cpout++;
		*cpout = hold = idrp_nlri_proto_id[i].proto_len;
		cpout ++;
		bcopy(idrp_nlri_proto_id[i].proto_val,cpout,hold);
		cpout +=hold;
	 	
		/* skip over total type length */	

		nlristart = cpout;	
		cpout += sizeof(nlriptr->addr_len);

		/* set up linked list for array headers */
	
		p_idrp_rt = p_irt_head = p_send_list->ann_nlri[i].head;
		p_irt_tail = p_send_list->ann_nlri[i].tail;

		/* start copying in the nlris */

		while (p_idrp_rt != p_irt_tail)
			{		
			/* prefix length */

			hold = PFX_BYTE_LENGTH(p_idrp_rt->nlri) + 1;
			*cpout = hold;
			cpout++;
		
			/* prefix */

			/* jgs 930208 -- this looks wrong.  I think that cpout */
			/* jgs 930208 -- (even if the above code were right) */
			/* jgs 930208 -- will be some value that's not a real */
			/* jgs 930208 -- address.  Param 2 is supposed to be */
			/* jgs 930208 -- a destination addr. */
			bcopy(&p_idrp_rt->nlri.pfx,cpout,hold);
			cpout += hold;
 			p_idrp_rt = p_idrp_rt->route_out[peer->id].p_next;
			}  /* done with nlri per family */

		/* copy in length */

		longhold = (cpout - nlristart);
		htonl(longhold);
		bcopy(&longhold,nlristart,sizeof(nlriptr->addr_len));
 
		} /* looping for packing nlri for all families */ 

	/* Finally, send it out. */

	tot_len = cpout - (u_char *)pdu;
	if (tot_len > peer->pdu_maxsendsize)
		{
		trace (TR_IDRP | TR_UPDATE,0,"IDRP XMIT peer %s update has too big pdu %d max size %d", peer->name,tot_len,peer->pdu_maxsendsize);
		return;
		}
	
	trace(TR_UPDATE, 0, "IDRP XMIT Update to %s", peer->name);
	idrp_send_pdu(peer, pdu, IDRP_UPDATE_TYPE, cpout - (u_char *) pdu);
}




/*
 * Send an error PDU
*/
void
idrp_send_error(peer, code)
idrpPeer	*peer;
u_int		code;
{
u_int subcode;
caddr_t *data;
u_int datalen;
idrpPdu *pdu;
idrp_error_pdu *errpdu;
int packetlen;

	if (code != peer->last_error_pdu.error_code)
		{
		trace (TR_IDRP,0,"error with send_error routine peer %s last_error_pdu code = %c call = %c", 
		peer->name,code,peer->last_error_pdu.error_code);
		}
	pdu = (idrpPdu *) task_mem_malloc(peer->task, packetlen);
 	errpdu = (idrp_error_pdu *) &(pdu->pdu.error);
	data = (caddr_t *) peer->last_error_pdu.data;
	datalen = peer->last_error_pdu.data_len; 
	packetlen = sizeof(idrp_fixed_header) + sizeof(idrp_error_header) + datalen;
	errpdu->header.error_code = peer->last_error_pdu.error_code;
	errpdu->header.error_subcode = peer->last_error_pdu.error_subcode;
	if (datalen) {
		bcopy(data, errpdu->data, datalen);
	}
	idrp_send_pdu(peer, pdu, IDRP_ERROR_TYPE, packetlen);
	begin_close(peer);
}


/*
 * Send a cease PDU
 */
static void send_cease(peer)
idrpPeer *peer;
{
	idrpPdu *pdu = (idrpPdu *) task_mem_malloc(peer->task, IDRP_CEASE_SIZE);
	idrp_send_pdu(peer, pdu, IDRP_CEASE_TYPE, IDRP_CEASE_SIZE);
}


/*
 * Send a keepalive PDU.  Ensure that this keepalive will be retransmitted soon if
 * the peer thinks he is flow blocked.
 */

static void send_keepalive(peer)
idrpPeer *peer;
{
	idrpPdu *pdu = (idrpPdu *) task_mem_malloc(peer->task, IDRP_KEEPALIVE_SIZE);
	idrp_send_pdu(peer, pdu, IDRP_KEEPALIVE_TYPE, IDRP_KEEPALIVE_SIZE);

	if (peer->cred_avail_zero_seq_last == 0) {	/* If we aren't currently retransmitting keepalives */
		if (peer->rcved_credit_avail == 0) {	/* If the other guy is flow blocked */
			peer->cred_avail_zero_seq_last = peer->cred_avail_zero_seq;
			start_rexmit_timer(peer);
		}
	}
}


#ifdef ECHOKLUDGE
static void send_echo(peer)
idrpPeer *peer;
{
	idrpPdu *pdu = (idrpPdu *) task_mem_malloc(peer->task, IDRP_ECHO_SIZE);
	idrp_send_pdu(peer, pdu, IDRP_ECHO_TYPE, IDRP_ECHO_SIZE);
}

static void send_echo_reply(peer)
idrpPeer *peer;
{
	idrpPdu *pdu = (idrpPdu *) task_mem_malloc(peer->task, IDRP_ECHO_REPLY_SIZE);
	idrp_send_pdu(peer, pdu, IDRP_ECHO_REPLY_TYPE, IDRP_ECHO_REPLY_SIZE);
}
#endif /* ECHOKLUDGE */

#endif /* PROTO_IDRP */
