#ifdef	PROTO_IDRP

#include "include.h"
#include "idrp.h"
#include <stdio.h>


/* idrp_rib_refresh.c
 *
 * rib refresh routines:
 *	rib_refresh_pdu_release(peer) - release all the Rib Refresh updates
 *	refresh_pdu_link(peer,results,pdu)
 *	rib_refresh_update_process(peer) - parse all the updates by walking 
 *		the Rib Refresh chain
 */ 
 

void rib_refresh_pdu_release(peer) 
idrpPeer *peer;
{
/* release all the Rib Refresh updates */

idrpRefresh	*p_ref;
idrpRefresh	*p_ref2;
	/* walk the chain and release the refresh structures */

	p_ref = peer->refresh.pdu_list_start;
	while (p_ref) 
		{ 
		task_mem_free(peer->task,p_ref->results->withdraw);
		task_mem_free(peer->task,p_ref->results);
		p_ref2 = p_ref;
		p_ref= p_ref->p_next;
		task_mem_free(peer->task,p_ref2);
		}
}

void refresh_pdu_link(peer,results,pdu)
idrpPeer *peer;
parse_results *results;
idrpPdu *pdu;
{
idrpRefresh	*p_refresh;	

	p_refresh = (idrpRefresh *) task_mem_malloc(peer->task,sizeof(idrpRefresh));
	p_refresh->p_next = (idrpRefresh *) 0;
	p_refresh->p_pdu = pdu;
	p_refresh->results = results;

	/* if this is not the first pdu on the list */

	if (peer->refresh.pdu_list_start)
		{
		peer->refresh.pdu_list_end->p_next = p_refresh;
		peer->refresh.cnt++;
		} 
	else
		{
		peer->refresh.pdu_list_end = peer->refresh.pdu_list_start = p_refresh;
		}
}


	
void rib_refresh_update_process(peer)
idrpPeer *peer;
{
idrpRefresh *p_ref; 

/* parse all the updates by walking the Rib Refresh */
/* chain */

	if (peer->refresh.pdu_list_start == (idrpRefresh *) 0) 
		{
		/* nothing in the link list - just return */
		trace(TR_IDRP,0,"Rib Refresh stop with no update pdus for peer %s  ",
			peer->name);
		return;
		}

	/* go through full list and update the gated routing table */
	
	for (p_ref = peer->refresh.pdu_list_start; p_ref == peer->refresh.pdu_list_end; p_ref = p_ref->p_next)
		{
		idrp_process_pdu_routes(peer,p_ref->results);
		peer->refresh.cnt--;

		/* clean up the results array */

		task_mem_free(peer->task,p_ref->results->withdraw);
		if (p_ref->results->free_pdu)
				{
				/* free idrp pdu - since attribute record duplicate */
				}	
		task_mem_free(peer->task,p_ref->results);
		task_mem_free(peer->task,p_ref);

		}
return;
}

#endif /* PROTO_IDRP */
