#ifdef	PROTO_IDRP

#include "include.h"
#include "idrp.h"
#include <stdio.h>


/* idrp_sm.c
 *  
 * IDRP State machine
 *	idrp_sm(peer,event) 
 */ 
  
 

int acked_something;			/* True if last received packet acknowledged something */
					/* jgs -- needs to be patched up */



int idrp_sm(peer,event)
idrpPeer *peer;
int	event;
{
int return_code = 0;
int old_state;

/* idrp state machine  */
    old_state = peer->state;
         
    switch (event) 
      {
      /* start event */

      case IDRP_START_EVENT:
	switch (peer->state) {
		case IDRP_CLOSED:
			peer->state=IDRP_OPEN_SENT;
			send_open(peer);
			break;

		case IDRP_OPEN_RCVD:
		case IDRP_OPEN_SENT:
		case IDRP_ESTABLISHED:
		case IDRP_CLOSE_WAIT:
		default:
			trace(TR_IDRP, LOG_ERR, "IDRP %s start event in state %s", peer->name, trace_state(peer->state,
			idrpStates));
			break;
		}
	break;
	
     /* stop event */

     case IDRP_STOP_EVENT:
	switch (peer->state) {
		case IDRP_OPEN_RCVD:
		case IDRP_OPEN_SENT:
		case IDRP_ESTABLISHED:
			peer->state = IDRP_CLOSE_WAIT;   /* %% more?? %% */
			send_cease(peer);
			break;

		case IDRP_CLOSED:
		case IDRP_CLOSE_WAIT:
		default:
			trace(TR_IDRP, LOG_ERR, "IDRP %s stop event in state %s", peer->name, trace_state(peer->state,
			idrpStates));
			break;
		}
	break;
  
    /* idrp Close wait timer expired event */
  
    case IDRP_CLOSE_TIMEOUT:
	switch (peer->state) {
        	case IDRP_CLOSE_WAIT:
			/* close down this peer and then restart */ 
			peer->state=IDRP_CLOSED;
               		close_peer(peer);
	                idrp_start_event(peer);
       		        break;

        	default:
	                trace(TR_IDRP, LOG_ERR, "IDRP %s close_wait timer expired in state %s", peer->name, trace_state(peer->state,
                        idrpStates));
       			break;
		}
	break;

    /* case Expiration of hold timer */

    case IDRP_HOLDTIME_TIMEOUT:
	switch(peer->state)
		{
		case IDRP_OPEN_RCVD:
		case IDRP_ESTABLISHED:
			peer->state = IDRP_CLOSE_WAIT;
			peer->last_error_pdu.error_code = IDRP_ERROR_HOLDTIME_EXPIRED;
			peer->last_error_pdu.error_subcode = (u_char)NULL;
			peer->last_error_pdu.data_len = 0;
			idrp_send_error(peer,IDRP_ERROR_HOLDTIME_EXPIRED); 
			break;	
	
        	case IDRP_OPEN_SENT:
			peer->state = IDRP_CLOSE_WAIT;
			break;

		case IDRP_CLOSED:
		case IDRP_CLOSE_WAIT:
			break;	
		}
	break;
	

    /* case BISPDU with errors */

     case IDRP_HEADER_ERROR:
		trace(TR_IDRP, LOG_ERR, "IDRP peer %s got pdu with header error in state %s", peer->name, peer->state);
	     	break;

    case IDRP_KEEPALIVE:
	{
	switch (peer->state) {
         
		case IDRP_CLOSED:
			trace(TR_IDRP, 0, "peer %s received keepalive in CLOSED state",peer->name);
			break;


		case IDRP_OPEN_RCVD:
			if (acked_something) {
				peer->state=IDRP_ESTABLISHED;
	                        path_up(peer);
       				}
			else
				{
				peer->state=IDRP_CLOSE_WAIT;
				peer->last_error_pdu.error_code = IDRP_ERROR_FSM;
				peer->last_error_pdu.error_subcode = IDRP_FSM_ERROR_OPEN_RCVED;
				peer->last_error_pdu.data_len = 0;
				begin_close(peer);
				idrp_send_error(peer,IDRP_ERROR_FSM);
				}
	                break;

        	case IDRP_OPEN_SENT:
			/* If we are in state OPEN_SENT, this is a FSM error.  */
			peer->state = IDRP_CLOSE_WAIT;
			peer->last_error_pdu.error_code = IDRP_ERROR_FSM;
			peer->last_error_pdu.error_subcode = IDRP_FSM_ERROR_OPEN_SENT;
			peer->last_error_pdu.data_len = 0;
                	idrp_send_error(peer, IDRP_ERROR_FSM);
			break;

		case IDRP_CLOSE_WAIT:
			trace(TR_IDRP, 0, "peer %s received keepalive in Close_wait state",peer->name);
			break;

		case IDRP_ESTABLISHED:
			trace(TR_IDRP, 0, "peer %s received keepalive in Established state",peer->name);
			peer->keepAlives_Since_last_Update++;
			break;
		}  /* end of switch statement */
	break;
	}	/* end of KEEPALIVE case */	

   /* CEASE With no errors */	
   case	IDRP_CEASE:
	{
	if (peer->state != IDRP_CLOSED)
		{
		begin_close(peer);
		peer->state = IDRP_CLOSE_WAIT;
		}
	break;
	} 


   case IDRP_OPEN_ERROR:

	switch (peer->state) {
		case IDRP_CLOSED:
			trace(TR_IDRP, 0, "peer %s received OPEN w/error in CLOSED state",peer->name);
			break;

		case IDRP_OPEN_RCVD:
		case IDRP_OPEN_SENT:
			peer->state = IDRP_CLOSE_WAIT;
			idrp_send_error(peer,IDRP_ERROR_OPEN_PDU);
			break;

		case IDRP_CLOSE_WAIT:
			trace(TR_IDRP, 0, "peer %s received OPEN w/error in CLOSE_WAIT state",peer->name);
			break;

		case IDRP_ESTABLISHED:
			peer->state = IDRP_CLOSE_WAIT;
			begin_close(peer);
			idrp_send_error(peer,IDRP_ERROR_OPEN_PDU);
			break;
		}
	break;

	/* valid IDRP open */

   case IDRP_OPEN:

        /*
         * If state is OPEN_SENT or OPEN_RCVD and the received open acked ours, bring up the
         * connection.  If it didn't ack, go to OPEN_RCVD (and wait for the retransmission timer
         * to go off and resend our OPEN).  If wrong state, ignore it.
         */
        switch (peer->state) {

        case IDRP_OPEN_SENT:
        case IDRP_OPEN_RCVD:
                if (acked_something) {
                        path_up(peer);
			peer->state = IDRP_ESTABLISHED;
			}
                else 	{
                        peer->state =IDRP_OPEN_RCVD;
			send_open(peer);
			}
                break;

	case IDRP_ESTABLISHED:
		peer->state=IDRP_CLOSE_WAIT;
		peer->last_error_pdu.error_code = IDRP_ERROR_FSM;
		peer->last_error_pdu.error_subcode = IDRP_FSM_ERROR_ESTABLISHED;
		idrp_send_error(peer,IDRP_ERROR_FSM);
		break;
        default:
		trace(TR_IDRP, 0, "OPEN received from peer %s in state %d", peer->name, peer->state);
                break;
	}
	break;

    case IDRP_UPDATE_ERROR:
	{		
	switch (peer->state) {
		case IDRP_OPEN_RCVD:
		case IDRP_ESTABLISHED:
			peer->state = IDRP_CLOSE_WAIT;
			idrp_send_error(peer,IDRP_UPDATE_ERROR);
			break;

		case IDRP_OPEN_SENT:
			peer->state = IDRP_CLOSE_WAIT;
			peer->last_error_pdu.error_code = IDRP_ERROR_FSM;
			peer->last_error_pdu.error_subcode = IDRP_FSM_ERROR_OPEN_SENT;
			peer->last_error_pdu.data_len = 0;
			idrp_send_error(peer,IDRP_ERROR_FSM);
			break;

		case IDRP_CLOSED:
		case IDRP_CLOSE_WAIT:
		default:
			trace(TR_IDRP, 0, "IDRP UPDATE header error and Close wait peer %s",peer->name);
			break;
		}  /* end of switch statement */
	break;
	}  /* end of Update w/error case */

      case IDRP_UPDATE:
	return_code = 0;
	switch (peer->state)
		{
		case IDRP_CLOSED:
			trace(TR_IDRP, 0, "IDRP UPDATE received in Close state peer %s",peer->name);
			break;

		case IDRP_OPEN_RCVD:
			/* if valid ack -> established */
			if (acked_something)
				{
				peer->state = IDRP_ESTABLISHED;
				return_code = 1;
				}
			else	{
				peer->state = IDRP_CLOSE_WAIT;
				peer->last_error_pdu.error_code = IDRP_ERROR_FSM;
				peer->last_error_pdu.error_subcode = IDRP_FSM_ERROR_OPEN_RCVED;  
				peer->last_error_pdu.data_len = 0;
				idrp_send_error(peer,IDRP_ERROR_FSM);
				}
				break;

		case IDRP_OPEN_SENT:
			peer->state = IDRP_CLOSE_WAIT;
			peer->last_error_pdu.error_code = IDRP_ERROR_FSM;
			peer->last_error_pdu.error_subcode = IDRP_FSM_ERROR_OPEN_SENT;  
			peer->last_error_pdu.data_len = 0;
			idrp_send_error(peer,IDRP_ERROR_FSM);
			break;

		case IDRP_CLOSE_WAIT:
			trace(TR_IDRP, 0, "IDRP UPDATE received in Close state peer %s",peer->name);
			break;

		case IDRP_ESTABLISHED:
			/* return code to process pdu */
			return_code = 1;
			break;
	
		default:
			trace(TR_IDRP, 0, "IDRP UPDATE and state machine in illegal state %d peer %s",peer->name);
			break;
      		}  /* end of switch */
	   break;

    /* logic for IDRP_ERROR and IDRP_ERROR_ERROR  is the same */ 
    
   case IDRP_ERROR:
   case IDRP_ERROR_ERROR:
	switch (peer->state)
		{
		case IDRP_OPEN_RCVD:
		case IDRP_OPEN_SENT:	
		case IDRP_ESTABLISHED:
			peer->state = IDRP_CLOSE_WAIT;
			send_cease(peer);
			begin_close(peer);
			break;

		case IDRP_CLOSED:
		case IDRP_CLOSE_WAIT:
		default:
			trace(TR_IDRP, 0, " IDRP ERROR pdu received from peer%s",peer->name);
			break;	 
		}
	break;

    case IDRP_REFRESH_ERROR:
	switch (peer->state)
		{
		case IDRP_OPEN_SENT:	
			peer->state = IDRP_CLOSE_WAIT;
			peer->last_error_pdu.error_code = IDRP_ERROR_FSM;
			peer->last_error_pdu.error_subcode = IDRP_FSM_ERROR_OPEN_SENT;
			peer->last_error_pdu.data_len = 0;
			idrp_send_error(peer,IDRP_REFRESH_ERROR);
			break;
		
		case IDRP_OPEN_RCVD:
			/* parsing of pdu has already set error codes */
			peer->state = IDRP_CLOSE_WAIT;
			idrp_send_error(peer,IDRP_REFRESH_ERROR);
			break;

		case IDRP_ESTABLISHED:
			/* parsing of pdu has already set error codes */
			/* clear down peer */

			peer->state = IDRP_CLOSE_WAIT;
			begin_close(peer);
			idrp_send_error(peer,IDRP_REFRESH_ERROR);
			break;

		case IDRP_CLOSED:
		case IDRP_CLOSE_WAIT:
		default:
			trace(TR_IDRP, 0, " IDRP RIB REFRESH pdu received from peer %s",peer->name);
			break;	 
		}
	break;

   /* IDRP RIB REFRESH packet */

   case IDRP_REFRESH:
	return_code = 0; 
	switch(peer->state)
		{
		case IDRP_CLOSE_WAIT:
		case IDRP_CLOSED:
			trace(TR_IDRP, 0, " IDRP RIB REFRESH pdu received from peer %s",peer->name);
			break;

		case IDRP_OPEN_RCVD:
			if (acked_something)
				{
				peer->state = IDRP_ESTABLISHED;
				path_up(peer);
				send_keepalive(peer); /* %% Send RIB contents */
				return_code = 1;
				}
			else	{
				peer->state = IDRP_CLOSE_WAIT;
				peer->last_error_pdu.error_code = IDRP_ERROR_FSM;
				peer->last_error_pdu.error_subcode = IDRP_FSM_ERROR_OPEN_RCVED;
				peer->last_error_pdu.data_len = 0;	
				idrp_send_error(peer,IDRP_ERROR_FSM);
				}
			break;
				
	
		case IDRP_OPEN_SENT:
			peer->state = IDRP_CLOSE_WAIT;
			begin_close(peer);
			peer->last_error_pdu.error_code = IDRP_ERROR_FSM;
			peer->last_error_pdu.error_subcode = IDRP_FSM_ERROR_OPEN_SENT;
			peer->last_error_pdu.data_len = 0;
			idrp_send_error(peer,IDRP_ERROR_FSM);
			break;
	
		case IDRP_ESTABLISHED:
			return_code = 1;
			break;
		}  /* end of switch within Rib Refresh w/no errors */
	break;

    } /* end of switch with events */

    if (old_state != peer->state)
	{	
	trace(TR_IDRP, 0, "IDRP %s state %s -> %s", peer->name, trace_state(old_state, idrpStates), 
		trace_state(peer->state, idrpStates));
	}
    return(return_code);
}

#endif /* PROTO_IDRP */
