#ifdef	PROTO_IDRP

#include "include.h"
#include "idrp.h"
#include <stdio.h>


/* idrp_transport.c
 *
 * idrp transport queue handling 
 *     	flush_xmit queue(peer)
 *	path_down(peer)
 *	free_acked_pdu(peer)
 *	flush_rexmit_queue(peer)
 *	close_peer(peer) - close all queues and reset timers
 *	begin_close(peer) - flush queues, reset timers for close wait
 * 	ack_pdu(peer,ack,credit,type)	 
 */ 
  
 
int acked_something;			/* True if last received packet acknowledged something */
					/* jgs -- need to patch this up.  Maybe put it in the  */
					/* jgs -- .h file. */



/*
 * Flush the transmit queue
 */
static void flush_xmit_queue(peer)
idrpPeer *peer;
{
	idrpBuffer *curbuf;
	while ((curbuf = peer->xmit_queue_head) != NULL) {
		peer->xmit_queue_head = curbuf->next;
		free_buffer(peer, curbuf);
	}
	peer->xmit_queue_tail = &peer->xmit_queue_head;
}


/*
 * Bring down the path to our neighbor
 */
static void path_down(peer)
idrpPeer *peer;
{
	trace(TR_ALL, LOG_ERR, "IDRP path to %s is down", peer->name);
	idrp_peer_down(peer);
}


/*
 * Free the head of the retransmit queue.
 */
static void free_acked_pdu(peer)
idrpPeer *peer;
{
	idrpBuffer *buf;
	if (peer->rexmit_queue_head) {
		buf = peer->rexmit_queue_head;
		peer->rexmit_queue_head = buf->next;
		if (buf->next == NULL)
			peer->rexmit_queue_tail = &peer->rexmit_queue_head;
		buf->next = NULL;
		free_buffer(peer, buf);
	}

	/* If the queue is empty, kill the timer. */

	if (!peer->rexmit_queue_head) {
		kill_rexmit_timer(peer);
	}
}


/*
 * Flush the retransmit queue
 */
static void flush_rexmit_queue(peer)
idrpPeer *peer;
{
	while (peer->rexmit_queue_head) {
		free_acked_pdu(peer);
	}
}

/* 
 * close peer, completes the clean-up of a connection 
 */
static void close_peer(peer)
idrpPeer *peer;
{
	int old_state = peer->state;
	flush_rexmit_queue(peer);
	flush_xmit_queue(peer);
	timer_reset(TASK_TIMER(peer->task, IDRPTIMER_CLOSEWAIT));
	timer_reset(TASK_TIMER(peer->task, IDRPTIMER_HOLDTIME));
	timer_reset(TASK_TIMER(peer->task, IDRPTIMER_KEEPALIVE));
	timer_reset(TASK_TIMER(peer->task, IDRPTIMER_OPENSENT));
	trace(TR_IDRP, 0, "IDRP %s connection closed", peer->name);
}

/* 
 * begin close sequence - flush transmit and re-transmitt queues, reset timers
 */ 
static void begin_close(peer)
idrpPeer *peer;
{
	flush_rexmit_queue(peer);
	flush_xmit_queue(peer);
	peer->cred_avail_zero_seq_last = 0;
	timer_set(TASK_TIMER(peer->task, IDRPTIMER_CLOSEWAIT), IDRP_CLOSE_WAIT_TIME, (time_t) 0);
	timer_reset(TASK_TIMER(peer->task, IDRPTIMER_KEEPALIVE));
	timer_reset(TASK_TIMER(peer->task, IDRPTIMER_HOLDTIME));
	kill_rexmit_timer(peer);
#ifdef ECHOKLUDGE
	timer_reset(TASK_TIMER(peer->task, IDRPTIMER_ECHO));
#endif
	trace(TR_IDRP, 0, "IDRP %s connection closing", peer->name);
}

/* 
 * ack_pdu (peer,ack,credit,type) 
 *     process end to end acknoweldgements in a pdu, set flag if acked a 
 *     sequence number.
 */
static void ack_pdu(peer, ack, credit,type)
idrpPeer *peer;
u_int ack;
u_int credit;
u_int type;
{
	idrpBuffer *buf;
	u_int old_upper_window_edge;

	/* Pull any acknowledged packets off of the retransmission queue */

	if (ack >= peer->send_unack) {
		acked_something = TRUE;
		while (ack >= peer->send_unack) {
			if (!peer->rexmit_queue_head) {
				trace(TR_ALL, LOG_ERR, "IDRP %s rexmit queue underflow\n", peer->name);
				break;
			} else {
				if (peer->rexmit_queue_head->seq != peer->send_unack) {
					trace(TR_ALL, LOG_ERR, "IDRP %s rexmit queue broken", peer->name);
				}
				free_acked_pdu(peer);
			}
			peer->send_unack++;
		}
	}

	/* Update the upper window edge.  If we were flow blocked, and now we're not, make sure
	 * our peer knows that we're not blocked any longer (send a keepalive if there's nothing
	 * otherwise queued to send).
	 */

	old_upper_window_edge = peer->upper_window_edge;
	peer->upper_window_edge = MAX(peer->upper_window_edge, ack + credit + 1);
	if (peer->state == IDRP_ESTABLISHED &&
	    old_upper_window_edge != peer->upper_window_edge &&
	    old_upper_window_edge <= peer->send_next &&
	    peer->xmit_queue_head == NULL) {
		send_keepalive(peer);
	}

	/* If there are any packets waiting in the transmission queue, post as many as possible. */
	/* CEASE, ERROR, and OPEN with errors do not see the light of this processing */
	/* ACK only processed for valid open, valid KEEPALIVE, any (??) update, any (??) refresh */
	/* %% wonder if only good refreshes should have ack processed?? - spec unclear */
	/* either one seems to be ok, since an error in either will soon shutdown peer */ 

	post_enqueued_pdus(peer);
}
		
#endif /* PROTO_IDRP */
